@@ -337,10 +337,19 @@ Please check that all strings which were not recognizable to the compiler
   ;; SAVE-LISP-AND-DIE.
   #-sb-devel (!unintern-init-only-stuff)
 
-  ;; A symbol whose INFO slot underwent any kind of manipulation
-  ;; such that it now has neither properties nor globaldb info,
-  ;; can have the slot set back to NIL if it wasn't already.
+
   (do-all-symbols (symbol)
+    ;; Don't futz with the header of static symbols.
+    ;; Technically LOGIOR-HEADER-BITS can only be used on an OTHER-POINTER-LOWTAG
+    ;; objects, so modifying NIL should not ever work, but it's especially wrong
+    ;; on ppc64 where OTHER- and LIST- pointer lowtags are 10 bytes apart instead
+    ;; of 8, so this was making a random alteration to the header.
+    (unless (eq (heap-allocated-p symbol) :static)
+      (sb-kernel:logior-header-bits symbol sb-vm::+symbol-initial-core+))
+
+    ;; A symbol whose INFO slot underwent any kind of manipulation
+    ;; such that it now has neither properties nor globaldb info,
+    ;; can have the slot set back to NIL if it wasn't already.
     (when (and (sb-kernel:symbol-%info symbol) ; "raw" value is something
                ;; but both "cooked" values are empty
                (null (sb-kernel:symbol-dbinfo symbol))
@@ -403,6 +412,12 @@ Please check that all strings which were not recognizable to the compiler
                        (sb-impl::package-external-symbol-count x)
                        (sb-impl::package-internal-symbol-count x)))
                (sort (list-all-packages) #'string< :key 'package-name))))
+  #-sb-devel
+  ;; Remove inline expansions
+  (do-symbols (symbol #.(find-package "SB-C"))
+    (when (equal (symbol-package symbol) #.(find-package "SB-C"))
+      (sb-int:clear-info :function :inlining-data symbol)
+      (sb-int:clear-info :function :inlinep symbol)))
   (sb-impl::shake-packages
    ;; Development mode: retain all symbols with any system-related properties
    #+sb-devel
@@ -433,11 +448,14 @@ Please check that all strings which were not recognizable to the compiler
       (#.(find-package "SB-ALIEN")
        (or (eq accessibility :external) (eq symbol 'sb-alien::alien-callback-p)))
       (#.(mapcar 'find-package
-                 '("SB-ASSEM" "SB-BROTHERTREE" "SB-C" "SB-DISASSEM" "SB-FORMAT"
+                 '("SB-ASSEM" "SB-BROTHERTREE" "SB-DISASSEM" "SB-FORMAT"
                    "SB-IMPL" "SB-KERNEL" "SB-MOP" "SB-PCL" "SB-PRETTY" "SB-PROFILE"
                    "SB-REGALLOC" "SB-SYS" "SB-UNICODE" "SB-UNIX" "SB-WALKER"))
        ;; Assume all and only external symbols must be retained
-       (eq accessibility :external))
+         (eq accessibility :external))
+      (#.(find-package "SB-C")
+       (or (eq accessibility :external)
+           (member symbol '(sb-c::tab sb-c::scramble))))
       (#.(find-package "SB-LOOP")
        (or (eq accessibility :external)
            ;; Retain some internals to keep CLSQL working.
@@ -515,63 +533,3 @@ Please check that all strings which were not recognizable to the compiler
 
 ;; See comments in 'readtable.lisp'
 (setf (readtable-base-char-preference *readtable*) :symbols)
-
-
-"done with warm.lisp, about to SAVE-LISP-AND-DIE"
-
-#|
-This is the actual "name" of a toplevel code component that gets dumped to fasl
-when compiling src/pcl/boot. Not only does it contain a format control string in
-its raw representation, this is a complete and utter waste of time to dump.
-We really ought to try a LOT harder not to produce such garbage metadata:
-
-* (progn (terpri)
-   (write (sb-c::compiled-debug-fun-toplevel-name
-           (elt (sb-c::compiled-debug-info-fun-map *cdi*) 0))
-          :level nil :length nil))
-
-(SB-C::TOP-LEVEL-FORM
- (LABELS ((WARN-PARSE (SPECIALIZER &OPTIONAL CONDITION)
-            (STYLE-WARN "~@<Cannot parse specializer ~S in ~S~@[: ~A~].~@:>"
-                        SPECIALIZER # CONDITION))
-          (WARN-FIND (CONDITION NAME PROTO-GENERIC-FUNCTION PROTO-METHOD)
-            (WARN CONDITION :FORMAT-CONTROL
-                  #<(SIMPLE-BASE-STRING
-                     228) ~@<Cannot find type for specializer ~
-                  ~/sb-ext:print-symbol-with-prefix/ when executing ~S ~
-                  for a ~/sb-impl:print-type-specifier/ of a ~
-                  ~/sb-imp... {1003A2BC8F}>
-                  :FORMAT-ARGUMENTS #))
-          (CLASS-NAME-TYPE-SPECIFIER
-              (NAME PROTO-GENERIC-FUNCTION PROTO-METHOD &OPTIONAL #)
-            (LET #
-              #)))
-   (DEFUN REAL-SPECIALIZER-TYPE-SPECIFIER/SYMBOL
-          (PROTO-GENERIC-FUNCTION PROTO-METHOD SPECIALIZER)
-     (LET (#)
-       (WHEN SPECIALIZER #)))
-   (DEFUN REAL-SPECIALIZER-TYPE-SPECIFIER/T
-          (PROTO-GENERIC-FUNCTION PROTO-METHOD SPECIALIZER)
-     (LET (#)
-       (WHEN SPECIALIZER #)))
-   (DEFUN REAL-SPECIALIZER-TYPE-SPECIFIER/CLASS-EQ-SPECIALIZER
-          (PROTO-GENERIC-FUNCTION PROTO-METHOD SPECIALIZER)
-     (SPECIALIZER-TYPE-SPECIFIER PROTO-GENERIC-FUNCTION PROTO-METHOD
-      (SPECIALIZER-CLASS SPECIALIZER)))
-   (DEFUN REAL-SPECIALIZER-TYPE-SPECIFIER/EQL-SPECIALIZER
-          (PROTO-GENERIC-FUNCTION PROTO-METHOD SPECIALIZER)
-     (DECLARE (IGNORE PROTO-GENERIC-FUNCTION PROTO-METHOD))
-     `(EQL SB-IMPL::COMMA))
-   (DEFUN REAL-SPECIALIZER-TYPE-SPECIFIER/STRUCTURE-CLASS
-          (PROTO-GENERIC-FUNCTION PROTO-METHOD SPECIALIZER)
-     (DECLARE (IGNORE PROTO-GENERIC-FUNCTION PROTO-METHOD))
-     (CLASS-NAME SPECIALIZER))
-   (DEFUN REAL-SPECIALIZER-TYPE-SPECIFIER/SYSTEM-CLASS
-          (PROTO-GENERIC-FUNCTION PROTO-METHOD SPECIALIZER)
-     (DECLARE (IGNORE PROTO-GENERIC-FUNCTION PROTO-METHOD))
-     (CLASS-NAME SPECIALIZER))
-   (DEFUN REAL-SPECIALIZER-TYPE-SPECIFIER/CLASS
-          (PROTO-GENERIC-FUNCTION PROTO-METHOD SPECIALIZER)
-     (LET (#)
-       (WHEN # #)))))
-|#
