@@ -108,19 +108,68 @@
         ;; Do not reload r9 (= NL9 = TMP-TN) as it just got moved from NL0 which holds
         ;; the result from C and which will be returned from this asm routine.
         (inst ldr (reg 8 'any-reg) (@ nsp-tn 64))
-        ;; I'm unsure how to utilize map-pairs on the others now :-(
-        ;; (It took more lines of code to alter MAP-PAIRS to accept another keyword
-        ;; asking it not to clobber TMP-TN than it did to just hand-write 4 instructions)
-        (inst ldp (reg 6 'any-reg) (reg 7 'any-reg) (@ nsp-tn 48))
-        (inst ldp (reg 4 'any-reg) (reg 5 'any-reg) (@ nsp-tn 32))
-        (inst ldp (reg 2 'any-reg) (reg 3 'any-reg) (@ nsp-tn 16))
-        (inst ldp (reg 0 'any-reg) (reg 1 'any-reg) (@ nsp-tn 80 :post-index))
+        (map-pairs ldp nsp-tn 48 (butlast nl-registers 2) :post-index 80 :delta -16)
         (inst ret)))))))
   (define-alloc-tramp alloc-tramp "alloc")
   (define-alloc-tramp list-alloc-tramp "alloc_list")
   (define-alloc-tramp listify-&rest "listify_rest_arg"
     (inst sub nl1 csp-tn tmp-tn)))
 
+#+debug-gc-barriers
+(define-assembly-routine (check-barrier (:return-style :none))
+ ((:temp nl0 unsigned-reg nl0-offset)
+  (:temp nl1 unsigned-reg nl1-offset)
+  (:temp nl2 unsigned-reg nl2-offset)
+  (:temp nl3 unsigned-reg nl3-offset))
+ (flet ((reg (offset sc)
+          (make-random-tn :kind :normal :sc (sc-or-lose sc) :offset offset))
+        (reverse-pairs (list)
+          (let (result)
+            (loop for (a b) on list by #'cddr
+                  do (push b result) (push a result))
+            result)))
+   (let ((nl-registers
+          (loop for i to 9
+                collect (reg i 'unsigned-reg)))
+         (lisp-registers
+          (loop for i from r0-offset to r9-offset
+                collect (reg i 'unsigned-reg)))
+         (float-registers
+          (loop for i below 32
+                collect (reg i 'complex-double-reg))))
+     (macrolet ((map-pairs (op base start offsets &key pre-index post-index (delta 16))
+                  `(let ((offsets
+                          ,(if (eq op 'ldp)
+                               `(reverse-pairs ,offsets)
+                               offsets)))
+                     (loop with offset = ,start
+                           for first = t then nil
+                           for (a b . next) on offsets by #'cddr
+                           for last = (not next)
+                           do (inst ,op a b
+                                    (@ ,base (or (and first ,pre-index) (and last ,post-index) offset)
+                                       (cond ((and last ,post-index) :post-index) ((and first ,pre-index) :pre-index) (t :offset)))) (incf offset ,delta)))))
+        (map-pairs stp nsp-tn 0 nl-registers :pre-index -80)
+       (pseudo-atomic (nl3)
+         (inst ldr nl0 (@ csp-tn (- n-word-bytes) :pre-index))
+         (inst ldr nl1 (@ csp-tn (- n-word-bytes) :pre-index))
+         (inst ldr nl2 (@ csp-tn (- n-word-bytes) :pre-index))
+         (inst stp cfp-tn csp-tn (@ thread-tn (* thread-control-frame-pointer-slot n-word-bytes)))
+         (inst add csp-tn csp-tn (+ 32 80))
+         (inst stp cfp-tn lr-tn (@ csp-tn -112))
+         (map-pairs stp csp-tn -80 lisp-registers)
+         (map-pairs stp nsp-tn 0 float-registers :pre-index -512 :delta 32)
+
+         (invoke-foreign-routine "check_barrier" nl3)
+
+         (map-pairs ldp nsp-tn 480 float-registers :post-index 512 :delta -32)
+         (map-pairs ldp csp-tn -16 lisp-registers :delta -16)
+         (inst ldr lr-tn (@ csp-tn -104))
+         (inst sub csp-tn csp-tn (+ 32 80))
+         (inst str zr-tn (@ thread-tn (* thread-control-stack-pointer-slot n-word-bytes))))
+        (map-pairs ldp nsp-tn 64 nl-registers :post-index 80 :delta -16)
+       (inst ret)))))
+
 (define-assembly-routine
     (undefined-tramp (:return-style :none))
     ()
