@@ -84,12 +84,11 @@
     (inst pop rdx-tn)))
 
 (define-assembly-routine (switch-to-arena (:return-style :raw)) ()
-  (inst mov rsi-tn (ea rsp-tn)) ; explicitly  pass the return PC
   ;; RSI and RDI are vop temps, so don't bother preserving them
   (with-registers-preserved (c :except (rsi rdi))
     (pseudo-atomic ()
       #-system-tlabs (inst break halt-trap)
-      #+system-tlabs (inst call (make-fixup "switch_to_arena" :foreign)))))
+      #+system-tlabs (call-c (make-fixup "switch_to_arena" :foreign) #+win32 rdi-tn #+win32 rsi-tn))))
 
 (macrolet ((def-routine-pair (name&options vars &body code)
              `(progn
@@ -105,56 +104,58 @@
 
 (def-routine-pair (alloc-tramp) ()
   (with-registers-preserved (c)
-    (inst mov rdi-tn (ea 16 rbp-tn))
-    (inst mov rsi-tn system-tlab-p)
-    (inst call (make-fixup "alloc" :foreign))
+    (call-c (make-fixup "alloc" :foreign)
+            (ea 16 rbp-tn)
+            system-tlab-p)
     (inst mov (ea 16 rbp-tn) rax-tn))) ; result onto stack
 
 (def-routine-pair (list-alloc-tramp) () ; CONS, ACONS, LIST, LIST*
   (with-registers-preserved (c)
-    (inst mov rdi-tn (ea 16 rbp-tn))
-    (inst mov rsi-tn system-tlab-p)
-    (inst call (make-fixup "alloc_list" :foreign))
+    (call-c (make-fixup "alloc_list" :foreign)
+            (ea 16 rbp-tn)
+            system-tlab-p)
     (inst mov (ea 16 rbp-tn) rax-tn))) ; result onto stack
 
 (def-routine-pair (listify-&rest (:return-style :none)) ()
   (with-registers-preserved (c)
-    (inst mov rdi-tn (ea 16 rbp-tn)) ; 1st C call arg
-    (inst mov rsi-tn (ea 24 rbp-tn)) ; 2nd C call arg
-    (inst mov rdx-tn system-tlab-p)
-    (inst call (make-fixup "listify_rest_arg" :foreign))
-    (inst mov (ea 24 rbp-tn) rax-tn)) ; result
+    (call-c (make-fixup "listify_rest_arg" :foreign)
+            (ea 16 rbp-tn)
+            (ea 24 rbp-tn)
+            system-tlab-p)
+    (inst mov (ea 24 rbp-tn) rax-tn))   ; result
   (inst ret 8)) ; pop one argument; the unpopped word now holds the result
 
 (def-routine-pair (make-list (:return-style :none)) ()
   (with-registers-preserved (c)
-    (inst mov rdi-tn (ea 16 rbp-tn)) ; 1st C call arg
-    (inst mov rsi-tn (ea 24 rbp-tn)) ; 2nd C call arg
-    (inst mov rdx-tn system-tlab-p)
-    (inst call (make-fixup "make_list" :foreign))
+    (call-c (make-fixup "make_list" :foreign)
+            (ea 16 rbp-tn)
+            (ea 24 rbp-tn)
+            system-tlab-p)
     (inst mov (ea 24 rbp-tn) rax-tn)) ; result
   (inst ret 8)) ; pop one argument; the unpopped word now holds the result
 )
 
 (define-assembly-routine (alloc-funinstance) ()
   (with-registers-preserved (c)
-    (inst mov rdi-tn (ea 16 rbp-tn))
-    (inst call (make-fixup "alloc_funinstance" :foreign))
+    (call-c (make-fixup "alloc_funinstance" :foreign)
+            (ea 16 rbp-tn))
     (inst mov (ea 16 rbp-tn) rax-tn)))
 
 ;;; These routines are for the deterministic consing profiler.
 ;;; The C support routine's argument is the return PC.
 (define-assembly-routine (enable-alloc-counter) ()
   (with-registers-preserved (c)
-    (inst lea rdi-tn (ea 8 rbp-tn))
     #+sb-thread
-    (pseudo-atomic () (inst call (make-fixup "allocation_tracker_counted" :foreign)))))
+    (pseudo-atomic ()
+      (call-c (make-fixup "allocation_tracker_counted" :foreign)
+              (* (ea 8 rbp-tn))))))
 
 (define-assembly-routine (enable-sized-alloc-counter) ()
   (with-registers-preserved (c)
-    (inst lea rdi-tn (ea 8 rbp-tn))
     #+sb-thread
-    (pseudo-atomic () (inst call (make-fixup "allocation_tracker_sized" :foreign)))))
+    (pseudo-atomic ()
+      (call-c (make-fixup "allocation_tracker_sized" :foreign)
+              (* (ea 8 rbp-tn))))))
 
 (define-assembly-routine (undefined-tramp (:return-style :none))
     ((:temp rax descriptor-reg rax-offset))
@@ -227,18 +228,22 @@
   (loadw rax-tn rax-tn fdefn-fun-slot other-pointer-lowtag)
   (inst jmp (object-slot-ea rax-tn closure-fun-slot fun-pointer-lowtag)))
 
-#-compact-instance-header
-(define-assembly-routine
-    (funcallable-instance-tramp (:return-style :none))
-    ()
-  (loadw rax-tn rax-tn funcallable-instance-function-slot fun-pointer-lowtag)
-  (inst jmp (object-slot-ea rax-tn closure-fun-slot fun-pointer-lowtag)))
-
-(define-assembly-routine (ensure-symbol-hash (:return-style :raw)) ()
-  (with-registers-preserved (lisp)
-    (inst mov rdx-tn (ea 16 rbp-tn)) ; arg
-    (call-static-fun 'ensure-symbol-hash 1)
-    (inst mov (ea 16 rbp-tn) rdx-tn))) ; result to arg passing loc
+#+debug-gc-barriers
+(define-assembly-routine (check-barrier (:return-style :none)) ()
+  (inst push rax-tn)
+  (inst mov rax-tn (ea 16 rsp-tn))
+  (inst not rax-tn)
+  (inst test :byte rax-tn 3)
+  (inst pop rax-tn)
+  (inst jmp :e check)
+  (inst ret 24)
+  check
+  (with-registers-preserved (c :except fp) ;; shouldn't have any fp operations
+    (call-c (make-fixup "check_barrier" :foreign)
+            (ea 16 rbp-tn)
+            (ea 24 rbp-tn)
+            (ea 32 rbp-tn)))
+  (inst ret 24))
 
 ;;; Perform a store to code, updating the GC card mark bit.
 ;;; This has two additional complications beyond the ordinary
@@ -282,7 +287,7 @@
       (inst mov rax object)
       (inst shr rax gencgc-card-shift)
       (inst and :dword rax card-index-mask)
-      (inst mov :byte (ea gc-card-table-reg-tn rax) 0)
+      (inst mov :byte (ea gc-card-table-reg-tn rax) CARD-MARKED)
       STORE
       (inst mov rdi object)
       (inst mov rdx word-index)
@@ -292,3 +297,21 @@
       ;; store newval into object
       (inst mov (ea (- other-pointer-lowtag) rdi rdx n-word-bytes) rax)))
   (inst ret 24)) ; remove 3 stack args
+
+#+ultrafutex
+(progn
+(define-assembly-routine (mutex-wake-waiter (:return-style :raw)) ()
+  (with-registers-preserved (lisp)
+    (inst call (make-fixup "lispmutex_wake_waiter" :foreign)))) ; no args!
+
+(define-assembly-routine (mutex-unlock (:return-style :raw)) ()
+  ;; There are no registers reserved for this asm routine
+  (inst push rax-tn)
+  (inst mov rax-tn (thread-tls-ea (load-time-tls-offset '*current-mutex*)))
+  (inst mov :qword (mutex-slot rax-tn %owner) 0)
+  (inst dec :lock :byte (mutex-slot rax-tn state))
+  (inst jmp :z uncontested) ; if ZF then previous value was 1, no waiters
+  (inst call (make-fixup 'mutex-wake-waiter :assembly-routine))
+  uncontested
+  (inst pop rax-tn))
+) ; end PROGN
