@@ -478,7 +478,7 @@
   (cond ((not (listp dims)) (values 1 (the index dims)))
         ((not dims) (values 0 1))
         (t (let ((rank 1) (product (car dims)))
-             (declare (array-rank rank) (index product))
+             (declare (%array-rank rank) (index product))
              (dolist (dim (cdr dims) (values rank product))
                (setq product (* product (the index dim)))
                (incf rank))))))
@@ -848,7 +848,7 @@ of specialized arrays is supported."
                                        ;; a cheaper one.
                                        (declare (optimize (speed 1)
                                                           (safety 1)))
-                                     (the* (,type :context 'sb-c::aref-context) new-value)))
+                                     (the* (,type :context sb-c::aref-context) new-value)))
                   ;; Low-level setters return no value
                   new-value)))
            (define-reffers (symbol deffer check-form slow-path)
@@ -924,7 +924,7 @@ of specialized arrays is supported."
 
 ;;; SUBSCRIPTS has a dynamic-extent list structure and is destroyed
 (defun %array-row-major-index (array &rest subscripts)
-  (declare (truly-dynamic-extent subscripts)
+  (declare (dynamic-extent subscripts)
            (array array))
   (let ((length (length subscripts)))
     (cond ((array-header-p array)
@@ -958,7 +958,7 @@ of specialized arrays is supported."
 
 (defun array-in-bounds-p (array &rest subscripts)
   "Return T if the SUBSCRIPTS are in bounds for the ARRAY, NIL otherwise."
-  (declare (truly-dynamic-extent subscripts))
+  (declare (dynamic-extent subscripts))
   (let ((length (length subscripts)))
     (cond ((array-header-p array)
            (let ((rank (%array-rank array)))
@@ -979,12 +979,12 @@ of specialized arrays is supported."
                      (length (truly-the (simple-array * (*)) array)))))))))
 
 (defun array-row-major-index (array &rest subscripts)
-  (declare (truly-dynamic-extent subscripts))
+  (declare (dynamic-extent subscripts))
   (apply #'%array-row-major-index array subscripts))
 
 (defun aref (array &rest subscripts)
   "Return the element of the ARRAY specified by the SUBSCRIPTS."
-  (declare (truly-dynamic-extent subscripts))
+  (declare (dynamic-extent subscripts))
   (row-major-aref array (apply #'%array-row-major-index array subscripts)))
 
 ;;; (setf aref/bit/sbit) are implemented using setf-functions,
@@ -993,7 +993,7 @@ of specialized arrays is supported."
 ;;; haven't found technical advantages or disadvantages for either
 ;;; scheme.
 (defun (setf aref) (new-value array &rest subscripts)
-  (declare (truly-dynamic-extent subscripts)
+  (declare (dynamic-extent subscripts)
            (type array array))
   (setf (row-major-aref array (apply #'%array-row-major-index array subscripts))
         new-value))
@@ -1020,14 +1020,14 @@ of specialized arrays is supported."
 (defun bit (bit-array &rest subscripts)
   "Return the bit from the BIT-ARRAY at the specified SUBSCRIPTS."
   (declare (type (array bit) bit-array)
-           (truly-dynamic-extent subscripts)
+           (dynamic-extent subscripts)
            (optimize (safety 1)))
   (row-major-aref bit-array (apply #'%array-row-major-index bit-array subscripts)))
 
 (defun (setf bit) (new-value bit-array &rest subscripts)
   (declare (type (array bit) bit-array)
            (type bit new-value)
-           (truly-dynamic-extent subscripts)
+           (dynamic-extent subscripts)
            (optimize (safety 1)))
   (setf (row-major-aref bit-array
                         (apply #'%array-row-major-index bit-array subscripts))
@@ -1036,7 +1036,7 @@ of specialized arrays is supported."
 (defun sbit (simple-bit-array &rest subscripts)
   "Return the bit from SIMPLE-BIT-ARRAY at the specified SUBSCRIPTS."
   (declare (type (simple-array bit) simple-bit-array)
-           (truly-dynamic-extent subscripts)
+           (dynamic-extent subscripts)
            (optimize (safety 1)))
   (row-major-aref simple-bit-array
                   (apply #'%array-row-major-index simple-bit-array subscripts)))
@@ -1044,7 +1044,7 @@ of specialized arrays is supported."
 (defun (setf sbit) (new-value bit-array &rest subscripts)
   (declare (type (simple-array bit) bit-array)
            (type bit new-value)
-           (truly-dynamic-extent subscripts)
+           (dynamic-extent subscripts)
            (optimize (safety 1)))
   (setf (row-major-aref bit-array
                         (apply #'%array-row-major-index bit-array subscripts))
@@ -1152,7 +1152,7 @@ of specialized arrays is supported."
          (fill-pointer-error vector))
         (t
          (let ((max (%array-available-elements vector)))
-           (when (> (the fixnum new) max)
+           (when (> (the (and unsigned-byte fixnum) new) max)
              (error 'simple-type-error
                     :datum new
                     :expected-type (list 'integer 0 max)
@@ -1826,7 +1826,7 @@ function to be removed without further warning."
 ;;; dispatching when called. It expects arguments that match PARAMS.
 ;;;
 (defmacro sb-impl::!define-array-dispatch (style dispatch-name params nil-array &body body)
-  #-(or x86 x86-64) (setq style :call)
+  #-(or x86 x86-64 arm64) (setq style :call)
   (let ((table-name (symbolicate "%%" dispatch-name "-FUNS%%"))
         (error-name (symbolicate "HAIRY-" dispatch-name "-ERROR")))
     (declare (ignorable table-name))
@@ -1869,7 +1869,9 @@ function to be removed without further warning."
        (defmacro ,dispatch-name (&rest args)
          (aver (symbolp (first args)))
          (let ((tag (gensym "TAG")))
-           `(funcall
+           `(,',(if (find '&rest params)
+                    'apply
+                    'funcall)
              (truly-the function
                (let ((,tag 0))
                  (when (%other-pointer-p ,(first args))
@@ -1897,12 +1899,15 @@ function to be removed without further warning."
                        collect i)
               (,error-name ,@params))
              ,@(loop
-                   for info across *specialized-array-element-type-properties*
-                   collect `(,(ash (saetp-typecode info) -2)
-                             (let ((,(first params)
-                                    (truly-the (simple-array ,(saetp-specifier info) (*))
-                                               ,(first params))))
-                               ,@body))))))))))))
+                 for info across *specialized-array-element-type-properties*
+                 for specifier = (saetp-specifier info)
+                 collect `(,(ash (saetp-typecode info) -2)
+                           (let ((,(first params)
+                                   (truly-the (simple-array ,specifier (*))
+                                              ,(first params))))
+                             ,@(if (null specifier)
+                                   nil-array
+                                   body)))))))))))))
 
 (defun sb-kernel::check-array-shape (array dimensions)
   (when (let ((dimensions dimensions))
@@ -1918,46 +1923,3 @@ function to be removed without further warning."
 ;;; Horrible kludge for the "static-vectors" system
 ;;; which uses an internal symbol in SB-IMPL.
 (import '%vector-widetag-and-n-bits-shift 'sb-impl)
-
-;; "new" weak vectors satisfy weak-pointer-p, not simple-vector-p
-#+weak-vector-readbarrier
-(progn
-  (defun weak-vector-p (x)
-    (and (weak-pointer-p x)
-         (= (logand (get-header-data x) #xFF) 0)))
-  (defun weak-vector-len (thing)
-    ;; FIXME: assert that it's a vector-like weak pointer, otherwise it'll see
-    ;; the weak-pointer-value slot.
-    (%array-fill-pointer thing))
-  (defun weak-vector-ref (vector index) ; TODO: needs dimension check and read barrier
-    (sb-vm::%weakvec-ref vector index))
-  (defun (setf weak-vector-ref) (newval vector index)
-    (sb-vm::%weakvec-set vector index newval)
-    newval))
-#-weak-vector-readbarrier
-;; legacy implementation of weak vector is basically SIMPLE-VECTOR
-(defun weak-vector-p (x)
-  (and (simple-vector-p x)
-       (test-header-data-bit x (ash vector-weak-flag array-flags-data-position))))
-
-(defun make-weak-vector (length &key (initial-contents nil contents-p)
-                                     (initial-element nil element-p))
-  (declare (index length))
-  (when (and element-p contents-p)
-    (error "Can't specify both :INITIAL-ELEMENT and :INITIAL-CONTENTS"))
-  ;; Since weak vectors are not in theory merely arrays any more, but potentially
-  ;; some kind of weak pointers with a varing enumber of slots, this isn't badly
-  ;; OAOO-violating in regard to make-array transforms.
-  (when contents-p
-    (let ((contents-length (length initial-contents)))
-      (unless (eql length contents-length)
-        (error "~S has ~D elements, vector length is ~D."
-               :initial-contents contents-length length))))
-  (let ((v (sb-c::allocate-weak-vector length)))
-    (if initial-contents
-        (dotimes (i length)
-          (setf (weak-vector-ref v i) (elt initial-contents i)))
-          ;; 0 is the usual default initial element for arrays, but all weak objects use NIL
-        (dotimes (i length)
-          (setf (weak-vector-ref v i) initial-element)))
-    v))
