@@ -27,14 +27,23 @@
 (defun double-float-high-bits (x)
   (ash (double-float-bits x) -32))
 
-(flet ((output-part (x stream)
+;;; This choice exists because cold-init reads from "output/sxhash-calls.lisp-expr"
+;;; using the ordinary definition of "#." which has to call the function at the car
+;;; of a list; but warm.lisp uses a purpose-made #. reader that handles only 2
+;;; symbols, reducing the size of xfloat-math.lisp-expr by abbreviating the float
+;;; constructors to single-character symbols.
+(defvar *proxy-sfloat-ctor* "MAKE-SINGLE-FLOAT")
+(defvar *proxy-dfloat-ctor* "MAKE-DOUBLE-FLOAT")
+(labels
+      ((stringify (bits) (if (= bits 0) 0 (format nil "#x~x" bits)))
+       (output-part (x stream)
          (typecase x
            (single-float
-            (format stream "(MAKE-SINGLE-FLOAT #x~x)" (flonum-%bits x)))
+            (format stream "(~A ~A)" *proxy-sfloat-ctor* (stringify (flonum-%bits x))))
            (double-float
-            (format stream "(MAKE-DOUBLE-FLOAT #x~x #x~x)"
-                    (double-float-high-bits x)
-                    (double-float-low-bits x)))
+            (format stream "(~A ~A ~A)" *proxy-dfloat-ctor*
+                    (stringify (double-float-high-bits x))
+                    (stringify (double-float-low-bits x))))
            (rational
             (prin1 x stream)))))
   (defmethod print-object ((self float) stream)
@@ -185,7 +194,7 @@
            (exp-marker (if (and marker-pos
                                 (char-not-equal (char string marker-pos) #\E))
                            (char-upcase (char string marker-pos))
-                           (ecase *read-default-float-format*
+                           (ecase cl:*read-default-float-format*
                             ((cl:single-float cl:short-float) #\F)
                             ((cl:double-float cl:long-float)  #\D))))
            (significand (if marker-pos (subseq string 0 marker-pos) string))
@@ -225,7 +234,7 @@
   (make-flonum :+infinity format)
   (make-flonum :-infinity format))
 
-(defun float-ops-cache-insert (key values table)
+(defun float-ops-cache-insert (table key values)
   ;; Verify results (when possible) prior to inserting into the hash-table.
   ;; If we were to support different floating-point formats across the various
   ;; backends, this check should confined to the scenarios where the host's
@@ -239,48 +248,86 @@
                (double-float (host-sb-kernel:make-double-float
                               (ash bits -32) (ldb (byte 32 0) bits))))))
       (let ((authoritative-answer
-             (apply (intern (string fun) "CL")
-                    (mapcar (lambda (x)
-                              (etypecase x
-                                (float (native-flonum-value x))
-                                (rational x)
-                                (symbol (intern (string x) "CL"))))
-                            (ensure-list args)))))
-        (unless (eql authoritative-answer
-                     (if (floatp (first values))
-                         (native-flonum-value (first values))
-                         (first values)))
-          (#+sb-devel cerror #+sb-devel "Ignore"
-           #-sb-devel format #-sb-devel t
+              (multiple-value-list
+               (host-sb-kernel::with-float-traps-masked (:overflow :divide-by-zero)
+                 (apply (intern (string fun) "CL")
+                        (mapcar (lambda (x)
+                                  (etypecase x
+                                    (float (native-flonum-value x))
+                                    (rational x)
+                                    (symbol (intern (string x) "CL"))))
+                                (ensure-list args)))))))
+        (unless (equal authoritative-answer
+                       (mapcar (lambda (value)
+                                 (if (floatp value)
+                                     (native-flonum-value value)
+                                     value))
+                               values))
+          (#+sb-devel-xfloat cerror #+sb-devel-xfloat "Ignore"
+           #-sb-devel-xfloat format #-sb-devel-xfloat t
            "~&//CROSS-FLOAT DISCREPANCY!
 // CACHE: ~S -> ~S~%// HOST : ~@[#x~X = ~]~S~%"
                   key values
                   (when (cl:floatp authoritative-answer)
                     (get-float-bits authoritative-answer))
                   authoritative-answer)))))
-  (setf (gethash key table) (if (singleton-p values) (car values) values)))
+  (setf (gethash key table) (if (singleton-p values) (car values) (cons '&values values))))
+
+(defun parse-xfloat-math-file (stream table mode)
+  (declare (type (member :all :merge) mode))
+  ;; Ensure that we're reading the correct variant of the file
+  ;; in case there is more than one set of floating-point formats.
+  (assert (eq (read stream) :default))
+  (let ((pkg (make-package "SB-FLOAT-MATH-GENIE" :use '("CL")))
+        (line 0))
+    (loop for (alias . actual) in '(("S" . sb-kernel:make-single-float)
+                                    ("D" . sb-kernel:make-double-float))
+          do (setf (fdefinition (intern alias pkg)) (fdefinition actual)))
+    (unwind-protect
+         (dolist (expr (let ((*package* pkg)) (read stream)))
+           (incf line)
+           (destructuring-bind (fun args . values) expr
+             ;; It seem so unnecessarily convoluted to remove &VALUES and re-insert
+             ;; in the table and remove for comparison below.
+             ;; Why not just NOT do any of that? After all, a list is a perfectly
+             ;; good representation of one or more values.
+             (let* ((key (cons fun args))
+                    (existsp (gethash key table))
+                    (values (if (and (symbolp (first values))
+                                     (string= (symbol-name (first values))
+                                              "&VALUES"))
+                                (rest values)
+                                values)))
+               (when existsp
+                 (case mode
+                   (:merge
+                    (when (and existsp
+                               (not (equal
+                                     (if (and (listp existsp) (eq (car existsp) '&values))
+                                         (cdr existsp)
+                                         existsp)
+                                     (if (singleton-p values) (car values) values))))
+                      (error "Got different answers for ~S: ~S and ~S"
+                             key existsp values)))
+                   (:all ; should not happen
+                    (error "Line ~D of float cache: ~S is repeated" line key))))
+               (float-ops-cache-insert table key values))))
+      (delete-package pkg))))
 
 (defun get-float-ops-cache (&aux (cache sb-cold::*math-ops-memoization*))
   (when (atom cache)
     (return-from get-float-ops-cache cache))
-  (let ((table (car cache)))
+  (let ((table (car cache))
+        (pathname))
     (when (zerop (hash-table-count table))
-      (with-open-file (stream "float-math.lisp-expr" :if-does-not-exist nil)
+      (with-open-file (stream (setq pathname (sb-cold::math-journal-pathname :input))
+                              :if-does-not-exist nil)
         (when stream
-          ;; Ensure that we're reading the correct variant of the file
-          ;; in case there is more than one set of floating-point formats.
-          (assert (eq (read stream) :default))
-          (let ((*package* (find-package "SB-KERNEL")))
-            (dolist (expr (read stream))
-              (destructuring-bind (fun args . values) expr
-                ;; some symbols, such as SQRT, read as XC-STRICT-CL:SQRT
-                ;; from the SB-KERNEL package, but the cache key should
-                ;; always use the symbol in the CL package.
-                (float-ops-cache-insert (cons (intern (string fun) "CL") args)
-                                        values table))))
+          (parse-xfloat-math-file stream table :all)
+          (setf (cdr cache) (hash-table-count table))
           (when cl:*compile-verbose*
-            (format t "~&; Float-ops cache prefill: ~D entries~%"
-                    (setf (cdr cache) (hash-table-count table)))))))
+            (format t "~&; Math journal: prefilled ~D entries from ~S~%"
+                    (cdr cache) pathname)))))
     table))
 
 (defun record-math-op (key &rest values)
@@ -289,17 +336,44 @@
          (fun (car key))
          (args (cdr key))
          ;; args list is potentially on stack, so copy it
-         (key (cons fun (if (listp args) (copy-list args) args))))
-    (float-ops-cache-insert key values table))
+         (key (cons fun (if (listp args) (copy-list args) args)))
+         (old-count (hash-table-count table)))
+    (float-ops-cache-insert table key values)
+    (unless (= (hash-table-count table) (1+ old-count))
+      (bug "Non-canonical math journal entry ~S" key)))
   (apply #'values values))
 
-(defmacro with-memoized-math-op ((name key-expr &optional (n-values 1)) calculation)
-  (unless (= n-values 1)
-    (error "multiple values not working with memoization yet"))
-  `(dx-let ((cache-key (cons ',(intern (string name) "CL") ,key-expr)))
+;;; Disallow non-canonical symbols in the float math cache,
+;;; or it gets very confusing as to whether the cache is dirty.
+(defun canonical-math-op-args (expr)
+  ;; Try to avoid consing a new list unless we have to.
+  (labels ((check (expr)
+             (cond ((consp expr) (and (check (car expr)) (check (cdr expr))))
+                   ((symbolp expr) (eq (cl:symbol-package expr) *cl-package*))
+                   (t)))
+           (recons (expr)
+             (cond ((consp expr) (cons (recons (car expr)) (recons (cdr expr))))
+                   ((symbolp expr) (intern (string expr) *cl-package*))
+                   (t expr))))
+    (if (check expr) expr (recons expr))))
+
+(defmacro with-memoized-math-op ((name key-expr) calculation)
+  (assert (symbolp name))
+  ;; In theory I could make this so that only a cache miss has to call SANIFY-MATH-OP-ARGS
+  ;; so that in the frequently-occuring cases we do not have to make an extra pass over
+  ;; the expression to determine whether was is canonical. But since only COERCE can have
+  ;; a problem of non-canononical symbols, it's easiest to just always canonicalize
+  ;; for COERCE, and nothing else.
+  `(dx-let ((cache-key
+             ,(if (string= name 'coerce)
+                  `(cons ',(intern (string name) "CL") (canonical-math-op-args ,key-expr))
+                  `(cons ',(intern (string name) "CL") ,key-expr))))
      (multiple-value-bind (answer foundp) (gethash cache-key (get-float-ops-cache))
        (if foundp
-           answer
+           (if (and (listp answer)
+                    (eq (first answer) '&values))
+               (values-list (rest answer))
+               answer)
            (multiple-value-call #'record-math-op cache-key ,calculation)))))
 
 ;;; REAL and IMAG are either host integers (therefore EQL-comparable)
@@ -320,6 +394,9 @@
   (logand (ash (flonum-%bits float)
                (- (1- (float-format-bits (flonum-format float)))))
           1))
+(defun float-sign-bit-set-p (float)
+  (declare (type float float))
+  (= (float-sign-bit float) 1))
 
 (defun calculate-flonum-value (x &optional (nan-errorp t) &aux (bits (flonum-%bits x)))
   ;; Convert the bits of a target float to an object with which we can perform
@@ -340,7 +417,7 @@
                                      mantissa)))
                ;; Subtract the exponent bias and account for discrepancy in binary
                ;; point placement in the IEEE representation and the CL function
-               (decf exp (+ (floor max-exp 2) mantissa-nbits))
+               (decf exp (+ (cl:floor max-exp 2) mantissa-nbits))
                (let ((value (cl:scale-float (cl:coerce mantissa host-type) exp)))
                  (if (logbitp (1- format-nbits) bits) (cl:- value) value))))))
     (ecase (flonum-format x)
@@ -371,7 +448,7 @@
                  ((null value) ; must not be one of the exceptional symbols
                   (setf (flonum-%value x) (calculate-flonum-value x)))
                  (t ; infinity or minus-zero
-                  (error "~S has no portable value" x)))))
+                  (error "~S (~D) has no portable value" value x)))))
         (t (error "Got host float"))))
 
 (defun realnumify* (args) (mapcar #'realnumify args))
@@ -407,6 +484,23 @@
           (when (cl:> arg-contagion result-contagion)
             (setq result-fmt arg-fmt result-contagion arg-contagion)))))))
 
+(defun pick-float-result-format (&rest args)
+  (flet ((target-num-fmt (num)
+           (cond ((rationalp num) 'single-float)
+                 ((floatp num) (flonum-format num))
+                 (t (error "What? ~S" num)))))
+    (let* ((result-fmt 'single-float)
+           (result-contagion 0))
+      (dolist (arg args result-fmt)
+        (let* ((arg-fmt (target-num-fmt arg))
+               ;; This is inadequate for complex numbers,
+               ;; but we don't need them.
+               (arg-contagion
+                (position arg-fmt
+                          '(short-float single-float double-float long-float))))
+          (when (cl:> arg-contagion result-contagion)
+            (setq result-fmt arg-fmt result-contagion arg-contagion)))))))
+
 (defmacro validate-args (&rest args)
   `(when (or ,@(mapcar (lambda (arg) `(typep ,arg '(or cl:float cl:complex))) args))
      (error "Unexpectedly got host float/complex args")))
@@ -424,6 +518,15 @@
               ;; to runtime use. So don't do it.
               (error "Won't do (RATIONAL ~S) due to possible precision loss" x))))))
 
+(defun rationalize (x)
+  (if (rationalp x)
+      x
+      (with-memoized-math-op (rationalize x)
+        (multiple-value-bind (whole frac) (cl:ftruncate (realnumify x))
+          (if (cl:zerop frac)
+              (cl:rationalize whole)
+              (error "Won't do (RATIONALIZE ~S) due to possible precision loss" x))))))
+
 (defun coerce (object type)
   ;; OBJECT is validated prior to testing the quick bail out case, because supposing
   ;; that we accidentally got a host complex number or float, and we accidentally got
@@ -441,7 +544,7 @@
                 (vector (destructuring-bind (et) (cdr type) et)))))
       (return-from coerce
         (sb-xc:make-array (length object) :element-type et
-                          :initial-contents object))))
+                                          :initial-contents object))))
   (unless (numberp object)
     (return-from coerce (cl:coerce object type)))
   (when (member type '(integer rational real))
@@ -456,15 +559,19 @@
     (return-from coerce object))
   (with-memoized-math-op (coerce (list object type))
     (if (realp object)
-        (let ((actual-type (if (member type '(double-float long-float))
-                               'double-float
-                               'single-float))
-              (source-value (realnumify object)))
-          (make-flonum (cl:coerce source-value
-                                  (ecase actual-type
-                                    (single-float 'cl:single-float)
-                                    (double-float 'cl:double-float)))
-                       actual-type))
+        (if (and (floatp object)
+                 (eq (flonum-%value object) :minus-zero)
+                 (member type '(double-float single-float)))
+            (make-flonum :minus-zero type)
+            (let ((actual-type (if (member type '(double-float long-float))
+                                   'double-float
+                                   'single-float))
+                  (source-value (realnumify object)))
+              (make-flonum (cl:coerce source-value
+                                      (ecase actual-type
+                                        (single-float 'cl:single-float)
+                                        (double-float 'cl:double-float)))
+                           actual-type)))
         (error "Can't COERCE ~S ~S" object type))))
 
 (macrolet ((define (name)
@@ -479,11 +586,17 @@
 
 (defun xfloat-abs (x)
   (with-memoized-math-op (abs x)
-    (if (= (float-sign-bit x) 0) x (sb-xc:- x))))
+    (if (not (float-sign-bit-set-p x)) x (sb-xc:- x))))
 
 ;;; Signum should return -0 of the correct type for -0 input.
 ;;; We don't need it currently.
-(defun xfloat-signum (x) (error "signum ~S: incomplete" x))
+(defun xfloat-signum (x)
+  (if (zerop x)
+      x
+      (coerce (if (float-sign-bit-set-p x)
+                  -1
+                  1)
+              (flonum-format x))))
 
 ;;; This is simple enough that it's not necessary to memoize all calls.
 (defun xfloat-zerop (x)
@@ -519,9 +632,44 @@
 (defun float-sign (float1 &optional (float2 (float 1 float1) float2p))
   (validate-args float1 float2)
   (with-memoized-math-op (float-sign (cons float1 (if float2p (list float2))))
-    (if (= (float-sign-bit float2) (float-sign-bit float1))
-        float2
-        (sb-xc:- float2))))
+    (let ((res (if (= (float-sign-bit float2) (float-sign-bit float1))
+                   float2
+                   (sb-xc:- float2))))
+      (if (eq (flonum-format float1) 'double-float)
+          (coerce res 'double-float)
+          res))))
+
+(macrolet ((define (name float-fun)
+             `(progn
+                (declaim (inline ,name))
+                (defun ,name (number &optional (divisor 1))
+                  (if (and (rationalp number) (rationalp divisor))
+                      (,(intern (string name) "CL") number divisor)
+                      (,float-fun number divisor)))
+                (defun ,float-fun (number divisor)
+                  (let ((type (if (or (eq (flonum-format number) 'double-float)
+                                      (and (floatp divisor)
+                                           (eq (flonum-format divisor) 'double-float)))
+                                  'double-float
+                                  'single-float)))
+                    (if (and (floatp number)
+                             (eq (flonum-%value number) :minus-zero))
+                        (values 0
+                                (make-flonum (if  (< divisor 0)
+                                                  0
+                                                  :minus-zero)
+                                             type))
+                        (with-memoized-math-op (,name (list number divisor))
+                          (multiple-value-bind (q r)
+                              (,(intern (string name) "CL")
+                               (realnumify number)
+                               (realnumify divisor))
+                            (values q
+                                    (make-flonum r type))))))))))
+  (define floor xfloat-floor)
+  (define ceiling xfloat-ceiling)
+  (define truncate xfloat-truncate)
+  (define round xfloat-round))
 
 (defun exp (x)
   (validate-args x)
@@ -533,13 +681,69 @@
   (cond ((and (rationalp base) (integerp power))
          (cl:expt base power))
         (t
-         (with-memoized-math-op (expt (list base power))
-           (cond ((and (eql base $2f0) (eql power 63))
-                  #.(make-flonum #x5F000000 'single-float))
-                 ((and (eql base $2d0) (eql power 63))
-                  #.(make-flonum #x43E0000000000000 'double-float))
-                 (t
-                  (error "Unhandled case in EXPT")))))))
+         (if (zerop power)
+             (coerce 1 (flonum-format base))
+             (with-memoized-math-op (expt (list base power))
+               (cond ((and (eql base $2f0) (eql power 63))
+                      #.(make-flonum #x5F000000 'single-float))
+                     ((and (eql base $2d0) (eql power 63))
+                      #.(make-flonum #x43E0000000000000 'double-float))
+                     ((and (eql base $10d0) (>= power 322))
+                      #.(make-flonum :+infinity 'double-float))
+                     ((and (eql base $10d0) (eql power -309))
+                      #.(make-flonum #xB8157268FDAF 'double-float))
+                     (t
+                      (make-flonum (cl:expt (realnumify base) power)
+                                   (pick-result-format base power)))))))))
+
+(defun %unary-truncate (number)
+  (typecase number
+    (integer number)
+    (ratio (values (truncate (numerator number) (denominator number))))
+    ((or single-float double-float #+long-float long-float)
+     (error "Unimplemented."))))
+
+(defun %unary-ftruncate (number)
+  (typecase number
+    (integer number)
+    (ratio (values (ftruncate (numerator number) (denominator number))))
+    ((or single-float double-float #+long-float long-float)
+     (error "Unimplemented."))))
+
+(defun %unary-round (number)
+  (typecase number
+    (integer number)
+    (ratio (values (round (numerator number) (denominator number))))
+    ((or single-float double-float #+long-float long-float)
+     (error "Unimplemented."))))
+
+(defun %unary-fround (number)
+  (typecase number
+    (integer number)
+    (ratio (values (fround (numerator number) (denominator number))))
+    ((or single-float double-float #+long-float long-float)
+     (error "Unimplemented."))))
+
+(defun cis (number)
+  (declare (ignore number))
+  (error "Unimplemented."))
+
+(defun scale-float (f ex)
+  (validate-args f)
+  (with-memoized-math-op (scale-float (list f ex))
+    (make-flonum (cl:scale-float (let ((val (realnumify f)))
+                                   (assert (cl:floatp val))
+                                   val)
+                                 ex)
+                 (flonum-format f))))
+
+(defun scale-single-float (f ex)
+  (validate-args f)
+  (scale-float f ex))
+
+(defun scale-double-float (f ex)
+  (validate-args f)
+  (scale-float f ex))
 
 (defun log (number &optional (base nil base-p))
   (validate-args number base)
@@ -551,18 +755,21 @@
     ;; * (log 2d0 2d0) => 1.0d0
     ;; * (log 2s0 2s0) => 1.0
     (let ((format (pick-result-format number (if base-p base 0))))
-      (make-flonum (if base-p
-                       (cl:log (realnumify number) (realnumify base))
-                       (cl:log (realnumify number)))
-                   (if (eq format 'rational) ; neither arg was floating-point
-                       'single-float
-                       format)))))
+      (if (zerop number)
+          (make-flonum :-infinity (if (eq format 'rational)
+                                      'single-float
+                                      format))
+          (make-flonum (if base-p
+                           (cl:log (realnumify number) (realnumify base))
+                           (cl:log (realnumify number)))
+                       (if (eq format 'rational) ; neither arg was floating-point
+                           'single-float
+                           format))))))
 
 ;;;;
 
-;;; There seems to be no portable way to mask float traps, but we
-;;; shouldn't encounter any float traps when cross-compiling SBCL
-;;; itself, anyway, so we just make this a no-op.
+;;; There seems to be no portable way to mask float traps, so right
+;;; now we ignore them and hardcode special cases.
 (defmacro sb-vm::with-float-traps-masked (traps &body body)
   (declare (ignore traps))
   #+nil
@@ -610,28 +817,35 @@
 ;;; These use "#." so that they are dumped as literals rather than having to
 ;;; call read-from-string at load-time (and failing) due to the reader intercept.
 ;;      #define __FLT_MAX__ 3.40282347e+38F
-(defconstant sb-xc:most-positive-single-float
+(defconstant most-positive-single-float
   #.(make-flonum (read-from-string "+3.40282347F38") 'single-float))
-(defconstant sb-xc:most-negative-single-float
+(defconstant most-negative-single-float
   #.(make-flonum (read-from-string "-3.40282347F38") 'single-float))
 
 ;;      #define __DBL_MAX__ 1.7976931348623157e+308
-(defconstant sb-xc:most-positive-double-float
+(defconstant most-positive-double-float
   #.(make-flonum (read-from-string "+1.7976931348623157D308") 'double-float))
-(defconstant sb-xc:most-negative-double-float
+(defconstant most-negative-double-float
   #.(make-flonum (read-from-string "-1.7976931348623157D308") 'double-float))
 
 ;;; PI is needed in order to build the cross-compiler mainly so that vm-fndb
 ;;; can define bounds on irrational functions.
-(defconstant sb-xc:pi
+(defconstant pi
   #.(make-flonum (read-from-string "3.14159265358979323846264338327950288419716939937511L0")
                  'double-float))
 
 (eval-when (:compile-toplevel :execute) (setq sb-cold::*choke-on-host-irrationals* t))
 
-;;; These two constants are used in 'late-type'
-(defconstant sb-xc:most-positive-long-float sb-xc:most-positive-double-float)
-(defconstant sb-xc:most-negative-long-float sb-xc:most-negative-double-float)
+;;; These two constants are used in 'type'
+(defconstant most-positive-long-float most-positive-double-float)
+(defconstant most-negative-long-float most-negative-double-float)
+
+(defun substitute-minus-zero (list)
+  (substitute $0.0d0
+              (make-flonum :minus-zero 'double-float)
+              (substitute $0.0f0
+                          (make-flonum :minus-zero 'single-float)
+                          list)))
 
 (macrolet ((intercept (symbols lambda-list body-form)
              `(progn ,@(mapcar (lambda (symbol)
@@ -643,25 +857,80 @@
              `(if (every #'rationalp args)
                   (apply #',f args)
                   (with-memoized-math-op (,f args) ,irrational)))
+           (dispatch-float (f irrational)
+             `(with-memoized-math-op (,f args) ,irrational))
            (flonums-eql-p ()
              `(let ((x (car args)) (y (cadr args)))
                 (and (floatp x) (floatp y) (eql (flonum-%bits x) (flonum-%bits y)))))
            (two-zeros-p ()
-             `(and (eql nargs 2) (zerop (car args)) (zerop (cadr args)))))
+             `(and (eql nargs 2) (zerop (car args)) (zerop (cadr args))))
+           (same-sign-infinities-p ()
+             `(and (eql nargs 2)
+                   (floatp (car args))
+                   (floatp (cadr args))
+                   (member (flonum-%value (car args)) '(:-infinity :+infinity))
+                   (eq (flonum-%value (cadr args)) (flonum-%value (car args)))))
+           (infinity-p (n kind)
+             `(let ((arg (nth ,n args)))
+                (and (floatp arg)
+                     (eq (flonum-%value arg) ,kind)))))
 
-  ;; Simple case of using the interceptor to return a boolean.
-  ;; If a signed zero or infinity leaks in, the host will choke
-  ;; since those are represented as symbols.
+  ;; Simple case of using the interceptor to return a boolean.  If
+  ;; infinity leaks in, the host will choke since those are
+  ;; represented as symbols.
   (intercept (= /=) (&rest args)
-    (dispatch :me (apply #':me (realnumify* args))))
+    (dispatch :me (apply #':me (realnumify* (substitute-minus-zero args)))))
 
   ;; Simple case of using the interceptor to return a float.
   ;; As above, no funky values allowed.
-  (intercept (max min + * / sqrt) (&rest args)
+  (intercept (max min +) (&rest args)
     (dispatch :me
      (make-flonum (apply #':me (realnumify* args))
                   (apply #'pick-result-format args))))
 
+  (intercept (acos acosh asin asinh atan atanh conjugate cos cosh phase sin sinh tan tanh) (&rest args)
+    (dispatch-float :me
+     (make-flonum (apply #':me (realnumify* args))
+                  (apply #'pick-float-result-format args))))
+
+  (intercept (fceiling ffloor fround ftruncate) (&rest args)
+             (dispatch :me
+                       (multiple-value-bind (a b) (apply #':me (realnumify* args))
+                         (let ((format (apply #'pick-result-format args)))
+                          (values (make-flonum a format)
+                                  (make-flonum b format))))))
+
+  (intercept (sqrt) (&rest args)
+    (destructuring-bind (x) args
+      (if (zerop x)
+          x
+          (dispatch :me
+                    (make-flonum (funcall #':me (realnumify x))
+                                 (pick-result-format x))))))
+
+  (intercept (*) (&rest args)
+    (dispatch :me
+              (let ((res (make-flonum (apply #':me (realnumify* (substitute-minus-zero args)))
+                                      (apply #'pick-result-format args))))
+                (if (evenp (+ (count $-0.0d0 args)
+                              (count $-0.0 args)))
+                    res
+                    (sb-xc:- res)))))
+
+  (intercept (/) (&rest args)
+    (if (and (= (length args) 2)
+             ;; ugly hack to avoid tripping up overflow trap
+             (eql (first args) $1.3407807929942596d154)
+             (eql (second args) $8.90029543402881d-308))
+        (make-flonum :+infinity (apply #'pick-result-format args))
+        (dispatch :me
+                  (let ((res (make-flonum (apply #':me (realnumify* (substitute-minus-zero args)))
+                                          (apply #'pick-result-format args))))
+                    (if (evenp (+ (count $-0.0d0 args)
+                                  (count $-0.0 args)))
+                        res
+                        (sb-xc:- res))))))
+
   (intercept (-) (&rest args)
     (dispatch :me
               (if (cdr args)
@@ -676,7 +945,7 @@
     (dispatch :me
       (if (and (eql nargs 2) (or (flonums-eql-p) (two-zeros-p)))
           nil ; if eql, or both = 0, then not '<'
-          (apply #':me (realnumify* args)))))
+          (apply #':me (realnumify* (substitute-minus-zero args))))))
 
   ;; '>' needs to permit some of the "inoperable" floats because making any constant
   ;; (as in 'early-float') calls CTYPE-OF -> CTYPE-OF-NUMBER -> MAKE-NUMERIC-TYPE
@@ -686,7 +955,7 @@
     (dispatch :me
       (if (and (eql nargs 2) (or (flonums-eql-p) (two-zeros-p)))
           nil ; if eql, or both = 0, then not '>'
-          (apply #':me (realnumify* args)))))
+          (apply #':me (realnumify* (substitute-minus-zero args))))))
 
   (intercept (>=) (&rest args  &aux (nargs (length args)))
     (dispatch :me
@@ -699,9 +968,24 @@
                          (eql b $2.0d0))))
              nil)
             ((two-zeros-p) t) ; signed zeros are equal
+            ((same-sign-infinities-p) t) ; infinities are =
+            ((and (eql nargs 2)
+                  (infinity-p 0 :+infinity))
+             t)
+            ((and (eql nargs 2)
+                  (infinity-p 1 :+infinity))
+             nil)
+            ((and (eql nargs 2)
+                  (infinity-p 0 :-infinity))
+             nil)
+            ((and (eql nargs 2)
+                  (infinity-p 1 :-infinity))
+             t)
             ((and (eql nargs 2) (zerop (cadr args)))
              ;; Need this case if the first arg is represented as bits
-             (= (float-sign-bit (car args)) 0)) ; anything positive is >= 0
+             (if (rationalp (car args))
+                 (plusp (car args))
+                 (not (float-sign-bit-set-p (car args))))) ; anything positive is >= 0
             ((eql nargs 2)
              (multiple-value-bind (a b) (collapse-zeros (car args) (cadr args))
                (:me a b)))
@@ -710,28 +994,34 @@
 
   (intercept (<=) (&rest args  &aux (nargs (length args)))
     (dispatch :me
-      (cond ((and (eql nargs 3) (every #'floatp args))
-             (destructuring-bind (a b c) args
-               (if (and (eq (flonum-format a) (flonum-format b))
-                        (eq (flonum-format b) (flonum-format c)))
-                   (cond ((and (eql a sb-xc:most-negative-single-float)
-                               (eql c sb-xc:most-positive-single-float))
-                          (not (float-infinity-p b)))
-                         ((and (eql a sb-xc:most-negative-double-float)
-                               (eql c sb-xc:most-positive-double-float))
-                          (not (float-infinity-p b)))
-                         (t
-                          (error "Unhandled")))
-                   (error "Unhandled"))))
-            ((two-zeros-p) t) ; signed zeros are equal
+      (cond ((two-zeros-p) t) ; signed zeros are equal
+            ((same-sign-infinities-p) t) ; infinities are =
+            ((and (eql nargs 2)
+                  (infinity-p 0 :+infinity))
+             nil)
+            ((and (eql nargs 2)
+                  (infinity-p 1 :+infinity))
+             t)
+            ((and (eql nargs 2)
+                  (infinity-p 0 :-infinity))
+             t)
+            ((and (eql nargs 2)
+                  (infinity-p 1 :-infinity))
+             nil)
             ((and (eql nargs 2) (zerop (cadr args)))
              ;; Need this case if the first arg is represented as bits
-             (= (float-sign-bit (car args)) 1)) ; anything negative is <= 0
+             (if (floatp (car args))
+                 (float-sign-bit-set-p (car args)) ; anything negative is <= 0
+                 (minusp (car args))))
             ((eql nargs 2)
              (multiple-value-bind (a b) (collapse-zeros (car args) (cadr args))
                (:me a b)))
+            ((eql nargs 1)
+             t)
             (t
-             (apply #':me (realnumify* args))))))
+             (loop for (a . rest) on args
+                   always (or (not rest)
+                              (sb-xc:<= a (car rest))))))))
 
 ) ; end MACROLET
 
@@ -742,6 +1032,11 @@
 (defun sb-c::maybe-exact-reciprocal (x)
   (cond ((eql x $2.0d0) $.5d0)
         ((eql x $10.0d0) nil)
+        ((or (eql x #.(make-flonum #x3FE62E42FEFA39EF 'double-float))
+             (eql x #.(make-flonum #x3FF71547652B82FE 'double-float)))
+         nil)
+        ((eql x #.(make-flonum #x49742400 'single-float))
+         nil)
         (t (error "MAYBE-EXACT-RECIPROCAL: didn't expect ~S" x))))
 
 ;;; This is a convenient utility to have - we use it within this file and
@@ -763,6 +1058,12 @@
                    ((single-float-p x) 2)
                    ((double-float-p x) 3)
                    (t (error "Unclassifiable arg ~S" x))))
+           (spelling-of (expr)
+             ;; MUST not write package prefixes !
+             ;; e.g. avoid writing a line like (COERCE (-33619991 SB-XC:DOUBLE-FLOAT) ...)
+             (let ((hex (write-to-string expr :pretty nil :base 16 :radix t :escape nil))
+                   (dec (write-to-string expr :pretty nil :base 10 :escape nil)))
+               (if (<= (length hex) (length dec)) hex dec)))
            (lessp (expr-a expr-b)
              (let ((f (string (car expr-a)))
                    (g (string (car expr-b))))
@@ -795,15 +1096,41 @@
     ;; if ~S is used here. The intent is to use only SBCL as host to compute
     ;; the table, since we assume that everybody's math routines suck.
     ;; But anyway, this does seem to work in most other lisps.
-    (let ((*print-pretty* nil) (*print-base* 16) (*print-radix* t))
+    (let ((*print-pretty* nil) (*proxy-sfloat-ctor* "S")  (*proxy-dfloat-ctor* "D"))
       (dolist (pair (sort (%hash-table-alist table) #'lessp :key #'car))
         (destructuring-bind ((fun . args) . result) pair
           (format stream "(~A ~A~{ ~A~})~%"
-                  fun args
+                  fun
+                  ;; Why do ABS and RATIONAL write the unary arg as an atom
+                  ;; but SQRT writes it as a singleton list?
+                  (if (listp args) (mapcar #'spelling-of args) (spelling-of args))
                   ;; Can't use ENSURE-LIST. We need NIL -> (NIL)
                   (if (consp result) result (list result)))))))
   (format stream ")~%"))
 
+;;; Recipe:
+;;; $ ./build-all-cores.sh
+;;; $ /path/to/host/sbcl
+;;; * (load "load-xc")
+;;; * (sb-impl:merge-all-xfloat-files)
+(export 'merge-all-xfloat-files)
+(defun merge-all-xfloat-files ()
+  (let ((ht sb-cold::*math-ops-memoization*)
+        (total-delta 0))
+    (dolist (pathname (directory "obj/xbuild/**/xfloat-math.lisp-expr"))
+      (let ((old-count (hash-table-count ht)))
+        (with-open-file (stream pathname)
+          (parse-xfloat-math-file stream ht :merge)
+          (let ((delta (- (hash-table-count ht) old-count)))
+            (incf total-delta delta)
+            (when (plusp delta)
+              (format t "~D new entr~@:P from ~S~%" delta pathname))))))
+    (if (zerop total-delta)
+        (format t "~&Nothing new~%")
+        (with-open-file (stream "xfloat-math.lisp-expr"
+                                :direction :output :if-exists :supersede)
+          (dump-math-memoization-table ht stream)))))
+
 (defun show-interned-numbers (stream)
   (flet ((to-native (x)
             (declare (ignorable x))
@@ -849,6 +1176,7 @@
   (assert (eq (make-flonum :+infinity format) (sb-xc:- (make-flonum :-infinity format))))
   (assert (eq (make-flonum :-infinity format) (sb-xc:- (make-flonum :+infinity format))))
   (assert (eq (sb-xc:- (coerce 0 format)) (make-flonum :minus-zero format)))
+  (let ((*break-on-signals* nil))
   (flet ((assert-not-number (x)
            (handler-case (realnumify x)
              (:no-error (x) (error "Expected an error, got ~S" x))
@@ -858,7 +1186,7 @@
       (assert-not-number nan)
       (assert (float-nan-p nan)))
     (dolist (symbol '(:+infinity :-infinity :minus-zero))
-      (assert-not-number (make-flonum symbol format)))))
+      (assert-not-number (make-flonum symbol format))))))
 
 #+host-quirks-sbcl ; Cross-check some more things if we can
 (loop for (our-symbol host-single host-double)
