@@ -36,13 +36,13 @@
       #+64-bit
       (let ((bits (double-float-bits f)))
         (if (dfloat-bits-subnormalp bits)
-            (ldb (byte 52 0) bits)
+            (ldb sb-vm:double-float-significand-byte bits)
             (return-from float-precision sb-vm:double-float-digits)))
       #-64-bit
       (let ((high (double-float-high-bits f)))
         (if (not (dfloat-high-bits-subnormalp high))
             (return-from float-precision sb-vm:double-float-digits)
-            (let ((n (integer-length (ldb sb-vm:double-float-significand-byte high))))
+            (let ((n (integer-length (ldb sb-vm:double-float-hi-significand-byte high))))
               (if (/= 0 n)
                   (return-from float-precision (+ n 32))
                   (double-float-low-bits f)))))))))
@@ -73,11 +73,13 @@
 (defun float-sign-bit (x) ; return 1 or 0, literally the sign bit
   (declare (explicit-check))
   (number-dispatch ((x float))
-    ((single-float)
-     (logand (ash (single-float-bits x) -31) 1))
-    ((double-float)
-     #-64-bit (logand (ash (double-float-high-bits x) -31) 1)
-     #+64-bit (ash (logand (double-float-bits x) most-positive-word) -63))))
+    ((single-float) (float-sign-bit x))
+    ((double-float) (float-sign-bit x))))
+(defun float-sign-bit-set-p (x)
+  (declare (explicit-check))
+  (number-dispatch ((x float))
+    ((single-float) (float-sign-bit-set-p x))
+    ((double-float) (float-sign-bit-set-p x))))
 
 (declaim (inline float-digits float-radix))
 
@@ -134,8 +136,8 @@
   (let* ((hi (double-float-high-bits x))
          (sign (if (minusp hi) -1 1))
          (lo (double-float-low-bits x))
-         (mantissa (logior (ash (ldb sb-vm:double-float-significand-byte hi) 32) lo))
-         (exp (ldb sb-vm:double-float-exponent-byte hi)))
+         (mantissa (logior (ash (ldb sb-vm:double-float-hi-significand-byte hi) 32) lo))
+         (exp (ldb sb-vm:double-float-hi-exponent-byte hi)))
     (cond ((zerop (logior (ldb (byte 31 0) hi) lo))
            (values 0 0 sign))
           ((< exp sb-vm:double-float-normal-exponent-min)
@@ -143,23 +145,20 @@
           ((> exp sb-vm:double-float-normal-exponent-max)
            (error float-decoding-error x))
           (t
-           ;; DOUBLE-FLOAT-HIDDEN-BIT is nonsense. It's 20 because it's the index
-           ;; within the high half. It should be an index within the entire fraction.
-           ;; If you want to manipulate the fraction as two 4-byte parts, that's on you.
-           (values (logior (ash sb-vm:double-float-hidden-bit 32) mantissa)
+           (values (logior sb-vm:double-float-hidden-bit mantissa)
                    (- exp sb-vm:double-float-bias sb-vm:double-float-digits)
                    sign))))
   #+64-bit ; don't split the high and low bits
   (let* ((bits (double-float-bits x))
-         (frac (ldb (byte 52 0) bits))
+         (frac (ldb sb-vm:double-float-significand-byte bits))
          (sign (if (minusp bits) -1 1))
-         (exp (dfloat-exponent-from-bits bits)))
+         (exp (ldb sb-vm:double-float-exponent-byte bits)))
     (cond ((= exp 0)
            (values frac (if (= frac 0) 0 subnormal-dfloat-exponent) sign))
           ((> exp sb-vm:double-float-normal-exponent-max)
            (error float-decoding-error x))
           (t
-           (values (logior (ash sb-vm:double-float-hidden-bit 32) frac)
+           (values (logior sb-vm:double-float-hidden-bit frac)
                    (- exp sb-vm:double-float-bias sb-vm:double-float-digits)
                    sign)))))
 
@@ -184,64 +183,63 @@
 ;;; that is FLOAT-DIGITS wide, and decrease the exponent.
 (defun decode-single-float (x)
   (declare (single-float x))
-  (multiple-value-bind (bits exp)
-      (let* ((bits (single-float-bits x))
-             (biased-exp (ldb sb-vm:single-float-exponent-byte bits)))
-        (if (> biased-exp sb-vm:single-float-normal-exponent-max)
-            (error float-decoding-error x)
-            (let ((frac (ldb sb-vm:single-float-significand-byte bits)))
-               (multiple-value-bind (new-exp new-frac lisp-exponent)
-                   (cond ((/= biased-exp 0) ; normal
-                          ;; SINGLE-FLOAT-BIAS as the stored exponent yields
-                          ;; an effective exponent of -1.
-                          (values sb-vm:single-float-bias frac
-                                  (- biased-exp sb-vm:single-float-bias)))
-                         ((= frac 0) (values 0 0 0))
-                         (t ; subnormal. Normalize it and unset the implied 1 bit
-                          (let ((prec (integer-length frac)))
-                            (values sb-vm:single-float-bias
-                                    (ldb (byte (1- sb-vm:single-float-digits) 0)
-                                         (ash frac (- sb-vm:single-float-digits prec)))
-                                    (+ subnormal-sfloat-exponent prec)))))
-                 (values (dpb new-exp sb-vm:single-float-exponent-byte new-frac)
-                         lisp-exponent)))))
-    (values (make-single-float bits) exp (float-sign x))))
+  (let* ((bits (single-float-bits x))
+         (biased-exp (ldb sb-vm:single-float-exponent-byte bits)))
+    (if (> biased-exp sb-vm:single-float-normal-exponent-max)
+        (error float-decoding-error x)
+        (let ((frac (ldb sb-vm:single-float-significand-byte bits)))
+          (multiple-value-bind (new-exp new-frac lisp-exponent)
+              (cond ((/= biased-exp 0) ; normal
+                     ;; SINGLE-FLOAT-BIAS as the stored exponent yields
+                     ;; an effective exponent of -1.
+                     (values sb-vm:single-float-bias frac
+                             (- biased-exp sb-vm:single-float-bias)))
+                    ((= frac 0) (values 0 0 0))
+                    (t ; subnormal. Normalize it and unset the implied 1 bit
+                     (let ((prec (integer-length frac)))
+                       (values sb-vm:single-float-bias
+                               (ldb (byte (1- sb-vm:single-float-digits) 0)
+                                    (ash frac (- sb-vm:single-float-digits prec)))
+                               (+ subnormal-sfloat-exponent prec)))))
+            (values (make-single-float (dpb new-exp sb-vm:single-float-exponent-byte new-frac))
+                    lisp-exponent
+                    (float-sign x)))))))
 
 ;;; The double-float logic mostly follows the skeleton of the above code,
 ;;; but there is a consed bignum or two on 32-bit architectures.
 ;;; Consing for the sake of code clarity is worth it as far as I'm concerned.
 (defun decode-double-float (x)
   (declare (double-float x))
-  (multiple-value-bind (high-bits low-bits exp)
-      (let* #+64-bit ((bits (double-float-bits x))
-                      (biased-exp (dfloat-exponent-from-bits bits)))
-            #-64-bit ((high (double-float-high-bits x))
-                      (biased-exp (ldb sb-vm:double-float-exponent-byte high)))
-        (if (> biased-exp sb-vm:double-float-normal-exponent-max)
-            (error float-decoding-error x)
-            (let ((frac #+64-bit (ldb (byte 52 0) bits)
-                        #-64-bit (logior (ash (ldb sb-vm:double-float-significand-byte high) 32)
-                                         (double-float-low-bits x))))
-               (multiple-value-bind (new-exp new-frac lisp-exponent)
-                   (cond ((/= biased-exp 0) ; normal
-                          ;; DOUBLE-FLOAT-BIAS as the stored exponent yields
-                          ;; an effective exponent of -1.
-                          (values sb-vm:double-float-bias frac
-                                  (- biased-exp sb-vm:double-float-bias)))
-                         ((= frac 0) (values 0 0 0))
-                         (t ; subnormal. Normalize it and unset the implied 1 bit
-                          (let ((prec (integer-length frac)))
-                            (values sb-vm:double-float-bias
-                                    (ldb (byte (1- sb-vm:double-float-digits) 0)
-                                         (ash frac (- sb-vm:double-float-digits prec)))
-                                    (+ subnormal-dfloat-exponent prec)))))
-                 ;; Now comes the dumb part for 64-bit machines -
-                 ;; splitting the fraction into halves for no good reason.
-                 (values (dpb new-exp sb-vm:double-float-exponent-byte
-                              (ldb (byte 32 32) new-frac))
-                         (ldb (byte 32 0) new-frac)
-                         lisp-exponent)))))
-    (values (make-double-float high-bits low-bits) exp (float-sign x))))
+  (let* #+64-bit ((bits (double-float-bits x))
+                  (biased-exp (ldb sb-vm:double-float-exponent-byte bits)))
+        #-64-bit ((high (double-float-high-bits x))
+                  (biased-exp (ldb sb-vm:double-float-hi-exponent-byte high)))
+    (if (> biased-exp sb-vm:double-float-normal-exponent-max)
+        (error float-decoding-error x)
+        (let ((frac #+64-bit (ldb sb-vm:double-float-significand-byte bits)
+                    #-64-bit (logior (ash (ldb sb-vm:double-float-hi-significand-byte high) 32)
+                                     (double-float-low-bits x))))
+          (multiple-value-bind (new-exp new-frac lisp-exponent)
+              (cond ((/= biased-exp 0)  ; normal
+                     ;; DOUBLE-FLOAT-BIAS as the stored exponent yields
+                     ;; an effective exponent of -1.
+                     (values sb-vm:double-float-bias frac
+                             (- biased-exp sb-vm:double-float-bias)))
+                    ((= frac 0) (values 0 0 0))
+                    (t ; subnormal. Normalize it and unset the implied 1 bit
+                     (let ((prec (integer-length frac)))
+                       (values sb-vm:double-float-bias
+                               (ldb (byte (1- sb-vm:double-float-digits) 0)
+                                    (ash frac (- sb-vm:double-float-digits prec)))
+                               (+ subnormal-dfloat-exponent prec)))))
+            (values #-64-bit
+                    (make-double-float (dpb new-exp sb-vm:double-float-hi-exponent-byte
+                                            (ldb (byte 32 32) new-frac))
+                                       (ldb (byte 32 0) new-frac))
+                    #+64-bit
+                    (%make-double-float (dpb new-exp sb-vm:double-float-exponent-byte new-frac))
+                    lisp-exponent
+                    (float-sign x)))))))
 
 ;;; Dispatch to the appropriate type-specific function.
 (defun decode-float (f)
@@ -266,37 +264,46 @@
 (macrolet ((def (type)
              `(defun ,(symbolicate 'scale- type '-maybe-underflow) (x exp)
                 (declare (inline ,(symbolicate 'integer-decode- type)))
-                (multiple-value-bind (sig old-exp sign) (,(symbolicate 'integer-decode- type) x)
-                  (let* ((digits (float-digits x))
-                         (new-exp (+ exp old-exp digits
-                                     ,(case type
-                                        (single-float 'sb-vm:single-float-bias)
-                                        (double-float 'sb-vm:double-float-bias))))
-                         ;; convert decoded values {-1,+1} into {1,0} respectively
-                         (sign (if (minusp sign) 1 0)))
-                    (cond
-                      ((< new-exp
-                          ,(case type
-                             (single-float 'sb-vm:single-float-normal-exponent-min)
-                             (double-float 'sb-vm:double-float-normal-exponent-min)))
-                       (when (sb-vm:current-float-trap :inexact)
-                         (error 'floating-point-inexact :operation 'scale-float
-                                                        :operands (list x exp)))
-                       (when (sb-vm:current-float-trap :underflow)
-                         (error 'floating-point-underflow :operation 'scale-float
-                                                          :operands (list x exp)))
-                       (let ((shift (1- new-exp)))
-                         (if (< shift (- (1- digits)))
-                             (float-sign x ,(case type
-                                              (single-float $0f0)
-                                              (double-float $0d0)))
-                             ,(case type
-                                (single-float '(single-from-bits sign 0 (ash sig shift)))
-                                (double-float '(double-from-bits sign 0 (ash sig shift)))))))
+                (cond ((float-infinity-p x)
+                       x)
+                      ((float-nan-p x)
+                       (when (and (float-trapping-nan-p x)
+                                  (sb-vm:current-float-trap :invalid))
+                         (error 'floating-point-invalid-operation :operation 'scale-float
+                                                                  :operands (list x exp)))
+                       x)
                       (t
-                       ,(case type
-                          (single-float '(single-from-bits sign new-exp sig))
-                          (double-float '(double-from-bits sign new-exp sig))))))))))
+                       (multiple-value-bind (sig old-exp sign) (,(symbolicate 'integer-decode- type) x)
+                         (let* ((digits (float-digits x))
+                                (new-exp (+ exp old-exp digits
+                                            ,(case type
+                                               (single-float 'sb-vm:single-float-bias)
+                                               (double-float 'sb-vm:double-float-bias))))
+                                ;; convert decoded values {-1,+1} into {1,0} respectively
+                                (sign (if (minusp sign) 1 0)))
+                           (cond
+                             ((< new-exp
+                                 ,(case type
+                                    (single-float 'sb-vm:single-float-normal-exponent-min)
+                                    (double-float 'sb-vm:double-float-normal-exponent-min)))
+                              (when (sb-vm:current-float-trap :inexact)
+                                (error 'floating-point-inexact :operation 'scale-float
+                                                               :operands (list x exp)))
+                              (when (sb-vm:current-float-trap :underflow)
+                                (error 'floating-point-underflow :operation 'scale-float
+                                                                 :operands (list x exp)))
+                              (let ((shift (1- new-exp)))
+                                (if (< shift (- (1- digits)))
+                                    (float-sign x ,(case type
+                                                     (single-float $0f0)
+                                                     (double-float $0d0)))
+                                    ,(case type
+                                       (single-float '(single-from-bits sign 0 (ash sig shift)))
+                                       (double-float '(double-from-bits sign 0 (ash sig shift)))))))
+                             (t
+                              ,(case type
+                                 (single-float '(single-from-bits sign new-exp sig))
+                                 (double-float '(double-from-bits sign new-exp sig))))))))))))
   (def single-float)
   (def double-float))
 
@@ -364,8 +371,7 @@
     (fixnum
      #+64-bit
      (let* ((bits (double-float-bits x))
-            (old-exp (ldb (byte (byte-size sb-vm:double-float-exponent-byte)
-                                (+ (byte-position sb-vm:double-float-exponent-byte) 32)) bits))
+            (old-exp (ldb sb-vm:double-float-exponent-byte bits))
             (new-exp (+ old-exp exp)))
        (cond
          ((zerop x) x)
@@ -376,13 +382,11 @@
               (> new-exp sb-vm:double-float-normal-exponent-max))
           (scale-double-float-maybe-overflow x exp))
          (t
-          (%make-double-float (dpb new-exp (byte (byte-size sb-vm:double-float-exponent-byte)
-                                                 (+ (byte-position sb-vm:double-float-exponent-byte) 32))
-                                   bits)))))
+          (%make-double-float (dpb new-exp sb-vm:double-float-exponent-byte bits)))))
      #-64-bit
      (let* ((hi (double-float-high-bits x))
             (lo (double-float-low-bits x))
-            (old-exp (ldb sb-vm:double-float-exponent-byte hi))
+            (old-exp (ldb sb-vm:double-float-hi-exponent-byte hi))
             (new-exp (+ old-exp exp)))
        (cond
          ((zerop x) x)
@@ -393,7 +397,7 @@
               (> new-exp sb-vm:double-float-normal-exponent-max))
           (scale-double-float-maybe-overflow x exp))
          (t
-          (make-double-float (dpb new-exp sb-vm:double-float-exponent-byte hi)
+          (make-double-float (dpb new-exp sb-vm:double-float-hi-exponent-byte hi)
                              lo)))))
     (unsigned-byte (scale-double-float-maybe-overflow x exp))
     ((integer * 0) (scale-double-float-maybe-underflow x exp))))
@@ -513,59 +517,6 @@
   (def double-float)
   (def single-float))
 
-;;; These might be useful if we ever have a machine without float/integer
-;;; conversion hardware. For now, we'll use special ops that
-;;; uninterruptibly frob the rounding modes & do ieee round-to-integer.
-#+nil
-(progn
-  ;; The compiler compiles a call to this when we are doing %UNARY-TRUNCATE
-  ;; and the result is known to be a fixnum. We can avoid some generic
-  ;; arithmetic in this case.
-  (defun %unary-truncate-single-float/fixnum (x)
-    (declare (single-float x) (values fixnum))
-    (locally (declare (optimize (speed 3) (safety 0)))
-      (let* ((bits (single-float-bits x))
-             (exp (ldb sb-vm:single-float-exponent-byte bits))
-             (frac (logior (ldb sb-vm:single-float-significand-byte bits)
-                           sb-vm:single-float-hidden-bit))
-             (shift (- exp sb-vm:single-float-digits sb-vm:single-float-bias)))
-        (when (> exp sb-vm:single-float-normal-exponent-max)
-          (error 'floating-point-invalid-operation :operator 'truncate
-                 :operands (list x)))
-        (if (<= shift (- sb-vm:single-float-digits))
-            0
-            (let ((res (ash frac shift)))
-              (declare (type (unsigned-byte 31) res))
-              (if (minusp bits)
-                  (- res)
-                  res))))))
-  ;; Double-float version of this operation (see above single op).
-  (defun %unary-truncate-double-float/fixnum (x)
-    (declare (double-float x) (values fixnum))
-    (locally (declare (optimize (speed 3) (safety 0)))
-      (let* ((hi-bits (double-float-high-bits x))
-             (exp (ldb sb-vm:double-float-exponent-byte hi-bits))
-             (frac (logior (ldb sb-vm:double-float-significand-byte hi-bits)
-                           sb-vm:double-float-hidden-bit))
-             (shift (- exp (- sb-vm:double-float-digits sb-vm:n-word-bits)
-                       sb-vm:double-float-bias)))
-        (when (> exp sb-vm:double-float-normal-exponent-max)
-          (error 'floating-point-invalid-operation :operator 'truncate
-                 :operands (list x)))
-        (if (<= shift (- sb-vm:n-word-bits sb-vm:double-float-digits))
-            0
-            (let* ((res-hi (ash frac shift))
-                   (res (if (plusp shift)
-                            (logior res-hi
-                                    (the fixnum
-                                      (ash (double-float-low-bits x)
-                                           (- shift sb-vm:n-word-bits))))
-                            res-hi)))
-              (declare (type (unsigned-byte 31) res-hi res))
-              (if (minusp hi-bits)
-                  (- res)
-                  res)))))))
-
 ;;; This function is called when we are doing a truncate without any funky
 ;;; divisor, i.e. converting a float or ratio to an integer. Note that we do
 ;;; *not* return the second value of truncate, so it must be computed by the
@@ -605,12 +556,20 @@
                `(<= ,(symbol-value (symbolicate 'most-negative-fixnum- type))
                     number
                     ,(symbol-value (symbolicate 'most-positive-fixnum- type))))
-             (shift (type integer count)
-               `(,(case type
-                    #-64-bit
-                    (double-float 'ash)
-                    (t 'bignum-ashift-left-fixnum))
-                 ,integer ,count)))
+             (shift (type)
+               (case type
+                 #-64-bit
+                 (double-float
+                  ;; Shifting negatives right is different
+                  `(let ((truncated (ash bits exp)))
+                     (if (minusp sign)
+                         (- truncated)
+                         truncated)))
+                 (t
+                  `(bignum-ashift-left-fixnum (if (minusp sign)
+                                                  (- bits)
+                                                  bits)
+                                              exp)))))
     (number-dispatch ((number real))
       ((integer) (values number 0))
       ((ratio)
@@ -624,11 +583,7 @@
                      (- number
                         (coerce truncated '(dispatch-type number)))))
            (multiple-value-bind (bits exp sign) (integer-decode-float number)
-             (let ((truncated (shift (dispatch-type number)
-                                     (if (minusp sign)
-                                         (- bits)
-                                         bits)
-                                     exp)))
+             (let ((truncated (shift (dispatch-type number))))
                (values
                 truncated
                 #+64-bit
@@ -643,14 +598,19 @@
                `(defun ,(symbolicate 'unary-truncate- type '-to-bignum) (number)
                   (declare (inline ,decode))
                   (multiple-value-bind (bits exp sign) (,decode number)
-                    (let ((truncated (,(case type
-                                         #-64-bit
-                                         (double-float 'ash)
-                                         (t 'bignum-ashift-left-fixnum))
-                                      (if (minusp sign)
-                                          (- bits)
-                                          bits)
-                                      exp)))
+                    (let ((truncated ,(case type
+                                        #-64-bit
+                                        (double-float
+                                         ;; Shifting negatives right is different
+                                         `(let ((truncated (ash bits exp)))
+                                            (if (minusp sign)
+                                                (- truncated)
+                                                truncated)))
+                                        (t
+                                         `(bignum-ashift-left-fixnum (if (minusp sign)
+                                                                         (- bits)
+                                                                         bits)
+                                                                     exp)))))
                       (values
                        truncated
                        ,(case type
@@ -666,12 +626,19 @@
               `(defun ,(symbolicate '%unary-truncate- type '-to-bignum) (number)
                  (declare (inline ,decode))
                  (multiple-value-bind (bits exp sign) (,decode number)
-                   (,(case type
-                       #-64-bit
-                       (double-float 'ash)
-                       (t 'bignum-ashift-left-fixnum))
-                    (if (minusp sign) (- bits) bits)
-                    exp))))))
+                   ,(case type
+                      #-64-bit
+                      (double-float
+                       ;; Shifting negatives right is different
+                       `(let ((truncated (ash bits exp)))
+                          (if (minusp sign)
+                              (- truncated)
+                              truncated)))
+                      (t
+                       `(bignum-ashift-left-fixnum (if (minusp sign)
+                                                       (- bits)
+                                                       bits)
+                                                   exp))))))))
   (def double-float)
   (def single-float))
 
@@ -772,20 +739,15 @@
                    rounded))))))))
 
 #-round-float
-(defun %unary-ftruncate (number)
-  (number-dispatch ((number real))
-    ((integer) (float number))
-    ((ratio) (float (truncate (numerator number) (denominator number))))
-    (((foreach single-float double-float #+long-float long-float))
-     (%unary-ftruncate number))))
-
-(declaim (inline first-bit-set))
-(defun first-bit-set (x)
-  #+x86-64
-  (truly-the (values (mod #.sb-vm:n-word-bits) &optional)
-             (%primitive sb-vm::unsigned-word-find-first-bit (the word x)))
-  #-x86-64
-  (1- (integer-length (logand x (- x)))))
+(macrolet ((def (name op)
+             `(defun ,name (number)
+                (number-dispatch ((number real))
+                                 ((integer) (float number))
+                                 ((ratio) (float (,op (numerator number) (denominator number))))
+                                 (((foreach single-float double-float #+long-float long-float))
+                                  (,name number))))))
+  (def %unary-ftruncate truncate)
+  (def %unary-fround round))
 
 (defun rational (x)
   "RATIONAL produces a rational number for any real numeric argument. This is
@@ -814,70 +776,6 @@
                    (t
                     (ash int exp)))))))
     ((rational) x)))
-
-#+64-bit
-(defun float-bignum-= (float bignum)
-  (declare (optimize speed))
-  (number-dispatch ((float))
-    (((foreach single-float double-float))
-     (multiple-value-bind (bits exp) (integer-decode-float float)
-       (if (or (eql bits 0)
-               (minusp exp))
-           nil
-           (let ((int (if (minusp float) (- bits) bits)))
-             (and (= (truly-the bignum-length (bignum-integer-length bignum))
-                     (+ (integer-length int) exp))
-                  (sb-bignum::bignum-lower-bits-zero-p bignum exp)
-                  (= int
-                     (truly-the fixnum
-                                (sb-bignum::last-bignum-part=>fixnum exp bignum))))))))))
-
-#+64-bit
-(defun float-bignum-< (float bignum)
-  (declare (optimize speed))
-  (number-dispatch ((float))
-    (((foreach single-float double-float))
-     (multiple-value-bind (bits exp) (integer-decode-float float)
-       (if (or (eql bits 0)
-               (minusp exp))
-           (bignum-plus-p bignum)
-           (let* ((int (if (minusp float) (- bits) bits))
-                  (length-diff (- (truly-the bignum-length (bignum-integer-length bignum))
-                                  (+ (integer-length int) exp))))
-             (cond
-               ((plusp length-diff) (bignum-plus-p bignum))
-               ((minusp length-diff) (minusp float))
-               (t
-                (let ((diff (- (truly-the fixnum
-                                          (sb-bignum::last-bignum-part=>fixnum exp bignum))
-                               int)))
-                  (cond ((plusp diff) t)
-                        ((minusp diff) nil)
-                        (t
-                         (not (sb-bignum::bignum-lower-bits-zero-p bignum exp)))))))))))))
-
-#+64-bit
-(defun float-bignum-> (float bignum)
-  (declare (optimize speed))
-  (number-dispatch ((float))
-    (((foreach single-float double-float))
-     (multiple-value-bind (bits exp) (integer-decode-float float)
-       (if (or (eql bits 0)
-               (minusp exp))
-           (not (bignum-plus-p bignum))
-           (let* ((int (if (minusp float) (- bits) bits))
-                  (length-diff (- (truly-the bignum-length (bignum-integer-length bignum))
-                                  (+ (integer-length int) exp))))
-             (cond
-               ((plusp length-diff) (not (bignum-plus-p bignum)))
-               ((minusp length-diff) (not (minusp float)))
-               (t
-                (let ((diff (- (truly-the fixnum
-                                          (sb-bignum::last-bignum-part=>fixnum exp bignum))
-                               int)))
-                  (cond ((plusp diff) nil)
-                        ((minusp diff) t)))))))))))
-
 ;;; This algorithm for RATIONALIZE, due to Bruno Haible, is included
 ;;; with permission.
 ;;;
