@@ -15,11 +15,6 @@
 
 (declaim (inline dynamic-usage))
 (defun dynamic-usage ()
-  #+mark-region-gc
-  (let ((bytes (extern-alien "bytes_allocated" os-vm-size-t)))
-    (values bytes
-            (- (* (pages-allocated) sb-vm:gencgc-page-bytes) bytes)))
-  #+gencgc
   (extern-alien "bytes_allocated" os-vm-size-t))
 
 (defun static-space-usage ()
@@ -304,6 +299,7 @@ used to specify the oldest generation guaranteed to be collected."
 
 (define-alien-routine scrub-control-stack void)
 
+(defglobal sb-unicode::*name->char-buffers* nil)
 (defun unsafe-clear-roots (gen)
   (declare (ignorable gen))
   ;; KLUDGE: Do things in an attempt to get rid of extra roots. Unsafe
@@ -313,6 +309,8 @@ used to specify the oldest generation guaranteed to be collected."
   ;; Power cache of the bignum printer: drops overly large bignums and
   ;; removes duplicate entries.
   (scrub-power-cache)
+  (setf sb-unicode::*name->char-buffers* nil)
+  (setf sb-c::*phash-lambda-cache* nil)
   ;; Clear caches depending on the generation being collected.
   (cond ((eql 0 gen)
          ;; Drop strings because the hash is pointer-hash
@@ -337,8 +335,11 @@ Note: currently changes to this value are lost when saving core."
 
 (defun (setf bytes-consed-between-gcs) (val)
   (declare (type (and fixnum unsigned-byte) val))
-  (setf (extern-alien "bytes_consed_between_gcs" os-vm-size-t)
-        val))
+  #+(or gencgc mark-region-gc)
+  (let ((current (extern-alien "bytes_consed_between_gcs" os-vm-size-t)))
+    (when (< val current)
+      (decf (extern-alien "auto_gc_trigger" os-vm-size-t) (- current val))))
+  (setf (extern-alien "bytes_consed_between_gcs" os-vm-size-t) val))
 
 (declaim (inline maybe-handle-pending-gc))
 (defun maybe-handle-pending-gc ()
@@ -403,8 +404,14 @@ statistics are appended to it."
             (sb-vm::flags (unsigned 8)) ; this named 'type' in C
             (sb-vm::gen (signed 8))))
 (define-alien-variable ("page_table" sb-vm:page-table) (* (struct sb-vm::page)))
-(declaim (inline sb-vm:find-page-index))
-(define-alien-routine ("ext_find_page_index" sb-vm:find-page-index) page-index-t (address unsigned))
+
+(declaim (ftype (sfunction (t) (integer -1 #.(floor most-positive-word sb-vm:gencgc-page-bytes)))
+                sb-vm:find-page-index)
+         (inline sb-vm:find-page-index))
+(defun sb-vm:find-page-index (address)
+  (with-alien ((find-page-index (function page-index-t unsigned) :extern "ext_find_page_index"))
+    (truly-the (integer -1 #.(floor most-positive-word sb-vm:gencgc-page-bytes))
+               (alien-funcall find-page-index address))))
 
 (defun pages-allocated ()
   (loop for n below (extern-alien "next_free_page" signed)
@@ -425,12 +432,13 @@ statistics are appended to it."
                (logand (if (fdefn-p object) (sap-ref-8 sap 1) (sap-ref-8 sap 3))
                        #xF)))))))
 
-#+mark-region-gc
-(define-alien-routine "gc_gen_of" char (address unsigned))
 #+mark-region-gc
 (defun generation-of (object)
-  (with-pinned-objects (object)
-    (gc-gen-of (get-lisp-obj-address object))))
+  (with-alien ((gc-gen-of (function char unsigned int) :extern))
+    (let ((result (with-pinned-objects (object)
+                    (alien-funcall gc-gen-of
+                                   (get-lisp-obj-address object) 127))))
+      (unless (= result 127) result))))
 
 (export 'page-protected-p)
 (macrolet ((addr->mark (addr)
@@ -524,6 +532,12 @@ Experimental: interface subject to change."
                     ((< sb-vm:read-only-space-start addr
                         (sap-int sb-vm:*read-only-space-free-pointer*))
                      :read-only)
+                    ;; Without immobile-space, the text range is
+                    ;; more-or-less an extension of static space.
+                    #-immobile-space
+                    ((< sb-vm:text-space-start addr
+                        (sap-int sb-vm:*text-space-free-pointer*))
+                     :static)
                     ((< sb-vm:static-space-start addr
                         (sap-int sb-vm:*static-space-free-pointer*))
                      :static))))
