@@ -323,10 +323,11 @@
 
 (defun nuke-fop-vector (vector)
   (declare (simple-vector vector)
-           #-gencgc (ignore vector)
+           (ignorable vector)
            (optimize speed))
   ;; Make sure we don't keep any garbage.
-  #+gencgc
+  ;; NOTE: for the work-in-progress concurrent GC, it is better *NOT* to 0-fill
+  ;; if the the deletion barrier is enabled.
   (fill vector 0))
 
 
@@ -377,9 +378,8 @@
 
 ;;;; the FOP database
 
-(eval-when (:compile-toplevel :load-toplevel :execute)
-  ;; The bottom 5 bits of the opcodes above 128 encode an implicit operand.
-  (defconstant n-ordinary-fops 128))
+;; The bottom 5 bits of the opcodes above 128 encode an implicit operand.
+(defconstant n-ordinary-fops 128)
 
 ;;; a vector indexed by a FaslOP that yields a function which performs
 ;;; the operation. Most functions take 0 arguments - they only manipulate
@@ -440,6 +440,7 @@
                  (error ,(format nil "Not-host fop invoked: ~A" name)))))
        (!%define-fop ',name ,fop-code ,(length operands) ,(if pushp 1 0)))))
 
+(defglobal *fop-name-to-opcode* (make-hash-table))
 (defun !%define-fop (name opcode n-operands pushp)
   (declare (type (mod 4) n-operands))
   (let ((function (svref **fop-funs** opcode)))
@@ -453,7 +454,7 @@
       (when (and existing-opcode (/= existing-opcode opcode))
         (error "multiple codes for fop name ~S: ~D and ~D"
                name opcode existing-opcode)))
-    (setf (get name 'opcode) opcode
+    (setf (gethash name *fop-name-to-opcode*) opcode
           (svref **fop-funs** opcode) (symbol-function name)
           (aref (car **fop-signatures**) opcode) n-operands
           (sbit (cdr **fop-signatures**) opcode) pushp))
@@ -723,8 +724,11 @@
                      n-operands arg1 arg2 arg3)
              (when (functionp function)
                (format *trace-output* " ~35t~(~a~)" (%fun-name function))
-               (when (eql (%fun-name function) 'fop-push)
-                 (format *trace-output* " ~(~A~)" (ref-fop-table fasl-input arg1)))))))))))
+               (case (%fun-name function)
+                 (fop-push (format *trace-output* " ~(~A~)" (ref-fop-table fasl-input arg1)))
+                 (fop-word-integer (format *trace-output* " ~V,'0X" (* 2 sb-vm:n-word-bytes) result))
+                 (fop-byte-integer (format *trace-output* " ~2,'0X" result))
+                 (fop-integer (format *trace-output* " ~X" result)))))))))))
 
 (defun load-as-fasl (stream verbose print)
   (when (zerop (file-length stream))
@@ -774,25 +778,15 @@
         (n-data-words (- size sb-vm:instance-data-start)))
     (with-fop-stack ((stack (operand-stack)) ptr n-data-words)
       (declare (type index ptr))
-      ;; FIXME: this is basically DO-LAYOUT-BITMAP, but probably not as efficient.
-      (let ((bitmap (sb-kernel::%layout-bitmap layout)))
-        ;; Values on the stack are in the same order as in the structure itself.
-        (do ((i sb-vm:instance-data-start (1+ i)))
-            ((>= i size))
-          (declare (type index i))
-          (let ((val (fop-stack-ref ptr)))
-            (if (logbitp i bitmap)
-                (%instance-set res i val)
-                (%raw-instance-set/word res i val))
-            (incf ptr)))))
+      ;; Values on the stack are in the same order as in the structure itself.
+      (do-layout-bitmap (i taggedp layout size)
+        (let ((val (fop-stack-ref ptr)))
+          (if taggedp
+              (%instance-set res i val)
+              (%raw-instance-set/word res i val)))
+        (incf ptr)))
     res))
 
-;;; Symbol-like entities
-(define-fop 49 :not-host (fop-debug-name-marker ((:operands kind)))
-  (ecase kind
-   (1 sb-c::*debug-name-sharp*)
-   (2 sb-c::*debug-name-ellipsis*)))
-
 (define-fop 45 :not-host (fop-layout ((:operands depthoid flags length)
                                        name bitmap inherits))
   (decf depthoid) ; was bumped by 1 since non-stack args can't encode negatives
@@ -852,17 +846,7 @@
            (multiple-value-bind (name length elt-type)
                (read-symbol-name length+flag fasl-input)
              (push-fop-table (%intern name length package elt-type t inherited)
-                             fasl-input)))
-         (ensure-hashed (symbol)
-           ;; ENSURE-SYMBOL-HASH when vop-translated is flushable since it is
-           ;; conceptually just a slot reader, however its actual effect is to fill in
-           ;; the hash if absent, so it's not quite flushable when called expressly
-           ;; to fill in the slot. In this case we need a full call to ENSURE-SYMBOL-HASH
-           ;; to ensure the side-effect happens.
-           ;; Careful if changing this again. There'a regression test thank goodness.
-           (declare (notinline ensure-symbol-hash))
-           (ensure-symbol-hash symbol)
-           symbol))
+                             fasl-input))))
 
   (define-fop 77 :not-host (fop-lisp-symbol-save ((:operands length+flag)))
     (aux-fop-intern length+flag *cl-package* t (fasl-input)))
@@ -875,12 +859,11 @@
 
   (define-fop 80 :not-host (fop-uninterned-symbol-save ((:operands length+flag)))
     (multiple-value-bind (name len) (read-symbol-name length+flag (fasl-input))
-      (push-fop-table (ensure-hashed (make-symbol (subseq name 0 len)))
+      (push-fop-table (make-symbol (subseq name 0 len))
                       (fasl-input))))
 
   (define-fop 81 :not-host (fop-copy-symbol-save ((:operands table-index)))
-    (push-fop-table (ensure-hashed
-                     (copy-symbol (ref-fop-table (fasl-input) table-index)))
+    (push-fop-table (copy-symbol (ref-fop-table (fasl-input) table-index))
                     (fasl-input))))
 
 (define-fop 82 (fop-package (pkg-designator))
@@ -939,7 +922,7 @@
   #-sb-xc-host (%make-ratio num den))
 
 (define-fop 71 (fop-complex (realpart imagpart))
-  #+sb-xc-host (number-pair-to-core realpart imagpart sb-vm:complex-widetag)
+  #+sb-xc-host (number-pair-to-core realpart imagpart sb-vm:complex-rational-widetag)
   #-sb-xc-host (%make-complex realpart imagpart))
 
 (macrolet ((fast-read-single-float ()
