@@ -59,13 +59,10 @@ Does not affect the cases that are already controlled by *PRINT-LENGTH*")
   "Suppress printer errors when the condition is of the type designated by this
 variable: an unreadable object representing the error is printed instead.")
 
-;; duplicate defglobal because this file is compiled before "reader"
-(define-load-time-global *standard-readtable* nil)
-
+(define-load-time-global sb-pretty::*standard-pprint-dispatch-table* nil)
 (defun %with-standard-io-syntax (function)
   (declare (type function function))
   (declare (dynamic-extent function))
-  (declare (special sb-pretty::*standard-pprint-dispatch-table*))
   (let ((*package* #.(find-package "COMMON-LISP-USER"))
         (*print-array* t)
         (*print-base* 10)
@@ -196,13 +193,18 @@ variable: an unreadable object representing the error is printed instead.")
            (%with-output-to-string (stream)
               (sb-pretty:output-pretty-object stream fun object))
            (let ((buffer-size (approx-chars-in-repr object)))
-             (let* ((string (make-string buffer-size :element-type 'base-char))
+             (let* ((string (make-string buffer-size :element-type 'base-char
+                                         :initial-element (code-char 0)))
                     (stream (%make-finite-base-string-output-stream string)))
                (declare (inline %make-finite-base-string-output-stream))
-               (declare (truly-dynamic-extent stream))
+               (declare (dynamic-extent stream))
                (output-integer object stream *print-base* *print-radix*)
-               (%shrink-vector string
-                               (finite-base-string-output-stream-pointer stream)))))))
+               ;; ASSUMPTION: we use pre-zeroed memory for unboxed objects.
+               ;; So we can avoid calling %SHRINK-VECTOR, and instead directly
+               ;; set the length.
+               (setf (%array-fill-pointer string)
+                     (finite-base-string-output-stream-pointer stream))
+               string)))))
     ;; Could do something for other numeric types, symbols, ...
     (t
      (%with-output-to-string (stream)
@@ -465,7 +467,7 @@ variable: an unreadable object representing the error is printed instead.")
 ;;; Automatically handle *PRINT-LEVEL* abbreviation. If we are too
 ;;; deep, then a #\# is printed to STREAM and BODY is ignored.
 (defmacro descend-into ((stream) &body body)
-  (let ((flet-name (sb-xc:gensym "DESCEND")))
+  (let ((flet-name (gensym "DESCEND")))
     `(flet ((,flet-name ()
               ,@body))
        (cond ((and (null *print-readably*)
@@ -578,12 +580,11 @@ variable: an unreadable object representing the error is printed instead.")
         (return-from output-ugly-object
           (print-unreadable-object (object stream :identity t)
             (prin1 'instance stream))))
-      (let* ((wrapper (layout-friend layout))
-             (classoid (wrapper-classoid wrapper)))
+      (let ((classoid (layout-classoid layout)))
         ;; Additionally, don't crash if the object is an obsolete thing with
         ;; no update protocol.
         (when (or (sb-kernel::undefined-classoid-p classoid)
-                  (and (wrapper-invalid wrapper)
+                  (and (layout-invalid layout)
                        (logtest (layout-flags layout)
                                 (logior +structure-layout-flag+
                                         +condition-layout-flag+))))
@@ -660,7 +661,7 @@ variable: an unreadable object representing the error is printed instead.")
             (unless (and accessible (eq found symbol))
               (output-token (or (package-local-nickname package current)
                                 (package-name package)))
-              (write-string (if (eql (find-external-symbol name package) 0) "::" ":")
+              (write-string (if (symbol-externalp symbol package) ":" "::")
                             stream)))))
         (output-token name)))))
 
@@ -1061,22 +1062,35 @@ variable: an unreadable object representing the error is printed instead.")
             (t
              (output-unreadable-array-readably vector stream))))))
 
-;;; This function outputs a string quoting characters sufficiently
-;;; so that someone can read it in again. Basically, put a slash in
-;;; front of an character satisfying NEEDS-SLASH-P.
+;;; Output a string, quoting characters to be readable by putting a slash in front
+;;; of any character satisfying NEEDS-SLASH-P.
 (defun quote-string (string stream)
   (macrolet ((needs-slash-p (char)
                ;; KLUDGE: We probably should look at the readtable, but just do
                ;; this for now. [noted by anonymous long ago] -- WHN 19991130
-               `(or (char= ,char #\\)
-                 (char= ,char #\"))))
+               `(let ((c ,char)) (or (char= c #\\) (char= c #\"))))
+             (scan (type)
+               ;; Pre-test for any escaping, and if needed, do char-at-a-time output.
+               ;; For 1 or 0 characters, always take the WRITE-CHAR branch.
+               `(let ((data (truly-the ,type data)))
+                  (declare (optimize (sb-c:insert-array-bounds-checks 0)))
+                  (when (or (<= (- end start) 1)
+                            (do ((index start (1+ index)))
+                                ((>= index end))
+                              (when (needs-slash-p (schar data index)) (return t))))
+                    (do ((index start (1+ index)))
+                        ((>= index end) (return-from quote-string))
+                      (let ((char (schar data index)))
+                        (when (needs-slash-p char) (write-char #\\ stream))
+                        (write-char char stream)))))))
     (with-array-data ((data string) (start) (end)
                       :check-fill-pointer t)
-      (do ((index start (1+ index)))
-          ((>= index end))
-        (let ((char (schar data index)))
-          (when (needs-slash-p char) (write-char #\\ stream))
-          (write-char char stream))))))
+      (if (simple-base-string-p data)
+          (scan simple-base-string)
+          #+sb-unicode (scan simple-character-string)))
+    ;; If no escaping needed, WRITE-STRING is way faster, up to 2x in my testing,
+    ;; than WRITE-CHAR because the stream layer will get so many fewer calls.
+    (write-string string stream)))
 
 (defun array-readably-printable-p (array)
   (and (eq (array-element-type array) t)
@@ -1331,7 +1345,7 @@ variable: an unreadable object representing the error is printed instead.")
                       ;; the character.
                       (write-char (schar chars r) stream)))))
       (cond ((typep integer 'word) ; Division vops can handle this all inline.
-             #+(and gencgc c-stack-is-control-stack) ; strings can be DX
+             #+c-stack-is-control-stack ; strings can be DX-allocated
              ;; For bases exceeding 10 we know how many characters (at most)
              ;; will be output. This allows for a single %WRITE-STRING call.
              ;; There's diminishing payback for other bases because the fixed array
@@ -1342,10 +1356,10 @@ variable: an unreadable object representing the error is printed instead.")
                  (let* ((ptr #.(length (write-to-string sb-ext:most-positive-word
                                                         :base 10)))
                         (buffer (make-array ptr :element-type 'base-char)))
-                   (declare (truly-dynamic-extent buffer))
+                   (declare (dynamic-extent buffer))
                    (iterative-algorithm)
                    (%write-string buffer stream ptr (length buffer))))
-             #-(and gencgc c-stack-is-control-stack) ; strings can not be DX
+             #-c-stack-is-control-stack ; strings can't be DX-allocated
              ;; Use the alien stack, which is not as fast as using the control stack
              ;; (when we can). Even the absence of 0-fill doesn't make up for it.
              ;; Since we've no choice in the matter, might as well allow
@@ -1523,10 +1537,6 @@ variable: an unreadable object representing the error is printed instead.")
 ;;; paper might not bring immediate illumination as CSR has attempted
 ;;; to turn idiomatic Scheme into idiomatic Lisp.
 ;;;
-;;; FIXME: figure 1 from Burger and Dybvig is the unoptimized
-;;; algorithm, noticeably slow at finding the exponent.  Figure 2 has
-;;; an improved algorithm, but CSR ran out of energy.
-;;;
 ;;; possible extension for the enthusiastic: printing floats in bases
 ;;; other than base 10.
 (defconstant single-float-min-e
@@ -1562,28 +1572,37 @@ variable: an unreadable object representing the error is printed instead.")
         (let ((shift (- float-digits (integer-length f))))
           (setq f (ash f shift)
                 e (- e shift))))
-      (let ( ;; FIXME: these even tests assume normal IEEE rounding
+      (let (;; FIXME: these even tests assume normal IEEE rounding
             ;; mode.  I wonder if we should cater for non-normal?
             (high-ok (evenp f))
             (low-ok (evenp f)))
-        (labels ((scale (r s m+ m-)
-                   (do ((r+m+ (+ r m+))
-                        (k 0 (1+ k))
-                        (s s (* s print-base)))
-                       ((not (or (> r+m+ s)
-                                 (and high-ok (= r+m+ s))))
-                        (do ((k k (1- k))
-                             (r r (* r print-base))
-                             (m+ m+ (* m+ print-base))
-                             (m- m- (* m- print-base)))
-                            ((not (and (> r m-) ; Extension to handle zero
-                                       (let ((x (* (+ r m+) print-base)))
-                                         (or (< x s)
-                                             (and (not high-ok)
-                                                  (= x s))))))
-                             (funcall prologue-fun k)
-                             (generate r s m+ m-)
-                             (funcall epilogue-fun k))))))
+        (labels ((expt2 (n)
+                   (svref #.(coerce (loop for i from 0 to 972 collect (expt 2 i))
+                                    'vector) n))
+                 (expt10 (n)
+                   (svref #.(coerce (loop for i from 0 to 323 collect (expt 10 i))
+                                    'vector) n))
+                 (scale (r s m+ m-)
+                   (let ((est (truly-the (integer -323 309)
+                                         (ceiling (- (* (+ e (integer-length (truly-the sb-kernel:double-float-significand f)) -1)
+                                                        (log $2d0 10))
+                                                     $1.0e-10)))))
+                     (if (>= est 0)
+                         (fixup r (* s (expt10 est)) m+ m- est)
+                         (let ((scale (expt10 (- est))))
+                           (fixup (* r scale)
+                                  s (* m+ scale) (* m- scale) est)))))
+
+                 (fixup (r s m+ m- k)
+                   (let ((r+m+ (+ r m+)))
+                     (when (if high-ok
+                               (>= r+m+ s)
+                               (> r+m+ s))
+                       (incf k)
+                       (setf s (* s 10)))
+                     (funcall prologue-fun k)
+                     (generate r s m+ m-)
+                     (funcall epilogue-fun k)))
                  (generate (r s m+ m-)
                    (let (d tc1 tc2)
                      (tagbody
@@ -1608,65 +1627,81 @@ variable: an unreadable object representing the error is printed instead.")
                                    (t
                                     (1+ d)))))
                           (funcall char-fun d)))))
-                 (initialize ()
-                   (let (r s m+ m-)
-                     (cond ((>= e 0)
-                            (let ((be (expt float-radix e)))
-                              (if (/= f (expt float-radix (1- float-digits)))
-                                  ;; multiply F by 2 first, avoding consing two bignums
-                                  (setf r (* f 2 be)
-                                        s 2
-                                        m+ be
-                                        m- be)
-                                  (setf m- be
-                                        m+ (* be float-radix)
-                                        r (* f 2 m+)
-                                        s (* float-radix 2)))))
-                           ((or (= e min-e)
-                                (/= f (expt float-radix (1- float-digits))))
-                            (setf r (* f 2)
-                                  s (expt float-radix (- 1 e))
-                                  m+ 1
-                                  m- 1))
-                           (t
-                            (setf r (* f float-radix 2)
-                                  s (expt float-radix (- 2 e))
-                                  m+ float-radix
-                                  m- 1)))
-                     (when position
-                       (when relativep
-                         (aver (> position 0))
-                         (do ((k 0 (1+ k))
-                              ;; running out of letters here
-                              (l 1 (* l print-base)))
-                             ((>= (* s l) (+ r m+))
-                              ;; k is now \hat{k}
-                              (if (< (+ r (* s (/ (expt print-base (- k position)) 2)))
-                                     (* s l))
-                                  (setf position (- k position))
-                                  (setf position (- k position 1))))))
-                       (let* ((x (/ (* s (expt print-base position)) 2))
-                              (low (max m- x))
-                              (high (max m+ x)))
-                         (when (<= m- low)
-                           (setf m- low)
-                           (setf low-ok t))
-                         (when (<= m+ high)
-                           (setf m+ high)
-                           (setf high-ok t))))
-                     (values r s m+ m-))))
-          (multiple-value-bind (r s m+ m-) (initialize)
-            (scale r s m+ m-)))))))
+                 (scale-p (r s m+ m-)
+                   (when relativep
+                     (aver (> position 0))
+                     (do ((k 0 (1+ k))
+                          ;; running out of letters here
+                          (l 1 (* l print-base)))
+                         ((>= (* s l) (+ r m+))
+                          ;; k is now \hat{k}
+                          (if (< (+ r (* s (/ (expt print-base (- k position)) 2)))
+                                 (* s l))
+                              (setf position (- k position))
+                              (setf position (- k position 1))))))
+                   (let* ((x (/ (* s (expt print-base position)) 2))
+                          (low (max m- x))
+                          (high (max m+ x)))
+                     (when (<= m- low)
+                       (setf m- low)
+                       (setf low-ok t))
+                     (when (<= m+ high)
+                       (setf m+ high)
+                       (setf high-ok t)))
+                   (do ((r+m+ (+ r m+))
+                        (k 0 (1+ k))
+                        (s s (* s print-base)))
+                       ((not (or (> r+m+ s)
+                                 (and high-ok (= r+m+ s))))
+                        (do ((k k (1- k))
+                             (r r (* r print-base))
+                             (m+ m+ (* m+ print-base))
+                             (m- m- (* m- print-base)))
+                            ((not (and (> r m-) ; Extension to handle zero
+                                       (let ((x (* (+ r m+) print-base)))
+                                         (or (< x s)
+                                             (and (not high-ok)
+                                                  (= x s))))))
+                             (funcall prologue-fun k)
+                             (generate r s m+ m-)
+                             (funcall epilogue-fun k)))))))
+          (let (r s m+ m-)
+            (cond ((>= e 0)
+                   (let ((be (expt2 e)))
+                     (setf m- be)
+                     (if (/= f (expt float-radix (1- float-digits)))
+                         (setf r (ash f (+ e 1))
+                               s 2
+                               m+ be)
+                         (setf m+ (expt2 (1+ e))
+                               r (ash f (+ e 2))
+                               s (* float-radix 2)))))
+                  ((or (= e min-e)
+                       (/= f (expt float-radix (1- float-digits))))
+                   (setf r (* f 2)
+                         s (expt float-radix (- 1 e))
+                         m+ 1
+                         m- 1))
+                  (t
+                   (setf r (* f float-radix 2)
+                         s (expt float-radix (- 2 e))
+                         m+ float-radix
+                         m- 1)))
+            (if position
+                (scale-p r s m+ m-)
+                (scale r s m+ m-))))))))
 
 (defun flonum-to-digits (float &optional position relativep)
-  (let ((digit-characters "0123456789"))
-    (with-push-char (:element-type base-char)
-      (%flonum-to-digits
-       (lambda (d)
-         (push-char (char digit-characters d)))
-       (lambda (k) k)
-       (lambda (k) (values k (get-pushed-string)))
-       float position relativep))))
+  (if (zerop float)
+      (values 0 "0")
+      (let ((digit-characters "0123456789"))
+        (with-push-char (:element-type base-char)
+          (%flonum-to-digits
+           (lambda (d)
+             (push-char (char digit-characters d)))
+           (lambda (k) k)
+           (lambda (k) (values k (get-pushed-string)))
+           float position relativep)))))
 
 (defun print-float (float stream)
   (let ((position 0)
@@ -1831,20 +1866,14 @@ variable: an unreadable object representing the error is printed instead.")
            (write-string (if (float-trapping-nan-p x) " trapping" " quiet") stream)
            (write-string " NaN" stream))))
     (t
-     (let ((x (cond ((minusp (float-sign x))
-                     (write-char #\- stream)
-                     (- x))
-                    (t
-                     x))))
-       (cond
-         ((zerop x)
-          (write-string "0.0" stream)
-          (print-float-exponent x 0 stream))
-         (t
-          (print-float x stream)))))))
-
-
-
+     (when (float-sign-bit-set-p x)
+       (write-char #\- stream))
+     (cond
+       ((zerop x)
+        (write-string "0.0" stream)
+        (print-float-exponent x 0 stream))
+       (t
+        (print-float x stream))))))
 
 ;;;; other leaf objects
 
@@ -1856,9 +1885,14 @@ variable: an unreadable object representing the error is printed instead.")
                            (standard-char-p char)))
             (name (char-name char)))
         (write-string "#\\" stream)
-        (if (and name (or (not graphicp) *print-readably*))
-            (quote-string name stream)
-            (write-char char stream)))
+        (cond
+          ((and name (or (not graphicp) *print-readably*)) (quote-string name stream))
+          (t
+           (write-char char stream)
+           ;; KLUDGE: arguably this should be in an :AROUND method
+           ;; specialized on ((EQL #\SPACE) T).
+           (when (and (eql char #\Space))
+             (sb-pretty:note-significant-space stream)))))
       (write-char char stream)))
 
 (defmethod print-object ((sap system-area-pointer) stream)
@@ -1868,6 +1902,20 @@ variable: an unreadable object representing the error is printed instead.")
          (print-unreadable-object (sap stream)
            (format stream "system area pointer: #X~8,'0X" (sap-int sap))))))
 
+#+weak-vector-readbarrier
+(defmethod print-object ((self weak-pointer) stream)
+  (let ((vectorp (weak-vector-p self)))
+    (print-unreadable-object (self stream :identity vectorp)
+      (if vectorp
+          (format stream "weak array [~d]" (weak-vector-len self))
+          (multiple-value-bind (value validp) (weak-pointer-value self)
+            (cond (validp
+                   (write-string "weak pointer: " stream)
+                   (write value :stream stream))
+                  (t
+                   (write-string "broken weak pointer" stream))))))))
+
+#-weak-vector-readbarrier
 (defmethod print-object ((weak-pointer weak-pointer) stream)
   (print-unreadable-object (weak-pointer stream)
     (multiple-value-bind (value validp) (weak-pointer-value weak-pointer)
@@ -1880,10 +1928,7 @@ variable: an unreadable object representing the error is printed instead.")
 (defmethod print-object ((component code-component) stream)
   (print-unreadable-object (component stream :identity t)
     (let (dinfo)
-      (cond ((code-obj-is-filler-p component)
-             (format stream "filler ~dw"
-                     (ash (code-object-size component) (- sb-vm:word-shift))))
-            ((eq (setq dinfo (%code-debug-info component)) :bpt-lra)
+      (cond ((eq (setq dinfo (%code-debug-info component)) :bpt-lra)
              (write-string "bpt-trap-return" stream))
             ((functionp dinfo)
              (format stream "trampoline ~S" dinfo))
@@ -1901,7 +1946,7 @@ variable: an unreadable object representing the error is printed instead.")
                       (write-string ", " stream)
                       (output-object (sb-c::debug-info-name dinfo) stream)))))))))
 
-#-(or x86 x86-64 arm64)
+#-(or x86 x86-64 arm64 riscv)
 (defmethod print-object ((lra lra) stream)
   (print-unreadable-object (lra stream :identity t)
     (write-string "return PC object" stream)))
@@ -1918,92 +1963,109 @@ variable: an unreadable object representing the error is printed instead.")
 #+sb-simd-pack
 (defmethod print-object ((pack simd-pack) stream)
   (cond ((and *print-readably* *read-eval*)
-         (multiple-value-bind (format maker extractor)
-             (etypecase pack
-               ((simd-pack double-float)
-                (values "#.(~S ~S ~S)"
-                        '%make-simd-pack-double #'%simd-pack-doubles))
-               ((simd-pack single-float)
-                (values "#.(~S ~S ~S ~S ~S)"
-                        '%make-simd-pack-single #'%simd-pack-singles))
-               (t
-                (values "#.(~S #X~16,'0X #X~16,'0X)"
-                        '%make-simd-pack-ub64 #'%simd-pack-ub64s)))
-           (multiple-value-call
-               #'format stream format maker (funcall extractor pack))))
+         (format stream "#.(~S #b~3,'0b #x~16,'0X #x~16,'0X)"
+                 '%make-simd-pack
+                 (%simd-pack-tag pack)
+                 (%simd-pack-low pack)
+                 (%simd-pack-high pack)))
         (*print-readably*
          (print-not-readable-error pack stream))
         (t
          (print-unreadable-object (pack stream)
-           (flet ((all-ones-p (value start end &aux (mask (- (ash 1 end) (ash 1 start))))
-                    (= (logand value mask) mask))
-                  (split-num (value start)
-                    (loop
-                       for i from 0 to 3
-                       and v = (ash value (- start)) then (ash v -8)
-                       collect (logand v #xFF))))
-             (multiple-value-bind (low high)
-                 (%simd-pack-ub64s pack)
-               (etypecase pack
-                 ((simd-pack double-float)
-                  (multiple-value-bind (v0 v1) (%simd-pack-doubles pack)
-                    (format stream "~S~@{ ~:[~,13E~;~*TRUE~]~}"
-                            'simd-pack
-                            (all-ones-p low 0 64) v0
-                            (all-ones-p high 0 64) v1)))
-                 ((simd-pack single-float)
-                  (multiple-value-bind (v0 v1 v2 v3) (%simd-pack-singles pack)
-                    (format stream "~S~@{ ~:[~,7E~;~*TRUE~]~}"
-                            'simd-pack
-                            (all-ones-p low 0 32) v0
-                            (all-ones-p low 32 64) v1
-                            (all-ones-p high 0 32) v2
-                            (all-ones-p high 32 64) v3)))
-                 (t
-                  (format stream "~S~@{ ~{ ~2,'0X~}~}"
-                          'simd-pack
-                          (split-num low 0) (split-num low 32)
-                          (split-num high 0) (split-num high 32))))))))))
+           (etypecase pack
+             ((simd-pack double-float)
+              (multiple-value-call #'format stream "~S~@{ ~,13E~}" 'simd-pack
+                (%simd-pack-doubles pack)))
+             ((simd-pack single-float)
+              (multiple-value-call #'format stream "~S~@{ ~,7E~}" 'simd-pack
+                (%simd-pack-singles pack)))
+             ((simd-pack (unsigned-byte 8))
+              (multiple-value-call #'format stream "~S~@{ ~3D~}" 'simd-pack
+                (%simd-pack-ub8s pack)))
+             ((simd-pack (unsigned-byte 16))
+              (multiple-value-call #'format stream "~S~@{ ~5D~}" 'simd-pack
+                (%simd-pack-ub16s pack)))
+             ((simd-pack (unsigned-byte 32))
+              (multiple-value-call #'format stream "~S~@{ ~10D~}" 'simd-pack
+                (%simd-pack-ub32s pack)))
+             ((simd-pack (unsigned-byte 64))
+              (multiple-value-call #'format stream "~S~@{ ~20D~}" 'simd-pack
+                (%simd-pack-ub64s pack)))
+             ((simd-pack (signed-byte 8))
+              (multiple-value-call #'format stream "~S~@{ ~4,@D~}" 'simd-pack
+                (%simd-pack-sb8s pack)))
+             ((simd-pack (signed-byte 16))
+              (multiple-value-call #'format stream "~S~@{ ~6,@D~}" 'simd-pack
+                (%simd-pack-sb16s pack)))
+             ((simd-pack (signed-byte 32))
+              (multiple-value-call #'format stream "~S~@{ ~11@D~}" 'simd-pack
+                (%simd-pack-sb32s pack)))
+             ((simd-pack (signed-byte 64))
+              (multiple-value-call #'format stream "~S~@{ ~20@D~}" 'simd-pack
+                (%simd-pack-sb64s pack))))))))
 
 #+sb-simd-pack-256
 (defmethod print-object ((pack simd-pack-256) stream)
   (cond ((and *print-readably* *read-eval*)
-         (multiple-value-bind (format maker extractor)
-             (etypecase pack
-               ((simd-pack-256 double-float)
-                (values "#.(~@{~S~^ ~})"
-                        '%make-simd-pack-256-double #'%simd-pack-256-doubles))
-               ((simd-pack-256 single-float)
-                (values "#.(~@{~S~^ ~})"
-                        '%make-simd-pack-256-single #'%simd-pack-256-singles))
-               (t
-                (values "#.(~S~@{ #X~16,'0X~})"
-                        '%make-simd-pack-256-ub64 #'%simd-pack-256-ub64s)))
-           (multiple-value-call
-               #'format stream format maker (funcall extractor pack))))
+         (format stream "#.(~S #b~3,'0B #x~16,'0D #x~16,'0D #x~16,'0D #x~16,'0D)"
+                 '%make-simd-pack-256
+                 (%simd-pack-256-tag pack)
+                 (%simd-pack-256-0 pack)
+                 (%simd-pack-256-1 pack)
+                 (%simd-pack-256-2 pack)
+                 (%simd-pack-256-3 pack)))
         (*print-readably*
          (print-not-readable-error pack stream))
         (t
          (print-unreadable-object (pack stream)
            (etypecase pack
              ((simd-pack-256 double-float)
-              (multiple-value-call #'format stream "~S~@{ ~,13E~}"
-                'simd-pack-256
+              (multiple-value-call #'format stream "~S~@{ ~,13E~}" 'simd-pack-256
                 (%simd-pack-256-doubles pack)))
              ((simd-pack-256 single-float)
-              (multiple-value-call #'format stream "~S~@{ ~,7E~}"
-                'simd-pack-256
+              (multiple-value-call #'format stream "~S~@{ ~,7E~}" 'simd-pack-256
                 (%simd-pack-256-singles pack)))
-             (t
-              (multiple-value-bind (p0 p1 p2 p3)
-                  (%simd-pack-256-ub64s pack)
-                (format stream "~S~@{ ~16,'0X~}"
-                        'simd-pack-256
-                        p0 p1 p2 p3))))))))
+             ((simd-pack-256 (unsigned-byte 8))
+              (multiple-value-call #'format stream "~S~@{ ~3D~}" 'simd-pack-256
+                (%simd-pack-256-ub8s pack)))
+             ((simd-pack-256 (unsigned-byte 16))
+              (multiple-value-call #'format stream "~S~@{ ~5D~}" 'simd-pack-256
+                (%simd-pack-256-ub16s pack)))
+             ((simd-pack-256 (unsigned-byte 32))
+              (multiple-value-call #'format stream "~S~@{ ~10D~}" 'simd-pack-256
+                (%simd-pack-256-ub32s pack)))
+             ((simd-pack-256 (unsigned-byte 64))
+              (multiple-value-call #'format stream "~S~@{ ~20D~}" 'simd-pack-256
+                (%simd-pack-256-ub64s pack)))
+             ((simd-pack-256 (signed-byte 8))
+              (multiple-value-call #'format stream "~S~@{ ~4@D~}" 'simd-pack-256
+                (%simd-pack-256-sb8s pack)))
+             ((simd-pack-256 (signed-byte 16))
+              (multiple-value-call #'format stream "~S~@{ ~6@D~}" 'simd-pack-256
+                (%simd-pack-256-sb16s pack)))
+             ((simd-pack-256 (signed-byte 32))
+              (multiple-value-call #'format stream "~S~@{ ~11@D~}" 'simd-pack-256
+                (%simd-pack-256-sb32s pack)))
+             ((simd-pack-256 (signed-byte 64))
+              (multiple-value-call #'format stream "~S~@{ ~20@D~}" 'simd-pack-256
+                (%simd-pack-256-sb64s pack))))))))
 
 ;;;; functions
 
 (defmethod print-object ((object function) stream)
+  (macrolet ((unprintable-instance-p (x)
+               ;; Guard against calling %FUN-FUN if it would return 0.
+               ;; %FUNCALLABLE-INSTANCE-FUN is known to return FUNCTION so determining
+               ;; whether it is actually assigned requires a low-level trick.
+               (let ((s (sb-vm::primitive-object-slot
+                         (sb-vm:primitive-object 'funcallable-instance)
+                         'function)))
+                 `(and (funcallable-instance-p ,x)
+                       (eql 0 (%primitive sb-alien:slot ,x 'function ,(sb-vm:slot-offset s)
+                                          ,sb-vm:fun-pointer-lowtag))))))
+    (when (unprintable-instance-p object)
+      (return-from print-object
+        (print-unreadable-object (object stream :type t :identity t)))))
   (let* ((name (%fun-name object))
          (proper-name-p (and (legal-fun-name-p name) (fboundp name)
                              (eq (fdefinition name) object))))
@@ -2017,16 +2079,20 @@ variable: an unreadable object representing the error is printed instead.")
 
 ;;;; catch-all for unknown things
 
-(declaim (inline lowtag-of))
-(defun lowtag-of (x) (logand (get-lisp-obj-address x) sb-vm:lowtag-mask))
-
 (defmethod print-object ((object t) stream)
   (when (eq object sb-pcl:+slot-unbound+)
     ;; If specifically the unbound marker with 0 data,
     ;; as opposed to any other unbound marker.
     (print-unreadable-object (object stream) (write-string "unbound" stream))
     (return-from print-object))
-  (when (eql (get-lisp-obj-address object) sb-vm:no-tls-value-marker-widetag)
+  ;; NO-TLS-VALUE was added here as a printable object type for #+ubsan which,
+  ;; among other things, detects read-before-write on a per-array-element basis.
+  ;; Git rev 22d8038118 caused uninitialized SIMPLE-VECTORs to get prefilled
+  ;; with NO_TLS_VALUE_MARKER, but a better choice would be
+  ;; (logior (mask-field (byte (- n-word-bits 8) 8) -1) unbound-marker-widetag).
+  ;; #+ubsan has probably bitrotted for other reasons, so this is untested.
+  #+ubsan
+  (when (eql (get-lisp-obj-address object) unwritten-vector-element-marker)
     (print-unreadable-object (object stream) (write-string "novalue" stream))
     (return-from print-object))
   (print-unreadable-object (object stream :identity t)
@@ -2036,8 +2102,9 @@ variable: an unreadable object representing the error is printed instead.")
          (let ((widetag (widetag-of object)))
            (case widetag
              (#.sb-vm:value-cell-widetag
-              (write-string "value cell " stream)
+              (write-string "value-cell " stream)
               (output-object (value-cell-ref object) stream))
+             #+nil
              (#.sb-vm:filler-widetag
               (write-string "pad " stream)
               (write (1+ (get-header-data object)) :stream stream)
