@@ -20,6 +20,9 @@
 
 (defvar *readtable*)
 
+(define-load-time-global *standard-readtable* (make-readtable))
+(declaim (readtable *standard-readtable*))
+
 (defun !readtable-cold-init ()
   (setq *empty-extended-char-table* (make-hash-table :rehash-size 1 :test #'eq)
         *readtable* (make-readtable)))
@@ -40,16 +43,16 @@
 
 ;;;; reader errors
 
-(defun reader-eof-error (stream context)
+(define-error-wrapper reader-eof-error (stream context)
   ;; Don't worry if STREAM isn't a valid stream; it's not a reason to fail now.
-  (declare (explicit-check) (optimize allow-non-returning-tail-call))
+  (declare (explicit-check))
   (error 'reader-eof-error
          :stream stream
          :context context))
 
-(defun simple-reader-error (stream control &rest args)
+(define-error-wrapper simple-reader-error (stream control &rest args)
   ;; Don't worry if STREAM isn't a valid stream; it's not a reason to fail now.
-  (declare (explicit-check) (optimize allow-non-returning-tail-call))
+  (declare (explicit-check))
   (error 'simple-reader-error
          :stream stream
          :format-control control
@@ -58,23 +61,26 @@
 ;;;; macros and functions for character tables
 
 ;;; As efficiently as possible look up the character attributes of CHAR.
-;;; If CAST is true then insert an assertion that CHARACTER is actually
-;;; a character. This seems to be necessary to ensure that gray stream
-;;; methods that should return characters don't return random objects.
-(defmacro char-syntax (char syntax-array extension-table &optional cast)
-  #-sb-unicode (declare (ignore cast))
-  `(progn
-     #-sb-unicode ,extension-table
-     (truly-the (unsigned-byte 8)
-                #-sb-unicode
-                (elt ,syntax-array (char-code ,char))
-                #+sb-unicode
-                (if (typep ,char 'base-char)
-                    (elt ,syntax-array (char-code ,char))
-                    (car (truly-the list
-                                    (gethash ,(if cast `(the character ,char) char)
-                                             ,extension-table
-                                             ',`(,+char-attr-constituent+))))))))
+;;; If TYPECHECK then insert an assertion that CHAR is actually a character,
+;;; which is needed sometimes when calling Gray stream methods that could return
+;;; random objects which must be of type CHARACTER.
+;;; (Naturally GETHASH doesn't care what its argument is.)
+(defmacro char-syntax (char syntax-array extension-table &optional typecheck)
+  (declare (ignorable extension-table typecheck))
+  `(truly-the (unsigned-byte 8)
+      ;; Unfortunately we require divergent code here for unicode/non-unicode builds.
+      ;; Non-unicode will deduce that the 'else' branch of the IF can't possibly satisfy
+      ;; the typecheck in `(THE CHARACTER) because CHARACTER = BASE-CHAR, and complains.
+      #+sb-unicode
+      (if (typep ,char 'base-char)
+          (elt ,syntax-array (char-code ,char))
+          (let ((found (gethash ,char ,extension-table +char-attr-constituent+)))
+            (cond ((fixnump found)
+                   ,@(if typecheck `((the character ,char))) ; a "drive-by" typecheck
+                   found)
+                  (t
+                   (car (truly-the cons found))))))
+      #-sb-unicode (elt ,syntax-array (char-code ,char))))
 
 ;;; Formerly known as SET-CAT-ENTRY and SET-CMT-ENTRY
 (defun assign-char-syntax (readtable char attributes function)
@@ -145,15 +151,18 @@
 (declaim (inline single-escape-p multiple-escape-p))
 (declaim (inline token-delimiterp))
 
-;;; the [1] and [2] here refer to ANSI glossary entries for
-;;; "whitespace".
-;; whitespace[2]p is the only predicate whose readtable is optional
-;; - other than whitespace[1]p which has a fixed readtable - due to
-;; callers not otherwise needing a readtable at all, and so not binding
-;; *READTABLE* into a local variable throughout their lifetime.
+;;; the [1] and [2] here refer to ANSI glossary entries for "whitespace".
 (defun whitespace[1]p (char)
-  (test-attribute char +char-attr-whitespace+ *standard-readtable*))
-(defun whitespace[2]p (char &optional (rt *readtable*))
+  ;; From *standard-readtable*
+  (case (char-code char)
+    (#.(list tab-char-code
+             (char-code #\Newline)
+             (char-code #\Space)
+             form-feed-char-code
+             return-char-code)
+     t)))
+
+(defun whitespace[2]p (char rt)
   (test-attribute char +char-attr-whitespace+ rt t))
 
 (defun constituentp (char rt)
@@ -450,12 +459,11 @@ standard Lisp readtable when NIL."
 
 (defconstant +EOF+ 0)
 
-(defun flush-whitespace (stream)
+(defun flush-whitespace (stream readtable)
   ;; This flushes whitespace chars, returning the last char it read (a
   ;; non-white one). It always gets an error on end-of-file.
-  (let* ((rt *readtable*)
-         (base (base-char-syntax-array rt))
-         (extended (extended-char-table rt)))
+  (let* ((base (base-char-syntax-array readtable))
+         #+sb-unicode (extended (extended-char-table readtable)))
     (macrolet ((done-p ()
                 '(neq (char-syntax char base extended) +char-attr-whitespace+)))
       (if (ansi-stream-p stream)
@@ -693,17 +701,17 @@ standard Lisp readtable when NIL."
 ;;; Like READ-PRESERVING-WHITESPACE, but doesn't check the read buffer
 ;;; for being set up properly.
 (defun %read-preserving-whitespace (stream eof-error-p eof-value recursive-p)
-  (declare (optimize (sb-c::check-tag-existence 0)))
   (if recursive-p
       ;; a loop for repeating when a macro returns nothing
       (let* ((tracking-p (form-tracking-stream-p stream))
+             (rt *readtable*)
              (outermost-p
               (and tracking-p
                    (null (form-tracking-stream-form-start-char-pos stream)))))
         (loop
          (let ((char (read-char stream eof-error-p +EOF+)))
            (cond ((eq char +EOF+) (return eof-value))
-                 ((whitespace[2]p char))
+                 ((whitespace[2]p char rt))
                  (t
                   (when outermost-p
                     ;; Calling FILE-POSITION at each token seems to slow down
@@ -780,7 +788,7 @@ standard Lisp readtable when NIL."
     (unless (or (eql result local-eof-val) recursive-p)
       (let ((next-char (read-char stream nil +EOF+)))
         (unless (or (eq next-char +EOF+)
-                    (whitespace[2]p next-char))
+                    (whitespace[2]p next-char *readtable*))
           (unread-char next-char stream))))
     (if (eq result local-eof-val) eof-value result)))
 
@@ -825,9 +833,10 @@ standard Lisp readtable when NIL."
     ((with-list-reader ((streamvar delimiter) &body body)
        `(let* ((thelist (list nil))
                (listtail thelist)
+               (rt *readtable*)
                (collectp (if *read-suppress* 0 -1)))
           (declare (dynamic-extent thelist))
-          (loop (let ((firstchar (flush-whitespace ,streamvar)))
+          (loop (let ((firstchar (flush-whitespace ,streamvar rt)))
                   (when (eq firstchar ,delimiter)
                     (return (cdr thelist)))
                   ,@body))))
@@ -850,8 +859,8 @@ standard Lisp readtable when NIL."
                         (unless (zerop collectp)
                           (simple-reader-error
                            stream "Nothing appears before . in list.")))
-                       ((whitespace[2]p nextchar)
-                        (setq nextchar (flush-whitespace stream))))
+                       ((whitespace[2]p nextchar rt)
+                        (setq nextchar (flush-whitespace stream rt))))
                  (rplacd (truly-the cons listtail)
                          (read-after-dot stream nextchar collectp))
                  ;; Check for improper ". ,@" or ". ,." now rather than
@@ -889,8 +898,9 @@ standard Lisp readtable when NIL."
 
 (defun read-after-dot (stream firstchar collectp)
   ;; FIRSTCHAR is non-whitespace!
-  (let ((lastobj ()))
-    (do ((char firstchar (flush-whitespace stream)))
+  (let ((lastobj ())
+        (rt *readtable*))
+    (do ((char firstchar (flush-whitespace stream rt)))
         ((eq char #\))
          (if (zerop collectp)
              (return-from read-after-dot nil)
@@ -901,7 +911,7 @@ standard Lisp readtable when NIL."
     ;; At least one thing appears after the dot.
     ;; Check for more than one thing following dot.
     (loop
-     (let ((char (flush-whitespace stream)))
+     (let ((char (flush-whitespace stream rt)))
        (cond ((eq char #\)) (return lastobj)) ;success!
              ;; Try reading virtual whitespace.
              ((not (zerop (logand (read-maybe-nothing stream char)
@@ -1225,7 +1235,7 @@ extended <package-name>::<form-in-package> syntax."
   (let* ((rt *readtable*)
          (base *read-base*)
          (attribute-array (base-char-syntax-array rt))
-         (attribute-hash-table (extended-char-table rt))
+         #+sb-unicode (attribute-hash-table (extended-char-table rt))
          (buf *read-buffer*)
          (package-designator nil)
          (colons 0)
@@ -1795,8 +1805,7 @@ extended <package-name>::<form-in-package> syntax."
 
 ;;;; General reader for dispatch macros
 
-(defun dispatch-char-error (stream sub-char ignore)
-  (declare (optimize allow-non-returning-tail-call))
+(define-error-wrapper dispatch-char-error (stream sub-char ignore)
   (declare (ignore ignore))
   (if *read-suppress*
       ;; This seems dubious. For comparison's sake, other implementations
@@ -1868,62 +1877,85 @@ extended <package-name>::<form-in-package> syntax."
 
 ;;;; PARSE-INTEGER
 
-(defun parse-integer (string &key (start 0) end (radix 10) junk-allowed)
-  "Examine the substring of string delimited by start and end
+(macrolet ((def (radix)
+             `(flet ((parse-error (format-control)
+                       (error 'simple-parse-error
+                              :format-control format-control
+                              :format-arguments (list string))))
+                (with-array-data ((string string :offset-var offset)
+                                  (start start)
+                                  (end end)
+                                  :check-fill-pointer t)
+                  (let ((radix ,(or radix 'radix))
+                        (index (do ((i start (1+ i)))
+                                   ((= i end)
+                                    (if junk-allowed
+                                        (return-from ,(symbolicate 'parse-integer
+                                                                   (if radix (princ-to-string radix) ""))
+                                          (values nil end))
+                                        (parse-error "no non-whitespace characters in string ~S.")))
+                                 (declare (fixnum i))
+                                 (unless (whitespace[1]p (char string i)) (return i))))
+                        (minusp nil)
+                        (found-digit nil))
+                    (declare (fixnum index)
+                             (inline digit-char-p))
+                    (let ((char (char string index)))
+                      (cond ((char= char #\-)
+                             (setq minusp t)
+                             (incf index))
+                            ((char= char #\+)
+                             (incf index))))
+                    (let ((final-result 0))
+                      (macrolet ((compute (type)
+                                   `(let ((result 0))
+                                      (declare (type ,type result))
+                                      (loop
+                                       (when (= index end) (return nil))
+                                       (let* ((char (char string index))
+                                              (weight (digit-char-p char radix)))
+                                         (cond (weight
+                                                (setq result (truly-the ,type
+                                                                        (+ weight
+                                                                           (truly-the ,type (* result radix))))
+                                                      found-digit t))
+                                               (junk-allowed (return nil))
+                                               ((whitespace[1]p char)
+                                                (loop
+                                                 (incf index)
+                                                 (when (= index end) (return))
+                                                 (unless (whitespace[1]p (char string index))
+                                                   (parse-error "junk in string ~S")))
+                                                (return nil))
+                                               (t
+                                                (parse-error "junk in string ~S"))))
+                                       (incf index))
+                                      (setf final-result
+                                            (if minusp
+                                                (- result)
+                                                result)))))
+                        ,(if radix
+                             `(if (<= (- end index) (floor (log most-positive-word ,radix)))
+                                  (compute word)
+                                  (compute t))
+                             `(compute t)))
+                      (values
+                       (if found-digit
+                           final-result
+                           (if junk-allowed
+                               nil
+                               (parse-error "no digits in string ~S")))
+                       (- index offset))))))))
+  (defun parse-integer (string &key (start 0) end (radix 10) junk-allowed)
+    "Examine the substring of string delimited by start and end
   (default to the beginning and end of the string)  It skips over
   whitespace characters and then tries to parse an integer. The
   radix parameter must be between 2 and 36."
-  (flet ((parse-error (format-control)
-           (declare (optimize allow-non-returning-tail-call))
-           (error 'simple-parse-error
-                  :format-control format-control
-                  :format-arguments (list string))))
-    (with-array-data ((string string :offset-var offset)
-                      (start start)
-                      (end end)
-                      :check-fill-pointer t)
-      (let ((index (do ((i start (1+ i)))
-                       ((= i end)
-                        (if junk-allowed
-                            (return-from parse-integer (values nil end))
-                            (parse-error "no non-whitespace characters in string ~S.")))
-                     (declare (fixnum i))
-                     (unless (whitespace[1]p (char string i)) (return i))))
-            (minusp nil)
-            (found-digit nil)
-            (result 0))
-        (declare (fixnum index))
-        (let ((char (char string index)))
-          (cond ((char= char #\-)
-                 (setq minusp t)
-                 (incf index))
-                ((char= char #\+)
-                 (incf index))))
-        (loop
-         (when (= index end) (return nil))
-         (let* ((char (char string index))
-                (weight (digit-char-p char radix)))
-           (cond (weight
-                  (setq result (+ weight (* result radix))
-                        found-digit t))
-                 (junk-allowed (return nil))
-                 ((whitespace[1]p char)
-                  (loop
-                   (incf index)
-                   (when (= index end) (return))
-                   (unless (whitespace[1]p (char string index))
-                     (parse-error "junk in string ~S")))
-                  (return nil))
-                 (t
-                  (parse-error "junk in string ~S"))))
-         (incf index))
-        (values
-         (if found-digit
-             (if minusp (- result) result)
-             (if junk-allowed
-                 nil
-                 (parse-error "no digits in string ~S")))
-         (- index offset))))))
+    (def nil))
+  (defun parse-integer10 (string start end junk-allowed)
+    (def 10))
+  (defun parse-integer16 (string start end junk-allowed)
+    (def 16)))
 
 ;;;; reader initialization code
 
