@@ -53,8 +53,7 @@
    for the new values, the setting function, and the accessing function."
   (named-let retry ((form form))
     (labels ((newvals (count)
-               (let ((*gensym-counter* 1))
-                 (make-gensym-list count "NEW")))
+               (make-gensym-list count "NEW"))
              ;; Produce the expansion of a SETF form that calls either
              ;; #'(SETF name) or an inverse given by short form DEFSETF.
              (call (call arg-maker &aux (vals (newvals 1)))
@@ -103,10 +102,8 @@
                       (vals (newvals 1)))
                   (return-from retry
                     (values (list instance) (list (cadr form)) vals
-                            (slot-access-transform
-                             :setf (list instance (car vals)) it)
-                            (slot-access-transform
-                             :read (list instance) it))))))
+                            (slot-access-transform :setf (list instance (car vals)) it)
+                            (slot-access-transform :read (list instance) it))))))
             (multiple-value-bind (expansion expanded)
                 (%macroexpand-1 form environment)
               (if expanded
@@ -186,8 +183,7 @@
               (return-from setf `(,(car inverse) ,@(cdr place) ,value-form))))
           (awhen (transformable-struct-setf-p place env)
             (return-from setf
-              (slot-access-transform
-               :setf (list (cadr place) value-form) it)))))
+              (slot-access-transform :setf (list (cadr place) value-form) it)))))
 
       (multiple-value-bind (temps vals newval setter)
           (get-setf-expansion place env)
@@ -220,7 +216,7 @@
                                 (thunk (cdr mv-bindings) (cdr getters) setters))
                    setters)))
       (let ((outputs (loop for i below (length (car (mv-bindings)))
-                           collect (sb-xc:gensym "OUT"))))
+                           collect (gensym "OUT"))))
         (car (gen-let* (reduce #'nconc (let*-bindings))
                        (gen-mv-bind outputs (car (getters))
                                     (thunk (mv-bindings) (cdr (getters))
@@ -403,7 +399,7 @@
 
 (sb-xc:defmacro define-modify-macro (name lambda-list function &optional doc-string)
   "Creates a new read-modify-write macro like PUSH or INCF."
-  (check-designator name define-modify-macro)
+  (check-designator name 'define-modify-macro)
   (binding* (((nil required optional rest)
               (parse-lambda-list
                lambda-list
@@ -460,7 +456,8 @@
 (sb-xc:defmacro defsetf (access-fn &rest rest)
   "Associates a SETF update function or macro with the specified access
   function or macro. The format is complex. See the manual for details."
-  (check-designator access-fn defsetf "access-function")
+  (check-designator access-fn 'defsetf
+      #'symbolp "symbol" "access-function")
   (typecase rest
     ((cons (and symbol (not null)) (or null (cons string null)))
      `(eval-when (:load-toplevel :compile-toplevel :execute)
@@ -578,7 +575,8 @@
 (sb-xc:defmacro define-setf-expander (access-fn lambda-list &body body)
   "Syntax like DEFMACRO, but creates a setf expander function. The body
   of the definition must be a form that returns five appropriate values."
-  (check-designator access-fn define-setf-expander "access-function")
+  (check-designator access-fn 'define-setf-expander
+      #'symbolp "symbol" "access-function")
   `(eval-when (:compile-toplevel :load-toplevel :execute)
      (%defsetf ',access-fn
                ,(make-macro-lambda `(setf-expander ,access-fn) lambda-list body
