@@ -78,8 +78,6 @@
     (def-accessor sap-ref-single)
     (def-accessor sap-ref-double))
   (def %byte-blt (src src-start dst dst-start count))
-  (def shift-towards-start (number count))
-  (def shift-towards-end (number count))
   (def get-header-data)
   (def set-header-data (x val))
   (def widetag-of)
@@ -113,7 +111,6 @@
   (def %caller-pc ())
   (def %code-debug-info)
   (def sb-vm::%code-fixups)
-  #+x86-64 (def pointerp)
   (def sb-bignum:%bignum-length)
 
   ;; instances
@@ -168,9 +165,6 @@
   (def symbol-package-id)
   (def symbol-hash)
   (def symbol-%info) ; primitive reader always needs a stub
-  ;; but the "wrapped" reader might not need a stub.
-  ;; If it's already a proper function, then it doesn't.
-  #.(if (fboundp 'symbol-dbinfo) (values) '(def symbol-dbinfo))
   #-(or x86 x86-64) (def lra-code-header)
   (def %make-lisp-obj)
   (def get-lisp-obj-address)
@@ -201,37 +195,11 @@
   "Hints the processor that the current thread is spin-looping."
   (spin-loop-hint))
 
-;;; The stub for sb-c::%structure-is-a should really use layout-id in the same way
-;;; that the vop does, however, because the all 64-bit architectures other than
-;;; x86-64 need to use with-pinned-objects to extract a layout-id, it is cheaper not to.
-;;; I should add a vop for uint32 access to raw slots.
-(defun sb-c::%structure-is-a (object-layout test-layout)
-  (or (eq object-layout test-layout)
-      (let ((depthoid (layout-depthoid test-layout))
-            (inherits (layout-inherits object-layout)))
-        (and (> (length inherits) depthoid)
-             (eq (svref inherits depthoid) test-layout)))))
-
-(defun sb-c::structure-typep (object test-layout)
-  (and (%instancep object)
-       (let ((object-layout (%instance-layout object)))
-        (or (eq object-layout test-layout)
-            (let ((depthoid (layout-depthoid test-layout))
-                  (inherits (layout-inherits object-layout)))
-              (and (> (length inherits) depthoid)
-                   (eq (svref inherits depthoid) test-layout)))))))
-
 (defun %other-pointer-subtype-p (x choices)
   (and (%other-pointer-p x)
        (member (%other-pointer-widetag x) choices)
        t))
 
-#+x86-64
-(defun symbol-hash* (x satisfies)
-  (declare (explicit-check)) ; actually, not
-  (declare (ignore satisfies))
-  (symbol-hash* x nil))
-
 ;;; TYPECASE could expand to contain a call to this function.
 ;;; The interpreter can ignore it, it is just compiler magic.
 (defun sb-c::%type-constraint (var type)
