@@ -13,6 +13,23 @@
 
 ;;;; structure frobbing primitives
 
+#+permgen
+(defun allocate-permgen-layout (nwords)
+  (flet ((thunk ()
+           (let ((freeptr sb-vm:*permgen-space-free-pointer*))
+             (setf sb-vm:*permgen-space-free-pointer*
+                   ;; round-to-odd, add the header word
+                   (sap+ freeptr (ash (1+ (logior nwords 1)) sb-vm:word-shift)))
+             (aver (<= (sap-int sb-vm:*permgen-space-free-pointer*)
+                       (+ sb-vm:permgen-space-start sb-vm:permgen-space-size)))
+             (setf (sap-ref-word freeptr 0)
+                   (logior (ash nwords sb-vm:instance-length-shift)
+                           sb-vm:instance-widetag))
+             (%make-lisp-obj (sap-int (sap+ freeptr sb-vm:instance-pointer-lowtag))))))
+    (if (sb-thread::mutex-p sb-vm::*allocator-mutex*)
+        (with-system-mutex (sb-vm::*allocator-mutex* :without-gcing t) (thunk))
+        (thunk))))
+
 ;;; For lack of any better to place to write up some detail surrounding
 ;;; layout creation for structure types, I'm putting here.
 ;;; When you issue a DEFSTRUCT at the REPL, there are *three* instances
@@ -28,20 +45,13 @@
 ;;; which contains copies of the length/depthoid/inherits etc
 ;;; that we compare against the installed one to make sure they match.
 ;;; The third one also gets thrown away.
-(define-load-time-global *layout-id-generator* (cons 0 nil))
-(declaim (type (cons fixnum) *layout-id-generator*))
-;;; NB: for #+metaspace this returns a WRAPPER, not a LAYOUT.
 (defun make-layout (clos-hash classoid
                     &key (depthoid -1) (length 0) (flags 0)
                          (inherits #())
                          (info nil)
                          (bitmap (if info (dd-bitmap info) 0))
-                         (invalid :uninitialized)
-                    &aux (id (or (atomic-pop (cdr *layout-id-generator*))
-                                 (atomic-incf (car *layout-id-generator*)))))
-  (unless (typep id '(and layout-id (not (eql 0))))
-    (error "Layout ID limit reached"))
-  (let* ((fixed-words (type-dd-length sb-vm:layout))
+                         (invalid :uninitialized))
+  (let* ((fixed-words (type-dd-length layout))
          (extra-id-words ; count of additional words needed to store ancestors
           (if (logtest flags +structure-layout-flag+)
               (calculate-extra-id-words depthoid)
@@ -49,80 +59,86 @@
          (bitmap-words (ceiling (1+ (integer-length bitmap)) sb-vm:n-word-bits))
          (nwords (+ fixed-words extra-id-words bitmap-words))
          (layout
-          (truly-the sb-vm:layout
-                     #-(or metaspace immobile-space) (%make-instance/mixed nwords)
-                     #+metaspace (sb-vm::allocate-metaspace-chunk (1+ nwords))
-                     #+(and immobile-space (not metaspace))
-                     (sb-vm::alloc-immobile-fixedobj
-                      (1+ nwords)
-                      (logior (ash nwords sb-vm:instance-length-shift)
-                              sb-vm:instance-widetag))))
-         (wrapper #-metaspace layout))
-    (%set-instance-layout layout
-          (wrapper-friend #.(find-layout #+metaspace 'sb-vm:layout
-                                         #-metaspace 'wrapper)))
-    #+metaspace
-    (setf wrapper (%make-wrapper :clos-hash clos-hash :classoid classoid
-                                 :%info info :invalid invalid :friend layout)
-          (layout-friend layout) wrapper)
+          (truly-the layout
+                     #+compact-instance-header
+                     (progn
+                       #+permgen (allocate-permgen-layout nwords)
+                       #+immobile-space
+                       (sb-vm::alloc-immobile-fixedobj
+                        (1+ nwords)
+                        (logior (ash nwords sb-vm:instance-length-shift)
+                                sb-vm:instance-widetag)))
+                     #-compact-instance-header
+                     (%make-instance/mixed nwords))))
+    (%set-instance-layout layout #.(find-layout 'layout))
     (setf (layout-flags layout) #+64-bit (pack-layout-flags depthoid length flags)
                                 #-64-bit flags)
     (setf (layout-clos-hash layout) clos-hash
-          (wrapper-classoid wrapper) classoid
-          (wrapper-invalid wrapper) invalid)
-    #-64-bit (setf (wrapper-depthoid wrapper) depthoid
-                   (wrapper-length wrapper) length)
-    #-metaspace (setf (wrapper-%info wrapper) info ; already set if #+metaspace
-                      (wrapper-slot-table wrapper) #(1 nil))
-    (set-layout-inherits wrapper inherits (logtest flags +structure-layout-flag+) id)
+          (layout-classoid layout) classoid
+          (layout-invalid layout) invalid)
+    #-64-bit (setf (layout-depthoid layout) depthoid
+                   (layout-length layout) length)
+    (setf (layout-%info layout) info
+          (layout-slot-table layout) #(1 nil))
+    ;; All layouts initially have an ID of 0. The ID is assigned on demand
+    ;; for the layouts that actually get registered, and only those.
+    ;; We do not need to or want to recycle IDs, because:
+    ;; - assembly code can use layout-ID fixups without reference to the layout.
+    ;;   So GCing a layout should *not* allow reuse of its ID because we have no
+    ;;   way of knowing where it was wired in.
+    ;; - only subtypes of STRUCTURE-OBJECT need an ID for TYPEP, and structures
+    ;;   are not redefinable, so you'd have to define 2^32 different structure
+    ;;   types to exhaust the space of IDS.
+    (set-layout-inherits layout inherits (logtest flags +structure-layout-flag+) 0)
     (let ((bitmap-base (+ fixed-words extra-id-words)))
       (dotimes (i bitmap-words)
         (%raw-instance-set/word layout (+ bitmap-base i)
               (ldb (byte sb-vm:n-word-bits (* i sb-vm:n-word-bits)) bitmap))))
     (aver (= (%layout-bitmap layout) bitmap)) ; verify it reads back the same
-    ;; It's not terribly important that we recycle layout IDs, but I have some other
-    ;; changes planned that warrant a finalizer per layout.
-    ;; FIXME: structure IDs should never be recycled because code blobs referencing
-    ;; th ID do not reference the layout, and so the layout could be GCd allowing
-    ;; reuse of an old ID for a new type.
-    (unless (built-in-classoid-p classoid)
-      (let ((layout-addr (- (get-lisp-obj-address layout) sb-vm:instance-pointer-lowtag)))
-        (declare (ignorable layout-addr))
-        (declare (sb-c::tlab :system))
-        (finalize wrapper
-                  (lambda ()
-                    #+metaspace (sb-vm::unallocate-metaspace-chunk layout-addr)
-                    (atomic-push id (cdr *layout-id-generator*)))
-                :dont-save t)))
-    ;; Rather than add and delete this line of debugging which I've done so many times,
-    ;; let's instead keep it but commented out.
-    #+nil
-    (alien-funcall (extern-alien "printf" (function void system-area-pointer unsigned unsigned
-                                                    unsigned system-area-pointer))
-                   (vector-sap #.(format nil "New wrapper ID=%d %p %p '%s'~%"))
-                   id (get-lisp-obj-address layout) (get-lisp-obj-address wrapper)
-                   (vector-sap (string (classoid-name classoid))))
-    wrapper))
+    layout))
 
-#+metaspace
-(defun make-temporary-wrapper (clos-hash classoid inherits)
-  (let* ((layout (%make-temporary-layout #.(find-layout t) clos-hash))
-         (wrapper (%make-wrapper :clos-hash clos-hash :classoid classoid
-                                 :inherits inherits :invalid nil
-                                 :friend layout)))
-    (setf (layout-friend layout) wrapper)
-    wrapper))
+;;; Extract the bitmap from 1 or more words of bits that have the same format
+;;; as a BIGNUM - least-significant word first, native bit order within each word,
+;;; all but the last are unsigned, and the last is signed.
+(defun %layout-bitmap (layout &aux (nwords (bitmap-nwords layout))
+                                   (start (bitmap-start layout)))
+  (declare (type layout layout))
+  (if (and (= nwords 1) (fixnump (%raw-instance-ref/signed-word layout start)))
+      (the fixnum (%raw-instance-ref/signed-word layout start))
+      (do* ((res (sb-bignum:%allocate-bignum nwords))
+            (i 0 (1+ i))
+            (j start (1+ j)))
+           ((= i nwords)
+            ;; not sure if this NORMALIZE call is needed, but it can't hurt
+            (sb-bignum::%normalize-bignum res nwords))
+       (setf (sb-bignum:%bignum-ref res i) (%raw-instance-ref/word layout j)))))
+
+(defun layout-bitmap (layout)
+  (declare (type layout layout))
+  ;; Whenever we call LAYOUT-BITMAP on a structure-object subtype,
+  ;; it's supposed to have the INFO slot populated, linking it to a defstruct-description.
+  (acond ((layout-info layout) (dd-bitmap it))
+         ;; Instances lacking DD-INFO are CLOS objects, which can't generally have
+         ;; raw slots, except that funcallable-instances may have 2 raw slots -
+         ;; the trampoline and the layout. The trampoline can have a tag, depending
+         ;; on the platform, and the layout is tagged but a special case.
+         ;; In any event, the bitmap is always 1 word, and there are no "extra ID"
+         ;; words preceding it.
+         (t
+          (aver (not (logtest +structure-layout-flag+ (layout-flags layout))))
+          (the fixnum
+               (%raw-instance-ref/signed-word layout (type-dd-length layout))))))
 
 ;;; This allocator is used by the expansion of MAKE-LOAD-FORM-SAVING-SLOTS
 ;;; when given a STRUCTURE-OBJECT.
 (defun allocate-struct (type)
-  (let* ((wrapper (classoid-wrapper (the structure-classoid (find-classoid type))))
-         (dd (wrapper-dd wrapper))
-         (structure (let ((len (wrapper-length wrapper)))
+  (let* ((layout (classoid-layout (the structure-classoid (find-classoid type))))
+         (dd (layout-dd layout))
+         (structure (let ((len (layout-length layout)))
                       (if (dd-has-raw-slot-p dd)
                           (%make-instance/mixed len)
                           (%make-instance len)))))
-    (%set-instance-layout structure wrapper)
+    (%set-instance-layout structure layout)
     (dolist (dsd (dd-slots dd) structure)
       (when (eq (dsd-raw-type dsd) 't)
         (%instance-set structure (dsd-index dsd) (make-unbound-marker))))))
@@ -159,32 +175,46 @@
                        sb-kernel::*raw-slot-data*))))
   (define-raw-slot-accessors))
 
+(define-load-time-global *layout-id-generator* (cons 0  nil))
+(declaim (type (cons fixnum) *layout-id-generator*))
+(define-load-time-global *layout-id-mutex* (sb-thread:make-mutex :name "LAYOUT-ID"))
 (macrolet ((id-bits-sap ()
              `(sap+ (int-sap (get-lisp-obj-address layout))
-                    ,(- (sb-vm::id-bits-offset) sb-vm:instance-pointer-lowtag))))
-(defun layout-id (layout)
+                    ,(- (sb-vm::id-bits-offset) sb-vm:instance-pointer-lowtag)))
+           (access-it ()
+             #-64-bit `(%raw-instance-ref/signed-word
+                        layout (+ (get-dsd-index layout id-word0) index))
+             ;; use SAP-ref, for lack of half-sized slots
+             #+64-bit `(signed-sap-ref-32 (id-bits-sap) (ash index 2))))
+(defun layout-id (layout &optional (assign t))
   ;; If a structure type at depthoid >= 2, then fetch the INDEXth id
   ;; where INDEX is depthoid - 2. Otherwise fetch the 0th id.
   ;; There are a few non-structure types at positive depthoid; those do not store
   ;; their ancestors in the vector; they only store self-id at index 0.
   ;; This isn't performance-critical. If it were, then we should store self-ID
   ;; at a fixed index. Using it for type-based dispatch remains a possibility.
-  (let* ((layout (cond #+metaspace ((typep layout 'wrapper) (wrapper-friend layout))
-                       (t layout)))
-         (depth (- (sb-vm::layout-depthoid layout) 2))
+  (let* ((depth (- (sb-vm::layout-depthoid layout) 2))
          (index (if (or (< depth 0) (not (logtest (layout-flags layout)
                                                   +structure-layout-flag+)))
-                    0 depth)))
-    (truly-the layout-id
-              #-64-bit (%raw-instance-ref/signed-word
-                        layout (+ (get-dsd-index sb-vm:layout id-word0) index))
-              #+64-bit ; use SAP-ref for lack of half-sized slots
-              (with-pinned-objects (layout)
-                (signed-sap-ref-32 (id-bits-sap) (ash index 2))))))
+                    0 depth))
+         (id (with-pinned-objects (layout)
+               (access-it))))
+    (truly-the
+     (or null layout-id)
+     (cond ((not (zerop id)) id)
+           (assign
+            (aver (logior +structure-layout-flag+ (layout-flags layout)))
+            (with-system-mutex (*layout-id-mutex*)
+              (let ((id (truly-the layout-id (access-it)))) ; double-check
+                (if (zerop id)
+                    (with-pinned-objects (layout)
+                      (setf (access-it)
+                            ;; doesn't really need ATOMIC- any moren
+                            (atomic-incf (car *layout-id-generator*))))
+                    id))))))))
 
-(defun set-layout-inherits (wrapper inherits structurep this-id
-                            &aux (layout (wrapper-friend wrapper)))
-  (setf (wrapper-inherits wrapper) inherits)
+(defun set-layout-inherits (layout inherits structurep this-id)
+  (setf (layout-inherits layout) inherits)
   ;;; If structurep, and *only* if, store all the inherited layout IDs.
   ;;; It looks enticing to try to always store "something", but that goes wrong,
   ;;; because only structure-object layouts are growable, and only structure-object
@@ -247,11 +277,11 @@
 
 ;;; the part of %DEFSTRUCT which makes sense only on the target SBCL
 ;;;
-(defmacro set-wrapper-equalp-impl (wrapper newval)
-  `(%instance-set ,wrapper (get-dsd-index wrapper equalp-impl) ,newval))
+(defmacro set-layout-equalp-impl (layout newval)
+  `(%instance-set ,layout (get-dsd-index layout equalp-impl) ,newval))
 
 (defun assign-equalp-impl (type-name function)
-  (set-wrapper-equalp-impl (find-layout type-name) function))
+  (set-layout-equalp-impl (find-layout type-name) function))
 
 ;;; This variable is just a somewhat hokey way to pass additional
 ;;; arguments to the defstruct hook (which renders the structure definition
@@ -276,8 +306,8 @@
     (setf (documentation (dd-name dd) 'structure) (dd-doc dd)))
 
   (let ((classoid (find-classoid (dd-name dd))))
-    (let ((layout (classoid-wrapper classoid)))
-      (set-wrapper-equalp-impl
+    (let ((layout (classoid-layout classoid)))
+      (set-layout-equalp-impl
           layout
           (cond ((compiled-function-p equalp) equalp)
                 ((eql (dd-bitmap dd) +layout-all-tagged+) #'sb-impl::instance-equalp)
@@ -380,11 +410,11 @@
   (loop for i from sb-vm:instance-data-start below (%instance-length to)
         do (%instance-set to i (%instance-ref from i)))
   to)
-(defun (setf %instance-wrapper) (newval x)
-  (%set-instance-layout x (wrapper-friend newval))
+(defun (setf %instance-layout) (newval x)
+  (%set-instance-layout x newval)
   newval)
-(defun (setf %fun-wrapper) (newval x)
-  (%set-fun-layout x (wrapper-friend newval))
+(defun (setf %fun-layout) (newval x)
+  (%set-fun-layout x newval)
   newval)
 
 ;;; default PRINT-OBJECT method
@@ -443,9 +473,9 @@
   (declare (ignore depth))
   (if (funcallable-instance-p structure)
       (print-unreadable-object (structure stream :identity t :type t))
-      (let* ((wrapper (%instance-wrapper structure))
-             (dd (wrapper-info wrapper))
-             (name (wrapper-classoid-name wrapper)))
+      (let* ((layout (%instance-layout structure))
+             (dd (layout-info layout))
+             (name (layout-classoid-name layout)))
         (cond ((not dd)
                ;; FIXME? this branch may be unnecessary as a consequence
                ;; of change f02bee325920166b69070e4735a8a3f295f8edfd which
@@ -496,22 +526,16 @@
                  (* (+ sb-vm:instance-slots-offset index)
                     sb-vm:n-word-bytes))))))))
 
-#+metaspace
-(defmethod print-object ((self sb-vm:layout) stream)
-  (print-unreadable-object (self stream :type t :identity t)
-    (write (layout-id self) :stream stream)))
-
 (defun id-to-layout (id)
   (maphash (lambda (k v)
              (declare (ignore k))
              (when (eql (layout-id v) id)
-               (return-from id-to-layout (wrapper-friend v))))
+               (return-from id-to-layout v)))
            (classoid-subclasses (find-classoid 't))))
 (export 'id-to-layout)
 
 (defun summarize-layouts ()
-  (flet ((flag-bits (x) (logand (layout-flags (wrapper-friend x))
-                                layout-flags-mask)))
+  (flet ((flag-bits (x) (logand (layout-flags x) layout-flags-mask)))
      (let ((prev -1))
        (dolist (layout (sort (loop for v being each hash-value
                                 of (classoid-subclasses (find-classoid 't))
@@ -522,5 +546,88 @@
              (format t "Layout flags = #b~10,'0b~%" flags)
              (setq prev flags)))
          (format t "  ~a~%" layout)))))
+
+(defun choose-smallest-element-type (seq &key (key #'identity))
+  (cond ((every (lambda (x) (typep (funcall key x) '(unsigned-byte 16))) seq)
+         '(unsigned-byte 16))
+        ((every (lambda (x) (typep (funcall key x) '(unsigned-byte 32))) seq)
+         '(unsigned-byte 32))
+        (t
+         't)))
+
+;;; Return a compiled function that maps keys to values based on SLOTS (an alist)
+;;; where keys are symbols. Though structure slots are stringlike (dups by STRING=
+;;; are disallowed), STANDARD-OBJECT has no such prohibition, so this employs SYMBOL-HASH
+;;; rather than SYMBOL-NAME-HASH to better distinguish slots whose symbol-names match.
+;;; We then attempt to perfectly hash the symbol-hashes.
+;;;
+;;; For 64-bit, SYMBOL-HASH contains excess precision, so it has to be cut down.
+;;; Technically we should pick which bits to feed into the perfect hash generator
+;;; but initially we assume that the low 32 bits are always ok, and that the need
+;;; for collision resolution is so infrequent that rather than resolving it by
+;;; chosing different input bits, the lambda expression wrapped around the
+;;; perfect hash should resolve collisions via another alist.
+(defun make-hash-based-slot-mapper (slots lambda-name)
+  (flet ((hash (s) (ldb (byte 32 0) (symbol-hash s))))
+    (binding* ((symbols (map 'vector #'car slots))
+               (hashes (map '(simple-array (unsigned-byte 32) (*))
+                            #'hash symbols))
+               (unique-hashes (remove-duplicates hashes))
+               (nil
+                (when (< (length unique-hashes) 3)
+                  ;; Return a simple-vector, all symbols first, followed by all data
+                  (let* ((n (length symbols))
+                         (a (make-array (* n 2))))
+                    (loop for j from 0 for slot in slots
+                          do (setf (svref a j) (car slot)
+                                   (svref a (+ n j)) (cdr slot)))
+                    (return-from make-hash-based-slot-mapper a))))
+               ;; power-of-2 sizing generally results in fewer instructions
+               (lexpr (sb-c:make-perfect-hash-lambda unique-hashes nil))
+               (nbuckets (power-of-two-ceiling (length unique-hashes)))
+               ((body decls) (parse-body (cddr lexpr) nil))
+               (optimize-decl (pop decls)))
+      (aver (eq (caadr optimize-decl) 'optimize))
+      (aver (eq (caadar decls) 'type))
+      (let* ((buckets (make-array nbuckets :initial-element nil))
+             (phash-fun (sb-c::compile-perfect-hash lexpr unique-hashes))
+             (resultform
+              (cond ((= (length unique-hashes) (length hashes))
+                     (let* ((et (choose-smallest-element-type slots :key #'cdr))
+                            (data (make-array nbuckets :element-type et)))
+                       (fill buckets 0)
+                       (loop for (key . value) in slots
+                             do (let ((phash (funcall phash-fun (hash key))))
+                                  (aver (eql (svref buckets phash) 0))
+                                  (setf (svref buckets phash) key
+                                        (aref data phash) value)))
+                       `(if (eq (svref ,buckets h) symbol) (aref ,data h))))
+                    (t ; Collisions
+                     (dolist (pair slots)
+                       (let ((phash (funcall phash-fun (hash (car pair)))))
+                         (push pair (svref buckets phash))))
+                     `(dolist (choice (svref ,buckets h))
+                        (when (eq (car choice) symbol)
+                          (return (cdr choice))))))))
+        ;; Bypass COMPILE-PERFECT-HASH here, as it can elect not to actually
+        ;; call COMPILE, but instead make an interpreted function.
+        ;; We don't need the extra check for perfectness that it performs
+        ;; since the above bucketing already asserted that hashing worked.
+        (values (compile
+                 nil
+                 ;; this resembles the ASSOC transform
+                 `(named-lambda ,lambda-name (symbol)
+                    ,optimize-decl
+                    (let* ((sb-c::val (ldb (byte 32 0) (symbol-hash symbol)))
+                           (h (progn ,@body)))
+                      (if (< h ,nbuckets) ,resultform)))))))))
+;;; Return a compiled function which takes a symbol
+;;; and returns the DSD-BITS for the slot in DD of that name.
+(defun make-struct-slot-map (dd)
+  (make-hash-based-slot-mapper
+   (mapcar (lambda (dsd) (cons (dsd-name dsd) (dsd-bits dsd)))
+           (dd-slots dd))
+   ;; Prevent random junk like (lambda (symbol) in "very-long-name-why-why-why")
+   `(slot-mapper ,(dd-name dd))))
 
 (/show0 "target-defstruct.lisp end of file")
