@@ -26,9 +26,10 @@
   (:args (object :scs (descriptor-reg))
          (value :scs (descriptor-reg any-reg zero)))
   (:info name offset lowtag)
-  (:ignore name)
   (:results)
+  (:vop-var vop)
   (:generator 1
+    (emit-gengc-barrier object nil tmp-tn (vop-nth-arg 1 vop) value name)
     (storew value object offset lowtag)))
 
 (define-vop (compare-and-swap-slot)
@@ -39,7 +40,9 @@
   (:ignore name)
   (:temporary (:sc non-descriptor-reg) lip)
   (:results (result :scs (descriptor-reg any-reg) :from :load))
+  (:vop-var vop)
   (:generator 5
+    (emit-gengc-barrier object nil lip (vop-nth-arg 2 vop) new)
     (inst add-sub lip object (- (* offset n-word-bytes) lowtag))
     (cond ((member :arm-v8.1 *backend-subfeatures*)
            (move result old)
@@ -81,6 +84,7 @@
                                       (symbol-always-has-tls-value-p (tn-value object)))))
            (value :scs (descriptor-reg any-reg zero)))
     (:temporary (:sc any-reg) tls-index)
+    (:vop-var vop)
     (:generator 4
       (sc-case object
         (constant
@@ -91,6 +95,7 @@
            (inst ldr tmp-tn (@ thread-tn tls-index))
            (compare-to-no-tls-value-marker tmp-tn)
            (inst b :ne LOCAL)
+           (emit-gengc-barrier object nil tls-index (vop-nth-arg 1 vop) value)
            (storew value object symbol-value-slot other-pointer-lowtag)
            (inst b DONE)
            LOCAL
@@ -211,17 +216,16 @@
   (:policy :fast-safe)
   (:translate symbol-hash)
   (:args (symbol :scs (descriptor-reg)))
-  (:arg-refs args)
-  (:results (res :scs (any-reg)))
+  (:results (res :scs (unsigned-reg)))
   (:result-types positive-fixnum)
   (:generator 2
-    ;; The symbol-hash slot of NIL holds NIL because it is also the
-    ;; car slot, so we have to strip off the fixnum-tag-mask to make sure
-    ;; it is a fixnum.  The lowtag selection magic that is required to
-    ;; ensure this is explained in the comment in objdef.lisp
     (loadw res symbol symbol-hash-slot other-pointer-lowtag)
-    (unless (not-nil-tn-ref-p args)
-      (inst and res res (bic-mask fixnum-tag-mask)))))
+    (inst lsr res res n-symbol-hash-discard-bits)))
+(define-vop (symbol-name-hash symbol-hash)
+  (:translate symbol-name-hash #-relocatable-static-space hash-as-if-symbol-name)
+  (:generator 1 ; ASSUMPTION: little-endian
+    (inst ldr (32-bit-reg res)
+          (@ symbol (- (+ 4 (ash symbol-hash-slot word-shift)) other-pointer-lowtag)))))
 
 (define-vop ()
   (:args (symbol :scs (descriptor-reg)))
@@ -244,6 +248,9 @@
       (loadw tmp-tn symbol symbol-name-slot other-pointer-lowtag)
       (inst and result tmp-tn (1- (ash 1 sb-impl::symbol-name-bits))))))
 
+;; Is it worth eliding the GC store barrier for a thread-local CAS?
+;; Not really, because why does such an operation even exists?
+;; No other thread can see our TLS except via SAP-REF.
 (define-vop (%compare-and-swap-symbol-value)
   (:translate %compare-and-swap-symbol-value)
   (:args (symbol :scs (descriptor-reg))
@@ -256,6 +263,7 @@
   (:policy :fast-safe)
   (:vop-var vop)
   (:generator 15
+    (emit-gengc-barrier symbol nil lip (vop-nth-arg 2 vop) new)
     (inst dsb)
     #+sb-thread
     (assemble ()
@@ -299,6 +307,7 @@
   (:vop-var vop)
   (:guard (member :arm-v8.1 *backend-subfeatures*))
   (:generator 14
+    (emit-gengc-barrier symbol nil lip (vop-nth-arg 2 vop) new)
     #+sb-thread
     (assemble ()
       (inst ldr (32-bit-reg tls-index) (tls-index-of symbol))
@@ -339,13 +348,12 @@
 
 (define-vop (set-fdefn-fun)
   (:policy :fast-safe)
-  (:translate (setf fdefn-fun))
-  (:args (function :scs (descriptor-reg) :target result)
+  (:args (function :scs (descriptor-reg))
          (fdefn :scs (descriptor-reg)))
   (:temporary (:scs (non-descriptor-reg)) lip)
   (:temporary (:scs (non-descriptor-reg)) type)
-  (:results (result :scs (descriptor-reg)))
   (:generator 38
+    (emit-gengc-barrier fdefn nil lip)
     (inst add-sub lip function (- (* simple-fun-insts-offset n-word-bytes)
                                   fun-pointer-lowtag))
     (load-type type function (- fun-pointer-lowtag))
@@ -354,8 +362,7 @@
     (load-asm-routine lip 'closure-tramp)
     SIMPLE-FUN
     (storew lip fdefn fdefn-raw-addr-slot other-pointer-lowtag)
-    (storew function fdefn fdefn-fun-slot other-pointer-lowtag)
-    (move result function)))
+    (storew function fdefn fdefn-fun-slot other-pointer-lowtag)))
 
 (define-vop (fdefn-makunbound)
   (:policy :fast-safe)
@@ -583,6 +590,36 @@
 (define-full-setter instance-index-set * instance-slots-offset
   instance-pointer-lowtag (descriptor-reg any-reg) * %instance-set)
 
+(define-vop (instance-set-multiple)
+  (:args (instance :scs (descriptor-reg))
+         (values :more t :scs (descriptor-reg any-reg)))
+  (:temporary (:sc descriptor-reg) val-temp)
+  (:info indices)
+  (:vop-var vop)
+  (:generator 1
+    (emit-gengc-barrier instance nil tmp-tn values)
+    (do ((tn-ref values (tn-ref-across tn-ref))
+         (indices indices (cdr indices)))
+        ((null tn-ref))
+      (let* ((value (tn-ref-tn tn-ref))
+             (source
+               (sc-case value
+                 (immediate
+                  (load-immediate vop value val-temp)
+                  val-temp)
+                 (constant
+                  (inst load-constant val-temp (tn-byte-offset value))
+                  val-temp)
+                 (control-stack
+                  (load-stack-tn val-temp value)
+                  val-temp)
+                 ((any-reg descriptor-reg)
+                  value))))
+        (inst str source
+              (@ instance (load-store-offset
+                           (- (ash (+ instance-slots-offset (car indices)) word-shift)
+                              instance-pointer-lowtag))))))))
+
 (define-vop (%instance-cas word-index-cas)
   (:policy :fast-safe)
   (:translate %instance-cas)
@@ -592,7 +629,7 @@
   (:args (object)
          (index)
          (old-value :scs (unsigned-reg))
-         (new-value :scs (unsigned-reg)))
+         (new-value :scs (unsigned-reg zero)))
   (:arg-types * tagged-num unsigned-num unsigned-num)
   (:results (result :scs (unsigned-reg) :from :load))
   (:result-types unsigned-num)
@@ -602,12 +639,37 @@
   (:args (object)
          (index)
          (old-value :scs (signed-reg))
-         (new-value :scs (signed-reg)))
+         (new-value :scs (signed-reg zero)))
   (:arg-types * tagged-num signed-num signed-num)
   (:results (result :scs (signed-reg) :from :load))
   (:result-types signed-num)
   (:translate %raw-instance-cas/signed-word))
 
+(define-vop (%instance-cas-v8.1 word-index-cas-v8.1)
+  (:policy :fast-safe)
+  (:translate %instance-cas)
+  (:variant instance-slots-offset instance-pointer-lowtag)
+  (:arg-types instance tagged-num * *))
+
+(define-vop (%raw-instance-cas/word-v8.1 %instance-cas-v8.1)
+  (:args (object)
+         (index)
+         (old-value :scs (unsigned-reg) :target result)
+         (new-value :scs (unsigned-reg zero)))
+  (:arg-types * tagged-num unsigned-num unsigned-num)
+  (:results (result :scs (unsigned-reg) :from (:argument 2)))
+  (:result-types unsigned-num)
+  (:translate %raw-instance-cas/word))
+
+(define-vop (%raw-instance-cas/signed-word-v8.1 %instance-cas-v8.1)
+  (:args (object)
+         (index)
+         (old-value :scs (signed-reg) :target result)
+         (new-value :scs (signed-reg zero)))
+  (:arg-types * tagged-num signed-num signed-num)
+  (:results (result :scs (signed-reg) :from (:argument 2)))
+  (:result-types signed-num)
+  (:translate %raw-instance-cas/signed-word))
 
 ;;;; Code object frobbing.
 
@@ -625,8 +687,6 @@
   (:temporary (:scs (non-descriptor-reg)) temp card)
   (:temporary (:sc non-descriptor-reg) pa-flag)
   (:generator 10
-    (load-foreign-symbol temp "gc_card_table_mask" :dataref t)
-    (inst ldr (32-bit-reg temp) (@ temp)) ; 4-byte int
     (pseudo-atomic (pa-flag)
       #+immobile-space
       (progn
@@ -661,6 +721,8 @@
                  (inst cbnz (32-bit-reg pa-flag) LOOP))))
         (inst b store))
       TRY-DYNAMIC-SPACE
+      (load-foreign-symbol temp "gc_card_table_mask" :dataref t)
+      (inst ldr (32-bit-reg temp) (@ temp)) ; 4-byte int
       ;; Compute card mark index
       (inst lsr card object gencgc-card-shift)
       (inst and card card temp)
