@@ -162,7 +162,9 @@
                                  (singleton-p (mv-combination-args dest)))
                             (let ((fun-ref (lvar-use (mv-combination-fun dest))))
                               (length (lambda-vars (ref-leaf fun-ref)))))))
-         (n-required (length (values-type-required dtype))))
+         (n-required (if (eq dtype *wild-type*)
+                         (return-from cast-check-types (values :too-hairy nil))
+                         (length (values-type-required dtype)))))
     (aver (not (eq ctype *wild-type*)))
     (cond ((and (null (values-type-optional dtype))
                 (not (values-type-rest dtype)))
@@ -195,66 +197,86 @@
 ;;; and should be checked externally -- that is, by the callee and not the caller.
 (defun cast-externally-checkable-p (cast)
   (declare (type cast cast))
-  (let* ((lvar (node-lvar cast))
-         (dest (and lvar (lvar-dest lvar))))
-    (and (combination-p dest)
-         (or (not (combination-fun-info dest))
-             ;; fixed-args functions do not check their arguments.
-             (not (ir1-attributep (fun-info-attributes (combination-fun-info dest))
-                                  fixed-args
-                                  always-translatable)))
-         ;; The theory is that the type assertion is from a declaration on the
-         ;; callee, so the callee should be able to do the check. We want to
-         ;; let the callee do the check, because it is possible that by the
-         ;; time of call that declaration will be changed and we do not want
-         ;; to make people recompile all calls to a function when they were
-         ;; originally compiled with a bad declaration.
-         ;;
-         ;; ALMOST-IMMEDIATELY-USED-P ensures that we don't delegate casts
-         ;; that occur before nodes that can cause observable side effects --
-         ;; most commonly other non-external casts: so the order in which
-         ;; possible type errors are signalled matches with the evaluation
-         ;; order.
-         ;;
-         ;; FIXME: We should let more cases be handled by the callee then we
-         ;; currently do, see: https://bugs.launchpad.net/sbcl/+bug/309104
-         ;; This is not fixable quite here, though, because flow-analysis has
-         ;; deleted the LVAR of the cast by the time we get here, so there is
-         ;; no destination. Perhaps we should mark cases inserted by
-         ;; ASSERT-CALL-TYPE explicitly, and delete those whose destination is
-         ;; deemed unreachable?
-         (and
-          (almost-immediately-used-p lvar cast)
-          (cond ((and (lvar-fun-is (combination-fun dest)
-                                   '(hairy-data-vector-set/check-bounds
-                                     hairy-data-vector-ref/check-bounds
-                                     hairy-data-vector-ref
-                                     hairy-data-vector-set))
-                      (eq (car (combination-args dest)) lvar)
-                      (type= (specifier-type 'vector)
-                             (single-value-type (cast-type-to-check cast))))
-                 (change-full-call dest
-                                   (getf '(hairy-data-vector-set/check-bounds vector-hairy-data-vector-set/check-bounds
-                                           hairy-data-vector-ref/check-bounds vector-hairy-data-vector-ref/check-bounds
-                                           hairy-data-vector-ref vector-hairy-data-vector-ref
-                                           hairy-data-vector-set vector-hairy-data-vector-set)
-                                         (lvar-fun-name (combination-fun dest) t))))
-                #+(or arm64 x86-64)
-                ((lvar-fun-is (combination-fun dest) '(values-list)))
-                ;; Not great
-                ((lvar-fun-is (combination-fun dest) '(%%primitive))
-                 (destructuring-bind (vop &rest args) (combination-args dest)
-                   (and (constant-lvar-p vop)
-                        (let ((name (vop-info-name (lvar-value vop))))
-                          (or (and (memq name '(sb-vm::overflow+t
-                                                sb-vm::overflow-t
-                                                sb-vm::overflow*t))
-                                   (eq lvar (car args)))
-                              (and (memq name '(sb-vm::overflow-t-y))
-                                   (eq lvar (cadr args))))))))
-                (t
-                 (values-subtypep (lvar-externally-checkable-type lvar)
-                                  (cast-type-to-check cast))))))))
+  (let ((lvar (node-lvar cast)))
+    (multiple-value-bind (dest lvar) (and lvar (immediately-used-let-dest lvar cast))
+      (cond ((and (basic-combination-p dest)
+                  (or (not (basic-combination-fun-info dest))
+                      ;; fixed-args functions do not check their arguments.
+                      (not (ir1-attributep (fun-info-attributes (basic-combination-fun-info dest))
+                                           fixed-args
+                                           always-translatable)))
+                  ;; The theory is that the type assertion is from a declaration on the
+                  ;; callee, so the callee should be able to do the check. We want to
+                  ;; let the callee do the check, because it is possible that by the
+                  ;; time of call that declaration will be changed and we do not want
+                  ;; to make people recompile all calls to a function when they were
+                  ;; originally compiled with a bad declaration.
+                  ;;
+                  ;; ALMOST-IMMEDIATELY-USED-P ensures that we don't delegate casts
+                  ;; that occur before nodes that can cause observable side effects --
+                  ;; most commonly other non-external casts: so the order in which
+                  ;; possible type errors are signalled matches with the evaluation
+                  ;; order.
+                  ;;
+                  ;; FIXME: We should let more cases be handled by the callee then we
+                  ;; currently do, see: https://bugs.launchpad.net/sbcl/+bug/309104
+                  ;; This is not fixable quite here, though, because flow-analysis has
+                  ;; deleted the LVAR of the cast by the time we get here, so there is
+                  ;; no destination. Perhaps we should mark cases inserted by
+                  ;; ASSERT-CALL-TYPE explicitly, and delete those whose destination is
+                  ;; deemed unreachable?
+                  (cond ((and (lvar-fun-is (basic-combination-fun dest)
+                                           '(hairy-data-vector-set/check-bounds
+                                             hairy-data-vector-ref/check-bounds
+                                             hairy-data-vector-ref
+                                             hairy-data-vector-set))
+                              (eq (car (basic-combination-args dest)) lvar)
+                              (type= (specifier-type 'vector)
+                                     (single-value-type (cast-type-to-check cast))))
+                         (change-full-call dest
+                                           (getf '(hairy-data-vector-set/check-bounds vector-hairy-data-vector-set/check-bounds
+                                                   hairy-data-vector-ref/check-bounds vector-hairy-data-vector-ref/check-bounds
+                                                   hairy-data-vector-ref vector-hairy-data-vector-ref
+                                                   hairy-data-vector-set vector-hairy-data-vector-set)
+                                                 (lvar-fun-name (basic-combination-fun dest) t))))
+                        #+(or arm64 x86-64)
+                        ((lvar-fun-is (basic-combination-fun dest) '(values-list)))
+                        ;; Not great
+                        ((lvar-fun-is (basic-combination-fun dest) '(%%primitive))
+                         (destructuring-bind (vop &rest args) (basic-combination-args dest)
+                           (and (constant-lvar-p vop)
+                                (let ((name (vop-info-name (lvar-value vop))))
+                                  (or (and (memq name '(sb-vm::overflow+t
+                                                        sb-vm::overflow-t
+                                                        sb-vm::overflow*t))
+                                           (eq lvar (car args)))
+                                      (and (memq name '(sb-vm::overflow-t-y))
+                                           (eq lvar (cadr args))))))))
+                        ((and (policy dest (= debug 3))
+                              (let ((leaf (nth-value 2 (lvar-fun-type (basic-combination-fun dest)))))
+                                (and leaf
+                                     (memq (leaf-where-from leaf) '(:declared-verify :defined-here)))))
+                         nil)
+                        (t
+                         (values-subtypep (lvar-externally-checkable-type lvar)
+                                          (cast-type-to-check cast))))))
+            ((and (cast-p dest)
+                  (cast-type-check dest)
+                  (atom (lvar-uses (node-lvar cast)))
+                  (atom (lvar-uses (cast-value dest)))
+                  (lvar-single-value-p (node-lvar cast))
+                  (cond ((and (values-type-p (cast-asserted-type dest))
+                              (values-type-p (cast-asserted-type cast)))
+                         (values-subtypep (cast-asserted-type dest)
+                                          (cast-asserted-type cast)))
+                        ((not (or (values-type-p (cast-asserted-type dest))
+                                  (values-type-p (cast-asserted-type cast))))
+                         (csubtypep (cast-asserted-type dest)
+                                    (cast-asserted-type cast)))))
+             (setf (cast-asserted-type cast) (cast-asserted-type dest)
+                   (cast-type-to-check cast) (cast-type-to-check dest)
+                   (cast-%type-check dest) nil)
+             nil)))))
 
 ;; Type specifiers handled by the general-purpose MAKE-TYPE-CHECK-FORM are often
 ;; trivial enough to have an internal error number assigned to them that can be
@@ -427,7 +449,7 @@
                                      lvar (combination-args dest))))
                            (format nil "~:[A possible~;The~] binding of ~S"
                                    (and (lvar-has-single-use-p lvar)
-                                        (eq (functional-kind lambda) :let))
+                                        (functional-kind-eq lambda let))
                                    (leaf-source-name (elt (lambda-vars lambda)
                                                           pos)))))))
             (cond ((and (ref-p use) (constant-p (ref-leaf use)))
@@ -483,16 +505,18 @@
             (when (and (cast-p node)
                        (cast-type-check node))
               (cast-check-uses node)
-              (cond ((cast-externally-checkable-p node)
-                     (setf (cast-%type-check node) :external))
-                    (t
-                     ;; it is possible that NODE was marked :EXTERNAL by
-                     ;; the previous pass
-                     (setf (cast-%type-check node) t)
-                     (casts node)))))
+              (let ((external (cast-externally-checkable-p node)))
+                (cond (external
+                       (setf (cast-%type-check node) :external))
+                      (t
+                       ;; it is possible that NODE was marked :EXTERNAL by
+                       ;; the previous pass
+                       (setf (cast-%type-check node) t)
+                       (casts node))))))
           (setf (block-type-check block) nil)))
       (dolist (cast (casts))
-        (unless (bound-cast-p cast)
+        ;; Disabled by cast-externally-checkable-p of a different cast.
+        (when (cast-type-check cast)
           (multiple-value-bind (check types) (cast-check-types cast)
             (ecase check
               (:simple
@@ -501,8 +525,8 @@
               (:too-hairy
                (when (policy cast (>= safety inhibit-warnings))
                  (let* ((*compiler-error-context* cast)
-                       (type (cast-asserted-type cast))
-                       (value-type (coerce-to-values type)))
+                        (type (cast-asserted-type cast))
+                        (value-type (coerce-to-values type)))
                    (compiler-notify
                     "Type assertion too complex to check:~@
                     ~/sb-impl:print-type/.~a"
@@ -520,6 +544,8 @@
                                    (make-values-type (append (values-type-required value-type)
                                                              (values-type-optional value-type)))))
                           ("")))))
+
                (setf (cast-type-to-check cast) *wild-type*)
                (setf (cast-%type-check cast) nil)))))))
     generated))
+
