@@ -52,6 +52,7 @@
 ;;; zero-length, non-zero-total-size vector-with-fill-pointer.
 (declaim (type (and (vector t) (not simple-array)) *constraint-universe*))
 (defvar *constraint-universe*)
+(defvar *blocks-to-terminate*)
 
 (defstruct (vector-length-constraint
             (:constructor make-vector-length-constraint (var))
@@ -81,7 +82,10 @@
   ;;
   ;; EQUALITY
   ;;     Relations between two variables.
-  (kind nil :type (member typep < > = >= <= eql equality))
+  ;;
+  ;; SET
+  ;;    A set indicating that the initial binding is not in effect.
+  (kind nil :type (member typep < > = >= <= eql equality set))
   ;; The operands to the relation.
   (x nil :type (or lambda-var vector-length-constraint))
   (y nil :type constraint-y)
@@ -92,9 +96,10 @@
 (defstruct (equality-constraint
             (:include constraint)
             (:constructor make-equality-constraint
-                (number operator x y not-p
+                (number operator x y not-p &optional (amount 0)
                  &aux (kind 'equality))))
-  (operator nil :type symbol))
+  (operator nil :type symbol)
+  (amount 0 :type integer))
 
 ;;; The basic interval type. It can handle open and closed intervals.
 ;;; A bound is open if it is a list containing a number, just like
@@ -177,6 +182,10 @@
     (min 0 :type fixnum)
     (max 0 :type fixnum))
 
+  #+sb-devel
+  (defprinter (conset)
+    vector)
+
   (defun conset-empty (conset)
     (or (= (conset-min conset) (conset-max conset))
         (not (find 1 (conset-vector conset)
@@ -206,6 +215,7 @@
       (%conset-grow conset new-size))
     (values))
 
+  (declaim (inline conset-member))
   (defun conset-member (constraint conset)
     (let ((number (%constraint-number constraint))
           (vector (conset-vector conset)))
@@ -305,7 +315,7 @@
                              (if position
                                  (1+ position)
                                  0))))))
-              (values))))
+              conset-1)))
     (defconsetop conset-union bit-ior)
     (defconsetop conset-intersection bit-and)
     (defconsetop conset-difference bit-andc2)))
@@ -400,6 +410,11 @@
         (when (lambda-var-p y)
           (register-constraint y new x))
         new)))
+
+(declaim (inline type-for-constraints-p))
+(defun type-for-constraints-p (type)
+  (not (or (eq type *universal-type*)
+           (contains-hairy-type-p type))))
 
 ;;; Actual conset interface
 ;;;
@@ -457,13 +472,16 @@
       `(flet ((body (,symbol)
                 (declare (type constraint ,symbol))
                 ,@body))
-         (declare (optimize speed)
-                  (inline body))
          (when ,constraints
            (let ((,min (conset-min ,conset))
-                 (,max (conset-max ,conset)))
-             (loop for constraint across ,constraints
-                   do (let ((number (constraint-number (the constraint constraint))))
+                 (,max (conset-max ,conset))
+                 (vector #-sb-xc-host (truly-the simple-vector (%array-data ,constraints))
+                          #+sb-xc-host ,constraints))
+             #-sb-xc-host
+             (declare (optimize (insert-array-bounds-checks 0)))
+             (loop for i below (length ,constraints)
+                   for constraint = (aref vector i)
+                   do (let ((number (truly-the index (constraint-number (truly-the constraint constraint)))))
                         (when (and (<= ,min number)
                                    (< number ,max)
                                    (conset-member constraint ,conset))
@@ -508,9 +526,6 @@
            (body-fun con))
          (do-conset-constraints-intersection
              (con (,conset (lambda-var-inheritable-constraints ,variable)) ,result)
-           (body-fun con))
-         (do-conset-constraints-intersection
-             (con (,conset (lambda-var-equality-constraints ,variable)) ,result)
            (body-fun con))))))
 
 (declaim (inline conset-lvar-lambda-var-eql-p conset-add-lvar-lambda-var-eql))
@@ -594,7 +609,14 @@
                   (conset-lvar-lambda-var-eql-p constraints lvar lambda-var)
                   lambda-var)))
           ((cast-p use)
-           (ok-lvar-lambda-var (cast-value use) constraints)))))
+           (ok-lvar-lambda-var (cast-value use) constraints))
+          ((set-p use)
+           (let ((leaf (set-var use)))
+             (when (and (lambda-var-p leaf)
+                        (lambda-var-constraints leaf))
+               (and leaf
+                    (conset-lvar-lambda-var-eql-p constraints lvar leaf)
+                    leaf)))))))
 ;;;; Searching constraints
 
 ;;; Add the indicated test constraint to TARGET.
@@ -646,49 +668,52 @@
                                          consequent-constraints
                                          alternative-constraints
                                          quick-p)
-  (flet ((add (fun x y &optional no-complement)
-           (if no-complement
-               (when x
-                 (add-test-constraint quick-p
-                                      fun x y nil
-                                      constraints
-                                      consequent-constraints))
-               (add-complement-constraints quick-p
-                                           fun x y nil
-                                           constraints
-                                           consequent-constraints
+  (flet ((add (fun lvar y &optional no-complement)
+           (let ((x (ok-lvar-lambda-var lvar constraints)))
+             (if no-complement
+                 (when x
+                   (add-test-constraint quick-p
+                                        fun x y nil
+                                        constraints
+                                        consequent-constraints))
+                 (add-complement-constraints quick-p
+                                             fun x y nil
+                                             constraints
+                                             consequent-constraints
+                                             alternative-constraints)))
+           (constraint-propagate-back lvar fun y constraints consequent-constraints
+                                      (and (not no-complement)
                                            alternative-constraints)))
          (prop (triples target)
            (map nil (lambda (constraint)
                       (destructuring-bind (kind x y &optional not-p)
                           constraint
                         (when (and kind x y)
-                          (add-test-constraint quick-p
-                                               kind x y
-                                               not-p constraints
-                                               target))))
+                          (let ((x (if (lvar-p x)
+                                       (ok-lvar-lambda-var x constraints)
+                                       x)))
+                            (when x
+                              (add-test-constraint quick-p
+                                                   kind x y
+                                                   not-p constraints
+                                                   target))))))
                 triples)))
     (when (eq (combination-kind use) :known)
       (binding* ((info (combination-fun-info use) :exit-if-null)
-                 (propagate (fun-info-constraint-propagate-if
-                             info)
-                            :exit-if-null))
+                 (propagate (fun-info-constraint-propagate-if info) :exit-if-null))
         (multiple-value-bind (lvar type if else no-complement)
             (funcall propagate use constraints)
           (prop if consequent-constraints)
           (prop else alternative-constraints)
           (when (and lvar type)
-            (add 'typep (ok-lvar-lambda-var lvar constraints)
-                 type no-complement)
+            (add 'typep lvar type no-complement)
             (return-from add-combination-test-constraints)))))
     (let* ((name (lvar-fun-name
                   (basic-combination-fun use)))
            (args (basic-combination-args use))
            (ptype (gethash name *backend-predicate-types*)))
       (when ptype
-        (add 'typep (ok-lvar-lambda-var (first args)
-                                        constraints)
-             ptype)))))
+        (add 'typep (first args) ptype)))))
 
 ;;; Add test constraints to the consequent and alternative blocks of
 ;;; the test represented by USE.
@@ -702,12 +727,14 @@
     (let ((consequent-constraints (make-conset))
           (alternative-constraints (make-conset))
           (quick-p (policy if (> compilation-speed speed))))
-      (labels ((add (fun x y not-p)
+      (labels ((add (fun x y not-p &optional lvar)
                  (add-complement-constraints quick-p
                                              fun x y not-p
                                              constraints
                                              consequent-constraints
-                                             alternative-constraints))
+                                             alternative-constraints)
+                 (when lvar
+                   (constraint-propagate-back lvar fun y constraints consequent-constraints alternative-constraints)))
                (process-node (node)
                  (typecase node
                    (ref
@@ -760,7 +787,8 @@
                                    ((constant-lvar-p arg2)
                                     (add 'eql var1
                                          (nth-value 1 (lvar-value arg2))
-                                         nil))
+                                         nil
+                                         arg1))
                                    (t
                                     (add-test-constraint quick-p
                                                          'typep var1 (lvar-type arg2)
@@ -1003,7 +1031,8 @@
                     (setq not-type (type-union not-type other)))))
             (eql
              (let ((other-type (leaf-type other)))
-               (cond ((not not-p)
+               (cond ((not (type-for-constraints-p other-type)))
+                     ((not not-p)
                       (intersect-result other-type))
                      ((constant-p other)
                       (cond ((member-type-p other-type)
@@ -1105,7 +1134,8 @@
         (not-numeric (alloc-xset))
         (not-fpz nil)
         (not-res *empty-type*)
-        (leaf (ref-leaf ref)))
+        (leaf (ref-leaf ref))
+        set)
     (declare (type lambda-var leaf))
     (flet ((note-not (x)
              (if (fp-zero-p x)
@@ -1119,64 +1149,40 @@
                (other (if (eq x leaf) y x))
                (kind (constraint-kind con)))
           (case kind
-            (equality
-             (unless (eq (ref-constraints ref)
-                         (pushnew con (ref-constraints ref)))
-               (labels ((reoptimize (node)
-                          (when (valued-node-p node)
-                            (let ((lvar (node-lvar node))
-                                  (principal-lvar (nth-value 1 (principal-lvar-end (node-lvar ref)))))
-                              (reoptimize-lvar lvar)
-                              (unless (eq lvar principal-lvar)
-                                (reoptimize-lvar principal-lvar)))))
-                        (try (x y)
-                          (when (and (vector-length-constraint-p x)
-                                     (eq (vector-length-constraint-var x) leaf))
-                            (when (and (constant-p y)
-                                       (not not-p)
-                                       (eq (equality-constraint-operator con) 'eq)
-                                       (not (types-equal-or-intersect res (specifier-type '(not simple-array)))))
-                              (setf res (type-intersection res (specifier-type `(simple-array * (,(constant-value y)))))))
-                            (reoptimize (node-dest ref))
-                            t)))
-                 (or (try x y)
-                     (try y x)
-                     (reoptimize ref)))))
             (typep
-             (if not-p
-                 (if (member-type-p other)
-                     (mapc-member-type-members #'note-not other)
-                     (setq not-res (type-union not-res other)))
-                 (setq res (type-intersection res other))))
+             (when (type-for-constraints-p other)
+               (if not-p
+                   (if (member-type-p other)
+                       (mapc-member-type-members #'note-not other)
+                       (setq not-res (type-union not-res other)))
+                   (setq res (type-intersection res other)))))
             (eql
              (let ((other-type (leaf-type other)))
-               (if not-p
-                   (when (constant-p other)
-                     (cond ((member-type-p other-type)
-                            (note-not (constant-value other)))
-                           ;; Numeric types will produce interesting
-                           ;; negations, other than just "not equal"
-                           ;; which can be handled by the equality
-                           ;; constraints.
-                           ((numeric-type-p other-type)
-                            (add-to-xset (constant-value other) not-numeric))))
-                   (let ((leaf-type (leaf-type leaf)))
+               (when (type-for-constraints-p other-type)
+                 (if not-p
+                     (when (constant-p other)
+                       (cond ((member-type-p other-type)
+                              (note-not (constant-value other)))
+                             ;; Numeric types will produce interesting
+                             ;; negations, other than just "not equal"
+                             ;; which can be handled by the equality
+                             ;; constraints.
+                             ((numeric-type-p other-type)
+                              (add-to-xset (constant-value other) not-numeric))))
                      (cond
-                       ((or (constant-p other)
-                            (and (leaf-refs other) ; protect from
-                                        ; deleted vars
-                                 (csubtypep other-type leaf-type)
-                                 (not (type= other-type leaf-type))
-                                 ;; Don't change to a LEAF not visible here.
-                                 (leaf-visible-from-node-p other ref)))
+                       ((constant-p other)
                         (change-ref-leaf ref other)
-                        (when (constant-p other) (return)))
+                        (return-from constrain-ref-type))
                        (t
                         (setq res (type-intersection res other-type))))))))
             ((< > <= >= =)
              (let ((type (type-after-comparison kind not-p res y)))
                (when type
-                 (setf res type))))))))
+                 (setf res type))))
+            (set
+             (setf set t))))))
+    (unless set
+      (setf (lambda-var-unused-initial-value leaf) nil))
     (cond ((and (if-p (node-dest ref))
                 (or (xset-member-p nil not-set)
                     (csubtypep (specifier-type 'null) not-res)))
@@ -1207,7 +1213,29 @@
                    (t
                     (derive-node-type ref
                                       (make-single-value-type type))
-                    (maybe-terminate-block ref nil)))))))
+                    (when (eq (node-derived-type ref) *empty-type*)
+                      ;; Terminating blocks early may leave loops with
+                      ;; just one entry point, resulting in
+                      ;; monotonically growing variables without a
+                      ;; starting point which will propagate new
+                      ;; constraints for each increment.
+                      (pushnew ref *blocks-to-terminate*)))))))
+    ;; Find unchanged eql refs to a set variable.
+    (when (lambda-var-sets leaf)
+      (let (mark
+            (eq (lambda-var-eq-constraints leaf)))
+        (when eq
+          (loop for other-ref in (leaf-refs leaf)
+                unless (eq other-ref ref)
+                do (let ((constraint (gethash (ref-lvar other-ref) eq)))
+                     (when (and constraint
+                                (conset-member constraint in))
+                       (unless mark
+                         (setf mark (list 0))
+                         (setf (ref-same-refs ref) mark))
+                       (setf (ref-same-refs other-ref) mark)))))
+        (when mark
+          (reoptimize-node ref)))))
   (values))
 
 ;;;; Flow analysis
@@ -1240,15 +1268,19 @@
     (typecase node
       (bind
        (let ((fun (bind-lambda node)))
-         (when (eq (functional-kind fun) :let)
-           (loop with call = (lvar-dest (node-lvar (first (lambda-refs fun))))
-                 for var in (lambda-vars fun)
-                 and val in (combination-args call)
-                 when (and val (lambda-var-constraints var))
-                 do (let ((type (lvar-type val)))
-                      (unless (eq type *universal-type*)
-                        (conset-add-constraint gen 'typep var type nil)))
-                    (maybe-add-eql-var-var-constraint var val gen)))))
+         (functional-kind-case fun
+           (let
+            (loop with call = (lvar-dest (node-lvar (first (lambda-refs fun))))
+                  for var in (lambda-vars fun)
+                  and val in (combination-args call)
+                  when (and val (lambda-var-constraints var))
+                  do (let ((type (lvar-type val)))
+                       (when (type-for-constraints-p type)
+                         (conset-add-constraint gen 'typep var type nil)))
+                     (maybe-add-eql-var-var-constraint var val gen)
+                     (add-var-result-constraints var val gen)))
+           (mv-let
+            (add-mv-let-result-constraints (lvar-dest (node-lvar (first (lambda-refs fun)))) fun gen)))))
       (ref
        (when (ok-ref-lambda-var node)
          (maybe-add-eql-var-lvar-constraint node gen)
@@ -1256,19 +1288,12 @@
            (constrain-ref-type node gen))))
       (cast
        (let* ((lvar (cast-value node))
-              (var (ok-lvar-lambda-var lvar gen)))
-         (when var
-           (let ((atype (single-value-type (cast-derived-type node)))) ;FIXME
-             (unless (eq atype *universal-type*)
-               (conset-add-constraint-to-eql gen 'typep var atype nil))))
-         (when (and (bound-cast-p node)
-                    (bound-cast-check node)
-                    (not (node-deleted (bound-cast-check node))))
-           (let ((check-bound (bound-cast-check node)))
-             (destructuring-bind (array dim index)
-                 (combination-args check-bound)
-               (declare (ignore array))
-               (add-equality-constraints '< (list index dim) gen gen nil))))))
+              (var (ok-lvar-lambda-var lvar gen))
+              (atype (single-value-type (cast-derived-type node))))
+         (when (and var
+                    (type-for-constraints-p atype))
+           (conset-add-constraint-to-eql gen 'typep var atype nil))
+         (constraint-propagate-back lvar 'typep atype gen gen nil)))
       (cset
        (binding* ((var (set-var node))
                   (nil (lambda-var-p var) :exit-if-null)
@@ -1276,35 +1301,59 @@
          (when (policy node (or (and (= speed 3) (> speed compilation-speed))
                                 (> debug 1)))
            (let ((type (lambda-var-type var)))
-             (unless (eq *universal-type* type)
+             (when (type-for-constraints-p type)
                (do-eql-vars (other (var gen))
                  (unless (eql other var)
                    (conset-add-constraint gen 'typep other type nil))))))
-         (conset-clear-lambda-var gen var)
+
+         (let ((new (add-set-constraints var (set-value node) gen)))
+           (conset-clear-lambda-var gen var)
+           (when new
+             (conset-union gen new)))
+
          (let ((type (single-value-type (node-derived-type node))))
-           (unless (eq type *universal-type*)
+           (when (type-for-constraints-p type)
              (conset-add-constraint gen 'typep var type nil)))
          (unless (policy node (> compilation-speed speed))
            (maybe-add-eql-var-var-constraint var (set-value node) gen))
-         (add-eq-constraint var (set-value node) gen)))
+         (add-eq-constraint var (set-value node) gen)
+         (conset-add-constraint gen 'set var var nil)
+         (when (node-lvar node)
+           (conset-add-lvar-lambda-var-eql gen (node-lvar node) var))))
       (combination
-       (when (eq (combination-kind node) :known)
-         (binding* ((info (combination-fun-info node) :exit-if-null)
-                    (propagate (fun-info-constraint-propagate info)
-                               :exit-if-null)
-                    (constraints (funcall propagate node gen))
-                    (register (if (policy node
-                                          (> compilation-speed speed))
-                                  #'conset-add-constraint
-                                  #'conset-add-constraint-to-eql)))
-           (map nil (lambda (constraint)
-                      (destructuring-bind (kind x y &optional not-p)
-                          constraint
-                        (when (and kind x y)
-                          (funcall register gen
-                                   kind x y
-                                   not-p))))
-                constraints))))))
+       (case (combination-kind node)
+         (:known
+          (unless (and preprocess-refs-p
+                       (try-equality-constraint node gen))
+            (let ((lvar (node-lvar node)))
+              (when lvar
+                (add-var-result-constraints lvar lvar gen)))
+            (binding* ((info (combination-fun-info node) :exit-if-null)
+                       (propagate (fun-info-constraint-propagate info)
+                                  :exit-if-null)
+                       (constraints (funcall propagate node gen))
+                       (register (if (policy node
+                                         (> compilation-speed speed))
+                                     #'conset-add-constraint
+                                     #'conset-add-constraint-to-eql)))
+              (map nil (lambda (constraint)
+                         (destructuring-bind (kind x y &optional not-p) constraint
+                           (when (and kind x y)
+                             (funcall register gen
+                                      kind x y
+                                      not-p))))
+                   constraints))))
+         (:local
+          (let ((fun (combination-lambda node))
+                (call-in (combination-constraints-in node)))
+
+            (when (and (functional-kind-eq fun nil assignment optional cleanup)
+                       (not (and call-in
+                                 (conset= call-in gen))))
+              (setf (combination-constraints-in node)
+                    (copy-conset gen))
+              (when (boundp '*constraint-blocks*)
+                (enqueue-block-for-constraints (lambda-block fun))))))))))
   gen)
 
 (defun constraint-propagate-if (block gen)
@@ -1385,7 +1434,16 @@
                            (lambda-var-no-constraints var))
                  (when (or (null (lambda-var-sets var))
                            (not (closure-var-p var)))
-                   (setf (lambda-var-constraints var) (make-conset)))))))
+                   (setf (lambda-var-constraints var) (make-conset))))
+               (when (and (lambda-var-constraints var)
+                          (lambda-var-sets var))
+                 (when (block-type-check (lambda-block fun))
+                   ;; This is optimistic, make sure it's going to be
+                   ;; processed.
+                   (setf (lambda-var-unused-initial-value var) t))
+                 (loop for ref in (lambda-var-refs var)
+
+                       do (setf (ref-same-refs ref) nil))))))
       (frob fun)
       (dolist (let (lambda-lets fun))
         (frob let)))))
@@ -1405,58 +1463,94 @@
 
 ;;; Join the constraints coming from the predecessors of BLOCK on
 ;;; every constrained variable into the constraint set IN.
-(defun join-type-constraints (in block)
+(defun join-type-constraints (in block &optional equality-only predecessor-outs)
   (let ((vars '())
+        (equality-vars)
         (predecessors (block-pred block)))
-    (progn
-      ;;unless (singleton-p predecessors) for some reason it makes the whole thing loop infinitely
-      ;; Find some set of constrained variables in the predecessors.
-      (dolist (pred predecessors)
-        (let ((out (block-out-for-successor pred block)))
-          (when out
-            (do-conset-elements (con out)
-              (let ((kind  (constraint-kind con))
-                    (y     (constraint-y con))
-                    (not-p (constraint-not-p con)))
-                (when (or (member kind '(typep < >))
-                          (and (eq kind 'eql) (or (not not-p)
-                                                  (constant-p y)))
-                          (and (eq kind '=) (and (numeric-type-p y)
-                                                 (not not-p))))
-                  (pushnew (constraint-x con) vars))))
-            (return))))
-      (dolist (var vars)
-        (let ((in-var-type *empty-type*))
-          (dolist (pred (block-pred block))
-            (let ((out (block-out-for-successor pred block)))
-              (when out
-                (setq in-var-type
-                      (type-union in-var-type
-                                  (type-from-constraints var out *universal-type*))))))
+    (flet ((find-vars (out)
+             (do-conset-elements (con out)
+               (let ((kind  (constraint-kind con))
+                     (y     (constraint-y con))
+                     (not-p (constraint-not-p con)))
+                 (when (and (not equality-only)
+                            (or (member kind '(typep < >))
+                                (and (eq kind 'eql) (or (not not-p)
+                                                        (constant-p y)))
+                                (and (eq kind '=) (and (numeric-type-p y)
+                                                       (not not-p)))))
+                   (pushnew (constraint-x con) vars))
+                 (when (and (eq kind 'equality)
+                            (/= (equality-constraint-amount con) 0))
+                   (pushnew (constraint-x con) equality-vars :test #'eq)
+                   (when (lambda-var/vector-length-p y)
+                     (pushnew y equality-vars)))))))
+      (cond (predecessor-outs
+             (find-vars (car predecessor-outs)))
+            (t
+             (loop for pred in predecessors
+                   do
+                   (let ((out (block-out-for-successor pred block)))
+                     (when out
+                       (find-vars out)
+                       (return)))))))
+    (dolist (var vars)
+      (let ((in-var-type *empty-type*))
+        (flet ((compute-type (out)
+                 (setq in-var-type
+                       (type-union in-var-type
+                                   (type-from-constraints var out *universal-type*)))))
+          (if predecessor-outs
+              (dolist (out predecessor-outs)
+                (compute-type out))
+              (dolist (pred predecessors)
+                (let ((out (block-out-for-successor pred block)))
+                  (when out
+                    (compute-type out))))))
 
-          (unless (eq in-var-type *universal-type*)
-            ;; Remove the existing constraints to avoid joining them again later.
-            (do-propagatable-constraints (con (in var))
-              (when (eq (constraint-kind con) 'typep)
-                (conset-delete con in)))
-            (conset-adjoin (find-or-create-constraint 'typep
-                                                      var
-                                                      in-var-type
-                                                      nil)
-                           in)))))))
+        (when (type-for-constraints-p in-var-type)
+          ;; Remove the existing constraints to avoid joining them again later.
+          (do-propagatable-constraints (con (in var))
+            (when (eq (constraint-kind con) 'typep)
+              (conset-delete con in)))
+          (conset-adjoin (find-or-create-constraint 'typep
+                                                    var
+                                                    in-var-type
+                                                    nil)
+                         in))))
+    (dolist (var equality-vars)
+      (join-equality-constraints var block in))))
 
 (defun compute-block-in (block join-types-p)
-  (let ((in nil))
-    (dolist (pred (block-pred block))
-      ;; If OUT has not been calculated, assume it to be the universal
-      ;; set.
-      (let ((out (block-out-for-successor pred block)))
-        (when out
-          (if in
-              (conset-intersection in out)
-              (setq in (copy-conset out))))))
-    (when join-types-p
-      (join-type-constraints in block))
+  (let ((in nil)
+        (bind (block-start-node block)))
+    (cond
+      ;; Use constraints from the local calls to this function
+      ((and (bind-p bind)
+            (functional-kind-eq (bind-lambda bind) nil assignment optional cleanup))
+       (let ((fun (bind-lambda bind))
+             (outs))
+         (loop for ref in (lambda-refs fun)
+               for call = (node-dest ref)
+               for call-in = (and call
+                                  (combination-constraints-in call))
+               when call-in
+               do (if in
+                      (conset-intersection in call-in)
+                      (setf in (copy-conset call-in)))
+                  (push call-in outs))
+         (when (rest outs)
+           (join-type-constraints in block (not join-types-p) outs))))
+      (t
+       (dolist (pred (block-pred block))
+         ;; If OUT has not been calculated, assume it to be the universal
+         ;; set.
+         (let ((out (block-out-for-successor pred block)))
+           (when out
+             (if in
+                 (conset-intersection in out)
+                 (setq in (copy-conset out))))))
+       (when (rest (block-pred block))
+         (join-type-constraints in block (not join-types-p)))))
     (or in (make-conset))))
 
 (defun update-block-in (block join-types-p)
@@ -1477,10 +1571,21 @@
   (declare (type component component))
   (collect ((leading-blocks) (rest-of-blocks))
     (do-blocks (block component)
-      (let ((leading t))
-        (dolist (pred (block-pred block))
-          (unless (block-flag pred)
-            (setq leading nil)))
+      (let ((leading t)
+            (bind (block-start-node block))
+            fun)
+        (if (and (bind-p bind)
+                 (functional-kind-eq (setf fun (bind-lambda bind)) nil assignment optional cleanup))
+            (loop for ref in (lambda-refs fun)
+                  for call = (node-dest ref)
+                  for call-block = (and call
+                                        (node-block call))
+                  do (unless (and call-block
+                                  (block-flag call-block))
+                       (setq leading nil)))
+            (dolist (pred (block-pred block))
+              (unless (block-flag pred)
+                (setq leading nil))))
         (setf (block-flag block) leading)
         (when (block-type-check block)
           (if leading
@@ -1501,49 +1606,55 @@
     (when (eql (car x) obj)
       (return-from nconc-new list))))
 
-(defun find-and-propagate-constraints (component)
-  (let ((blocks-to-process ()))
-    (flet ((enqueue (blocks)
-             (dolist (block blocks)
-               (when (block-type-check block)
-                 (setq blocks-to-process (nconc-new block blocks-to-process))))))
-      (clear-flags component)
-      (multiple-value-bind (leading-blocks rest-of-blocks)
-          (leading-component-blocks component)
-        ;; Update every block once to account for changes in the
-        ;; IR1. The constraints of the lead blocks cannot be changed
-        ;; after the first pass so we might as well use them and skip
-        ;; USE-RESULT-CONSTRAINTS later.
-        (dolist (block leading-blocks)
-          (setf (block-in block) (compute-block-in block t))
-          (find-block-type-constraints block t))
-        ;; We can only start joining types on blocks in which
-        ;; constraint propagation might have to run multiple times (to
-        ;; fixpoint) once all type constraints are definitely
-        ;; correct. They may not be the first time around because EQL
-        ;; constraint propagation is optimistic, i.e. un-EQL variables
-        ;; may be considered EQL before constraint propagation is
-        ;; done, hence any inherited type constraints from such
-        ;; constraints will be wrong as well.
-        (dolist (join-types-p '(nil t))
-          (setq blocks-to-process (copy-list rest-of-blocks))
-          ;; The rest of the blocks.
-          (dolist (block rest-of-blocks)
-            (aver (eq block (pop blocks-to-process)))
-            (setf (block-in block) (compute-block-in block join-types-p))
-            (enqueue (find-block-type-constraints block nil)))
-          ;; Propagate constraints
-          (loop for block = (pop blocks-to-process)
-                while block do
-                  (unless (eq block (component-tail component))
-                    (when (update-block-in block join-types-p)
-                      (enqueue (find-block-type-constraints block nil))))))
+(defvar *constraint-blocks*)
 
-        rest-of-blocks))))
+(defun enqueue-block-for-constraints (block)
+  (when (block-type-check block)
+    (setq *constraint-blocks* (nconc-new block *constraint-blocks*))))
+
+(defun find-and-propagate-constraints (component)
+  (clear-flags component)
+  (multiple-value-bind (leading-blocks rest-of-blocks)
+      (leading-component-blocks component)
+    ;; Update every block once to account for changes in the
+    ;; IR1. The constraints of the lead blocks cannot be changed
+    ;; after the first pass so we might as well use them and skip
+    ;; USE-RESULT-CONSTRAINTS later.
+    (dolist (block leading-blocks)
+      (setf (block-in block) (compute-block-in block t))
+      (find-block-type-constraints block t))
+    ;; We can only start joining types on blocks in which
+    ;; constraint propagation might have to run multiple times (to
+    ;; fixpoint) once all type constraints are definitely
+    ;; correct. They may not be the first time around because EQL
+    ;; constraint propagation is optimistic, i.e. un-EQL variables
+    ;; may be considered EQL before constraint propagation is
+    ;; done, hence any inherited type constraints from such
+    ;; constraints will be wrong as well.
+    (dolist (join-types-p '(nil t))
+      (let ((*constraint-blocks* (copy-list rest-of-blocks)))
+        ;; The rest of the blocks.
+        (dolist (block rest-of-blocks)
+          (aver (eq block (pop *constraint-blocks*)))
+          (setf (block-in block) (compute-block-in block join-types-p))
+          (mapc #'enqueue-block-for-constraints
+                (find-block-type-constraints block nil)))
+        ;; Propagate constraints
+        (loop for block = (pop *constraint-blocks*)
+              while block do
+              (unless (or (block-delete-p block)
+                          (eq block (component-tail component)))
+                (when (update-block-in block join-types-p)
+                  (mapc #'enqueue-block-for-constraints
+                        (find-block-type-constraints block nil)))))))
+
+    rest-of-blocks))
 
 (defun constraint-propagate (component)
   (declare (type component component))
+  (setf (block-out (component-head component)) (make-conset))
   (init-var-constraints component)
+
   ;; Previous results can confuse propagation and may loop forever
   (do-blocks (block component)
     (setf (block-out block) nil)
@@ -1551,9 +1662,11 @@
       (when (if-p last)
         (setf (if-alternative-constraints last) nil)
         (setf (if-consequent-constraints last) nil))))
-  (setf (block-out (component-head component)) (make-conset))
-  (dolist (block (find-and-propagate-constraints component))
-    (unless (block-delete-p block)
-      (use-result-constraints block)))
 
+  (let (*blocks-to-terminate*)
+    (dolist (block (find-and-propagate-constraints component))
+      (unless (block-delete-p block)
+        (use-result-constraints block)))
+    (loop for node in *blocks-to-terminate*
+          do (maybe-terminate-block node nil)))
   (values))
