@@ -280,15 +280,23 @@
   (when arg3p (dump-varint arg3 fasl-output)))
 
 ;;; Dump the FOP code for the named FOP to the specified FASL-OUTPUT.
-(defmacro dump-fop (fs-expr file &rest args)
-  (let* ((fs (eval fs-expr))
-         (val (or (get fs 'opcode)
+;;; This macro is supposed to look functional in that it evals its args,
+;;; but it wants to evaluate the first arg at compile-time. For this reason
+;;; it should really not be a quoted symbol, but I think this used to actually
+;;; be a function which had to look up the fop's opcode every time called.
+(defmacro dump-fop (fop-symbol file &rest args)
+  (let* ((fop-symbol
+          ;; EVAL is too much. Just ascertain we have a quoted symbol
+          (if (typep fop-symbol '(cons (eql quote) (cons symbol null)))
+              (cadr fop-symbol)
+              (error "Bad 1st arg to DUMP-FOP: ~S" fop-symbol)))
+         (val (or (gethash fop-symbol *fop-name-to-opcode*)
                   (error "compiler bug: ~S is not a legal fasload operator."
-                         fs-expr)))
+                         fop-symbol)))
          (fop-argc (aref (car **fop-signatures**) val)))
     (cond
       ((not (eql (length args) fop-argc))
-       (error "~S takes ~D argument~:P" fs fop-argc))
+       (error "~S takes ~D argument~:P" fop-symbol fop-argc))
       ((eql fop-argc 0)
        `(dump-byte ,val ,file))
       (t
@@ -469,11 +477,6 @@
               (dump-object 'values-specifier-type file)
               (dump-object (type-specifier x) file)
               (dump-fop 'fop-funcall file 1))
-             (sb-c::debug-name-marker ; these are atoms, much like symbols
-              (dump-fop 'fop-debug-name-marker file
-                        (cond ((eq x sb-c::*debug-name-sharp*) 1)
-                              ((eq x sb-c::*debug-name-ellipsis*) 2)
-                              (t (bug "Bogus debug name marker")))))
              (instance
               (multiple-value-bind (slot-names slot-names-p)
                   (gethash x (fasl-output-saved-slot-names file))
@@ -570,15 +573,16 @@
                (dump-fop 'fop-nthcdr file i))
             (declare (type index i)))))
 
-      (dump-byte (ecase (circularity-type info)
-                   (:rplaca     #.(get 'fop-rplaca 'opcode))
-                   (:rplacd     #.(get 'fop-rplacd 'opcode))
-                   (:svset      #.(get 'fop-svset 'opcode))
-                   (:struct-set #.(get 'fop-structset 'opcode))
-                   (:slot-set
-                    (dump-object (circularity-slot-name info) file)
-                    #.(get 'fop-slotset 'opcode)))
-                 file)
+      (macrolet ((fop-op (symbol) (gethash symbol *fop-name-to-opcode*)))
+        (dump-byte (ecase (circularity-type info)
+                     (:rplaca     (fop-op fop-rplaca))
+                     (:rplacd     (fop-op fop-rplacd))
+                     (:svset      (fop-op fop-svset))
+                     (:struct-set (fop-op fop-structset))
+                     (:slot-set
+                      (dump-object (circularity-slot-name info) file)
+                      (fop-op fop-slotset)))
+                   file))
       (dump-varint (gethash (circularity-object info) table) file)
       (dump-varint (circularity-index info) file))))
 
@@ -1032,7 +1036,7 @@
   (values))
 
 (eval-when (:compile-toplevel)
-  (assert (<= (length +fixup-kinds+) 8))) ; fixup-kind fits in 3 bits
+  (assert (<= (length +fixup-kinds+) 16))) ; fixup-kind fits in 4 bits
 
 (defconstant-eqx +fixup-flavors+
   #(:assembly-routine :assembly-routine*
@@ -1050,30 +1054,33 @@
 ;;; so the fixup can be reduced to one word instead of an integer and a symbol.
 (declaim (inline !pack-fixup-info))
 (defun !pack-fixup-info (offset kind flavor data)
-  (logior ;; 3 bits
-          (the (mod 8) (or (position kind +fixup-kinds+)
+  (logior ;; 4 bits
+          (the (mod 16) (or (position kind +fixup-kinds+)
                            (error "Bad fixup kind ~s" kind)))
           ;; 4 bits
           (ash (the (mod 16) (or (position flavor +fixup-flavors+)
                                  (error "Bad fixup flavor ~s" flavor)))
-               3)
+               4)
           ;; 8 bits
-          (ash (the (mod 256) data) 7)
+          (ash (the (mod 256) data) 8)
           ;; whatever it needs
-          (ash offset 15)))
+          (ash offset 16)))
 
 ;;; Unpack an integer from DUMP-FIXUPs. Shared by genesis and target fasloader
 (declaim (inline !unpack-fixup-info))
 (defun !unpack-fixup-info (packed-info) ; Return (VALUES offset kind flavor data)
-  (values (ash packed-info -15)
-          (aref +fixup-kinds+ (ldb (byte 3 0) packed-info))
-          (aref +fixup-flavors+ (ldb (byte 4 3) packed-info))
-          (ldb (byte 8 7) packed-info)))
+  (values (ash packed-info -16)
+          (aref +fixup-kinds+ (ldb (byte 4 0) packed-info))
+          (aref +fixup-flavors+ (ldb (byte 4 4) packed-info))
+          (ldb (byte 8 8) packed-info)))
 
-#+(or arm arm64 riscv sparc) ; these don't have any retained packed fixups (yet)
-(defun sb-c::pack-retained-fixups (fixup-notes)
-  (declare (ignore fixup-notes))
-  0) ; as if from PACK-CODE-FIXUP-LOCS
+#-(or x86 x86-64) ; these two architectures provide an overriding definition
+(defun pack-fixups-for-reapplication (fixup-notes)
+  (let (result)
+    (dolist (note fixup-notes (sb-c:pack-code-fixup-locs nil nil result))
+      (let ((fixup (fixup-note-fixup note)))
+        (when (eq (fixup-flavor fixup) :card-table-index-mask)
+          (push (fixup-note-position note) result))))))
 
 ;;; Dump all the fixups.
 ;;;  - foreign (C) symbols: named by a string
@@ -1084,7 +1091,7 @@
   ;; "retained" fixups are those whose offset in the code needs to be
   ;; remembered for subsequent reapplication by the garbage collector,
   ;; or in some cases, on core startup.
-  (dump-object (sb-c::pack-retained-fixups fixup-notes) fasl-output)
+  (dump-object (pack-fixups-for-reapplication fixup-notes) fasl-output)
   (dump-object alloc-points fasl-output)
   (dolist (note fixup-notes nelements)
     (let* ((fixup (fixup-note-fixup note))
