@@ -14,22 +14,6 @@
       (vector-length-constraint-var thing)
       thing))
 
-(defun find-equality-constraint (operator x y not-p)
-  (let ((constraints (lambda-var-equality-constraints (constraint-var x))))
-    (flet ((constraints-eq-p (a b)
-             (or (eq a b)
-                 (and (vector-length-constraint-p a)
-                      (vector-length-constraint-p b)
-                      (eq (vector-length-constraint-var a)
-                          (vector-length-constraint-var b))))))
-     (when constraints
-       (loop for con across constraints
-             when (and (eq (equality-constraint-operator con) operator)
-                       (eq (constraint-not-p con) not-p)
-                       (constraints-eq-p (constraint-x con) x)
-                       (constraints-eq-p (constraint-y con) y))
-             return con)))))
-
 (defun invert-operator (operator)
   (case operator
     (< '>)
@@ -38,39 +22,91 @@
     (>= '<=)
     (t operator)))
 
-(defmacro do-equality-constraints ((con op not-p) var constraints &body body)
+(defun not-operator (operator)
+  (case operator
+    (> '<=)
+    (< '>=)
+    (>= '<)
+    (<= '>)))
+
+(defmacro do-equality-constraints ((con op not-p &optional amount) var constraints &body body)
   (once-only ((var var)
               (constraints constraints))
     `(do-conset-constraints-intersection
          (,con (,constraints (lambda-var-equality-constraints (constraint-var ,var))))
-       (flet ((body (,con ,op ,not-p) ,@body))
+       (flet ((body (,con ,op ,@(and amount (list amount)) ,not-p) ,@body))
          (cond ((vector-constraint-eq-p (constraint-x ,con) ,var)
                 (body (constraint-y ,con)
                       (equality-constraint-operator ,con)
+                      ,@(and amount
+                             `((equality-constraint-amount ,con)))
                       (equality-constraint-not-p ,con)))
                ((vector-constraint-eq-p (constraint-y ,con) ,var)
                 (body (constraint-x ,con)
                       (invert-operator (equality-constraint-operator ,con))
+                      ,@(and amount
+                             `((equality-constraint-amount ,con)))
                       (equality-constraint-not-p ,con))))))))
 
-(defun find-or-create-equality-constraint (operator x y not-p)
-  (or (find-equality-constraint operator x y not-p)
-      (let ((new (make-equality-constraint (length *constraint-universe*)
-                                           operator
-                                           x y not-p)))
-        (vector-push-extend new *constraint-universe* (1+ (length *constraint-universe*)))
-        (flet ((add (var)
-                 (conset-adjoin new (lambda-var-constraints var))
-                 (macrolet ((ensure-vec (place)
-                              `(or ,place
-                                   (setf ,place
-                                         (make-array 8 :adjustable t :fill-pointer 0)))))
-                   (vector-push-extend new (ensure-vec (lambda-var-equality-constraints var))))))
-          (add (constraint-var x))
-          (let ((y (constraint-var y)))
-            (when (lambda-var-p y)
-              (add y))))
-        new)))
+(defun find-equality-constraint (operator amount x x-var y y-key not-p)
+  (let ((constraints (lambda-var-equality-constraints-hash x-var)))
+    (when constraints
+      (let ((constraints (gethash y-key constraints)))
+        (if (typep y-key 'sb-kernel::type-class)
+            (loop for con in constraints
+                  when (and (eq (equality-constraint-operator con) operator)
+                            (eq (constraint-not-p con) not-p)
+                            (eql (equality-constraint-amount con) amount)
+                            (vector-constraint-eq-p (constraint-x con) x)
+                            (type= (constraint-y con) y))
+                  return con)
+            (loop for con in constraints
+                  when (and (eq (equality-constraint-operator con) operator)
+                            (eq (constraint-not-p con) not-p)
+                            (vector-constraint-eq-p (constraint-x con) x)
+                            (vector-constraint-eq-p (constraint-y con) y)
+                            (eql (equality-constraint-amount con) amount))
+                  return con))))))
+
+(defun find-or-create-equality-constraint (operator x y not-p &optional (amount 0))
+  (unless amount
+    (setf amount 0))
+  (let ((x-var (constraint-var x))
+        (cache-key (typecase y
+                     (numeric-type ;; eq-comparable
+                      y)
+                     (ctype
+                      (sb-kernel::type-class y))
+                     (vector-length-constraint y
+                      (vector-length-constraint-var y))
+                     (t
+                      y))))
+    (or (find-equality-constraint operator amount x x-var y cache-key not-p)
+        (let ((new (make-equality-constraint (length *constraint-universe*)
+                                             operator
+                                             x y not-p
+                                             amount)))
+          (vector-push-extend new *constraint-universe* (1+ (length *constraint-universe*)))
+          (flet ((add (var)
+                   (conset-adjoin new (lambda-var-constraints var))
+                   (macrolet ((ensure-vec (place)
+                                `(or ,place
+                                     (setf ,place
+                                           (make-array 8 :adjustable t :fill-pointer 0)))))
+                     (vector-push-extend new (ensure-vec (lambda-var-equality-constraints var))))))
+            (add x-var)
+            (let ((hash (or (lambda-var-equality-constraints-hash x-var)
+                            (setf (lambda-var-equality-constraints-hash x-var)
+                                  (make-hash-table :test #'eq)))))
+              (push new (gethash cache-key hash)))
+            (let ((y (constraint-var y)))
+              (when (lambda-var-p y)
+                (add y))))
+          new))))
+
+(defun conset-add-equality-constraint (conset operator x y not-p &optional (amount 0))
+  (conset-adjoin (find-or-create-equality-constraint operator x y not-p amount)
+                 conset))
 
 (defun add-eq-constraint (var lvar gen)
   (let ((var2 (ok-lvar-lambda-var lvar gen)))
@@ -91,10 +127,6 @@
     (and array-var
          (make-vector-length-constraint array-var))))
 
-(defun conset-add-equality-constraint (conset operator x y not-p)
-  (conset-adjoin (find-or-create-equality-constraint operator x y not-p)
-                 conset))
-
 (defun vector-constraint-eq-p (con1 con2)
   (or (eq con1 con2)
       (and (vector-length-constraint-p con1)
@@ -112,23 +144,39 @@
 
 (defconstant-eqx +eq-ops+ '(eq eql =) #'equal)
 
-(defun inherit-equality-p (new from not-p)
+(defun inherit-equality-p (new from not-p &optional (min-amount 0) max-amount (from-amount 0) set)
+  (unless min-amount
+    (setf min-amount 0))
+  (unless from-amount
+    (setf from-amount 0))
   (case new
-    ((char= two-arg-char-equal . #.+eq-ops+)
+    ((char= char-equal . #.+eq-ops+)
      (unless not-p
-       from))
+       (values from from-amount)))
     (<
      (if not-p
          (case from
-           ((> >=) '<))
+           ((> >=) (values '< (+ from-amount min-amount))))
          (case from
-           ((< <= . #.+eq-ops+) '<))))
+           ((< <= . #.+eq-ops+) (values '< (+ from-amount min-amount)))
+           (>
+            (when (and max-amount
+                       (not set))
+              (let ((diff (- from-amount max-amount)))
+                (cond ((> diff 0)
+                       (values '> diff)))))))))
     (>
      (if not-p
          (case from
-           ((< <=) '>))
+           ((< <=) (values '> (+ from-amount min-amount))))
          (case from
-           ((> >= . #.+eq-ops+) '>))))
+           ((> >= . #.+eq-ops+) (values '> (+ from-amount min-amount)))
+           (<
+            (when (and max-amount
+                       (not set))
+              (let ((diff (- from-amount max-amount)))
+                (cond ((> diff 0)
+                       (values '< diff)))))))))
     (<=
      (if not-p
          (case from
@@ -142,103 +190,122 @@
          (case from
            ((> >= . #.+eq-ops+) '>=))))))
 
-(defun add-equality-constraint (operator first second constraints consequent-constraints alternative-constraints)
-  (let* ((constant-x (and (lvar-p first)
-                          (constant-lvar-p first)))
-         (constant-y (and (lvar-p second)
-                          (constant-lvar-p second)))
-         (x (cond (constant-x
-                   (nth-value 1 (lvar-value first)))
-                  ((lambda-var/vector-length-p first)
-                   first)
-                  ((ok-lvar-lambda-var/vector-length first constraints))
-                  (t
-                   first)))
-         (y (cond (constant-y
-                   (nth-value 1 (lvar-value second)))
-                  ((lambda-var/vector-length-p second)
-                   second)
-                  ((ok-lvar-lambda-var/vector-length second constraints))
-                  (t
-                   second)))
-         (x-type (etypecase first
-                   (lvar (lvar-type first))
-                   (lambda-var (lambda-var-type first))
-                   (vector-length-constraint (specifier-type 'index))))
-         (y-type (etypecase second
-                   (lvar (lvar-type second))
-                   (lambda-var (lambda-var-type second))
-                   (vector-length-constraint (specifier-type 'index)))))
-    (when (lvar-p first)
-      (constraint-propagate-back first operator second
-                                 constraints
-                                 consequent-constraints
-                                 alternative-constraints))
-    (unless (lambda-var/vector-length-p x)
-      (unless (lambda-var/vector-length-p y)
-        (return-from add-equality-constraint))
-      (rotatef first second)
-      (rotatef x y)
-      (rotatef x-type y-type)
-      (rotatef constant-x constant-y)
-      (setf operator (invert-operator operator)))
-    (flet ((replace-var (var with)
-             (cond ((eq var with)
-                    var)
-                   ((vector-length-constraint-p var)
-                    (make-vector-length-constraint with))
-                   (t
-                    with)))
-           (add (x y &optional (operator operator) (alternative-constraints alternative-constraints))
-             (unless (lambda-var/vector-length-p x)
-               (unless (lambda-var/vector-length-p y)
-                 (return-from add))
-               (rotatef x y)
-               (setf operator (invert-operator operator)))
-             (conset-add-equality-constraint consequent-constraints operator x y nil)
-             (when alternative-constraints
-               (conset-add-equality-constraint alternative-constraints operator x y t))))
-      (do-eql-vars (eql-x ((constraint-var x) constraints))
-        (let ((x (replace-var x eql-x)))
-          (add x y)
-          (when (and (vector-length-constraint-p x)
-                     (not constant-y)
-                     (neq y-type *universal-type*))
-            (add x y-type))))
-      (if (lambda-var/vector-length-p y)
-          (do-eql-vars (eql-y ((constraint-var y) constraints))
-            (let ((y (replace-var y eql-y)))
-              (add x y)
-              (when (and (vector-length-constraint-p y)
-                         (not constant-x)
-                         (neq x-type *universal-type*))
-                (add x-type y))))
-          (add x y))
-      (flet ((inherit (x y x-type operator)
-               (when (lambda-var/vector-length-p y)
-                 (do-equality-constraints (in-y in-op in-not-p) y constraints
-                   (unless (eq in-y y)
-                     (let ((inherit (inherit-equality-p operator in-op in-not-p)))
-                       (when inherit
-                         (add x in-y inherit nil)
-                         (when (and (vector-length-constraint-p in-y)
-                                    (not (constant-p x)))
-                           (add x-type in-y inherit nil)))))))))
-        (inherit x y x-type operator)
-        (inherit y x y-type (invert-operator operator))
-        (when (lvar-p y)
-          (loop for (in-op in-lvar) in (lvar-result-constraints y)
-                do
-                (let ((inherit (inherit-equality-p operator in-op nil)))
-                  (when inherit
-                    (add-equality-constraint inherit x in-lvar
-                                             constraints consequent-constraints nil)))))))))
+(defun add-equality-constraint (operator first second constraints consequent-constraints alternative-constraints
+                                &optional (min-amount 0) max-amount)
+  (case operator
+    (vector-length
+     (let ((var (if (lambda-var-p first)
+                    first
+                    (lvar-dest-var first))))
+       (when (and var
+                  (lambda-var-constraints var))
+         (add-equality-constraint 'eq (make-vector-length-constraint var) second
+                                  constraints consequent-constraints alternative-constraints))))
+    (t
+     (let* ((x (cond ((lambda-var/vector-length-p first)
+                      first)
+                     ((ok-lvar-lambda-var/vector-length first constraints))
+                     (t
+                      first)))
+            (y (cond ((lambda-var/vector-length-p second)
+                      second)
+                     ((ok-lvar-lambda-var/vector-length second constraints))
+                     (t
+                      second)))
+            (x-type (etypecase first
+                      (lvar (lvar-type first))
+                      (lambda-var (lambda-var-type first))
+                      (vector-length-constraint (specifier-type 'index))))
+            (y-type (etypecase second
+                      (lvar (lvar-type second))
+                      (lambda-var (lambda-var-type second))
+                      (vector-length-constraint (specifier-type 'index)))))
+       (when (lvar-p first)
+         (constraint-propagate-back first operator second
+                                    constraints
+                                    consequent-constraints
+                                    alternative-constraints))
+       (unless (lambda-var/vector-length-p x)
+         (unless (lambda-var/vector-length-p y)
+           (return-from add-equality-constraint))
+         (rotatef first second)
+         (rotatef x y)
+         (rotatef x-type y-type)
+         (setf operator (invert-operator operator)))
+       (flet ((replace-var (var with)
+                (cond ((eq var with)
+                       var)
+                      ((vector-length-constraint-p var)
+                       (make-vector-length-constraint with))
+                      (t
+                       with)))
+              (add (x y &key (operator operator)
+                             (consequent consequent-constraints)
+                             (alternative alternative-constraints)
+                             (amount min-amount))
+                (when (eq x y)
+                  (return-from add))
+                (unless (lambda-var/vector-length-p x)
+                  (unless (lambda-var/vector-length-p y)
+                    (return-from add))
+                  (rotatef x y)
+                  (setf operator (invert-operator operator)))
+                (conset-add-equality-constraint consequent operator x y nil amount)
+                (when alternative
+                  (conset-add-equality-constraint alternative operator x y t amount))))
+         (do-eql-vars (eql-x ((constraint-var x) constraints))
+           (let ((x (replace-var x eql-x)))
+             (when (neq y-type *universal-type*)
+               (add x y-type :amount 0))
+             (if (lambda-var/vector-length-p y)
+                 (do-eql-vars (eql-y ((constraint-var y) constraints))
+                   (let ((y (replace-var y eql-y)))
+                     (add x y)
+                     (when (neq x-type *universal-type*)
+                       (add x-type y :amount 0))))
+                 (add x y))
+             (add x y)))
+         (flet ((inherit (x y x-type operator)
+                  (when (lambda-var/vector-length-p y)
+                    (do-equality-constraints (in-y in-op in-not-p in-amount) y constraints
+                      (unless (eq in-y y)
+                        (flet ((add (operator target)
+                                 (multiple-value-bind (inherit inherit-amount)
+                                     (inherit-equality-p operator in-op in-not-p min-amount max-amount in-amount)
+                                   (when inherit
+                                     (add x in-y :operator inherit
+                                                 :amount inherit-amount
+                                                 :consequent target
+                                                 :alternative nil)
+                                     (when (neq x-type *universal-type*)
+                                       (add x-type in-y :operator inherit
+                                                        :alternative nil
+                                                        :consequent target
+                                                        :amount 0))))))
+                          (add operator consequent-constraints)
+                          (when alternative-constraints
+                            (add (not-operator operator) alternative-constraints))))))))
+           (inherit x y x-type operator)
+           (inherit y x y-type (invert-operator operator))
+           (when (lvar-p y)
+             (loop for (in-op in-lvar in-min-amount) in (lvar-result-constraints y)
+                   do
+                   (unless (eq in-lvar first)
+                     (flet ((add (operator target)
+                              (multiple-value-bind (inherit inherit-amount)
+                                  (inherit-equality-p operator in-op nil min-amount max-amount in-min-amount)
+                                (when inherit
+                                  (add-equality-constraint inherit x in-lvar
+                                                           constraints target nil inherit-amount)))))
+                       (add operator consequent-constraints)
+                       (when alternative-constraints
+                         (add (not-operator operator) alternative-constraints))))))))))))
 
 (defun add-equality-constraints (operator args constraints
                                  consequent-constraints
                                  alternative-constraints)
   (case operator
-    ((char= two-arg-char-equal > <  <= >=
+    ((char= char-equal > <  <= >=
             . #.+eq-ops+)
      (when (= (length args) 2)
        (add-equality-constraint operator (first args) (second args)
@@ -263,8 +330,53 @@
                                               (if replace-x
                                                   (constraint-y con)
                                                   (replace-var (constraint-y con) var))
-                                              (constraint-not-p con))
+                                              (constraint-not-p con)
+                                              (equality-constraint-amount con))
           target))))))
+
+;;; Ignore AMOUNT
+(defun join-equality-constraints (var block in)
+  (let* ((constraints (make-hash-table :test #'equal))
+         (pred (block-pred block))
+         (i -1))
+    (loop for pred in pred
+          do
+          (let ((out (block-out-for-successor pred block)))
+            (when out
+              (incf i)
+              (do-equality-constraints (in-con in-op not-p amount) var out
+                (let ((existing (gethash (list in-con in-op not-p) constraints)))
+                  (cond ((= (if existing
+                                (car existing)
+                                -1)
+                            (1- i))
+                         (setf (gethash (list in-con in-op not-p) constraints)
+                               (list i
+                                     (if existing
+                                         (min amount (second existing))
+                                         amount)
+                                     (second existing))))
+                        ((and existing
+                              (= (car existing) i))
+                         ;; Maximize the current block value while
+                         ;; not exceeding the overall minimal amount.
+                         (let ((overall-min (third existing)))
+                           (setf (gethash (list in-con in-op not-p) constraints)
+                                 (list i
+                                       (if overall-min
+                                           (min (max amount (second existing)) overall-min)
+                                           (max amount (second existing)))
+                                       overall-min))))))))))
+    (dohash ((key value) constraints)
+      (when (= (car value) i)
+        (destructuring-bind (y op not-p) key
+          (conset-adjoin
+           (find-or-create-equality-constraint op
+                                               var
+                                               y
+                                               not-p
+                                               (second value))
+           in))))))
 
 (defun try-equality-constraint (call gen)
   (let ((constraint (fun-info-equality-constraint (basic-combination-fun-info call)))
@@ -276,45 +388,43 @@
             (replace-combination-with-constant result call))
           t)))))
 
-(defun map-equality-constraints (x y constraints
-                                 function)
-  (let* ((constant-x (and (lvar-p x)
-                          (constant-lvar-p x)))
-         (constant-y (and (lvar-p y)
-                          (constant-lvar-p y)))
-         (x (cond (constant-x
-                   (nth-value 1 (lvar-value x)))
-                  ((lambda-var/vector-length-p x)
-                   x)
-                  ((ok-lvar-lambda-var/vector-length x constraints t))
+(defun map-equality-constraints (lvar-x lvar-y constraints function)
+  (let* ((x (cond ((lambda-var/vector-length-p lvar-x)
+                   lvar-x)
+                  ((ok-lvar-lambda-var/vector-length lvar-x constraints t))
                   (t
-                   x)))
-         (y (cond (constant-y
-                   (nth-value 1 (lvar-value y)))
-                  ((ok-lvar-lambda-var/vector-length y constraints t))
+                   lvar-x)))
+         (y (cond ((ok-lvar-lambda-var/vector-length lvar-y constraints t))
                   (t
-                   y)))
+                   lvar-y)))
          (invert))
     (unless (lambda-var/vector-length-p x)
       (unless (lambda-var/vector-length-p y)
         (return-from map-equality-constraints))
-
       (rotatef x y)
-      (rotatef constant-x constant-y)
+      (rotatef lvar-x lvar-y)
       (setf invert t))
     (do-equality-constraints (in-y op not-p) x constraints
       (when (or (vector-constraint-eq-p in-y y)
-                (and constant-y
-                     (constant-p in-y)
-                     (let ((a (constant-value in-y))
-                           (b (constant-value y)))
-                       (and (realp a)
-                            (realp b)
-                            (case op
-                              (< (unless not-p
-                                   (sb-xc:< a b)))
-                              (> (unless not-p
-                                   (sb-xc:> a b))))))))
+                (and (ctype-p in-y)
+                     (lvar-p lvar-y)
+                     (or (and (type-singleton-p in-y)
+                              (type= in-y (lvar-type lvar-y)))
+                         (let ((interval-in-y (type-approximate-interval in-y))
+                               (interval-y (type-approximate-interval (lvar-type lvar-y))))
+                           (when (and interval-y
+                                      interval-in-y)
+                             (let* ((a interval-in-y)
+                                    (b interval-y))
+                               (case op
+                                 (< (unless not-p
+                                      (interval-< a b)))
+                                 (> (unless not-p
+                                      (interval-< b a)))
+                                 (eq
+                                  (unless not-p
+                                    (cond ((interval-< b a)
+                                           (setf op '>))))))))))))
         (funcall function (if invert
                               (invert-operator op)
                               op)
@@ -333,7 +443,7 @@
                                   ((> <)
                                    (unless not-p
                                      (return nil)))
-                                  ((= <= >= two-arg-char-equal)
+                                  ((= <= >= char-equal)
                                    (when not-p
                                      (return nil))))))
     :give-up))
@@ -348,7 +458,7 @@
                                   ((> <)
                                    (unless not-p
                                      (return nil)))
-                                  ((= <= >= two-arg-char-equal)
+                                  ((= <= >= char-equal)
                                    (when not-p
                                      (return nil))))))
     :give-up))
@@ -391,17 +501,17 @@
                                 (case op
                                   ((char= eq eql)
                                    (return (not not-p)))
-                                  (two-arg-char-equal
+                                  (char-equal
                                    (when not-p
                                      (return nil))))))
     :give-up))
 
-(defoptimizer (two-arg-char-equal equality-constraint) ((x y)  node gen)
+(defoptimizer (char-equal equality-constraint) ((x y)  node gen)
   (block nil
     (map-equality-constraints x y gen
                               (lambda (op not-p)
                                 (case op
-                                  (two-arg-char-equal
+                                  (char-equal
                                    (return (not not-p)))
                                   ((char= eq eql)
                                    (unless not-p
@@ -483,7 +593,6 @@
     :give-up))
 
 (defoptimizer (- equality-constraint) ((x y) node gen)
-  gen
   (when (and (csubtypep (lvar-type x) (specifier-type 'integer))
              (csubtypep (lvar-type y) (specifier-type 'integer)))
     (block nil
@@ -507,6 +616,15 @@
                  (< (derive (integer * (0)))))))))))
   :give-up)
 
+(defoptimizer (%check-bound constraint-propagate) ((array dimension index) node gen)
+  (add-equality-constraint '< index dimension gen gen nil)
+  (let ((var (ok-lvar-lambda-var index gen))
+        (type (if (constant-lvar-p dimension)
+                  (specifier-type `(integer 0 (,(lvar-value dimension))))
+                  (specifier-type 'index))))
+    (when var
+      (list (list 'typep var type nil)))))
+
 (defoptimizer (%check-bound equality-constraint) ((array dimension index) node gen)
   (let ((array-var (ok-lvar-lambda-var array gen)))
     (when (and array-var
@@ -522,7 +640,11 @@
   :give-up)
 
 (deftransform %check-bound ((array dimension index) ((simple-array * (*)) t t) * :node node)
-  (if (eq (combination-info node) 'array-in-bounds-p)
+  (if (or (eq (combination-info node) 'array-in-bounds-p)
+          (let ((index (type-approximate-interval (lvar-type index)))
+                (dim (type-approximate-interval (lvar-type dimension))))
+            (and index dim
+                 (interval-< index dim))))
       ;; It's in bounds but it may be of the wrong type
       `(progn (the (and fixnum unsigned-byte) index)
               (values))
@@ -581,7 +703,9 @@
                (when (and (ref-p ref1) (ref-p ref2))
                  (let ((var1 (ok-ref-lambda-var ref1))
                        (var2 (ok-ref-lambda-var ref2)))
-                   (when (and var1 var2)
+                   (when (and var1 var2
+                              (not (lambda-var-sets var1))
+                              (not (lambda-var-sets var2)))
                      (labels ((normalize-not (op not)
                                 (if not
                                     (case op
@@ -620,34 +744,86 @@
                  r)))))))
 
 (defun add-var-result-constraints (var lvar constraints &optional (target constraints))
-  (loop for (operator second) in (lvar-result-constraints lvar)
+  (loop for (operator second min-amount max-amount) in (lvar-result-constraints lvar)
         do
-        (add-equality-constraint operator var second constraints target nil))
+        (add-equality-constraint operator var second constraints target nil min-amount max-amount))
   (when (lambda-var-p var)
     (let ((vector-length (vector-length-var-p lvar constraints)))
       (when vector-length
         (conset-add-equality-constraint target 'eq var vector-length nil)))))
 
+(defun add-mv-let-result-constraints (call fun constraints &optional (target constraints))
+  (let ((vars (lambda-vars fun))
+        (lvars (basic-combination-args call)))
+    (when (= (length lvars) 1)
+      (loop for (nth-value operator second min-amount max-amount) in (nth-value 1 (lvar-result-constraints (car lvars)))
+            do
+            (add-equality-constraint operator (elt vars nth-value) second constraints target nil min-amount max-amount)))))
+
+;;; Need a separate function because a set clears the constraints of the var
+(defun add-set-constraints (var lvar constraints)
+  (let (gen)
+    (loop for (operator second min-amount max-amount) in (lvar-result-constraints lvar)
+          for y = (if (lambda-var-p second)
+                      second
+                      (ok-lvar-lambda-var second constraints))
+          when y
+          do
+          (do-eql-vars (eql-y (y constraints))
+            (when (eq eql-y var)
+              (do-equality-constraints (in-y in-op in-not-p in-amount) var constraints
+                (unless (eq in-y var)
+                  (multiple-value-bind (inherit inherit-amount)
+                      ;; Avoid changing the direction of inequalities
+                      ;; because it might be done in a loop.
+                      (inherit-equality-p operator in-op in-not-p min-amount max-amount in-amount t)
+                    (when inherit
+                      (conset-add-equality-constraint (or gen
+                                                          (setf gen (make-conset)))
+                                                      inherit var in-y nil
+                                                      inherit-amount))))))))
+    gen))
+
 (defoptimizer (- constraint-propagate-result) ((a b) node)
-  (and (csubtypep (lvar-type a) (specifier-type 'integer))
-       (csubtypep (lvar-type b) (specifier-type '(integer 0)))
-       (let ((plusp (csubtypep (lvar-type b) (specifier-type '(integer 1)))))
-         (list (list (if plusp
-                         '<
-                         '<=)
-                     a)))))
+  (when (csubtypep (lvar-type a) (specifier-type 'integer))
+    (cond ((csubtypep (lvar-type b) (specifier-type '(integer 0)))
+           (let ((interval (type-approximate-interval (lvar-type b))))
+             (when interval
+               (let ((low (interval-low interval)))
+                 (list (if (> low 0)
+                           (list '< a low (interval-high interval))
+                           (list '<= a)))))))
+          ((csubtypep (lvar-type b) (specifier-type '(integer * -1)))
+           (let ((interval (type-approximate-interval (lvar-type b))))
+             (when interval
+               (let* ((interval (interval-neg interval))
+                      (low (interval-low interval)))
+                 (list (if (> low 0)
+                           (list '> a low (interval-high interval))
+                           (list '>= a))))))))))
 
 (defoptimizer (+ constraint-propagate-result) ((x y) node)
   (let (r)
     (flet ((try (a b)
-             (and (csubtypep (lvar-type a) (specifier-type 'integer))
-                  (csubtypep (lvar-type b) (specifier-type '(integer 0)))
-                  (let ((plusp (csubtypep (lvar-type b) (specifier-type '(integer 1)))))
-                    (push (list (if plusp
-                                    '>
-                                    '>=)
-                                a)
-                          r)))))
+             (when (csubtypep (lvar-type a) (specifier-type 'integer))
+               (cond ((csubtypep (lvar-type b) (specifier-type '(integer 0)))
+                      (let ((interval (type-approximate-interval (lvar-type b))))
+                        (when interval
+                          (let ((low (interval-low interval)))
+                            (push (if (> low 0)
+                                      (list '> a low (interval-high interval))
+                                      (list '>= a))
+                                  r)))))
+                     ((csubtypep (lvar-type b) (specifier-type '(integer * -1)))
+                      (let ((interval (type-approximate-interval (lvar-type b))))
+                        (when interval
+                          (let* ((interval (interval-neg interval))
+                                 (low (interval-low interval)))
+                            (push
+                             (if (> low 0)
+                                 (list '< a low (interval-high interval))
+                                 (list '<= a))
+                             r)))))))))
       (try x y)
       (try y x))
     r))
@@ -676,3 +852,17 @@
 
 (defoptimizer (/ constraint-propagate-result) ((x y) node)
   (div-constraints x y))
+
+(defoptimizer (allocate-vector constraint-propagate-result) ((#+ubsan poisoned type length words) node)
+  (list (list 'vector-length length)))
+
+(defoptimizer (%make-array constraint-propagate-result) ((length &rest args) node)
+  (list (list 'vector-length length)))
+
+(defoptimizer (make-sequence constraint-propagate-result) ((type length &rest args) node)
+  (list (list 'vector-length length)))
+
+(defoptimizer (floor constraint-propagate-result) ((x y) node)
+  (when (csubtypep (lvar-type y) (specifier-type '(real 0)))
+    (values nil
+            (list (list 1 '< y)))))
