@@ -86,11 +86,6 @@
 (defknown (eq eql) (t t) boolean
   (movable foldable flushable commutative))
 (defknown (equal equalp) (t t) boolean (foldable flushable recursive))
-
-#+(or x86 x86-64 arm arm64)
-(defknown fixnum-mod-p (t fixnum) boolean
-  (movable flushable always-translatable))
-
 
 ;;;; classes
 
@@ -157,6 +152,7 @@
 ;;; represented by a call to VALUES.
 (defknown values (&rest t) * (movable flushable))
 (defknown values-list (list) * (movable foldable unsafely-flushable))
+(defknown reverse-values-list (list index) * (movable foldable unsafely-flushable))
 
 ;;;; from the "Macros" chapter:
 
@@ -186,10 +182,10 @@
 (defknown symbol-name (symbol) simple-string (movable foldable flushable))
 (defknown make-symbol (string) symbol (flushable))
 ;; %make-symbol is the internal API, but the primitive object allocator
-;; is %%make-symbol, because when immobile space feature is present,
+;; is %alloc-symbol, because when immobile space feature is present,
 ;; we dispatch to either the C allocator or the Lisp allocator.
 (defknown %make-symbol (fixnum simple-string) symbol (flushable))
-(defknown sb-vm::%%make-symbol (simple-string) symbol (flushable))
+(defknown sb-vm::%alloc-symbol (simple-string) symbol (flushable))
 (defknown copy-symbol (symbol &optional t) symbol (flushable))
 (defknown gensym (&optional (or string unsigned-byte)) symbol ())
 (defknown symbol-package (symbol) (or package null) (flushable))
@@ -282,6 +278,10 @@
     (fixnum real fixnum) boolean
     (foldable flushable movable no-verify-arg-count))
 
+(defknown (check-range<=)
+    (fixnum t fixnum) boolean
+    (foldable flushable movable no-verify-arg-count))
+
 (defknown (two-arg-gcd two-arg-lcm two-arg-and two-arg-ior two-arg-xor two-arg-eqv)
   (integer integer) integer
   (no-verify-arg-count))
@@ -303,6 +303,10 @@
 
 (defknown expt (number number) number
   (movable foldable flushable recursive))
+
+(defknown sb-kernel::intexp ((or rational (complex rational)) integer) rational
+  (movable foldable flushable recursive no-verify-arg-count))
+
 (defknown log (number &optional real) irrational
   (movable foldable flushable recursive))
 (defknown sqrt (number) irrational
@@ -356,11 +360,10 @@
 (defknown unary-truncate (real) (values integer real)
   (movable foldable flushable no-verify-arg-count))
 
-;;; KLUDGE: Don't fold these, the compiler may call it on floats that
-;;; do not truncate into a bignum, and the functions do not check
-;;; their input values and produce corrupted memory.
 (defknown unary-truncate-single-float-to-bignum (single-float) (values bignum (eql $0f0))
-    (#+(or) foldable movable flushable fixed-args))
+    (foldable movable flushable fixed-args)
+  :folder #'truncate)
+
 (defknown unary-truncate-double-float-to-bignum (double-float)
     (values #+64-bit bignum #-64-bit integer
             (and
@@ -368,12 +371,15 @@
                     (not (or riscv ppc64))) ;; they can't survive cold-init
              (eql $0d0)
              double-float))
-   (#+(or) foldable movable flushable fixed-args))
+   (foldable movable flushable fixed-args)
+  :folder #'truncate)
 
 (defknown %unary-truncate-single-float-to-bignum (single-float) bignum
-   (#+(or) foldable movable flushable fixed-args))
+   (foldable movable flushable fixed-args)
+  :folder (lambda (float) (values (truncate float))))
 (defknown %unary-truncate-double-float-to-bignum (double-float) bignum
-   (#+(or) foldable movable flushable fixed-args))
+   (foldable movable flushable fixed-args)
+  :folder (lambda (float) (values (truncate float))))
 
 (defknown (subtract-bignum add-bignums) (bignum bignum) integer
     (movable flushable no-verify-arg-count))
@@ -407,7 +413,11 @@
   (movable foldable unsafely-flushable))
 (defknown scale-float (float integer) float
   (movable foldable unsafely-flushable))
-(defknown float-radix (float) float-radix
+(defknown float-radix (float) sb-kernel::%float-radix
+  (movable foldable unsafely-flushable))
+(defknown float-sign-bit (float) bit
+  (movable foldable unsafely-flushable))
+(defknown float-sign-bit-set-p (float) boolean
   (movable foldable unsafely-flushable))
 ;;; This says "unsafely flushable" as if to imply that there is a possibility
 ;;; of signaling a condition in safe code, but in practice we can never trap
@@ -429,7 +439,7 @@
                        (t
                         (specifier-type 'float)))))
 
-(defknown (float-digits float-precision) (float) float-digits
+(defknown (float-digits float-precision) (float) sb-kernel::%float-digits
   (movable foldable unsafely-flushable))
 (defknown integer-decode-float (float)
     (values double-float-significand float-int-exponent (member -1 1))
@@ -497,8 +507,8 @@
 (defknown random-state-p (t) boolean (movable foldable flushable))
 
 ;;;; from the "Characters" chapter:
-(defknown (standard-char-p graphic-char-p alpha-char-p
-                           upper-case-p lower-case-p both-case-p alphanumericp)
+(defknown (graphic-char-p alpha-char-p standard-char-p
+           upper-case-p lower-case-p both-case-p alphanumericp)
   (character) boolean (movable foldable flushable))
 
 (defknown digit-char-p (character &optional (integer 2 36))
@@ -512,19 +522,16 @@
 ;; generated by the compiler, and successive args checked by hand.
 ;; The case-insensitive functions don't need any checks, since the underlying
 ;; two-arg case-insensitive function does it, except when it isn't called.
-(defknown (char=)
+(defknown (char= char-equal)
   (character &rest character) boolean (movable foldable flushable commutative))
 
 (defknown (char/= char< char> char<= char>= char-not-equal)
   (character &rest character) boolean (movable foldable flushable))
-(defknown (char-equal char-lessp char-greaterp char-not-greaterp char-not-lessp)
+(defknown (char-lessp char-greaterp char-not-greaterp char-not-lessp)
   (character &rest character) boolean (movable foldable flushable))
 
-(defknown (two-arg-char-equal)
-    (character character) boolean
-  (movable foldable flushable commutative no-verify-arg-count))
-
-(defknown (two-arg-char-not-equal
+(defknown (two-arg-char-equal
+           two-arg-char-not-equal
            two-arg-char-lessp
            two-arg-char-not-lessp
            two-arg-char-greaterp
@@ -533,17 +540,17 @@
   (movable foldable flushable no-verify-arg-count))
 
 (defknown character (t) character (movable foldable unsafely-flushable))
-(defknown char-code (character) char-code (movable foldable flushable))
+(defknown char-code (character) %char-code (movable foldable flushable))
 (defknown (char-upcase char-downcase) (character) character
   (movable foldable flushable))
 (defknown digit-char (unsigned-byte &optional (integer 2 36))
   (or character null) (movable foldable flushable))
-(defknown char-int (character) char-code (movable foldable flushable))
-(defknown char-name (character) (or simple-string null)
+(defknown char-int (character) %char-code (movable foldable flushable))
+(defknown char-name (character) (or simple-base-string null)
   (movable foldable flushable))
 (defknown name-char (string-designator) (or character null)
   (movable foldable flushable))
-(defknown code-char (char-code) character
+(defknown code-char (%char-code) character
   ;; By suppressing constant folding on CODE-CHAR when the
   ;; cross-compiler is running in the cross-compilation host vanilla
   ;; ANSI Common Lisp, we can use CODE-CHAR expressions to delay until
@@ -575,6 +582,9 @@
   :derive-type (sequence-result-nth-arg 0 :preserve-dimensions t
                                           :preserve-vector-type t))
 
+(defknown list-reverse-into-vector (proper-list) simple-vector
+  (flushable no-verify-arg-count))
+
 (defknown make-sequence (type-specifier index
                                         &key
                                         (:initial-element t))
@@ -586,17 +596,29 @@
   :derive-type (creation-result-type-specifier-nth-arg 0))
 
 (defknown %concatenate-to-string (&rest sequence) simple-string
-  (flushable no-verify-arg-count))
+  (flushable))
 (defknown %concatenate-to-base-string (&rest sequence) simple-base-string
-  (flushable no-verify-arg-count))
+  (flushable))
 (defknown %concatenate-to-list (&rest sequence) list
-    (flushable no-verify-arg-count))
+    (flushable))
 (defknown %concatenate-to-simple-vector (&rest sequence) simple-vector
-  (flushable no-verify-arg-count))
+  (flushable))
 (defknown %concatenate-to-vector ((unsigned-byte #.sb-vm:n-widetag-bits) &rest sequence)
     vector
   (flushable no-verify-arg-count))
 
+(defknown %concatenate-to-string-subseq (&rest t) simple-string
+  (flushable))
+(defknown %concatenate-to-base-string-subseq (&rest t) simple-base-string
+  (flushable))
+(defknown %concatenate-to-list-subseq (&rest t) list
+    (flushable))
+(defknown %concatenate-to-simple-vector-subseq (&rest t) simple-vector
+  (flushable))
+(defknown %concatenate-to-vector-subseq ((unsigned-byte #.sb-vm:n-widetag-bits) &rest t)
+    vector
+  (flushable no-verify-arg-count))
+
 (defknown (possibly-base-stringize possibly-base-stringize-to-heap) (string) simple-string
   (flushable no-verify-arg-count))
 
@@ -628,7 +650,7 @@
                   (intern (concatenate 'string "VECTOR-MAP-INTO/"
                                        (string (sb-vm:saetp-primitive-type-name info)))
                           :sb-impl))
-    (simple-array index index function list)
+    (simple-array index index function &rest sequence)
     index
     (call no-verify-arg-count))
 
@@ -973,9 +995,8 @@
 ;;; - the number of bytes to allocate should be a fixnum
 ;;; - type-checking can use use efficient bit-masking approach
 ;;;   to combine the fixnum + range test into one instruction
-(eval-when (:compile-toplevel :load-toplevel :execute)
-  (defconstant make-list-limit
-    (ash most-positive-fixnum (- (+ sb-vm:word-shift 1)))))
+(defconstant make-list-limit
+  (ash most-positive-fixnum (- (+ sb-vm:word-shift 1))))
 (defknown make-list ((integer 0 #.make-list-limit) &key (:initial-element t)) list
   (movable flushable))
 (defknown %make-list ((integer 0 #.make-list-limit) t) list
@@ -995,6 +1016,8 @@
   :call-type-deriver #'append-call-type-deriver)
 
 (defknown copy-list (proper-or-dotted-list) list (flushable))
+(defknown sb-impl::copy-list-to (proper-or-dotted-list cons) cons
+  (flushable no-verify-arg-count))
 (defknown copy-alist (proper-list) list (flushable))
 (defknown copy-tree (t) t (flushable recursive))
 (defknown revappend (proper-list t) t (flushable))
@@ -1149,6 +1172,7 @@
 (defknown (sxhash psxhash) (t) hash-code (foldable flushable))
 (defknown hash-table-equalp (hash-table hash-table) boolean (foldable flushable))
 (defknown sb-impl::install-hash-table-lock (hash-table) sb-thread:mutex ())
+(defknown sb-vm::quick-try-mutex (sb-thread:mutex) boolean)
 ;; To avoid emitting code to test for nil-function-returned
 (defknown (sb-impl::signal-corrupt-hash-table
            sb-impl::signal-corrupt-hash-table-bucket)
@@ -1213,19 +1237,19 @@
 
 (defknown array-element-type (array) (or list symbol)
   (foldable flushable))
-(defknown array-rank (array) array-rank (foldable flushable))
+(defknown array-rank (array) %array-rank (foldable flushable))
 ;; FIXME: there's a fencepost bug, but for all practical purposes our
 ;; ARRAY-RANK-LIMIT is infinite, thus masking the bug. e.g. if the
 ;; exclusive limit on rank were 8, then your dimension numbers can
 ;; be in the range 0 through 6, not 0 through 7.
-(defknown array-dimension (array array-rank) index (foldable flushable))
+(defknown array-dimension (array %array-rank) index (foldable flushable))
 (defknown array-dimensions (array) list (foldable flushable))
 (defknown array-in-bounds-p (array &rest integer) boolean (foldable flushable)
   :call-type-deriver #'array-call-type-deriver)
-(defknown array-row-major-index (array &rest index) array-total-size
+(defknown array-row-major-index (array &rest index) sb-kernel::%array-total-size
   (foldable flushable)
   :call-type-deriver #'array-call-type-deriver)
-(defknown array-total-size (array) array-total-size (foldable flushable))
+(defknown array-total-size (array) sb-kernel::%array-total-size (foldable flushable))
 (defknown adjustable-array-p (array) boolean (movable foldable flushable))
 
 (defknown svref (simple-vector index) t (foldable flushable))
@@ -1509,6 +1533,11 @@
           (:junk-allowed t))
   (values (or integer null ()) index))
 
+(defknown (sb-impl::parse-integer10 sb-impl::parse-integer16)
+    (string index sequence-end t)
+    (values (or integer null ()) index)
+    (no-verify-arg-count))
+
 (defknown read-byte (stream &optional t t) t ()
   :derive-type (read-elt-type-deriver nil 'integer nil))
 
@@ -1638,39 +1667,39 @@
   (recursive))
 
 (defknown merge-pathnames
-  (pathname-designator &optional pathname-designator pathname-version)
+  (pathname-designator &optional pathname-designator sb-impl::%pathname-version)
   pathname
   ())
 
 (defknown make-pathname
  (&key (:defaults pathname-designator)
-       (:host (or string pathname-host))
-       (:device (or string pathname-device))
-       (:directory (or pathname-directory string (member :wild)))
-       (:name (or pathname-name string (member :wild)))
-       (:type (or pathname-type string (member :wild)))
-       (:version pathname-version) (:case pathname-component-case))
+       (:host (or string sb-impl::%pathname-host))
+       (:device (or string sb-impl::%pathname-device))
+       (:directory (or sb-impl::%pathname-directory string (member :wild)))
+       (:name (or sb-impl::%pathname-name string (member :wild)))
+       (:type (or sb-impl::%pathname-type string (member :wild)))
+       (:version sb-impl::%pathname-version) (:case pathname-component-case))
   pathname (unsafely-flushable))
 
 (defknown pathnamep (t) boolean (movable flushable))
 
 (defknown pathname-host (pathname-designator
                          &key (:case pathname-component-case))
-  pathname-host (flushable))
+  sb-impl::%pathname-host (flushable))
 (defknown pathname-device (pathname-designator
                            &key (:case pathname-component-case))
-  pathname-device (flushable))
+  sb-impl::%pathname-device (flushable))
 (defknown pathname-directory (pathname-designator
                               &key (:case pathname-component-case))
-  pathname-directory (flushable))
+  sb-impl::%pathname-directory (flushable))
 (defknown pathname-name (pathname-designator
                          &key (:case pathname-component-case))
-  pathname-name (flushable))
+  sb-impl::%pathname-name (flushable))
 (defknown pathname-type (pathname-designator
                          &key (:case pathname-component-case))
-  pathname-type (flushable))
+  sb-impl::%pathname-type (flushable))
 (defknown pathname-version (pathname-designator)
-  pathname-version (flushable))
+  sb-impl::%pathname-version (flushable))
 
 (defknown pathname= (pathname pathname) boolean (movable foldable flushable))
 
@@ -1730,6 +1759,7 @@
 (defknown (signal warn) (condition-designator-head &rest t) null)
 (defknown error (condition-designator-head &rest t) nil)
 (defknown %simple-error (condition-designator-head t &rest t) nil)
+(defknown %simple-type-error (condition-designator-head t t t &rest t) nil)
 (defknown cerror (format-control condition-designator-head &rest t) null)
 (defknown invalid-method-error (t format-control &rest t) *) ; FIXME: first arg is METHOD
 (defknown method-combination-error (format-control &rest t) *)
@@ -1866,6 +1896,23 @@
 
 (defknown array-storage-vector (array) (simple-array * (*))
     (any))
+
+(defknown octets-to-string ((vector (unsigned-byte 8))
+                            &key
+                            (:external-format t)
+                            (:start index)
+                            (:end sequence-end))
+  (or (simple-array character (*)) (simple-array base-char (*)))
+  (flushable))
+(defknown string-to-octets (string
+                            &key
+                            (:external-format t)
+                            (:start index)
+                            (:end sequence-end)
+                            (:null-terminate t))
+  (simple-array (unsigned-byte 8) (*))
+  (flushable))
+
 
 ;;;; magical compiler frobs
 
@@ -1901,7 +1948,7 @@
 (defknown %dynamic-extent-start () t)
 (defknown %special-bind (t t) t)
 (defknown %special-unbind (&rest symbol) t)
-(defknown %listify-rest-args (t index) list (flushable))
+(defknown %listify-rest-args (t index) list (flushable always-translatable))
 (defknown %more-arg-context (t t) (values t index) (flushable))
 (defknown %more-arg (t index) t (flushable))
 (defknown %more-kw-arg (t fixnum) (values t t))
@@ -1939,7 +1986,9 @@
   (movable foldable flushable no-verify-arg-count))
 (defknown %negate (number) number
   (movable foldable flushable no-verify-arg-count))
-(defknown (%check-bound check-bound) (array index t) index
+(defknown check-bound (array index t) index
+  (dx-safe))
+(defknown %check-bound (array index t) (values)
   (dx-safe))
 (defknown data-vector-ref (simple-array index) t
   (foldable flushable always-translatable))
@@ -1973,7 +2022,7 @@
 (defknown %set-symbol-package (symbol t) t ())
 (defknown (%coerce-callable-to-fun %coerce-callable-for-call)
     (function-designator)
-    function (flushable))
+    function (flushable no-verify-arg-count))
 (defknown array-bounding-indices-bad-error (t t t) nil)
 (defknown sequence-bounding-indices-bad-error (t t t) nil)
 (defknown %find-position
@@ -2226,19 +2275,22 @@
 (defknown (sb-impl::%with-standard-io-syntax
            sb-impl::%with-rebound-io-syntax
            sb-impl::call-with-sane-io-syntax)
-    (function) *)
-(defknown sb-debug::funcall-with-debug-io-syntax (function &rest t) *)
-(defknown sb-impl::%print-unreadable-object (t t t &optional function) null)
+    ((function ())) *)
+(defknown sb-debug::funcall-with-debug-io-syntax ((function ((rest-args))) &rest t) *)
+(defknown sb-impl::%print-unreadable-object (t t t &optional (function ())) null)
 
 #+sb-thread
 (progn
-(defknown (sb-thread::call-with-mutex sb-thread::call-with-recursive-lock)
-    (function t t t) *)
+(defknown (sb-thread::call-with-mutex-timed sb-thread::call-with-recursive-lock-timed)
+    ((function ()) t t t) *)
+(defknown (sb-thread::call-with-mutex sb-thread::call-with-recursive-lock
+           sb-thread::fast-call-with-mutex sb-thread::fast-call-with-recursive-lock)
+    ((function ()) t) *)
 (defknown (sb-thread::call-with-system-mutex
            sb-thread::call-with-system-mutex/allow-with-interrupts
            sb-thread::call-with-system-mutex/without-gcing
            sb-thread::call-with-recursive-system-lock)
-    (function t) *))
+    ((function ()) t) *))
 
 #+round-float
 (progn
@@ -2259,3 +2311,7 @@
   (integer t)
   integer
   (movable always-translatable))
+
+(defknown case-to-jump-table (t list &optional sequence t t) * (always-translatable))
+(defknown %jump-table (t &rest t) * (always-translatable))
+
