@@ -64,7 +64,7 @@
                ;; except in forcing discontiguous addresses for testing.
                ;; And of course, don't use them if unsupported.
                ((:fixedobj-space-start fixedobj-space-start*))
-               ((:fixedobj-space-size  fixedobj-space-size*) (* 24 1024 1024))
+               ((:fixedobj-space-size  fixedobj-space-size*) (* 48 1024 1024))
                ((:text-space-start text-space-start*))
                ((:text-space-size  text-space-size*) (* 160 1024 1024))
                (small-space-size #x100000)
@@ -86,10 +86,13 @@
                            #+(and win32 x86-64)
                            (seh-data ,(symbol-value '+backend-page-bytes+) win64-seh-data-addr)
                            #-immobile-space (alien-linkage-table ,alien-linkage-table-space-size)
-                           #+sb-safepoint
+                           ;; safepoint on 64-bit uses a relocatable trap page just below the card mark
+                           ;; table, which works nicely assuming a register is wired to the card table
+                           #+(and sb-safepoint (not x86-64))
                            ;; Must be just before NIL.
-                           (safepoint ,(symbol-value '+backend-page-bytes+) gc-safepoint-page-addr)
+                           (safepoint ,(symbol-value '+backend-page-bytes+))
                            (static ,small-space-size)
+                           (permgen 8388608) ; 8MiB
                            #+darwin-jit
                            (static-code ,small-space-size))
                          #+immobile-space
@@ -101,10 +104,11 @@
            (loop for (space size var-name) in spaces
                  appending
                  (let* ((relocatable
-                          ;; READONLY is usually movable now.
-                          (member space '(fixedobj text #+relocatable-static-space static
+                          (member space '(fixedobj text permgen
+                                          #+relocatable-static-space safepoint
+                                          #+relocatable-static-space static
                                           #+immobile-space alien-linkage-table
-                                          #-darwin-jit read-only)))
+                                          read-only)))
                         (start ptr)
                         (end (+ ptr size)))
                    (setf ptr end)
@@ -139,7 +143,7 @@
                  (32 (expt 2 29))
                  (64 (expt 2 30)))))
          ;; an arbitrary value to avoid kludging genesis
-         #+(and sb-xc-host (not darwin-jit))
+         #+sb-xc-host
          (defparameter read-only-space-end read-only-space-start)
          #-soft-card-marks (defconstant cards-per-page 1)
          (defconstant gencgc-card-bytes (/ gencgc-page-bytes cards-per-page))
@@ -188,23 +192,24 @@
 ;;; On #+sb-thread builds, these are not static, because access to them
 ;;; is via the TLS, not the symbol.
 (defconstant-eqx per-thread-c-interface-symbols
-  `((*free-interrupt-context-index* 0)
-    (sb-sys:*allow-with-interrupts* t)
-    (sb-sys:*interrupts-enabled* t)
-    sb-sys:*interrupt-pending*
-    #+sb-safepoint sb-sys:*thruption-pending*
-    *in-without-gcing*
-    *gc-inhibit*
-    *gc-pending*
-    #+sb-safepoint sb-impl::*in-safepoint*
-    #+sb-thread *stop-for-gc-pending*
-    sb-impl::*unweakened-vectors*
-    *pinned-objects*
-    (*gc-pin-code-pages* 0)
-    ;; things needed for non-local-exit
-    (*current-catch-block* 0)
-    (*current-unwind-protect-block* 0)
-    )
+    (hash-cons
+     '((*free-interrupt-context-index* 0)
+       (sb-sys:*allow-with-interrupts* t)
+       (sb-sys:*interrupts-enabled* t)
+       sb-sys:*interrupt-pending*
+       #+sb-safepoint sb-sys:*thruption-pending*
+       *in-without-gcing*
+       *gc-inhibit*
+       *gc-pending*
+       #+sb-safepoint sb-impl::*in-safepoint*
+       #+sb-thread *stop-for-gc-pending*
+       sb-impl::*unweakened-vectors*
+       *pinned-objects*
+       (*gc-pin-code-pages* 0)
+       ;; things needed for non-local-exit
+       #+ultrafutex (*current-mutex* 0)
+       (*current-catch-block* 0)
+       (*current-unwind-protect-block* 0)))
   #'equal)
 
 (defconstant-eqx +common-static-symbols+
@@ -370,7 +375,7 @@
   (defconstant +pseudo-static-generation+ 6))
 
 (defparameter *runtime-asm-routines* nil)
-(defparameter *linkage-space-predefined-entries* nil)
+(defparameter *alien-linkage-table-predefined-entries* nil)
 
 ;;; Floating-point related constants, both format descriptions and FPU
 ;;; control register descriptions.  These don't exactly match up with
@@ -413,5 +418,16 @@
 (defconstant double-float-digits 53)
 )
 
+;;; Reserve some bits of SYMBOL-HASH slot for future use
+#+64-bit
+(progn (defconstant n-symbol-hash-prng-bits 10) ; how many to randomize
+       (defconstant n-symbol-hash-discard-bits
+         (let ((precision (+ 32 n-symbol-hash-prng-bits))) ; total N bits
+           (- 64 precision)))
+       (defconstant-eqx sb-impl::symbol-hash-prng-byte
+         (byte n-symbol-hash-prng-bits (- 32 n-symbol-hash-prng-bits))
+         #'equal))
+#-64-bit (defconstant-eqx sb-impl::symbol-hash-prng-byte (byte 3 0) #'equal)
+
 (push '("SB-VM" +c-callable-fdefns+ +common-static-symbols+)
       *!removable-symbols*)
