@@ -14,7 +14,6 @@
 ;;; We need to define these predicates, since the TYPEP source
 ;;; transform picks whichever predicate was defined last when there
 ;;; are multiple predicates for equivalent types.
-(define-source-transform short-float-p (x) `(single-float-p ,x))
 #-long-float
 (define-source-transform long-float-p (x) `(double-float-p ,x))
 
@@ -121,7 +120,7 @@
 
                  ;; Definitely do this if copying to stack
                  ;; (Allocation has to be inlined, otherwise there's no way to DX it)
-                 (or (lvar-dynamic-extent result)
+                 (or (node-stack-allocate-p node)
                      (and dd-copier
                           (eq (sb-int:info :function :inlinep dd-copier) 'inline))
                      ;; Or if it's a small fixed number of words
@@ -177,18 +176,16 @@
      (%instance-set .instance. .index. .newval.)
      .newval.))
 
-;;; *** These transforms should be the only code, aside from the C runtime
-;;;     with knowledge of the layout index.
 #+compact-instance-header
 (define-source-transform function-with-layout-p (x) `(functionp ,x))
 #-compact-instance-header
 (progn
+  (define-source-transform function-with-layout-p (x) `(funcallable-instance-p ,x))
+  ;; Nothing but these transforms should assume that slot 0 holds a layout
   (define-source-transform %instance-layout (x)
     `(truly-the layout (%instance-ref ,x 0)))
   (define-source-transform %set-instance-layout (instance layout)
-    `(%instance-set ,instance 0 (the layout ,layout)))
-  (define-source-transform function-with-layout-p (x)
-    `(funcallable-instance-p ,x)))
+    `(%instance-set ,instance 0 (the layout ,layout))))
 
 ;;;; simplifying HAIRY-DATA-VECTOR-REF and HAIRY-DATA-VECTOR-SET
 
@@ -309,20 +306,24 @@
          (element-ctype (array-type-upgraded-element-type type))
          (declared-element-ctype (declared-array-element-type type)))
     (declare (type ctype element-ctype))
-    (when (eq *wild-type* element-ctype)
-      (give-up-ir1-transform
-       "Upgraded element type of array is not known at compile time."))
-    (let ((element-type-specifier (type-specifier element-ctype)))
-      `(multiple-value-bind (array index)
-           (%data-vector-and-index array index)
-         (declare (type (simple-array ,element-type-specifier 1) array)
-                  (type ,element-type-specifier new-value))
-         ,(if (type= element-ctype declared-element-ctype)
-              '(progn (data-vector-set array index new-value)
-                      new-value)
-              `(progn (data-vector-set array index
-                       ,(the-unwild declared-element-ctype 'new-value))
-                      ,(truly-the-unwild declared-element-ctype 'new-value)))))))
+    (cond ((eq *wild-type* element-ctype)
+           ;; The new value is only suitable for a simple-vector
+           (if (csubtypep (lvar-type new-value) (specifier-type '(not (or number character))))
+               `(hairy-data-vector-set (the simple-vector array) index new-value)
+               (give-up-ir1-transform
+                "Upgraded element type of array is not known at compile time.")))
+          (t
+           (let ((element-type-specifier (type-specifier element-ctype)))
+             `(multiple-value-bind (array index)
+                  (%data-vector-and-index array index)
+                (declare (type (simple-array ,element-type-specifier 1) array)
+                         (type ,element-type-specifier new-value))
+                ,(if (type= element-ctype declared-element-ctype)
+                     '(progn (data-vector-set array index new-value)
+                       new-value)
+                     `(progn (data-vector-set array index
+                                              ,(the-unwild declared-element-ctype 'new-value))
+                             ,(truly-the-unwild declared-element-ctype 'new-value)))))))))
 
 ;;; Transform multi-dimensional array to one dimensional data vector
 ;;; access.
@@ -448,42 +449,43 @@
 ;;; of the last word to be operated on, so they are effectively
 ;;; in an indeterminate state which is why equality testing, COUNT,
 ;;; and FIND have to ignore them.
-(deftransform bit-op->word-op ((bit-array-1 bit-array-2 result-bit-array)
-                               (simple-bit-vector simple-bit-vector simple-bit-vector)
-                               *
-                               :node node :defun-only t :info wordfun)
-  `(let ((length (vector-length result-bit-array)))
-     ,@(unless (policy node (zerop safety))
-         `((unless (= length
-                      ,@(unless (same-leaf-ref-p bit-array-1 result-bit-array)
-                          '((vector-length bit-array-1)))
-                      (vector-length bit-array-2))
-             (error "Argument and/or result bit arrays are not the same length:~
+(defun make-bit-op->word-op-transform (wordfun)
+  (deftransform bit-op->word-op ((bit-array-1 bit-array-2 result-bit-array)
+                                 (simple-bit-vector simple-bit-vector simple-bit-vector)
+                                 *
+                                 :node node :defun-only lambda)
+    `(let ((length (vector-length result-bit-array)))
+       ,@(unless (policy node (zerop safety))
+           `((unless (= length
+                        ,@(unless (same-leaf-ref-p bit-array-1 result-bit-array)
+                            '((vector-length bit-array-1)))
+                        (vector-length bit-array-2))
+               (error "Argument and/or result bit arrays are not the same length:~
                          ~%  ~S~%  ~S  ~%  ~S"
-                    bit-array-1 bit-array-2 result-bit-array))))
-     (dotimes (index (ceiling length sb-vm:n-word-bits))
-       (declare (optimize (speed 3) (safety 0)) (type index index))
-       (setf (%vector-raw-bits result-bit-array index)
-             (,wordfun (%vector-raw-bits bit-array-1 index)
-                       (%vector-raw-bits bit-array-2 index))))
-     result-bit-array))
+                      bit-array-1 bit-array-2 result-bit-array))))
+       (dotimes (index (ceiling length sb-vm:n-word-bits))
+         (declare (optimize (speed 3) (safety 0)) (type index index))
+         (setf (%vector-raw-bits result-bit-array index)
+               (,wordfun (%vector-raw-bits bit-array-1 index)
+                         (%vector-raw-bits bit-array-2 index))))
+       result-bit-array)))
 
 (flet ((policy-test (node) (policy node (>= speed space))))
-(macrolet ((def (bitfun wordfun)
-             `(%deftransform ',bitfun #'policy-test
-                             '(function (simple-bit-vector simple-bit-vector simple-bit-vector)
-                                        *)
-                             (cons #'bit-op->word-op ',wordfun))))
- (def bit-and word-logical-and)
- (def bit-ior word-logical-or)
- (def bit-xor word-logical-xor)
- (def bit-eqv word-logical-eqv)
- (def bit-nand word-logical-nand)
- (def bit-nor word-logical-nor)
- (def bit-andc1 word-logical-andc1)
- (def bit-andc2 word-logical-andc2)
- (def bit-orc1 word-logical-orc1)
- (def bit-orc2 word-logical-orc2)))
+  (macrolet ((def (bitfun wordfun)
+               `(%deftransform ',bitfun #'policy-test
+                               '(function (simple-bit-vector simple-bit-vector simple-bit-vector)
+                                 *)
+                               (make-bit-op->word-op-transform ',wordfun))))
+    (def bit-and word-logical-and)
+    (def bit-ior word-logical-or)
+    (def bit-xor word-logical-xor)
+    (def bit-eqv word-logical-eqv)
+    (def bit-nand word-logical-nand)
+    (def bit-nor word-logical-nor)
+    (def bit-andc1 word-logical-andc1)
+    (def bit-andc2 word-logical-andc2)
+    (def bit-orc1 word-logical-orc1)
+    (def bit-orc2 word-logical-orc2)))
 
 (deftransform bit-not
               ((bit-array result-bit-array)
@@ -543,49 +545,6 @@
           (if (zerop (lvar-value item)) '(- length count) 'count)
           '(if (zerop item) (- length count) count))))
 
-;;; This transform does not require that ITEM be derived as BIT,
-;;; but at runtime it has to be.
-(deftransform fill ((sequence item) (simple-bit-vector t) *
-                    :policy (>= speed space))
-  `(let ((value (logand (- (the bit item)) most-positive-word)))
-     ;; Unlike for the SIMPLE-BASE-STRING case, we are allowed to touch
-     ;; bits beyond LENGTH with impunity.
-     (dotimes (index (ceiling (vector-length sequence) sb-vm:n-word-bits))
-       (declare (optimize (speed 3) (safety 0))
-                (type index index))
-       (setf (%vector-raw-bits sequence index) value))
-     sequence))
-
-(deftransform fill ((sequence item) (simple-base-string t) *
-                                    :policy (>= speed space))
-  (let ((multiplier (logand #x0101010101010101 most-positive-word)))
-    `(let* ((value ,(if (and (constant-lvar-p item)
-                             (typep (lvar-value item) 'base-char))
-                        (* multiplier (char-code (lvar-value item)))
-                        ;; Use multiplication if it's known to be cheap
-                        #+(or x86 x86-64)
-                        `(* ,multiplier (char-code (the base-char item)))
-                        #-(or x86 x86-64)
-                        '(let ((code (char-code (the base-char item))))
-                          (setf code (dpb code (byte 8 8) code))
-                          (setf code (dpb code (byte 16 16) code))
-                          #+64-bit (dpb code (byte 32 32) code))))
-            (len (vector-length sequence))
-            (words (truncate len sb-vm:n-word-bytes)))
-       (dotimes (index words)
-         (declare (optimize (speed 3) (safety 0))
-                  (type index index))
-         (setf (%vector-raw-bits sequence index) value))
-       ;; For 64-bit:
-       ;;  if 1 more byte should be written, then shift-towards-start 56
-       ;;  if 2 more bytes ...               then shift-towards-start 48
-       ;;  etc
-       ;; This correctly rewrites the trailing null in its proper place.
-       (let ((bits (ash (mod len sb-vm:n-word-bytes) 3)))
-         (when (plusp bits)
-           (setf (%vector-raw-bits sequence words)
-                 (shift-towards-start value (- bits)))))
-       sequence)))
 
 ;;;; %BYTE-BLT
 
@@ -596,7 +555,7 @@
 ;;; currently (ca. sbcl-0.6.12.30) the main interface for code in
 ;;; SB-KERNEL and SB-SYS (e.g. i/o code). It's not clear that it's the
 ;;; ideal interface, though, and it probably deserves some thought.
-(deftransform %byte-blt ((src src-start dst dst-start dst-end)
+(deftransform %byte-blt ((src src-start dst dst-start nbytes)
                          ((or (simple-unboxed-array (*)) system-area-pointer)
                           index
                           (or (simple-unboxed-array (*)) system-area-pointer)
@@ -630,19 +589,21 @@
               (sap-ref-8 dst (1- dst-end)) (sap-ref-8 dst (1- dst-end))))
       (memmove (sap+ (sapify dst) dst-start)
                (sap+ (sapify src) src-start)
-               (- dst-end dst-start)))
+               nbytes))
      (values)))
 
 ;;;; transforms for EQL of floating point values
 (unless (vop-existsp :named sb-vm::eql/single-float)
-(deftransform eql ((x y) (single-float single-float))
-  '(= (single-float-bits x) (single-float-bits y))))
+(deftransform eql ((x y) (single-float single-float) * :node node)
+  (delay-ir1-transform node :ir1-phases)
+  '(eql (single-float-bits x) (single-float-bits y))))
 
 (unless (vop-existsp :named sb-vm::eql/double-float)
-(deftransform eql ((x y) (double-float double-float))
-  #-64-bit '(and (= (double-float-low-bits x) (double-float-low-bits y))
-                  (= (double-float-high-bits x) (double-float-high-bits y)))
-  #+64-bit '(= (double-float-bits x) (double-float-bits y))))
+(deftransform eql ((x y) (double-float double-float) * :node node)
+  (delay-ir1-transform node :ir1-phases)
+  #-64-bit '(and (eql (double-float-low-bits x) (double-float-low-bits y))
+             (eql (double-float-high-bits x) (double-float-high-bits y)))
+  #+64-bit '(eql (double-float-bits x) (double-float-bits y))))
 
 
 ;;;; modular functions
