@@ -32,8 +32,6 @@ otherwise evaluate ELSE and return its values. ELSE defaults to NIL."
          (then-block (ctran-starts-block then-ctran))
          (else-ctran (make-ctran))
          (else-block (ctran-starts-block else-ctran))
-         (maybe-instrument *instrument-if-for-code-coverage*)
-         (*instrument-if-for-code-coverage* t)
          (node (make-if :test pred-lvar
                         :consequent then-block
                         :alternative else-block)))
@@ -54,17 +52,76 @@ otherwise evaluate ELSE and return its values. ELSE defaults to NIL."
       (link-blocks start-block else-block))
 
     (let ((path (best-sub-source-path test)))
-      (ir1-convert (if (and path maybe-instrument)
+      (ir1-convert (if path
                        (let ((*current-path* path))
                          (instrument-coverage then-ctran :then test))
                        then-ctran)
                    next result then)
-      (ir1-convert (if (and path maybe-instrument)
+      (ir1-convert (if path
                        (let ((*current-path* path))
                          (instrument-coverage else-ctran :else test))
                        else-ctran)
                    next result else))))
 
+;;; Hiding jump-table from third party code walkers
+(define-source-transform %jump-table (index &rest targets)
+  `(jump-table ,index
+               ,@(loop for (target . next) on targets
+                       for i from 0
+                       collect `(,(if next
+                                      i
+                                      'otherwise)
+                                 (funcall ,target)))))
+
+(def-ir1-translator jump-table ((index &rest targets) start next result)
+  (aver targets)
+  (let* ((index-ctran (make-ctran))
+         (index-lvar (make-lvar))
+         (node (make-jump-table index-lvar)))
+    (setf (lvar-dest index-lvar) node)
+    (ir1-convert start index-ctran index-lvar index)
+    (link-node-to-previous-ctran node index-ctran)
+
+    (let ((start-block (ctran-block index-ctran)))
+      (setf (block-last start-block) node)
+      (ctran-starts-block next)
+      (setf (jump-table-targets node)
+            (loop for (index . target) in targets
+                  for ctran = (make-ctran)
+                  for block = (ctran-starts-block ctran)
+                  do
+                  (ir1-convert-progn-body ctran next result target)
+                  (link-blocks start-block block)
+                  collect (cons index block))))))
+
+;;; then or else can be already converted blocks
+(def-ir1-translator if-to-blocks ((test then &optional else) start next result)
+  (flet ((to-block (x)
+           (if (block-p x)
+               (values nil x)
+               (let ((ctran (make-ctran)))
+                 (values ctran (ctran-starts-block ctran))))))
+    (multiple-value-bind (then-ctran then-block) (to-block then)
+      (multiple-value-bind (else-ctran else-block) (to-block else)
+        (let* ((pred-ctran (make-ctran))
+               (pred-lvar (make-lvar))
+               (node (make-if :test pred-lvar
+                              :consequent then-block
+                              :alternative else-block)))
+          (setf (lvar-dest pred-lvar) node)
+          (ir1-convert start pred-ctran pred-lvar test)
+          (link-node-to-previous-ctran node pred-ctran)
+
+          (let ((start-block (ctran-block pred-ctran)))
+            (setf (block-last start-block) node)
+            (ctran-starts-block next)
+            (link-blocks start-block then-block)
+            (link-blocks start-block else-block))
+          (when then-ctran
+            (ir1-convert then-ctran next result then))
+          (when else-ctran
+            (ir1-convert else-ctran next result else)))))))
+
 ;;; To get even remotely sensible results for branch coverage
 ;;; tracking, we need good source paths. If the macroexpansions
 ;;; interfere enough the TEST of the conditional doesn't actually have
@@ -231,6 +288,8 @@ constrained to be used only within the dynamic extent of the TAGBODY."
                                     tag)))
          (entry (first found))
          (exit (make-exit :entry entry)))
+    (when (ctran-deleted-p (second found))
+      (throw 'locall-already-let-converted (second found)))
     (push exit (entry-exits entry))
     (link-node-to-previous-ctran exit start)
     (let ((home-lambda (ctran-home-lambda-or-null start)))
@@ -337,7 +396,13 @@ Evaluate the FORMS in the specified SITUATIONS (any of :COMPILE-TOPLEVEL,
 
 (defun funcall-in-macrolet-lexenv (definitions fun context)
   (%funcall-in-foomacrolet-lexenv
-   (macrolet-definitionize-fun context (make-restricted-lexenv *lexenv*))
+   (macrolet-definitionize-fun
+    context
+    (make-restricted-lexenv *lexenv*
+                            ;; Avoid efficiency notes
+                            ;; and in general there's no need to optimizes macrolets
+                            (augment-policy speed 1
+                                            (lexenv-policy *lexenv*))))
    :funs
    definitions
    fun))
@@ -714,7 +779,7 @@ be a lambda expression."
               (ecase operator
                 (function (find-or-convert-fun-leaf definition start))
                 (global-function (values (find-global-fun definition t) start)))
-            (ir1-convert start next result `(,leaf ,@args))))
+            (ir1-convert-common-functoid start next result `(,leaf ,@args) leaf)))
         (let ((ctran (make-ctran))
               (fun-lvar (make-lvar)))
           (ir1-convert start ctran fun-lvar `(the function ,function))
@@ -829,7 +894,9 @@ have been evaluated."
                            forms
                            vars
                            :post-binding-lexenv post-binding-lexenv
-                           :debug-name (debug-name 'let bindings))))
+                           :debug-name (debug-name 'let
+                                                   (mapcar #'leaf-source-name
+                                                           vars)))))
                  (reference-leaf start ctran fun-lvar fun)))
              (ir1-convert-combination-args fun-lvar ctran next result values
                                            :arg-source-forms bindings))))))
@@ -928,13 +995,18 @@ also processed as top level forms."
     (cond ((some #'leaf-dynamic-extent funs)
            (ctran-starts-block next)
            (let* ((enclose (ctran-use enclose-ctran))
+                  (dynamic-extent (make-dynamic-extent))
                   (cleanup (make-cleanup :kind :dynamic-extent
-                                         :mess-up enclose))
-                  (cleanup-ctran (make-ctran)))
-             (setf (enclose-cleanup enclose) cleanup)
+                                         :mess-up dynamic-extent))
+                  (dynamic-extent-ctran (make-ctran)))
+             (setf (enclose-dynamic-extent enclose) dynamic-extent)
+             (setf (dynamic-extent-cleanup dynamic-extent) cleanup)
+             (link-node-to-previous-ctran dynamic-extent enclose-ctran)
+             (use-ctran dynamic-extent dynamic-extent-ctran)
+             (push dynamic-extent
+                   (lambda-dynamic-extents (node-home-lambda dynamic-extent)))
              (let ((*lexenv* (make-lexenv :cleanup cleanup)))
-               (ir1-convert enclose-ctran cleanup-ctran nil '(%cleanup-point))
-               (ir1-convert-progn-body cleanup-ctran next result body))))
+               (ir1-convert-progn-body dynamic-extent-ctran next result body))))
           (t
            (ir1-convert-progn-body enclose-ctran next result body)))))
 
@@ -1113,13 +1185,12 @@ care."
                ;; Would be great for IR1-CONVERT to return the uses it creates
                (let ((new-uses (lvar-uses result)))
                  (derive-node-type (cond ((consp new-uses)
-                                          ;; Handle just a single use for now,
-                                          ;; doubt it'll be useful for multiple uses.
-                                          (aver (= (1- (length new-uses))
-                                                   (if (consp before-uses)
-                                                       (length before-uses)
-                                                       1)))
-                                          (car new-uses))
+                                          (loop for other in (cdr new-uses)
+                                                do (aver (or (exit-p other)
+                                                             (eq before-uses other)
+                                                             (and (consp before-uses)
+                                                                  (memq other before-uses)))))
+                                          (the (not exit) (car new-uses)))
                                          (t
                                           new-uses))
                                    value-type)))))
@@ -1161,33 +1232,6 @@ care."
   (let ((*current-path* source-path))
     (ir1-convert start next result form)))
 
-(def-ir1-translator bound-cast ((array bound index) start next result)
-  (let ((check-bound-tran (make-ctran))
-        (index-ctran (make-ctran))
-        (index-lvar (make-lvar)))
-    ;; CHECK-BOUND transform ensures that INDEX won't be evaluated twice
-    (ir1-convert start check-bound-tran nil `(%check-bound ,array ,bound ,index))
-    (ir1-convert check-bound-tran index-ctran index-lvar index)
-    (let* ((check-bound-combination (ctran-use check-bound-tran))
-           (array (first (combination-args check-bound-combination)))
-           (bound (second (combination-args check-bound-combination)))
-           (derived (constant-lvar-p bound))
-           (type (specifier-type (if derived
-                                     `(integer 0 (,(lvar-value bound)))
-                                     '(and unsigned-byte fixnum))))
-           (cast (make-bound-cast :value index-lvar
-                                  :asserted-type type
-                                  :type-to-check type
-                                  :derived-type (coerce-to-values type)
-                                  :check check-bound-combination
-                                  :derived derived
-                                  :array array
-                                  :bound bound)))
-      (push cast (lvar-dependent-nodes array))
-      (link-node-to-previous-ctran cast index-ctran)
-      (setf (lvar-dest index-lvar) cast)
-      (use-continuation cast next result))))
-
 #-sb-xc-host
 (setf (info :function :macro-function 'truly-the)
       (lambda (whole env)
@@ -1200,7 +1244,7 @@ care."
       (info :function :macro-function 'with-source-form)
       (lambda (whole env)
         (declare (ignore env))
-                `(progn ,@(cddr whole))))
+        `(progn ,@(cddr whole))))
 
 ;;;; SETQ
 
@@ -1324,7 +1368,7 @@ to TAG."
                    (return-from ,tag (%unknown-values)))
                 :debug-name (debug-name 'escape-fun tag))))
         (ctran (make-ctran)))
-    (setf (functional-kind fun) :escape)
+    (setf (functional-kind fun) (functional-kind-attributes escape))
     (enclose start ctran (list fun))
     (reference-leaf ctran next result fun)))
 
@@ -1336,7 +1380,7 @@ to TAG."
   ;; (SETF FOO) here?
   (let ((fun (lexenv-find name funs)))
     (aver (lambda-p fun))
-    (setf (functional-kind fun) :cleanup)
+    (setf (functional-kind fun) (functional-kind-attributes cleanup))
     (reference-leaf start next result fun)))
 
 (def-ir1-translator catch ((tag &body body) start next result)
@@ -1532,8 +1576,8 @@ values from the first VALUES-FORM making up the first argument, etc."
 ;;; to interpose a DELAY node using RESULT immediately so that the
 ;;; result continuation can assume that it is immediately used. This
 ;;; is important here because MULTIPLE-VALUE-PROG1 is the only special
-;;; form which receives unknown values with multiple uses, some (in
-;;; this case one) of which are not immediate.
+;;; form that produces code where lvar substitution is potentially
+;;; incorrect.
 (def-ir1-translator multiple-value-prog1
     ((values-form &rest forms) start next result)
   "MULTIPLE-VALUE-PROG1 values-form form*
