@@ -131,7 +131,8 @@
                                           :post-binding-lexenv post-binding-lexenv
                                           :debug-name (debug-name
                                                        '&aux-bindings
-                                                       aux-vars)
+                                                       (mapcar #'leaf-source-name
+                                                               aux-vars))
                                           :value-source-forms (rest value-source-forms))))
         (reference-leaf start ctran fun-lvar fun)
         (ir1-convert-combination-args fun-lvar ctran next result
@@ -285,7 +286,7 @@
 (defun register-entry-point (entry dispatcher)
   (declare (type clambda entry)
            (type optional-dispatch dispatcher))
-  (setf (functional-kind entry) :optional)
+  (setf (functional-kind entry) (functional-kind-attributes optional))
   (setf (leaf-ever-used entry) t)
   (setf (lambda-optional-dispatch entry) dispatcher)
   entry)
@@ -392,7 +393,7 @@
     (let ((name (or debug-name source-name)))
       (if (or force
               supplied-p-p ; this entry will be of kind NIL
-              (and (lambda-p ep) (eq (lambda-kind ep) nil)))
+              (and (lambda-p ep) (functional-kind-eq ep nil)))
           (convert-optional-entry ep
                                   default-vars default-vals
                                   (if supplied-p (list default nil) (list default))
@@ -1128,7 +1129,12 @@
       (if (eq (car fun) 'lambda-with-lexenv)
           (cdr fun)
           `(() . ,(cdr fun)))
-    (let* ((*lexenv*
+    (let* ((notinlines
+             (loop for fun in (lexenv-funs *lexenv*)
+                   when (and (defined-fun-p (cdr fun))
+                             (defined-fun-inlinep (cdr fun)))
+                   collect fun))
+           (*lexenv*
              (if decls
                  (make-lexenv
                   :default (process-decls decls nil nil
@@ -1148,9 +1154,13 @@
                   (lexenv-lambda *lexenv*)
                   *lexenv*)))
            (*inlining* (1+ *inlining*))
-           (clambda (ir1-convert-lambda `(lambda ,@body)
-                                        :source-name source-name
-                                        :debug-name debug-name)))
+           (clambda (progn
+                      (when notinlines
+                        (setf (lexenv-funs *lexenv*)
+                              notinlines))
+                      (ir1-convert-lambda `(lambda ,@body)
+                                          :source-name source-name
+                                          :debug-name debug-name))))
       (setf (functional-inline-expanded clambda) t)
       clambda)))
 
@@ -1259,18 +1269,29 @@
                      (or (fun-info-transforms info)
                          (fun-info-templates info)
                          (fun-info-ir2-convert info))))
-      (if (block-compile *compilation*)
-          (progn
-            (substitute-leaf fun var)
-            ;; If in a simple environment, then we can allow backward
-            ;; references to this function from following top-level
-            ;; forms.
-            (when simple-lexenv-p
-              (setf (defined-fun-functional var) fun)))
-          (substitute-leaf-if
-           (lambda (ref)
-             (policy ref (> recognize-self-calls 0)))
-           fun var)))
+      (let (type)
+        (if (block-compile *compilation*)
+            (progn
+              (substitute-leaf fun var)
+              ;; If in a simple environment, then we can allow backward
+              ;; references to this function from following top-level
+              ;; forms.
+              (when simple-lexenv-p
+                (setf (defined-fun-functional var) fun)))
+            (substitute-leaf-if
+             (lambda (ref)
+               (if (policy ref (> recognize-self-calls 0))
+                   t
+                   (let ((type (or type
+                                   (setf type (definition-type fun))))
+                         (call (node-dest ref)))
+                     (when (and (combination-p call)
+                                (eq (combination-fun call)
+                                    (ref-lvar ref))
+                                (fun-type-p type))
+                       (assert-call-type call type t :defined-here))
+                     nil)))
+             fun var))))
     fun))
 
 ;;; Convert a lambda for global inline expansion.
