@@ -67,6 +67,15 @@
         (abort-ir1-transform)
         expansion)))
 
+(progn
+  (defun type-other-pointer-p (type)
+    (csubtypep type (specifier-type '(not #1=(or fixnum #+64-bit single-float
+                                                 list function instance character
+                                                 extended-sequence)))))
+
+  (defun type-not-other-pointer-p (type)
+    (csubtypep type (specifier-type '#1#))))
+
 ;;; If the lvar OBJECT definitely is or isn't of the specified
 ;;; type, then return T or NIL as appropriate. Otherwise quietly
 ;;; GIVE-UP-IR1-TRANSFORM.
@@ -79,41 +88,106 @@
            (return-from ir1-transform-type-predicate t))
           ((eq type *empty-type*)
            (return-from ir1-transform-type-predicate nil)))
-    (let ((intersect (type-intersection2 type otype)))
-      ;; I guess the theory here is that an intersection type
-      ;; is never a singleton, because if we could see that it was
-      ;; a singleton, it wouldn't be an intersection.
-      (when (and intersect (not (intersection-type-p intersect)))
-        (multiple-value-bind (constantp value) (type-singleton-p intersect)
-          (when constantp
-            (return-from ir1-transform-type-predicate `(eql object ',value)))))
+    (let ((intersect (type-intersection type otype))
+          (current-predicate (combination-fun-source-name node)))
       ;; If the object type is known to be (OR NULL <type>),
       ;; it is almost always cheaper to test for not EQ to NIL.
       ;; There is one exception:
-      ;;  - FIXNUMP is possibly cheapear than comparison to NIL, or definitely
+      ;;  - FIXNUMP is possibly cheaper than comparison to NIL, or definitely
       ;;    not worse. For x86, NIL is a 4-byte immediate operand,
       ;;    for lack of a null-tn register. FIXNUM-TAG-MASK is only 1 byte.
       (when (type= otype (type-union (specifier-type 'null) type))
         (let ((difference (type-difference type (specifier-type 'null))))
           (unless (type= difference (specifier-type 'fixnum))
             (return-from  ir1-transform-type-predicate `(not (null object))))))
-      (cond ((typep type 'alien-type-type)
-             ;; We don't transform alien type tests until here, because
-             ;; once we do that the rest of the type system can no longer
-             ;; reason about them properly -- so we'd miss out on type
-             ;; derivation, etc.
-             (delay-ir1-transform node :ir1-phases)
-             (let ((alien-type (alien-type-type-alien-type type)))
-               ;; If it's a lisp-rep-type, the CTYPE should be one already.
-               (aver (not (compute-lisp-rep-type alien-type)))
-               `(sb-alien::alien-value-typep object ',alien-type)))
-            ((let ((intersect (type-intersection otype type))
-                   (fixnum (specifier-type 'fixnum)))
-               (when (and (csubtypep intersect fixnum)
-                          (not (csubtypep type fixnum)))
-                 `(typep object ',(type-specifier intersect)))))
-            (t
-             (give-up-ir1-transform))))))
+      (flet ((memory-type-test-p (type)
+               (and (types-equal-or-intersect
+                     type
+                     (specifier-type
+                      '(not (or fixnum #+64-bit single-float
+                                       boolean character
+                             function list))))
+                    (not (type= type (specifier-type 'instance))))))
+        (cond ((typep type 'alien-type-type)
+               ;; We don't transform alien type tests until here, because
+               ;; once we do that the rest of the type system can no longer
+               ;; reason about them properly -- so we'd miss out on type
+               ;; derivation, etc.
+               (delay-ir1-transform node :ir1-phases)
+               (let ((alien-type (alien-type-type-alien-type type)))
+                 ;; If it's a lisp-rep-type, the CTYPE should be one already.
+                 (aver (not (compute-lisp-rep-type alien-type)))
+                 `(sb-alien::alien-value-typep object ',alien-type)))
+              ;; (typep (the (or list fixnum) x) 'integer) =>
+              ;; (typep x 'fixnum)
+              ((let ((new-predicate
+                       (or
+                        (backend-type-predicate intersect)
+                        ;; Remove bounds from numeric types
+                        (and (csubtypep intersect (specifier-type 'real))
+                             (macrolet ((up (&rest types)
+                                          `(cond ,@(loop for (type predicate) on types by #'cddr
+                                                         collect
+                                                         `((and (csubtypep intersect (specifier-type ',type))
+                                                                (csubtypep (specifier-type ',type) type))
+                                                           ',predicate)))))
+                               (up fixnum fixnump
+                                   integer integerp
+                                   rational rationalp
+                                   single-float single-float-p
+                                   double-float double-float-p
+                                   float floatp
+                                   real realp))))))
+                 (when (and new-predicate
+                            (neq new-predicate current-predicate)
+                            ;; Some subtypes are more expensive to check
+                            (not (and (eq current-predicate 'listp)
+                                      (eq new-predicate 'consp)))
+                            (not (and (eq current-predicate 'functionp)
+                                      (eq new-predicate 'compiled-function-p)))
+                            (not (eq current-predicate 'characterp))
+                            (not (and (eq current-predicate 'non-null-symbol-p)
+                                      (eq new-predicate 'keywordp)))
+                            (not (eq new-predicate #+64-bit 'signed-byte-64-p
+                                                   #-64-bit 'signed-byte-32-p))
+                            (not (eq new-predicate #+64-bit 'unsigned-byte-64-p
+                                                   #-64-bit 'unsigned-byte-32-p)))
+                   `(,new-predicate object))))
+              ;; (typep (the float x) 'double-float) =>
+              ;; (typep x 'single-float)
+              ((let* ((diff (type-difference otype type))
+                      (pred (and (or (eq current-predicate 'sequencep) ;; always expensive
+                                     (not (memory-type-test-p diff)))
+                                 (or (backend-type-predicate diff)
+                                     ;; Remove bounds from numeric types
+                                     (and (csubtypep diff (specifier-type 'real))
+                                          (macrolet ((up (&rest types)
+                                                       `(cond ,@(loop for (type predicate) on types by #'cddr
+                                                                      collect
+                                                                      `((and (csubtypep diff (specifier-type ',type))
+                                                                             (not (types-equal-or-intersect type (specifier-type ',type))))
+                                                                        ',predicate)))))
+                                            (up fixnum fixnump
+                                                integer integerp
+                                                rational rationalp
+                                                single-float single-float-p
+                                                double-float double-float-p
+                                                float floatp
+                                                real realp)))))))
+                 (cond ((and pred
+                             ;; Testing for fixnum is usually the cheapest
+                             (or (eq pred 'fixnump)
+                                 (memory-type-test-p type)))
+                        `(not (,pred object)))
+                       ((and (memory-type-test-p type)
+                             (cond ((and (type-not-other-pointer-p diff)
+                                         (type-other-pointer-p type))
+                                    `(%other-pointer-p object))
+                                   ((and (type-other-pointer-p type)
+                                         (type-not-other-pointer-p type))
+                                    `(not (%other-pointer-p object)))))))))
+              (t
+               (give-up-ir1-transform)))))))
 
 ;;; Flush %TYPEP tests whose result is known at compile time.
 (deftransform %typep ((object type) * * :node node)
@@ -316,30 +390,11 @@
          (high (numeric-type-high type)))
     (ecase (numeric-type-complexp type)
       (:real
-       (cond ((and (eql (numeric-type-class type) 'integer)
-                   (and (fixnump low)
-                        (fixnump high)
-                        #+(or x86 x86-64 arm arm64)
-                        (/= low 0)
-                        (< (- high low) 2)))
-              ;; The fixnum-mod-p case is worse than just EQ testing with
-              ;; only 2 values in the range. (INTEGER 1 2) would have become
-              ;;   (and (not (eq x 0)) (fixnump x) (not (> x 2))).
-              ;; If exactly 1 value, it should have been picked off by TYPE-SINGLETON-P
-              ;; in %SOURCE-TRANSFORM-TYPEP, but even if it wasn't,
-              ;; the OR will drop out due to constraint propagation.
-              `(or (eq ,object ,low) (eq ,object ,high)))
-             #+(or x86 x86-64 arm arm64) ;; Not implemented elsewhere yet
-             ((and (eql (numeric-type-class type) 'integer)
-                   (or (eql low 0)
-                       (and (eql low 1)
-                            (not (eql high most-positive-fixnum))))
+       (cond ((and (vop-existsp :translate check-range<=)
+                   (eql (numeric-type-class type) 'integer)
+                   (fixnump low)
                    (fixnump high))
-              (let ((mod-p `(fixnum-mod-p ,object ,high)))
-                (if (eql low 1)
-                    `(and (not (eq ,object 0))
-                          ,mod-p)
-                    mod-p)))
+              `(check-range<= ,low ,object ,high))
              ((type= type (specifier-type '(or word sb-vm:signed-word)))
               `(or (typep ,object 'sb-vm:signed-word)
                    (typep ,object 'word)))
@@ -347,17 +402,14 @@
               `(and (typep ,object ',base)
                     ,(transform-numeric-bound-test object type base)))))
       (:complex
-       `(and (complexp ,object)
-             ,(once-only ((n-real `(realpart (truly-the complex ,object)))
-                          (n-imag `(imagpart (truly-the complex ,object))))
-                `(progn
-                   ,n-imag ; ignorable
-                   (and (typep ,n-real ',base)
-                        ,@(when (eq class 'integer)
-                            `((typep ,n-imag ',base)))
-                        ,(transform-numeric-bound-test n-real type base)
-                        ,(transform-numeric-bound-test n-imag type
-                                                       base)))))))))
+       (let ((part-type (second (type-specifier type))))
+         `(and (typep ,object '(complex ,(case base
+                                           ((double-float single-float rational) base)
+                                           (t (if (eq class 'integer)
+                                                  'rational
+                                                  '*)))))
+               (typep (realpart ,object) ',part-type)
+               (typep (imagpart ,object) ',part-type)))))))
 
 ;;; Do the source transformation for a test of a hairy type.
 ;;; SATISFIES is converted into the obvious. Otherwise, we convert
@@ -524,8 +576,8 @@
               ;; don't test a range.
               ;; (and subtype-of-integer (not (integer x y)))
               (destructuring-bind (b a) types
-                (and (eq (numeric-type-class a) 'integer)
-                     (eq (numeric-type-class b) 'integer)
+                (and (integer-type-p a)
+                     (integer-type-p b)
                      (flet ((check (a b)
                               (let* ((a-hi (numeric-type-high a))
                                      (a-lo (numeric-type-low a))
@@ -556,8 +608,8 @@
               ;; (or (integer * fixnum-x) (integer fixnum-y))
               ;; only check for bignump and not its value.
               (destructuring-bind (b a) types
-                (and (eq (numeric-type-class a) 'integer)
-                     (eq (numeric-type-class b) 'integer)
+                (and (integer-type-p a)
+                     (integer-type-p b)
                      (flet ((check (a b)
                               (let* ((a-hi (numeric-type-high a))
                                      (a-lo (numeric-type-low a))
@@ -616,6 +668,22 @@
                      (widetags
                       `(or (%other-pointer-subtype-p ,object ',widetags)
                            (typep ,object '(or ,@(mapcar #'type-specifier more-types)))))
+                     ((and (cdr more-types)
+                           (every #'intersection-type-p more-types)
+                           (let ((common (intersection-type-types (car more-types))))
+                             (loop for type in (cdr more-types)
+                                   for types = (intersection-type-types type)
+                                   for int = (intersection common types :test #'type=)
+                                   always int
+                                   do (setf common int)
+                                   finally
+                                   (return `(and
+                                             (typep ,object '(and ,@(mapcar #'type-specifier common)))
+                                             (or ,@(loop for type in more-types
+                                                         for types = (intersection-type-types type)
+                                                         collect
+                                                         `(typep ,object '(and ,@(mapcar #'type-specifier
+                                                                                  (set-difference types common))))))))))))
                      (t
                       `(or
                         ,@(mapcar (lambda (x)
@@ -808,32 +876,88 @@
                        `((if ,header-test
                              (= (%array-dimension ,obj 0) ,@dims)
                              (= (vector-length ,obj) ,@dims))))
-                     nil))))))
+                     nil
+                     (car dims)))))))
 
 ;;; Return forms to test that OBJ has the element-type specified by type
 ;;; specified by TYPE, where STYPE is the type we have checked against (which
 ;;; is the same but for dimensions and element type). If HEADERP is true, OBJ
 ;;; is guaranteed to be an array-header.
-(defun test-array-element-type (obj type stype headerp)
+(defun test-array-element-type (obj type stype headerp pred length)
   (declare (type array-type type stype))
-  (let ((obj `(truly-the ,(type-specifier stype) ,obj))
-        (eltype (array-type-specialized-element-type type)))
+  (let ((eltype (array-type-specialized-element-type type)))
     (unless (or (type= eltype (array-type-specialized-element-type stype))
                 (eq eltype *wild-type*))
-      (let ((typecode (sb-vm:saetp-typecode (find-saetp-by-ctype eltype))))
-        (with-unique-names (data)
-         (if (and headerp (not (array-type-complexp stype)))
-             ;; If we know OBJ is an array header, and that the array is
-             ;; simple, we also know there is exactly one indirection to
-             ;; follow.
-             `(#-x86-64
-               (eq (%other-pointer-widetag (%array-data ,obj)) ,typecode)
-               #+x86-64
-               (widetag= (%array-data ,obj) ,typecode))
-             `((do ((,data ,(if headerp `(%array-data ,obj) obj)
-                           (%array-data ,data)))
-                   ((not (array-header-p ,data))
-                    (eq (%other-pointer-widetag ,data) ,typecode))))))))))
+      (let* ((typecode (sb-vm:saetp-typecode (find-saetp-by-ctype eltype))))
+        (cond ((and headerp (not (array-type-complexp stype)))
+               (let ((obj `(truly-the ,(type-specifier stype) ,obj)))
+                 ;; If we know OBJ is an array header, and that the array is
+                 ;; simple, we also know there is exactly one indirection to
+                 ;; follow.
+                 `(#-x86-64
+                   (eq (%other-pointer-widetag (%array-data ,obj)) ,typecode)
+                   #+x86-64
+                   (widetag= (%array-data ,obj) ,typecode))))
+              ((not (array-type-complexp stype))
+               (values
+                `((and (%other-pointer-p ,obj)
+                       (let ((widetag (%other-pointer-widetag ,obj)))
+                         (or (eq widetag ,typecode)
+                             (and (eq widetag sb-vm:simple-array-widetag)
+                                  (eq (%other-pointer-widetag (%array-data ,obj)) ,typecode))))))
+                ;; skip checking for array.
+                t))
+              (t
+               (ecase pred
+                 (arrayp
+                  (values `((and (%other-pointer-p ,obj)
+                                 (let ((data ,obj))
+                                   (loop
+                                    (let ((widetag (%other-pointer-widetag data)))
+                                      (if (eq widetag ,typecode)
+                                          (return t)
+                                          (if (or (eq widetag sb-vm:simple-array-widetag)
+                                                  (>= widetag sb-vm:complex-base-string-widetag))
+                                              (setf data (%array-data data))
+                                              (return nil))))))))
+                          t))
+                 (vectorp
+                  (if length
+                      (values `((and (%other-pointer-p ,obj)
+                                     (let ((widetag (%other-pointer-widetag ,obj)))
+                                       (if (eq widetag ,typecode)
+                                           (= (vector-length (truly-the (simple-array * (*)) ,obj)) ,length)
+                                           (and (= widetag sb-vm:complex-vector-widetag)
+                                                (= (%array-dimension (truly-the (and (array * (*))
+                                                                                     (not simple-array)) ,obj) 0)
+                                                   ,length)
+                                                (let ((data ,obj))
+                                                  (loop
+                                                   (setf data (%array-data data))
+                                                   (let ((widetag (%other-pointer-widetag data)))
+                                                     (if (eq widetag ,typecode)
+                                                         (return t)
+                                                         (unless (or (eq widetag sb-vm:simple-array-widetag)
+                                                                     (>= widetag sb-vm:complex-vector-widetag))
+                                                           (return nil)))))))))))
+                              t
+                              t)
+                      (values `((and (%other-pointer-p ,obj)
+                                     (let ((widetag (%other-pointer-widetag ,obj)))
+                                       (if (eq widetag ,typecode)
+                                           t
+                                           (and (= widetag sb-vm:complex-vector-widetag)
+                                                (let ((data ,obj))
+                                                  (loop
+                                                   (setf data (%array-data data))
+                                                   (let ((widetag (%other-pointer-widetag data)))
+                                                     (if (eq widetag ,typecode)
+                                                         (return t)
+                                                         (unless (or
+                                                                  (eq widetag sb-vm:simple-array-widetag)
+                                                                  (>= widetag sb-vm:complex-vector-widetag))
+                                                           (return nil)))))))))))
+                              t))))))))))
 
 ;;; If we can find a type predicate that tests for the type without
 ;;; dimensions, then use that predicate and test for dimensions.
@@ -876,34 +1000,40 @@
                           (eq (%other-pointer-widetag ,object) ,complex-tag)
                           ,@(unless (eq (car dims) '*)
                               `((= (%array-dimension ,object 0) ,(car dims)))))
-                    (multiple-value-bind (tests headerp)
+                    (multiple-value-bind (dim-tests headerp length)
                         (test-array-dimensions object type stype
                                                simple-array-header-p)
-                      `(and
-                        ,@(cond ((and (eql pred 'vectorp)
-                                      complexp)
-                                 `((%other-pointer-subtype-p ,object
-                                                             ',(list sb-vm:complex-base-string-widetag
-                                                                     #+sb-unicode sb-vm:complex-character-string-widetag
-                                                                     sb-vm:complex-bit-vector-widetag
-                                                                     sb-vm:complex-vector-widetag))))
-                                ((and (eql pred 'arrayp)
-                                      complexp)
-                                 `((%other-pointer-subtype-p ,object
-                                                             ',(list sb-vm:complex-base-string-widetag
-                                                                     #+sb-unicode sb-vm:complex-character-string-widetag
-                                                                     sb-vm:complex-bit-vector-widetag
-                                                                     sb-vm:complex-vector-widetag
-                                                                     sb-vm:complex-array-widetag))))
-                                (t
-                                 `(,@(unless (or (and headerp (eql pred 'arrayp))
-                                                 simple-array-header-p)
-                                       ;; ARRAY-HEADER-P from TESTS will test for that
-                                       `((,pred ,object)))
-                                   ,@(when complexp
-                                       `((typep ,object '(not simple-array)))))))
-                        ,@tests
-                        ,@(test-array-element-type object type stype headerp)))))
+                      (multiple-value-bind (type-test no-check-for-array length-checked)
+                          (test-array-element-type object type stype headerp pred length)
+                        (if no-check-for-array
+                            `(and ,@type-test
+                                  ,@(unless length-checked
+                                      dim-tests))
+                            `(and
+                              ,@(cond ((and (eql pred 'vectorp)
+                                            complexp)
+                                       `((%other-pointer-subtype-p ,object
+                                                                   ',(list sb-vm:complex-base-string-widetag
+                                                                           #+sb-unicode sb-vm:complex-character-string-widetag
+                                                                           sb-vm:complex-bit-vector-widetag
+                                                                           sb-vm:complex-vector-widetag))))
+                                      ((and (eql pred 'arrayp)
+                                            complexp)
+                                       `((%other-pointer-subtype-p ,object
+                                                                   ',(list sb-vm:complex-base-string-widetag
+                                                                           #+sb-unicode sb-vm:complex-character-string-widetag
+                                                                           sb-vm:complex-bit-vector-widetag
+                                                                           sb-vm:complex-vector-widetag
+                                                                           sb-vm:complex-array-widetag))))
+                                      (t
+                                       `(,@(unless (or (and headerp (eql pred 'arrayp))
+                                                       simple-array-header-p)
+                                             ;; ARRAY-HEADER-P from DIM-TESTS will test for that
+                                             `((,pred ,object)))
+                                         ,@(when complexp
+                                             `((typep ,object '(not simple-array)))))))
+                              ,@dim-tests
+                              ,@type-test))))))
               `(%typep ,object ',(type-specifier type)))))))
 
 ;;; Transform a type test against some instance type. The type test is
@@ -1437,12 +1567,12 @@
        `(character x))
       ;; Handle specialized element types for 1D arrays.
       ((multiple-value-bind (result already-type-p dimension specialization)
-           (cond ((or (and (array-type-p tspec)
-                           (neq (array-type-complexp tspec) t) ; :MAYBE and NIL are good
-                           (not (contains-unknown-type-p (array-type-element-type tspec)))
-                           ;; just for requesting (array nil (*)), you lose
-                           (neq (array-type-specialized-element-type tspec) *empty-type*)
-                           (consp (array-type-dimensions tspec))))
+           (cond ((and (array-type-p tspec)
+                       (neq (array-type-complexp tspec) t) ; :MAYBE and NIL are good
+                       (not (contains-unknown-type-p (array-type-element-type tspec)))
+                       ;; just for requesting (array nil (*)), you lose
+                       (neq (array-type-specialized-element-type tspec) *empty-type*)
+                       (consp (array-type-dimensions tspec)))
                   (values tspec
                           (source-transform-array-typep 'x tspec)
                           (car (array-type-dimensions tspec))
@@ -1486,7 +1616,12 @@
                            `(make-array ,dimension ,@specialization :initial-contents x))))))))
       ((type= tspec (specifier-type 'list))
        `(coerce-to-list x))
-      ((csubtypep tspec (specifier-type 'function))
+      ((csubtypep tspec (specifier-type 'extended-sequence))
+       (let ((class (and (symbolp tval) (find-class tval nil))))
+         (if (null class)
+             (give-up-ir1-transform)
+             `(coerce-to-extended-sequence x (load-time-value (find-class ',tval) t)))))
+      ((type= tspec (specifier-type 'function))
        (if (csubtypep (lvar-type x) (specifier-type 'symbol))
            `(coerce-symbol-to-fun x)
            ;; if X can later be derived as FUNCTION then we don't want
@@ -1519,6 +1654,11 @@
 ;;; BIGNUMP is simpler than INTEGERP, so if we can rule out FIXNUM then ...
 (deftransform integerp ((x) ((not fixnum)) * :important nil) '(bignump x))
 
+(deftransform structure-typep ((object type) (t t) * :node node)
+  (if (types-equal-or-intersect (lvar-type object) (specifier-type 'instance))
+      (give-up-ir1-transform)
+      nil))
+
 (deftransform structure-typep ((object type) (t (constant-arg t)))
   (let* ((layout (lvar-value type))
          (type (case layout
@@ -1529,9 +1669,14 @@
                   (layout-classoid layout))))
          (diff (type-difference (lvar-type object) type))
          (pred (backend-type-predicate diff)))
-    (if pred
-        `(not (,pred object))
-        (give-up-ir1-transform))))
+    (cond ((not (types-equal-or-intersect (lvar-type object) type))
+           nil)
+          ((csubtypep (lvar-type object) type)
+           t)
+          (pred
+           `(not (,pred object)))
+          (t
+           (give-up-ir1-transform)))))
 
 (deftransform classoid-cell-typep ((cell object) ((constant-arg t) t))
   (let* ((type (specifier-type (classoid-cell-name (lvar-value cell))))
@@ -1540,3 +1685,13 @@
     (if pred
         `(not (,pred object))
         (give-up-ir1-transform))))
+
+(when-vop-existsp (:translate signed-byte-8-p)
+  (macrolet ((def (bits)
+               `(deftransform ,(symbolicate "SIGNED-BYTE-" (princ-to-string bits) "-P")
+                    ((x) (unsigned-byte) * :important nil)
+                  '(typep x '(unsigned-byte ,(1- bits))))))
+    (def 8)
+    (def 16)
+    #+64-bit
+    (def 32)))
