@@ -77,70 +77,40 @@
              (dolist (flag flags)
                (inst jmp flag dest)))))))
 
-(define-vop (multiway-branch-if-eq)
-  ;; TODO: also accept signed-reg, unsigned-reg, character-reg
-  ;; also, could probably tighten up the TN lifetime to avoid a move
-  (:args (x :scs (any-reg descriptor-reg)))
-  (:info labels otherwise key-type keys test-vop-name)
+(define-vop (jump-table)
+  (:args (index :scs (signed-reg unsigned-reg any-reg)
+                :target offset))
+  (:info targets otherwise min max)
+  (:temporary (:sc any-reg :from (:argument 0)) offset)
   (:temporary (:sc unsigned-reg) table)
-  (:temporary (:sc unsigned-reg) temp)
-  (:arg-refs x-tn-ref)
-  (:generator 10
-    (let* ((key-derived-type (tn-ref-type x-tn-ref))
-           (ea)
-           (min (car keys)) ; keys are sorted
-           (max (car (last keys)))
-           (vector (make-array (1+ (- max min)) :initial-element otherwise))
-           ;; This fixnumize won't overflow because ir2opt won't use
-           ;; a multiway-branch unless all keys are char or fixnum.
-           ;; But what if MIN is MOST-NEGATIVE-FIXNUM ????
-           (-min (fixnumize (- min))))
-      (mapc (lambda (key label) (setf (aref vector (- key min)) label))
-            keys labels)
-      (ecase key-type
-        (fixnum
-           (cond
-             ((and (typep (* min (- sb-vm:n-word-bytes)) '(signed-byte 32))
-                   (typep key-derived-type 'numeric-type)
-                   (csubtypep key-derived-type (specifier-type 'fixnum))
-                   ;; There could be some dead code if the ranges don't line up.
-                   (>= (numeric-type-low key-derived-type) min)
-                   (<= (numeric-type-high key-derived-type) max))
-              (setq ea (ea (* min (- sb-vm:n-word-bytes)) table x 4)))
-             (t
-              ;; First exclude out-of-bounds values because there's no harm
-              ;; in doing that up front regardless of the argument's lisp type.
-              (typecase -min
-                ;; TODO: if min is 0, use X directly, don't move into temp
-                ((eql 0) (move temp x))
-                ((signed-byte 32) (inst lea temp (ea -min x)))
-                (t (inst mov temp x)
-                   (inst add :qword temp (constantize -min))))
-              (inst cmp temp (constantize (fixnumize (- max min))))
-              (inst jmp :a otherwise)
-              ;; We have to check the type here because a chain of EQ tests
-              ;; does not impose a type constraint.
-              ;; If type of X was derived as fixnum, then elide this test.
-              (unless (eq test-vop-name 'sb-vm::fast-if-eq-fixnum/c)
-                (inst test :byte x fixnum-tag-mask)
-                (inst jmp :ne otherwise))
-              (setq ea (ea table temp 4))))
-            (inst lea table (register-inline-constant :jump-table vector))
-            (inst jmp ea))
-        (character
-           ;; Same as above, but test the widetag before shifting it out.
-           (unless (member test-vop-name '(fast-char=/character/c
-                                           fast-if-eq-character/c))
-             (inst cmp :byte x character-widetag)
-             (inst jmp :ne otherwise))
-           (inst mov :dword temp x)
-           (inst shr :dword temp n-widetag-bits)
-           (unless (= min 0)
-             (inst sub :dword temp min))
-           (inst cmp temp (- max min))
-           (inst jmp :a otherwise)
-           (inst lea table (register-inline-constant :jump-table vector))
-           (inst jmp (ea table temp 8)))))))
+  (:generator 0
+    (let ((fixnump (sc-is index any-reg)))
+      (flet ((fix (x)
+               (if fixnump
+                   (fixnumize x)
+                   x)))
+        (let (disp)
+          (unless (zerop min)
+            (let ((byte-disp (- (* min n-word-bytes))))
+              (if (and (not otherwise)
+                       (typep byte-disp '(signed-byte 32)))
+                  (setf disp byte-disp)
+                  (let ((diff (- (fix min))))
+                    (unless (typep diff '(signed-byte 32))
+                      (inst mov table diff)
+                      (setf diff table))
+                    (cond ((location= offset index)
+                           (inst add offset diff))
+                          (t
+                           (inst lea offset (ea diff index))
+                           (setf index offset)))))))
+          (when otherwise
+            (inst cmp index (fix (- max min)))
+            (inst jmp :a otherwise))
+          (inst lea table (register-inline-constant :jump-table targets))
+          (inst jmp (if disp
+                        (ea disp table index (if fixnump 4 8))
+                        (ea table index (if fixnump 4 8)))))))))
 
 (defun convert-conditional-move-p (dst-tn)
   (sc-case dst-tn
@@ -327,7 +297,7 @@
 ;;; Note: a constant-tn is allowed in CMP; it uses an EA displacement,
 ;;; not immediate data.
 (define-vop (if-eq)
-  (:args (x :scs (any-reg descriptor-reg control-stack immediate))
+  (:args (x :scs (any-reg descriptor-reg control-stack))
          (y :scs (any-reg descriptor-reg control-stack immediate constant)))
   (:conditional :e)
   (:policy :fast-safe)
@@ -419,10 +389,10 @@
   (assert (eql other-pointer-lowtag #b1111))
   ;; This is also assumed in src/runtime/x86-64-assem.S
   (assert (eql (min bignum-widetag ratio-widetag single-float-widetag double-float-widetag
-                    complex-widetag complex-single-float-widetag complex-double-float-widetag)
+                    complex-rational-widetag complex-single-float-widetag complex-double-float-widetag)
                bignum-widetag))
   (assert (eql (max bignum-widetag ratio-widetag single-float-widetag double-float-widetag
-                    complex-widetag complex-single-float-widetag complex-double-float-widetag)
+                    complex-rational-widetag complex-single-float-widetag complex-double-float-widetag)
                complex-double-float-widetag)))
 
 ;;; Most uses of EQL are transformed into a non-generic form, but when we need
@@ -460,10 +430,8 @@
 
     ;; If the widetags are not the same, return false.
     ;; Using a :dword compare gets us the bignum length check almost for free
-    ;; unless the length's representation requires more than 3 bytes.
-    ;; It sounds like a :qword compare would be the right thing, but remember
-    ;; one header bit acts as a concurrent GC mark bit in all headered objects,
-    ;; though we're not really using it yet. (We are, but not concurrently)
+    ;; unless the length's representation requires more 4 bytes.
+    ;; I bet nobody would mind if MAXIMUM-BIGNUM-LENGTH were #xFFFFFF.
     (inst mov :dword rax (ea (- other-pointer-lowtag) x))
     (inst cmp :dword rax (ea (- other-pointer-lowtag) y))
     (inst jmp :ne done) ; negative
