@@ -43,59 +43,64 @@
 
 ;; Numerics including fixnum, excluding short-float. (INTEGER,RATIONAL)
 (defun %test-fixnum-and-headers (value temp target not-p headers
-                                 &key value-tn-ref)
+                                 &key value-tn-ref immediate-tested)
   (let ((drop-through (gen-label)))
     (case n-fixnum-tag-bits
-     (1 (%lea-for-lowtag-test temp value other-pointer-lowtag :qword)
+     (1
+      (%lea-for-lowtag-test temp value other-pointer-lowtag :qword)
+      (when (types-equal-or-intersect (tn-ref-type value-tn-ref)
+                                      (specifier-type 'fixnum))
         (inst test :byte temp 1)
-        (inst jmp :nz (if not-p drop-through target)) ; inverted
-        (%test-headers value temp target not-p nil headers
-                       :drop-through drop-through :compute-temp nil
-                       :value-tn-ref value-tn-ref))
+        (inst jmp :nz (if not-p drop-through target))) ; inverted
+      (%test-headers value temp target not-p nil headers
+                     :drop-through drop-through :compute-temp nil
+                     :value-tn-ref value-tn-ref
+                     :immediate-tested immediate-tested))
      (t
       (generate-fixnum-test value)
       (inst jmp :z (if not-p drop-through target))
       (%test-headers value temp target not-p nil headers
-                     :drop-through drop-through)))))
-
-;; I can see no reason this would ever be used.
-;; (or fixnum character|unbound-marker) is implausible.
-(defun %test-fixnum-and-immediate (value target not-p immediate)
-  (error "WAT")
-  (let ((drop-through (gen-label)))
-    (generate-fixnum-test value)
-    (inst jmp :z (if not-p drop-through target))
-    (%test-immediate value target not-p immediate drop-through)))
+                     :drop-through drop-through
+                     :immediate-tested immediate-tested)))))
 
 ;; Numerics
 (defun %test-fixnum-immediate-and-headers (value temp target not-p immediate headers
-                                           &key value-tn-ref)
+                                           &key value-tn-ref immediate-tested)
   (let ((drop-through (gen-label)))
     (case n-fixnum-tag-bits
-     (1 (%lea-for-lowtag-test temp value other-pointer-lowtag :qword)
+     (1
+      (%lea-for-lowtag-test temp value other-pointer-lowtag :qword)
+      (when (types-equal-or-intersect (tn-ref-type value-tn-ref)
+                                      (specifier-type 'fixnum))
         (inst test :byte temp 1)
-        (inst jmp :nz (if not-p drop-through target)) ; inverted
+        (inst jmp :nz (if not-p drop-through target))) ; inverted
+      (when (or (/= immediate single-float-widetag)
+                (types-equal-or-intersect (tn-ref-type value-tn-ref)
+                                          (specifier-type 'single-float)))
         (inst cmp :byte temp (- immediate other-pointer-lowtag))
-        (inst jmp :e (if not-p drop-through target))
-        (%test-headers value temp target not-p nil headers
-                       :drop-through drop-through :compute-temp nil
-                       :value-tn-ref value-tn-ref))
+        (inst jmp :e (if not-p drop-through target)))
+
+      (%test-headers value temp target not-p nil headers
+                     :drop-through drop-through :compute-temp nil
+                     :value-tn-ref value-tn-ref
+                     :immediate-tested immediate-tested))
      (t (generate-fixnum-test value)
         (inst jmp :z (if not-p drop-through target))
         (%test-immediate-and-headers value temp target not-p immediate headers
-                                     :drop-through drop-through)))))
+                                     :drop-through drop-through
+                                     :immediate-tested immediate-tested)))))
 
 (defun %test-immediate (value temp target not-p immediate
-                        &optional (drop-through (gen-label)))
-  (declare (ignore temp))
+                        &key value-tn-ref)
+  (declare (ignore temp value-tn-ref))
   (inst cmp :byte value immediate)
-  (inst jmp (if not-p :ne :e) target)
-  (emit-label drop-through))
+  (inst jmp (if not-p :ne :e) target))
 
 ;; Numerics including short-float, excluding fixnum
 (defun %test-immediate-and-headers (value temp target not-p immediate headers
                                     &key (drop-through (gen-label))
-                                         value-tn-ref)
+                                         value-tn-ref
+                                         immediate-tested)
   ;; Code a single instruction byte test if possible.
   (cond ((sc-is value any-reg descriptor-reg)
          (inst cmp :byte value immediate))
@@ -105,9 +110,11 @@
   (inst jmp :e (if not-p drop-through target))
   (%test-headers value temp target not-p nil headers
                  :drop-through drop-through
-                 :value-tn-ref value-tn-ref))
+                 :value-tn-ref value-tn-ref
+                 :immediate-tested immediate-tested))
 
-(defun %test-lowtag (value temp target not-p lowtag)
+(defun %test-lowtag (value temp target not-p lowtag &key value-tn-ref)
+  (declare (ignore value-tn-ref))
   (%lea-for-lowtag-test temp value lowtag)
   (inst test :byte temp lowtag-mask)
   (inst jmp (if not-p :nz :z) target))
@@ -117,7 +124,8 @@
                            (drop-through (gen-label))
                            (load-widetag t)
                            (compute-temp load-widetag)
-                           value-tn-ref)
+                           value-tn-ref
+                           immediate-tested)
   (let* ((lowtag (if function-p fun-pointer-lowtag other-pointer-lowtag))
          ;; It is preferable (smaller and faster code) to directly
          ;; compare the value in memory instead of loading it into
@@ -149,7 +157,7 @@
       (cond ((not load-widetag))
             ((and value-tn-ref
                   (eq lowtag other-pointer-lowtag)
-                  (other-pointer-tn-ref-p value-tn-ref t))) ; best case: lowtag is right
+                  (other-pointer-tn-ref-p value-tn-ref t immediate-tested))) ; best case: lowtag is right
             ((and value-tn-ref
                   ;; If HEADERS contains a range, then list pointers have to be
                   ;; disallowed - consider a list whose CAR has a fixnum that
@@ -202,12 +210,15 @@
           (let ((header (car remaining))
                 (last (null (cdr remaining))))
             (cond
-             ((atom header)
-              (inst cmp :byte widetag-tn header)
-              (if last
-                  (inst jmp equal target)
-                  (inst jmp :e when-true)))
-             (t
+              ((and (eql header simple-array-widetag)
+                    value-tn-ref
+                    (csubtypep (tn-ref-type value-tn-ref) (specifier-type 'string))))
+              ((atom header)
+               (inst cmp :byte widetag-tn header)
+               (if last
+                   (inst jmp equal target)
+                   (inst jmp :e when-true)))
+              (t
                (let ((start (car header))
                      (end (cdr header)))
                  (cond
@@ -321,42 +332,59 @@
     (inst test value value)))
 
 (macrolet ((define (name src-size)
-             `(progn
-                (define-vop (,name)
-                  (:translate ,name)
-                  (:args (value :scs (any-reg descriptor-reg)))
-                  (:conditional :z)
-                  (:arg-refs arg-ref)
-                  (:policy :fast-safe)
-                  (:temporary (:sc unsigned-reg) temp temp2)
-                  (:generator 6
-                    ;; Optimistically assume that the argument is a fixnum.
-                    ;; If it is, then we're shifting out the tag bit, otherwise
-                    ;; putting random bits in the low byte, which is harmless.
-                    (move temp value)
-                    (inst sar temp n-fixnum-tag-bits)
-                    (inst movsx '(,src-size :qword) temp2 temp)
-                    (inst cmp temp2 temp)
-                    ;; If the input wasn't known to be a fixnum, and it passed
-                    ;; the sign-extension test, then check FIXNUMP now.
-                    (unless (csubtypep (tn-ref-type arg-ref) (specifier-type 'fixnum))
-                      (inst jmp :nz OUT) ; did NOT pass the sign-extension test
-                      (inst test :byte value fixnum-tag-mask))
-                    OUT))
-                (define-vop (,(symbolicate name "/SIGNED"))
-                  (:translate ,name)
-                  (:args (value :scs (signed-reg)))
-                  (:arg-types signed-num)
-                  (:conditional :z)
-                  (:policy :fast-safe)
-                  (:temporary (:sc unsigned-reg) temp)
-                  (:generator 2
-                    (inst movsx '(,src-size :qword) temp value)
-                    (inst cmp temp value))))))
+             `(define-vop (,(symbolicate name "/SIGNED"))
+                (:translate ,name)
+                (:args (value :scs (signed-reg)))
+                (:arg-types signed-num)
+                (:conditional :z)
+                (:policy :fast-safe)
+                (:temporary (:sc unsigned-reg) temp)
+                (:generator 2
+                  (inst movsx '(,src-size :qword) temp value)
+                  (inst cmp temp value)))))
   (define signed-byte-8-p :byte)
   (define signed-byte-16-p :word)
   (define signed-byte-32-p :dword))
 
+(define-vop (signed-byte-8-p)
+  (:translate signed-byte-8-p)
+  (:args (value :scs (any-reg descriptor-reg)))
+  (:conditional :z)
+  (:arg-refs arg-ref)
+  (:policy :fast-safe)
+  (:temporary (:sc unsigned-reg) temp)
+  (:generator 6
+    (inst lea temp (ea (ash (expt 2 7) n-fixnum-tag-bits) value))
+    (inst test temp (lognot (fixnumize (1- (expt 2 8)))))))
+
+(define-vop (signed-byte-16-p)
+  (:translate signed-byte-16-p)
+  (:args (value :scs (any-reg descriptor-reg)))
+  (:conditional :z)
+  (:arg-refs arg-ref)
+  (:policy :fast-safe)
+  (:temporary (:sc unsigned-reg) temp)
+  (:generator 6
+    (inst lea temp (ea (ash (expt 2 15) n-fixnum-tag-bits) value))
+    (inst test temp (lognot (fixnumize (1- (expt 2 16)))))))
+
+(define-vop (signed-byte-32-p)
+  (:translate signed-byte-32-p)
+  (:args (value :scs (any-reg descriptor-reg)))
+  (:conditional :z)
+  (:arg-refs arg-ref)
+  (:policy :fast-safe)
+  (:temporary (:sc unsigned-reg) temp temp2)
+  (:generator 6
+    (move temp value)
+    (inst sar temp n-fixnum-tag-bits)
+    (inst movsx '(:dword :qword) temp2 temp)
+    (inst cmp temp2 temp)
+    (unless (csubtypep (tn-ref-type arg-ref) (specifier-type 'fixnum))
+      (inst jmp :nz out)
+      (inst test :byte value fixnum-tag-mask))
+    out))
+
 ;;; Sign bit and fixnum tag bit.
 (defconstant non-negative-fixnum-mask-constant
   #x8000000000000001)
@@ -435,66 +463,6 @@
 
         (emit-label not-target)))))
 
-(defun power-of-two-limit-p (x)
-  (and (fixnump x)
-       (= (logcount (1+ x)) 1)))
-
-(define-vop (test-fixnum-mod-power-of-two)
-  (:args (value :scs (any-reg descriptor-reg
-                              unsigned-reg signed-reg)))
-  (:arg-types *
-              (:constant (satisfies power-of-two-limit-p)))
-  (:translate fixnum-mod-p)
-  (:conditional :e)
-  (:info hi)
-  (:policy :fast-safe)
-  (:arg-refs value-ref)
-  (:vop-var vop)
-  (:generator 4
-    (let ((fixnum-hi (if (sc-is value unsigned-reg signed-reg)
-                         hi
-                         (fixnumize hi))))
-      (cond ((sc-is value descriptor-reg)
-             (let* ((mask (lognot fixnum-hi))
-                    (constant
-                      (if (= (ldb (byte 64 0) mask) non-negative-fixnum-mask-constant)
-                          (ea non-negative-fixnum-mask-constant-wired-address)
-                          (constantize mask))))
-               (inst test value constant)))
-            (t
-             (change-vop-flags vop '(:be))
-             (inst cmp value (constantize fixnum-hi)))))))
-
-(define-vop (test-fixnum-mod-tagged-unsigned)
-  (:args (value :scs (any-reg unsigned-reg signed-reg)))
-  (:arg-types (:or tagged-num unsigned-num signed-num)
-              (:constant fixnum))
-  (:translate fixnum-mod-p)
-  (:conditional :be)
-  (:info hi)
-  (:policy :fast-safe)
-  (:generator 5
-     (let ((fixnum-hi (if (sc-is value unsigned-reg signed-reg)
-                          hi
-                          (fixnumize hi))))
-       (inst cmp value (constantize fixnum-hi)))))
-
-(define-vop (test-fixnum-mod-*)
-  (:args (value :scs (any-reg descriptor-reg)))
-  (:arg-types * (:constant fixnum))
-  (:translate fixnum-mod-p)
-  (:conditional)
-  (:info target not-p hi)
-  (:policy :fast-safe)
-  (:generator 6
-     (let* ((fixnum-hi (fixnumize hi))
-            (skip (gen-label)))
-       (generate-fixnum-test value)
-       (inst jmp :ne (if not-p target skip))
-       (inst cmp value (constantize fixnum-hi))
-       (inst jmp (if not-p :a :be) target)
-       (emit-label skip))))
-
 ;;; SINGLE-FLOAT-P, CHARACTERP, UNBOUND-MARKER-P produce a flag result
 ;;; and never need a temporary.
 (macrolet ((define (name widetag)
@@ -547,7 +515,7 @@
                   out))))
   (define bignump (bignum-widetag))
   (define ratiop (ratio-widetag))
-  (define complex-rational-p (complex-widetag))
+  (define complex-rational-p (complex-rational-widetag))
   (define complex-single-float-p (complex-single-float-widetag))
   (define complex-double-float-p (complex-double-float-widetag))
   (define double-float-p (double-float-widetag))
@@ -723,10 +691,8 @@
           (value :scs (any-reg descriptor-reg)))
    (:vop-var vop)
    (:temporary (:sc unsigned-reg) temp)
-   (:node-var node)
    (:generator 1
-     (unless (sb-c::set-slot-old-p node)
-       (emit-gengc-barrier object nil temp (vop-nth-arg 1 vop) value))
+     (emit-gengc-barrier object nil temp (vop-nth-arg 1 vop) value)
      (inst mov :dword (ea (- 4 instance-pointer-lowtag) object) value)))
  (define-vop (%fun-layout %instance-layout)
    (:translate %fun-layout)
@@ -734,8 +700,7 @@
  (define-vop (%set-fun-layout %set-instance-layout)
    (:translate %set-fun-layout)
    (:generator 1
-     (unless (sb-c::set-slot-old-p node)
-       (emit-gengc-barrier object nil temp (vop-nth-arg 1 vop) value))
+     (emit-gengc-barrier object nil temp (vop-nth-arg 1 vop) value)
      (inst mov :dword (ea (- 4 fun-pointer-lowtag) object) value)))
  (define-vop ()
   (:translate sb-c::layout-eq)
@@ -780,7 +745,7 @@
 (define-vop (fixnump simple-type-predicate)
   (:translate fixnump)
   (:arg-refs arg-ref)
-  (:args (value :scs (any-reg descriptor-reg) :load-if (tn-ref-memory-access arg-ref)))
+  (:args (value :scs (any-reg descriptor-reg)))
   (:conditional :z)
   ;; the compiler is very sensitive to this cost here as regards boxing. DON'T TOUCH !!!
   (:generator 3
@@ -883,24 +848,42 @@
         (assemble ()
           (when (types-equal-or-intersect (tn-ref-type integer-ref) (specifier-type 'fixnum))
             (generate-fixnum-test integer)
-            (inst jmp :nz BIGNUM)
-            (cond ((eql fixnum 0)
-                   (inst test integer integer))
-                  ((or (and (eql fixnum #.(fixnumize -1))
-                            (or (and (eql comparison :le)
-                                     (setf comparison :l))
-                                (and (eql comparison :g)
-                                     (setf comparison :ge))))
-                       (and (eql fixnum #.(fixnumize 1))
-                            (or (and (eql comparison :ge)
-                                     (setf comparison :g))
-                                (and (eql comparison :l)
-                                     (setf comparison :le)))))
-                   (inst test integer integer))
-                  (t
-                   (inst cmp integer fixnum)))
-            (inst jmp comparison yep)
-            (inst jmp nope))
+            (cond
+              ((and (eql fixnum 0)
+                    (memq comparison '(:ge :l))
+                    (csubtypep (type-intersection (tn-ref-type integer-ref)
+                                                  (specifier-type 'fixnum))
+                               (specifier-type '(integer * -1))))
+               (inst jmp :z (if (eq comparison :l)
+                                yep
+                                nope)))
+              ((and (eql fixnum 0)
+                    (memq comparison '(:ge :l))
+                    (csubtypep (type-intersection (tn-ref-type integer-ref)
+                                                  (specifier-type 'fixnum))
+                               (specifier-type '(integer 0))))
+               (inst jmp :z (if (eq comparison :l)
+                                nope
+                                yep)))
+              (t
+               (inst jmp :nz BIGNUM)
+               (cond ((eql fixnum 0)
+                      (inst test integer integer))
+                     ((or (and (eql fixnum #.(fixnumize -1))
+                               (or (and (eql comparison :le)
+                                        (setf comparison :l))
+                                   (and (eql comparison :g)
+                                        (setf comparison :ge))))
+                          (and (eql fixnum #.(fixnumize 1))
+                               (or (and (eql comparison :ge)
+                                        (setf comparison :g))
+                                   (and (eql comparison :l)
+                                        (setf comparison :le)))))
+                      (inst test integer integer))
+                     (t
+                      (inst cmp integer fixnum)))
+               (inst jmp comparison yep)
+               (inst jmp nope))))
           bignum
           (unless (fixnum-or-other-pointer-tn-ref-p integer-ref t)
             (test-type integer temp nope t (other-pointer-lowtag)))
@@ -1102,7 +1085,6 @@
   (:args (layout :scs (descriptor-reg)))
   (:arg-types * (:constant t))
   (:policy :fast-safe)
-  (:conditional)
   (:info target not-p test-layout)
   (:generator 4
     (structure-is-a layout test-layout target not-p done)
