@@ -84,7 +84,7 @@ open_binary(char *filename, int mode)
     return open(filename, mode);
 }
 
-#if defined LISP_FEATURE_LINUX && defined LISP_FEATURE_X86_64
+#if defined LISP_FEATURE_LINUX && (defined LISP_FEATURE_X86_64 || defined LISP_FEATURE_ARM64)
 #define ELFCORE 1
 #elif !defined(ELFCORE)
 #define ELFCORE 0
@@ -330,13 +330,13 @@ static void adjust_pointers(lispobj *where, sword_t n_words, struct heap_adjust*
 }
 
 #include "var-io.h"
-#include "unaligned.h"
+#include "align.h"
 static void
 adjust_code_refs(struct heap_adjust __attribute__((unused)) *adj,
                  struct code __attribute__((unused)) *code,
                  lispobj __attribute__((unused)) original_vaddr)
 {
-#ifdef LISP_FEATURE_IMMOBILE_SPACE
+#if defined LISP_FEATURE_PERMGEN || defined LISP_FEATURE_IMMOBILE_SPACE
     // Dynamic space always gets relocated before immobile space does,
     // and dynamic space code does not use fixups (except on 32-bit x86).
     // So if we're here, it must be to relocate an immobile object.
@@ -388,27 +388,7 @@ static inline void fix_fun_header_layout(lispobj __attribute__((unused)) *fun,
 #endif
 }
 
-/* Refer to the figures above CALCULATE-DD-BITMAP in src/code/defstruct
- * for the meaning of these constants.
- * I think this fixes what may have been a latent bug in space relocation.
- * Due to inconsistent use of the bitmap, it could have processed funinstance
- * LAYOUT slots twice.  Layout is supposed to have a 0 bit, meaning do NOT scan
- * in a loop over slots, but instead externally to the loop. By marking it as 1,
- * it could be processed both outside and inside the loop. In the extremely rare
- * case of partial overlap of the desired and actual heap addresses, a pointer
- * somewhere in the middle can not be discerned as having been fixed up or not
- * fixed up based on its value. You have to know based on code flow.
- * Therefore layouts slots always have to be treated as raw in the bitmap,
- * but I did not see where it is enforced in lisp for funinstances. */
-#ifndef LISP_FEATURE_EXECUTABLE_FUNINSTANCES
-# define FUNINSTANCE_BITMAP -4
-#elif defined LISP_FEATURE_COMPACT_INSTANCE_HEADER
-# define FUNINSTANCE_BITMAP -8
-#else
-# define FUNINSTANCE_BITMAP -24
-#endif
-
-static void relocate_space(uword_t start, lispobj* end, struct heap_adjust* adj)
+static void fix_space(uword_t start, lispobj* end, struct heap_adjust* adj)
 {
     int widetag;
     long nwords;
@@ -429,6 +409,12 @@ static void relocate_space(uword_t start, lispobj* end, struct heap_adjust* adj)
         widetag = header_widetag(word);
         nwords = sizetab[widetag](where);
         switch (widetag) {
+        case FUNCALLABLE_INSTANCE_WIDETAG:
+            /* If non-executable funinstance, then word index 1 points at read-only space,
+             * hence needs no adjustment. Otherwise, the word points within the funinstance.
+             * Either way, adjust_word_at will do the right thing */
+            adjust_word_at(where+1, adj);
+            /* FALLTHRU */
         case INSTANCE_WIDETAG:
             layout = layout_of(where);
             adjusted_layout = adjust_word(adj, layout);
@@ -442,33 +428,18 @@ static void relocate_space(uword_t start, lispobj* end, struct heap_adjust* adj)
                 if (bitmap_logbitp(i, bitmap)) adjust_pointers(slots+i, 1, adj);
             }
             continue;
-        case FUNCALLABLE_INSTANCE_WIDETAG:
-            /* If non-executable funinstance, then word index 1 points at read-only space,
-             * hence needs no adjustment. Otherwise, the word points within the funinstance.
-             * Either way, adjust_word_at will do the right thing */
-            adjust_word_at(where+1, adj);
-            /* We have to adjust the layout, but its bitmap can be disregarded.
-             * There is no variation in bitmap by particular subtype of funinstance */
-            layout = funinstance_layout(where);
-            adjusted_layout = adjust_word(adj, layout);
-            if (adjusted_layout != layout) funinstance_layout(where) = adjusted_layout;
-            const int mask = FUNINSTANCE_BITMAP;
-            gc_assert((mask & (1<<0)) == 0); // bitmap must reflect it as untagged
-            {
-            lispobj* slots = where+1;
-            // slots[0] is where[1] which we just adjusted, so skip it.
-            for (i=1; i<(nwords-1); ++i) // -1 from nwords because 'slots' is +1 from 'where'
-                if ((1<<i) & mask) adjust_pointers(slots+i, 1, adj);
-            }
-            continue;
         case SYMBOL_WIDETAG:
-            { // Copied from scav_symbol() in gc-common
+            { // Modeled on scav_symbol() in gc-common
             struct symbol* s = (void*)where;
-            adjust_pointers(&s->value, 3, adj); // value, fdefn, info
+#ifdef LISP_FEATURE_64_BIT
+            adjust_pointers(where + 1, 4, adj);
             lispobj name = decode_symbol_name(s->name);
             lispobj adjusted_name = adjust_word(adj, name);
             // writeback the name if it changed
             if (adjusted_name != name) FIXUP(set_symbol_name(s, adjusted_name), &s->name);
+#else
+            adjust_pointers(&s->fdefn, 4, adj); // fdefn, value, info, name
+#endif
             }
             continue;
         case FDEFN_WIDETAG:
@@ -491,7 +462,7 @@ static void relocate_space(uword_t start, lispobj* end, struct heap_adjust* adj)
                 adjust_word_at(fdefns_start+i, adj);
 #endif
 #if defined LISP_FEATURE_X86 || defined LISP_FEATURE_X86_64 || \
-    defined LISP_FEATURE_PPC || defined LISP_FEATURE_PPC64
+    defined LISP_FEATURE_PPC || defined LISP_FEATURE_PPC64 || defined LISP_FEATURE_ARM64
             // Fixup absolute jump table
             lispobj* jump_table = code_jumptable_start(code);
             int count = jumptable_count(jump_table);
@@ -571,7 +542,7 @@ static void relocate_space(uword_t start, lispobj* end, struct heap_adjust* adj)
         case VALUE_CELL_WIDETAG:
         case WEAK_POINTER_WIDETAG:
         case RATIO_WIDETAG:
-        case COMPLEX_WIDETAG:
+        case COMPLEX_RATIONAL_WIDETAG:
             break;
 
         // Other
@@ -592,8 +563,9 @@ static void relocate_space(uword_t start, lispobj* end, struct heap_adjust* adj)
         adjust_pointers(where+1, nwords-1, adj);
     }
 #if SHOW_SPACE_RELOCATION
-    fprintf(stderr, "space @ %p..%p: fixed %d absolute + %d relative pointers\n",
-            (lispobj*)start, end, adj->n_relocs_abs, adj->n_relocs_rel);
+    if (end)
+        fprintf(stderr, "space @ %p..%p: fixed %d absolute + %d relative pointers\n",
+                (lispobj*)start, end, adj->n_relocs_abs, adj->n_relocs_rel);
 #endif
 }
 
@@ -610,21 +582,24 @@ static void relocate_heap(struct heap_adjust* adj)
                         (char*)adj->range[i].end + adj->range[i].delta);
     }
 #ifdef LISP_FEATURE_RELOCATABLE_STATIC_SPACE
-    relocate_space(READ_ONLY_SPACE_START, read_only_space_free_pointer, adj);
+    fix_space(READ_ONLY_SPACE_START, read_only_space_free_pointer, adj);
     // Relocate the CAR slot of nil-as-a-list, which needs to point to
     // itself.
     adjust_pointers((void*)(NIL - LIST_POINTER_LOWTAG), 1, adj);
 #endif
-    relocate_space(NIL_SYMBOL_SLOTS_START, (lispobj*)NIL_SYMBOL_SLOTS_END, adj);
-    relocate_space(STATIC_SPACE_OBJECTS_START, static_space_free_pointer, adj);
-#ifdef LISP_FEATURE_IMMOBILE_SPACE
-    relocate_space(FIXEDOBJ_SPACE_START, fixedobj_free_pointer, adj);
+    fix_space(NIL_SYMBOL_SLOTS_START, (lispobj*)NIL_SYMBOL_SLOTS_END, adj);
+    fix_space(STATIC_SPACE_OBJECTS_START, static_space_free_pointer, adj);
+#ifdef LISP_FEATURE_PERMGEN
+    fix_space(PERMGEN_SPACE_START, permgen_space_free_pointer, adj);
 #endif
-    relocate_space(DYNAMIC_SPACE_START, (lispobj*)dynamic_space_highwatermark(),
+#ifdef LISP_FEATURE_IMMOBILE_SPACE
+    fix_space(FIXEDOBJ_SPACE_START, fixedobj_free_pointer, adj);
+#endif
+    fix_space(DYNAMIC_SPACE_START, (lispobj*)dynamic_space_highwatermark(),
                    adj);
     // FIXME: This fixes pointers in the space but I think it won't work
     // if the space itself fails to map where requested.
-    relocate_space(TEXT_SPACE_START, text_space_highwatermark, adj);
+    fix_space(TEXT_SPACE_START, text_space_highwatermark, adj);
 }
 
 #if defined(LISP_FEATURE_ELF) && defined(LISP_FEATURE_IMMOBILE_SPACE)
@@ -649,7 +624,10 @@ void calc_asm_routine_bounds()
                 asm_routines_start = (uword_t)where;
                 break;
             }
-        if (!asm_routines_start && TEXT_SPACE_START != 0) {
+        /* I thought it would be possible to erase the static-space copy of asm routines
+         * after converting to text space but maybe not. So they'll still be here.
+         * But I want to erase them. This disabled block represents such aspiration */
+        if (0) { // !asm_routines_start && TEXT_SPACE_START != 0) {
             lispobj *where = (lispobj*)TEXT_SPACE_START;
             where += object_size(where);
             gc_assert(widetag_of(where) == CODE_HEADER_WIDETAG);
@@ -718,7 +696,15 @@ static os_vm_address_t reserve_space(int space_id, int attr,
     }
 #endif
     if (size == 0) return addr;
-    addr = os_alloc_gc_space(space_id, attr, addr, size);
+    // 64-bit already allocated a trap page when the GC card mark table was made
+#if defined(LISP_FEATURE_SB_SAFEPOINT) && !defined(LISP_FEATURE_X86_64)
+    if (space_id == STATIC_CORE_SPACE_ID) {
+        // Allocate space for the safepoint page.
+        addr = os_alloc_gc_space(space_id, attr, addr - BACKEND_PAGE_BYTES, size + BACKEND_PAGE_BYTES) + BACKEND_PAGE_BYTES;
+    }
+    else
+#endif
+        addr = os_alloc_gc_space(space_id, attr, addr, size);
     if (!addr) lose("Can't allocate %#"OBJ_FMTX" bytes for space %d", size, space_id);
     return addr;
 }
@@ -784,18 +770,14 @@ process_directory(int count, struct ndir_entry *entry,
     }
 
     for ( ; --count >= 0 ; ++entry) {
-        long id = entry->identifier;
+        long id = entry->identifier; // FIXME: is this 'long' and not 'int' for a reason?
         uword_t addr = entry->address;
         int compressed = id & DEFLATED_CORE_SPACE_ID_FLAG;
         id -= compressed;
         if (id < 1 || id > MAX_CORE_SPACE_ID)
             lose("unknown space ID %ld addr %p", id, (void*)addr);
 
-#ifdef LISP_FEATURE_DARWIN_JIT
-        int enforce_address = (id == READ_ONLY_CORE_SPACE_ID);
-#else
         int enforce_address = 0;
-#endif
 #ifndef LISP_FEATURE_RELOCATABLE_STATIC_SPACE
         enforce_address |= (id == STATIC_CORE_SPACE_ID);
 #endif
@@ -815,7 +797,6 @@ process_directory(int count, struct ndir_entry *entry,
                  (unsigned long)len >> 10,
                  (unsigned long)dynamic_space_size >> 10);
         }
-#ifndef LISP_FEATURE_DARWIN_JIT
 #ifdef LISP_FEATURE_RELOCATABLE_STATIC_SPACE
         if (id == READ_ONLY_CORE_SPACE_ID || id == STATIC_CORE_SPACE_ID) {
 #else
@@ -830,7 +811,6 @@ process_directory(int count, struct ndir_entry *entry,
                 READ_ONLY_SPACE_START = READ_ONLY_SPACE_END = addr;
             }
         }
-#endif
         if (len != 0) {
             spaces[id].len = len;
             size_t request = spaces[id].desired_size;
@@ -843,18 +823,19 @@ process_directory(int count, struct ndir_entry *entry,
             addr = (uword_t)reserve_space(id, sub_2gb_flag ? MOVABLE_LOW : MOVABLE,
                                           (os_vm_address_t)addr, request);
             switch (id) {
+            case PERMGEN_CORE_SPACE_ID:
+                PERMGEN_SPACE_START = addr;
+                break;
             case STATIC_CORE_SPACE_ID:
 #ifdef LISP_FEATURE_RELOCATABLE_STATIC_SPACE
                 STATIC_SPACE_START = addr;
                 STATIC_SPACE_END = addr + len;
 #endif
                 break;
-#ifndef LISP_FEATURE_DARWIN_JIT
             case READ_ONLY_CORE_SPACE_ID:
                 READ_ONLY_SPACE_START = addr;
                 READ_ONLY_SPACE_END = addr + len;
                 break;
-#endif
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
             case IMMOBILE_FIXEDOBJ_CORE_SPACE_ID:
             case IMMOBILE_TEXT_CORE_SPACE_ID:
@@ -936,17 +917,15 @@ process_directory(int count, struct ndir_entry *entry,
 
     calc_asm_routine_bounds();
     struct cons spaceadj = {(lispobj)spaces, (lispobj)adj};
-#ifndef LISP_FEATURE_DARWIN_JIT
     set_adjustment(&spaceadj, READ_ONLY_CORE_SPACE_ID, READ_ONLY_SPACE_START);
-#endif
     set_adjustment(&spaceadj, DYNAMIC_CORE_SPACE_ID, DYNAMIC_SPACE_START);
 #ifdef LISP_FEATURE_RELOCATABLE_STATIC_SPACE
     set_adjustment(&spaceadj, STATIC_CORE_SPACE_ID, STATIC_SPACE_START);
 #endif
+#ifdef LISP_FEATURE_PERMGEN
+    set_adjustment(&spaceadj, PERMGEN_CORE_SPACE_ID, PERMGEN_SPACE_START);
+#endif
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
-    if (lisp_code_in_elf() && TEXT_SPACE_START != spaces[IMMOBILE_TEXT_CORE_SPACE_ID].base) {
-        lose("code-in-elf + PIE not supported");
-    }
     set_adjustment(&spaceadj, IMMOBILE_FIXEDOBJ_CORE_SPACE_ID, FIXEDOBJ_SPACE_START);
     if (!apply_pie_relocs(TEXT_SPACE_START
                           - spaces[IMMOBILE_TEXT_CORE_SPACE_ID].base,
@@ -1009,6 +988,9 @@ bool gc_allocate_ptes()
     ++page_table;
     gc_page_pins = calloc(page_table_pages, 1);
     gc_assert(gc_page_pins);
+#ifdef LISP_FEATURE_DARWIN_JIT
+    page_execp = calloc(page_table_pages, 1);
+#endif
 
     // The card table size is a power of 2 at *least* as large
     // as the number of cards. These are the default values.
@@ -1037,7 +1019,17 @@ bool gc_allocate_ptes()
     num_gc_cards = 1L << gc_card_table_nbits;
 
     gc_card_table_mask =  num_gc_cards - 1;
+#if defined LISP_FEATURE_SB_SAFEPOINT && defined LISP_FEATURE_X86_64
+    /* The card table is hardware-page-aligned. Preceding it and occupying a whole
+     * "backend page" - which by the way is overkill - is the global safepoint trap page.
+     * The dummy TEST instruction for safepoints encodes shorter this way */
+    void* result = os_alloc_gc_space(0, MOVABLE, 0,
+                                     ALIGN_UP(num_gc_cards, BACKEND_PAGE_BYTES) + BACKEND_PAGE_BYTES);
+    gc_card_mark = (unsigned char*)result + BACKEND_PAGE_BYTES;
+#else
     gc_card_mark = successful_malloc(num_gc_cards);
+#endif
+
     /* The mark array used to work "by accident" if the numeric value of CARD_MARKED
      * is 0 - or equivalently the "WP'ed" state - which is the value that calloc()
      * fills with. If using malloc() we have to fill with CARD_MARKED,
@@ -1106,7 +1098,7 @@ static void gengcbarrier_patch_code_range(uword_t start, lispobj* limit)
  * it's not written to after changing its protection flags.
  * Touch every page... */
 void darwin_jit_code_pages_kludge () {
-    THREAD_JIT(0);
+    THREAD_JIT_WP(0);
     page_index_t page;
     for (page = 0; page  < next_free_page; page++) {
         if(is_code(page_table[page].type)) {
@@ -1117,7 +1109,7 @@ void darwin_jit_code_pages_kludge () {
             }
         }
     }
-    THREAD_JIT(1);
+    THREAD_JIT_WP(1);
 }
 #endif
 
@@ -1218,8 +1210,14 @@ void gc_load_corefile_ptes(int card_table_nbits,
         gengcbarrier_patch_code_range(TEXT_SPACE_START, text_space_highwatermark);
     }
 
-    // Apply physical page protection as needed.
-    // The non-soft-card-mark code is disgusting and I do not understand it.
+    // Toggle page protection bits as needed. There are essentially two major cases:
+    // - when soft card marking is used, all cards are "clean" but since that's
+    //   not the default state of the mark bit, we have to set it.
+    //   Additionally, with darwin-jit the mapping type on on pages of code
+    //   needs to change from 'rw-' to 'rwx'
+    // - when MMU-based marking is used, then all pages except code get
+    //   mprotected; but code pages writes are tracked with soft marking.
+
     if (gen != 0 && ENABLE_PAGE_PROTECTION) {
 #ifdef LISP_FEATURE_SOFT_CARD_MARKS
         page_index_t p;
@@ -1230,9 +1228,11 @@ void gc_load_corefile_ptes(int card_table_nbits,
         page_index_t start = 0, end;
         while (start  < next_free_page) {
             if (is_code(page_table[start].type)) {
+                set_page_executable(start, 1);
                 for (end = start + 1; end < next_free_page; end++) {
                     if (!page_words_used(end) || !is_code(page_table[end].type))
                         break;
+                    set_page_executable(end, 1);
                 }
                 os_protect(page_address(start), npage_bytes(end - start), OS_VM_PROT_ALL);
                 start = end+1;
@@ -1245,26 +1245,16 @@ void gc_load_corefile_ptes(int card_table_nbits,
 #else
         // coreparse can avoid hundreds to thousands of mprotect() calls by
         // treating the whole range from the corefile as protectable, except
-        // that soft-marked code pages must NOT be subject to mprotect.
-        // So just watch out for empty pages and code.  Unboxed object pages
-        // will get unprotected on demand.
+        // that code pages must NOT be subject to mprotect.
+        // So just watch out for empty pages and code.
+        // Unboxed pages do not technically need to be mprotected since we don't
+        // need to track writes, but they are lumped in with other pages for simplicity
+        // of this loop. Optimistically assume that doing so won't cause too many
+        // extra page faults if the unboxed pages do get written.
 #define non_protectable_page_p(x) !page_words_used(x) || is_code(page_table[x].type)
         page_index_t start = 0, end;
         // cf. write_protect_generation_pages()
         while (start  < next_free_page) {
-#ifdef LISP_FEATURE_DARWIN_JIT
-            if(is_code(page_table[start].type)) {
-              SET_PAGE_PROTECTED(start,1);
-                for (end = start + 1; end < next_free_page; end++) {
-                    if (!page_words_used(end) || !is_code(page_table[end].type))
-                        break;
-                    SET_PAGE_PROTECTED(end,1);
-                }
-                os_protect(page_address(start), npage_bytes(end - start), OS_VM_PROT_ALL);
-                start = end+1;
-                continue;
-            }
-#endif
             if (non_protectable_page_p(start)) {
                 ++start;
                 continue;
@@ -1275,7 +1265,7 @@ void gc_load_corefile_ptes(int card_table_nbits,
                     break;
                 SET_PAGE_PROTECTED(end,1);
             }
-            os_protect(page_address(start), npage_bytes(end - start), OS_VM_PROT_JIT_READ);
+            os_protect(page_address(start), npage_bytes(end - start), OS_VM_PROT_READ);
             start = end;
         }
 #endif
@@ -1347,6 +1337,10 @@ load_core_file(char *file, os_vm_offset_t file_offset, int merge_core_pages)
     struct coreparse_space defined_spaces[] = {
         {READ_ONLY_CORE_SPACE_ID, 0, 0, READ_ONLY_SPACE_START, &read_only_space_free_pointer},
         {STATIC_CORE_SPACE_ID, 0, 0, STATIC_SPACE_START, &static_space_free_pointer},
+#ifdef LISP_FEATURE_PERMGEN
+        {PERMGEN_CORE_SPACE_ID, PERMGEN_SPACE_SIZE | 1, 0,
+         PERMGEN_SPACE_START, &permgen_space_free_pointer},
+#endif
 #ifdef LISP_FEATURE_DARWIN_JIT
         {STATIC_CODE_CORE_SPACE_ID, 0, 0, STATIC_CODE_SPACE_START, &static_code_space_free_pointer},
 #endif
@@ -1424,6 +1418,7 @@ load_core_file(char *file, os_vm_offset_t file_offset, int merge_core_pages)
 }
 
 #include "genesis/hash-table.h"
+#include "genesis/split-ordered-list.h"
 #include "genesis/vector.h"
 #include "genesis/cons.h"
 char* get_asm_routine_by_name(const char* name, int *index)
@@ -1505,6 +1500,14 @@ static void graph_visit(lispobj referer, lispobj ptr, struct grvisit_context* co
             struct bitmap bitmap = get_layout_bitmap(LAYOUT(layout));
             for (i=0; i<(nwords-1); ++i)
                 if (bitmap_logbitp(i, bitmap)) RECURSE(obj[1+i]);
+            if (layout && finalizer_node_layout_p(LAYOUT(layout))) {
+                struct solist_node* node = (void*)obj;
+                // _node_next might have no lowtag, and so_key never does
+                if (node->_node_next && !lowtag_of(node->_node_next))
+                    RECURSE(node->_node_next | INSTANCE_POINTER_LOWTAG);
+                if (node->so_key)
+                    RECURSE(compute_lispobj((lispobj*)node->so_key));
+            }
             break;
         case CODE_HEADER_WIDETAG:
             nwords = code_header_words((struct code*)obj);
@@ -1640,7 +1643,9 @@ static uword_t visit_range(lispobj* where, lispobj* limit, uword_t arg)
         }
         lispobj ptr = compute_lispobj(obj);
         tally(ptr, v);
-        if (!hopscotch_get(v->reached, ptr, 0)) printf("unreachable: %p\n", (void*)ptr);
+        // Perhaps it's reachable via some path that this slighty-deficient
+        // tracer is unable to discover. e.g. installed Lisp signal handlers
+        if (!hopscotch_get(v->reached, ptr, 0)) printf("unreached: %p\n", (void*)ptr);
         obj = next_object(obj, object_size(obj), limit);
     }
     return 0;
