@@ -37,7 +37,9 @@
 #define PREFETCH_DISTANCE 32
 
 //#define LOG_COLLECTIONS
-#define COMPACT
+#ifndef ENABLE_COMPACTION
+#define ENABLE_COMPACTION 1
+#endif
 
 /* The idea of the mark-region collector is to avoid copying where
  * possible, and instead reclaim as much memory in-place as possible.
@@ -54,7 +56,7 @@ static struct {
   _Atomic(uword_t) consider, scavenge, prefix;
   _Atomic(uword_t) trace, trace_alive, trace_running;
   _Atomic(uword_t) sweep, weak, sweep_lines, sweep_pages;
-  _Atomic(uword_t) compact, copy, fix, raise;
+  _Atomic(uword_t) compact, copy, fix, compact_resweep, raise;
   uword_t fresh_pointers; uword_t pinned_pages;
   uword_t compacts;
 } meters = { 0 };
@@ -70,14 +72,14 @@ void mr_print_meters() {
           "collection %d (%.0f%% compacting):\n"
           "  %ldus consider\n"
           "  %ld scavenge (%ld prefixes) %ld trace (%ld alive %ld running)\n"
-          "  %ld sweep (%ld lines %ld pages) %ld compact (%ld copy %ld fix)\n"
+          "  %ld sweep (%ld lines %ld pages) %ld compact (%ld copy %ld fix %ld resweep)\n"
           "  %ld raise; %ldB fresh %ldpg pinned\n",
           collection,
           collection ? 100.0 *  (float)meters.compacts / collection : 0.0,
           NORM(consider), NORM(scavenge), NORM(prefix),
           NORM(trace), NORM(trace_alive), NORM(trace_running),
           NORM(sweep), NORM(sweep_lines), NORM(sweep_pages),
-          NORM(compact), NORM(copy), NORM(fix),
+          NORM(compact), NORM(copy), NORM(fix), NORM(compact_resweep),
           NORM(raise), NORM(fresh_pointers), NORM(pinned_pages));
 #undef NORM
 }
@@ -99,7 +101,7 @@ static void allocate_bitmap(uword_t **bitmap, uword_t size,
 /* Initialisation */
 uword_t *allocation_bitmap;
 _Atomic(uword_t) *mark_bitmap;
-unsigned char *line_bytemap;             /* 1 if line used, 0 if not used */
+unsigned char *line_bytemap;
 line_index_t line_count;
 uword_t mark_bitmap_size;
 
@@ -148,9 +150,43 @@ generation_index_t gc_gen_of(lispobj obj, int defaultval) {
       if (test) return where;                           \
     return fail; }
 
-DEF_FINDER(find_free_line, line_index_t, !line_bytemap[where], -1);
+static line_index_t find_free_line(line_index_t start, line_index_t end) {
+  /* memchr tends to have some vectorisation (e.g. GNU) or SWAR behind
+   * it (musl, FreeBSD 12), so we use that instead of a naive loop. */
+  unsigned char *where = memchr(line_bytemap + start, 0, end - start);
+  if (!where) return end;
+  return where - line_bytemap;
+}
 DEF_FINDER(find_used_line, line_index_t, line_bytemap[where], end);
 
+/* Try to find a page which could fit a new object. This should be
+ * be called before the caller locks and calls
+ * try_allocate_small_from_pages, to minimise the time spent locking. */
+void pre_search_for_small_space(sword_t nbytes, int page_type,
+                                struct allocator_state *state, page_index_t end) {
+  sword_t nlines = ALIGN_UP(nbytes, LINE_SIZE) / LINE_SIZE;
+  for (page_index_t page = state->page; page < end; page++) {
+    if (page_bytes_used(page) <= GENCGC_PAGE_BYTES - nbytes &&
+        !target_pages[page] &&
+        ((state->allow_free_pages && page_free_p(page)) ||
+         (page_table[page].type == page_type &&
+          page_table[page].gen != PSEUDO_STATIC_GENERATION))) {
+      line_index_t where = address_line(page_address(page));
+      line_index_t last_line = address_line(page_address(page + 1));
+      while (where < last_line) {
+        line_index_t chunk_start = find_free_line(where, last_line);
+        if (chunk_start == -1) break;
+        line_index_t chunk_end = find_used_line(chunk_start, last_line);
+        if (chunk_end - chunk_start >= nlines) {
+          state->page = page;
+          return;
+        }
+        where = chunk_end + 1;
+      }
+    }
+  }
+}
+
 /* Try to find space to fit a new object in the lines between `start`
  * and `end`. Updates `region` and returns true if we succeed, keeps
  * `region` untouched and returns false if we fail. The caller must
@@ -161,7 +197,7 @@ bool try_allocate_small(sword_t nbytes, struct alloc_region *region,
   line_index_t where = start;
   while (1) {
     line_index_t chunk_start = find_free_line(where, end);
-    if (chunk_start == -1) return 0;
+    if (chunk_start == -1) return false;
     line_index_t chunk_end = find_used_line(chunk_start, end);
     if (chunk_end - chunk_start >= nlines) {
       region->start_addr = line_address(chunk_start);
@@ -175,9 +211,9 @@ bool try_allocate_small(sword_t nbytes, struct alloc_region *region,
        * to the page and its state in the page table.. */
       for (line_index_t c = chunk_start; c < chunk_end; c++)
         line_bytemap[c] = gc_active_p ? 0 : FRESHEN_GEN(0);
-      return 1;
+      return true;
     }
-    if (chunk_end == end) return 0;
+    if (chunk_end == end) return false;
     where = chunk_end;
   }
 }
@@ -192,33 +228,33 @@ bool try_allocate_small_after_region(sword_t nbytes, struct alloc_region *region
   return try_allocate_small(nbytes, region, address_line(region->end_addr), end);
 }
 
-/* We try not to allocate small objects from free pages, to reduce
- * fragmentation. Something like "wilderness preservation". */
-bool allow_free_pages[8] = {0};
-void set_all_allow_free_pages(bool value) {
-  for (int i = 0; i < 8; i++) allow_free_pages[i] = value;
-}
+extern generation_index_t get_alloc_generation();
 
 /* try_allocate_small_from_pages updates the start pointer to after the
  * claimed page. */
 bool try_allocate_small_from_pages(sword_t nbytes, struct alloc_region *region,
                                    int page_type, generation_index_t gen,
-                                   page_index_t *start, page_index_t end) {
+                                   struct allocator_state *start, page_index_t end) {
   gc_assert(gen != SCRATCH_GENERATION);
  again:
-  for (page_index_t where = *start; where < end; where++) {
+  for (page_index_t where = start->page; where < end; where++) {
     if (page_bytes_used(where) <= GENCGC_PAGE_BYTES - nbytes &&
         !target_pages[where] &&
-        ((allow_free_pages[page_type & PAGE_TYPE_MASK] && page_free_p(where)) ||
+        ((start->allow_free_pages && page_free_p(where)) ||
          (page_table[where].type == page_type &&
           page_table[where].gen != PSEUDO_STATIC_GENERATION)) &&
         try_allocate_small(nbytes, region,
                            address_line(page_address(where)),
                            address_line(page_address(where + 1)))) {
-      page_table[where].type = page_type | OPEN_REGION_PAGE_FLAG;
+      // mark-region has a different way of zeroing, so just tell prepare_pages
+      // that the page is unboxed if it's boxed, so that it doesn't try to zero.
+      if (!page_table[where].type)
+          prepare_pages(1, where, where, page_type==PAGE_TYPE_CODE?page_type:0,
+                        get_alloc_generation());
+      set_page_type(page_table[where], page_type | OPEN_REGION_PAGE_FLAG);
       page_table[where].gen = 0;
       set_page_scan_start_offset(where, 0);
-      *start = where + 1;
+      start->page = where + 1;
       /* Update residency statistics. mr_update_closed_region will
        * enliven all lines on this page, so it's correct to set the
        * page bytes used like this. */
@@ -227,15 +263,14 @@ bool try_allocate_small_from_pages(sword_t nbytes, struct alloc_region *region,
       generations[gen].bytes_allocated += claimed;
       set_page_bytes_used(where, GENCGC_PAGE_BYTES);
       if (where + 1 > next_free_page) next_free_page = where + 1;
-      return 1;
+      return true;
     }
   }
-  if (!allow_free_pages[page_type & PAGE_TYPE_MASK]) {
-    allow_free_pages[page_type & PAGE_TYPE_MASK] = 1;
-    *start = 0;
+  if (!start->allow_free_pages) {;
+    *start = (struct allocator_state){0, true};
     goto again;
   }
-  return 0;
+  return false;
 }
 
 /* Large object allocation */
@@ -245,22 +280,30 @@ DEF_FINDER(find_used_page, page_index_t, !page_free_p(where), end);
 
 page_index_t try_allocate_large(uword_t nbytes,
                                 int page_type, generation_index_t gen,
-                                page_index_t *start, page_index_t end,
+                                struct allocator_state *start, page_index_t end,
                                 uword_t *largest_hole) {
   gc_assert(gen != SCRATCH_GENERATION);
   uword_t pages_needed = ALIGN_UP(nbytes, GENCGC_PAGE_BYTES) / GENCGC_PAGE_BYTES;
   uword_t remainder = nbytes % GENCGC_PAGE_BYTES;
-  page_index_t where = *start;
+  page_index_t where = start->page;
   uword_t largest_hole_seen = 0;
   while (1) {
     page_index_t chunk_start = find_free_page(where, end);
     if (chunk_start == -1) return -1;
+    /* TODO: this is suboptimal - the full extent of the free space is irrelevant
+     * as long as it's at _least_ pages_needed. So find_used_page is a poor choice
+     * of algorithm for this. It's not wrong, though I've seen it say (via added
+     * printing) that it needed 2 pages but found 40000 pages. So it scanned
+     * the entire page table before deciding yup, we have enough to work with..
+     */
     page_index_t chunk_end = find_used_page(chunk_start, end);
     uword_t hole_size = chunk_end - chunk_start;
     if (hole_size >= pages_needed) {
       page_index_t last_page = chunk_start + pages_needed - 1;
+      prepare_pages(1, chunk_start, last_page, page_type==PAGE_TYPE_CODE?page_type:0,
+                    get_alloc_generation());
       for (page_index_t p = chunk_start; p <= last_page; p++) {
-        page_table[p].type = SINGLE_OBJECT_FLAG | page_type;
+        set_page_type(page_table[p], SINGLE_OBJECT_FLAG | page_type);
         page_table[p].gen = gen;
         set_page_bytes_used(p,
                             (p == last_page && remainder > 0) ? remainder
@@ -268,7 +311,7 @@ page_index_t try_allocate_large(uword_t nbytes,
         set_page_scan_start_offset(p,
                                    GENCGC_PAGE_BYTES * (p - chunk_start));
       }
-      *start = chunk_start + pages_needed;
+      start->page = chunk_start + pages_needed;
       bytes_allocated += nbytes;
       generations[gen].bytes_allocated += nbytes;
       if (last_page + 1 > next_free_page) next_free_page = last_page + 1;
@@ -421,7 +464,7 @@ static void mark(lispobj object, lispobj *where, enum source source_type) {
       lispobj *base = (lispobj*)fun_code_header((struct simple_fun*)np);
       object = make_lispobj(base, OTHER_POINTER_LOWTAG);
     }
-#ifdef COMPACT
+#if ENABLE_COMPACTION
     if (where)
       log_slot(object, where, source_object, source_type);
 #endif
@@ -481,7 +524,7 @@ static void trace_object(lispobj object) {
           next = make_lispobj(base, OTHER_POINTER_LOWTAG);
           np = base;
         }
-#ifdef COMPACT
+#if ENABLE_COMPACTION
         /* Inlined logic from mark() */
         log_slot(c->cdr, &c->cdr, native_pointer(object), SOURCE_NORMAL);
 #endif
@@ -837,7 +880,8 @@ static void __attribute__((noinline)) sweep_pages() {
         allocation_bitmap[mark_bitmap_word_index(page_address(p))] = 0;
       /* Why is reset_page_flags(p) much slower here? It does other stuff
        * for gencgc, sure, but not that much more stuff. */
-      page_table[p].type = FREE_PAGE_FLAG;
+      set_page_need_to_zero(p, 1);
+      set_page_type(page_table[p], FREE_PAGE_FLAG);
       page_table[p].scan_start_offset_ = 0;
     } else {
       bytes_allocated += page_bytes_used(p);
@@ -887,6 +931,10 @@ static void trace_static_roots() {
   trace_other_object((lispobj*)NIL_SYMBOL_SLOTS_START);
   mr_trace_bump_range((lispobj*)STATIC_SPACE_OBJECTS_START,
                       static_space_free_pointer);
+  mr_trace_bump_range((lispobj*)PERMGEN_SPACE_START, permgen_space_free_pointer);
+
+  // TODO: use an explicit remembered set of modified objects in this range
+  if (TEXT_SPACE_START) mr_trace_bump_range((lispobj*)TEXT_SPACE_START, text_space_highwatermark);
 #ifdef LISP_FEATURE_SYSTEM_TLABS
   extern void gc_scavenge_arenas();
   gc_scavenge_arenas();
@@ -934,12 +982,6 @@ void mr_preserve_leaf(lispobj obj) {
   }
 }
 
-/* Trace through an object outside dynamic space.
- * Used by immobile space. */
-void mr_trace_object(lispobj *obj) {
-  trace_object(compute_lispobj(obj));
-}
-
 /* Preserve an exact pointer, trace and pin it.
  * Used by pin_exact_root. */
 void mr_preserve_object(lispobj obj) {
@@ -968,7 +1010,7 @@ static void watch_deferred(lispobj *where, uword_t start, uword_t end) {
   generation_index_t gen = dirty_generation_source;
   for (uword_t i = start; i < end; i++) {
     if (is_lisp_pointer(where[i])) {
-#ifdef COMPACT
+#if ENABLE_COMPACTION
       log_slot(where[i], where + i, where, SOURCE_NORMAL);
 #endif
       if (gc_gen_of(where[i], 0) < gen) {
@@ -999,14 +1041,11 @@ static void CPU_SPLIT scavenge_root_gens_worker(void) {
           case SIMPLE_VECTOR_WIDETAG:
           case WEAK_POINTER_WIDETAG: {
             /* Scavenge a page of a vector. */
-            if (page_starts_contiguous_block_p(i))
-              /* This page has the start of a large vector, and later pages
-               * will be part of this vector. */
-              source_object = (lispobj*)page_address(i);
-            /* The only time that page_address + page_words_used actually
-             * demarcates the end of a (sole) object on the page, with this
-             * heap layout. */
-            generation_index_t gen = page_table[i].gen;
+            source_object = (lispobj*)(page_address(i) - page_scan_start_offset(i));
+            dirty_generation_source = page_table[i].gen;
+            /* page_address(i) + page_words_used(i) only demarcates
+             * the end of a (sole) object on the page with this heap
+             * layout when the object is large. */
             lispobj *limit = (lispobj*)page_address(i) + page_words_used(i);
             lispobj *start = (lispobj*)page_address(i);
             for (int j = 0, card = addr_to_card_index(start);
@@ -1015,7 +1054,7 @@ static void CPU_SPLIT scavenge_root_gens_worker(void) {
               if (card_dirtyp(card)) {
                 lispobj *card_end = start + WORDS_PER_CARD;
                 lispobj *end = (limit < card_end) ? limit : card_end;
-                dirty_generation_source = gen, dirty = 0;
+                dirty = 0;
                 for (lispobj *p = start; p < end; p++)
                   mark(*p, p, SOURCE_NORMAL);
                 update_card_mark(card, dirty);
@@ -1133,14 +1172,6 @@ static void CPU_SPLIT raise_survivors(void) {
 /* Main entrypoints into GC */
 
 void mrgc_init() {
-#ifndef LISP_FEATURE_GCC_TLS
-  int ok = !pthread_key_create(&recycle_list_key, 0) &&
-    !pthread_key_create(&output_block_key, 0) &&
-    !pthread_key_create(&dirty_generation_source_key, 0) &&
-    !pthread_key_create(&dirty_key, 0) &&
-    !pthread_key_create(&source_object_key, 0);
-  gc_assert(ok);
-#endif
   allocate_bitmaps();
   thread_pool_init();
   compactor_init();
@@ -1155,27 +1186,28 @@ void mr_pre_gc(generation_index_t generation) {
 #endif
   generation_to_collect = generation;
   reset_statistics();
-#ifdef COMPACT
+#if ENABLE_COMPACTION
   if (generation != PSEUDO_STATIC_GENERATION)
     METER(consider, consider_compaction(generation_to_collect));
 #endif
 }
 
 void mr_collect_garbage(bool raise) {
+  extern void reset_alloc_start_pages(bool allow_free_pages);
   if (generation_to_collect != PSEUDO_STATIC_GENERATION) {
     METER(scavenge, mr_scavenge_root_gens());
   }
   trace_static_roots();
   METER(trace, trace_everything());
   METER(sweep, sweep());
-#ifdef COMPACT
+#if ENABLE_COMPACTION
   if (compacting) {
     meters.compacts++;
     /* This isn't a lot of work to wake up every thread for. Perhaps
      * we could snoop TLS of each GC thread instead. */
     run_on_thread_pool(commit_thread_local_remset);
-    set_all_allow_free_pages(1);
-    METER(compact, run_compaction(&meters.copy, &meters.fix));
+    reset_alloc_start_pages(true);
+    METER(compact, run_compaction(&meters.copy, &meters.fix, &meters.compact_resweep));
   }
 #endif
   /* scan_finalizers checks forwarding pointers, so we need to
@@ -1192,8 +1224,8 @@ void mr_collect_garbage(bool raise) {
           dirty_root_objects, root_objects_checked,
           next_free_page, raise ? ", raised" : "");
 #endif
-  if (gencgc_verbose) mr_print_meters();
-  set_all_allow_free_pages(0);
+  if (gencgc_verbose>1) mr_print_meters();
+  reset_alloc_start_pages(false);
   reset_pinned_pages();
 }
 
@@ -1252,26 +1284,6 @@ void prepare_lines_for_final_gc() {
   }
 }
 
-/* Core file I/O */
-
-void load_corefile_bitmaps(int fd, core_entry_elt_t n_ptes) {
-  sword_t size = bitmap_size(n_ptes);
-  lseek(fd, ALIGN_UP(lseek(fd, 0, SEEK_CUR), N_WORD_BYTES), SEEK_SET);
-  if (read(fd, allocation_bitmap, size) != size)
-    lose("failed to read allocation bitmap from core");
-
-  /* Mark pseudo-static lines as if everything is live. */
-  for (page_index_t p = 0; p <= page_table_pages; p++)
-    if (!page_free_p(p) && !page_single_obj_p(p)) {
-      set_page_bytes_used(p, GENCGC_PAGE_BYTES);
-      for_lines_in_page(l, p) line_bytemap[l] = ENCODE_GEN(PSEUDO_STATIC_GENERATION);
-    }
-  bytes_allocated = 0;
-  for (page_index_t p = 0; p <= page_table_pages; p++)
-    bytes_allocated += page_bytes_used(p);
-  generations[PSEUDO_STATIC_GENERATION].bytes_allocated = bytes_allocated;
-}
-
 /* Useful hacky stuff */
 
 void find_references_to(lispobj something) {
@@ -1290,11 +1302,11 @@ void find_references_to(lispobj something) {
 void draw_page_table(int from, int to) {
   fprintf(stderr, "\n       ");
   for (int i = 0; i < 50; i++)
-    fprintf(stderr, "%3d", i);
+    fprintf(stderr, "%3d ", i);
   for (int i = from; i < to; i++) {
     if (i % 50 == 0) fprintf(stderr, "\n%6d ", i);
     fprintf(stderr,
-            "\033[%c;9%cm%c%c%c\033[0m",
+            "\033[%c;9%cm%c%c%c\033[0m ",
             (page_table[i].type & SINGLE_OBJECT_FLAG) ? '1' : '0',
             '0' + (page_table[i].type & 7),
             64 + page_table[i].type,
