@@ -13,9 +13,10 @@
  * files for more information.
  */
 
+#include <stdatomic.h>
 #include <stdlib.h>
 #include <stdio.h>
-#include "sbcl.h"
+#include "genesis/sbcl.h"
 #include "lispregs.h"
 #include "gc.h"
 #include "code.h"
@@ -26,6 +27,7 @@
 #include "genesis/gc-tables.h"
 #include "genesis/split-ordered-list.h"
 #include "thread.h"
+#include "sys_mmap.inc"
 
 /* forward declarations */
 extern FILE *gc_activitylog();
@@ -101,12 +103,7 @@ page_scan_start(page_index_t page_index)
  * Particularly the 'need_zerofill' bit MUST remain as-is */
 void reset_page_flags(page_index_t page) {
     page_table[page].scan_start_offset_ = 0;
-#ifdef LISP_FEATURE_DARWIN_JIT
-    // Whenever a page was mapped as code, it potentially needs to be remapped on the next use.
-    // This avoids any affect of pthread_jit_write_protect_np when next used.
-    if (page_table[page].type == PAGE_TYPE_CODE) set_page_need_to_zero(page, 1);
-#endif
-    page_table[page].type = 0;
+    set_page_type(page_table[page], FREE_PAGE_FLAG);
     gc_page_pins[page] = 0;
 }
 
@@ -149,62 +146,9 @@ static pthread_mutex_t free_pages_lock = PTHREAD_MUTEX_INITIALIZER;
 void acquire_gc_page_table_lock() { ignore_value(mutex_acquire(&free_pages_lock)); }
 void release_gc_page_table_lock() { ignore_value(mutex_release(&free_pages_lock)); }
 
-extern os_vm_size_t gencgc_release_granularity;
-os_vm_size_t gencgc_release_granularity = GENCGC_RELEASE_GRANULARITY;
-
-
-/* Zero the memory at ADDR for LENGTH bytes, but use mmap/munmap instead
- * of zeroing it ourselves, i.e. in practice give the memory back to the
- * OS. Generally done after a large GC.
- */
-static void __attribute__((unused))
-zero_range_with_mmap(os_vm_address_t addr, os_vm_size_t length) {
-#ifdef LISP_FEATURE_WIN32
-    os_decommit_mem(addr, length);
-#elif defined LISP_FEATURE_LINUX
-    // We use MADV_DONTNEED only on Linux due to differing semantics from BSD.
-    // Linux treats it as a demand that the memory be 0-filled, or refreshed
-    // from a file that backs the range. BSD takes it as a hint that you don't
-    // care if the memory has to brought in from swap when next accessed,
-    // i.e. it's not a request to make a user-visible alteration to memory.
-    // So in theory this can bring a page in from the core file, if we happen
-    // to hit a page that resides in the portion of memory mapped by coreparse.
-    // In practice this should not happen because objects from a core file can't
-    // become garbage. Except in save-lisp-and-die they can, and we must be
-    // cautious not to resurrect bytes that originally came from the file.
-    if ((os_vm_address_t)addr >= anon_dynamic_space_start) {
-        if (madvise(addr, length, MADV_DONTNEED) != 0)
-            lose("madvise failed");
-    } else { // See doc/internals-notes/zero-with-mmap-bug.txt
-        // Trying to see how often this happens.
-        // fprintf(stderr, "zero_range_with_mmap: fallback to memset()\n");
-        memset(addr, 0, length);
-    }
-#else
-    void *new_addr;
-    os_deallocate(addr, length);
-    new_addr = os_alloc_gc_space(DYNAMIC_CORE_SPACE_ID, NOT_MOVABLE, addr, length);
-    if (new_addr == NULL || new_addr != addr) {
-        lose("remap_free_pages: page moved, %p ==> %p",
-             addr, new_addr);
-    }
-#endif
-}
-
-/* Zero the pages from START to END (inclusive). Generally done just after
- * a new region has been allocated.
- */
-static inline void zero_pages(page_index_t start, page_index_t end) {
-    if (start <= end)
-#ifdef LISP_FEATURE_DARWIN_JIT
-        zero_range_with_mmap(page_address(start), npage_bytes(1+end-start));
-#else
-        memset(page_address(start), 0, npage_bytes(1+end-start));
-#endif
-}
-
 /* The generation currently being allocated to. */
 static generation_index_t gc_alloc_generation;
+generation_index_t get_alloc_generation() { return gc_alloc_generation; }
 
 /*
  * To support quick and inline allocation, regions of memory can be
@@ -255,9 +199,8 @@ static generation_index_t gc_alloc_generation;
 /* We use five regions for the current newspace generation. */
 struct alloc_region gc_alloc_region[6];
 
-static page_index_t
-  alloc_start_pages[8], // one for each value of PAGE_TYPE_x
-  max_alloc_start_page; // the largest of any array element
+static _Atomic struct allocator_state alloc_start_states[8]; // one for each value of PAGE_TYPE_x
+static page_index_t max_alloc_start_page; // the largest of any array element
 page_index_t gencgc_alloc_start_page; // initializer for the preceding array
 
 /* Each 'start_page' informs the region-opening logic where it should
@@ -269,47 +212,42 @@ page_index_t gencgc_alloc_start_page; // initializer for the preceding array
  * It's kind of dumb that there is one start_page per type,
  * other than it serves its purpose for picking up where it left off
  * on a partially full page during GC */
-#define RESET_ALLOC_START_PAGES() \
-        alloc_start_pages[0] = gencgc_alloc_start_page; \
-        alloc_start_pages[1] = gencgc_alloc_start_page; \
-        alloc_start_pages[2] = gencgc_alloc_start_page; \
-        alloc_start_pages[3] = gencgc_alloc_start_page; \
-        alloc_start_pages[4] = gencgc_alloc_start_page; \
-        alloc_start_pages[5] = gencgc_alloc_start_page; \
-        alloc_start_pages[6] = gencgc_alloc_start_page; \
-        alloc_start_pages[7] = gencgc_alloc_start_page; \
-        max_alloc_start_page = gencgc_alloc_start_page;
+void reset_alloc_start_pages(bool allow_free_pages) {
+  for (int i = 0; i < 8; i++)
+    alloc_start_states[i] = (struct allocator_state){gencgc_alloc_start_page, allow_free_pages};
+  max_alloc_start_page = gencgc_alloc_start_page;
+}
 
-static page_index_t
+static struct allocator_state
 get_alloc_start_page(unsigned int page_type)
 {
     if (page_type > 7) lose("bad page_type: %d", page_type);
     struct thread* th = get_sb_vm_thread();
-    page_index_t global_start = alloc_start_pages[page_type];
+    struct allocator_state global_start = alloc_start_states[page_type];
     page_index_t hint;
     switch (page_type) {
     case PAGE_TYPE_MIXED:
-        if ((hint = thread_extra_data(th)->mixed_page_hint) > 0 && hint <= global_start) {
+        if ((hint = thread_extra_data(th)->mixed_page_hint) > 0 && hint <= global_start.page) {
             thread_extra_data(th)->mixed_page_hint = - 1;
-            return hint;
+            return (struct allocator_state){hint, global_start.allow_free_pages};
         }
         break;
     case PAGE_TYPE_CONS:
-        if ((hint = thread_extra_data(th)->cons_page_hint) > 0 && hint <= global_start) {
+        if ((hint = thread_extra_data(th)->cons_page_hint) > 0 && hint <= global_start.page) {
             thread_extra_data(th)->cons_page_hint = - 1;
-            return hint;
+            return (struct allocator_state){hint, global_start.allow_free_pages};
         }
         break;
     }
     return global_start;
 }
 
-static inline void
-set_alloc_start_page(unsigned int page_type, page_index_t page)
+static void
+set_alloc_start_page(unsigned int page_type, struct allocator_state state)
 {
     if (page_type > 7) lose("bad page_type: %d", page_type);
-    if (page > max_alloc_start_page) max_alloc_start_page = page;
-    alloc_start_pages[page_type] = page;
+    if (state.page > max_alloc_start_page) max_alloc_start_page = state.page;
+    alloc_start_states[page_type] = state;
 }
 #include "private-cons.inc"
 
@@ -323,7 +261,7 @@ void gc_close_region(struct alloc_region *alloc_region,
  * when it can't fit in the open region.
  * This entry point is only for use within the GC itself. */
 void *collector_alloc_fallback(struct alloc_region* region, sword_t nbytes, int page_type) {
-    page_index_t alloc_start = get_alloc_start_page(page_type);
+    struct allocator_state alloc_start = get_alloc_start_page(page_type);
     void *new_obj;
     if ((uword_t)nbytes >= (GENCGC_PAGE_BYTES / 4 * 3)) {
         uword_t largest_hole;
@@ -433,6 +371,63 @@ int lowtag_ok_for_page_type(__attribute__((unused)) lispobj ptr,
 
 int sb_introspect_pinnedp(__attribute__((unused)) lispobj obj) { return 0; }
 
+/* Additional logic for soft marks: any word that is potentially a
+ * tagged pointer to a page being written must preserve the mark regardless
+ * of what update_writeprotection() thinks. That's because the mark is set
+ * prior to storing. If GC occurs in between setting the mark and storing,
+ * then resetting the mark would be wrong if the subsequent store
+ * creates an old->young pointer.
+ * Mark stickiness is checked only once per invocation of collect_garbage(),
+ * when scanning interrupt contexts for generation 0 but not higher gens.
+ * There are two cases:
+ * (1) tagged pointer to a large simple-vector, but we scan card-by-card
+ *     for specifically the marked cards.  This has to be checked first
+ *     so as not to fail to see subsequent cards if the first is marked.
+ * (2) tagged pointer to an object that marks only the page containing
+ *     the object base.
+ * And note a subtle point: only an already-marked card can acquire sticky
+ * status. So we can ignore any unmarked (a/k/a WRITEPROTECTED_P) card
+ * regardless of a context register pointing to it, because if a mark was not
+ * stored, then the pointer was not stored. Without examining the next few
+ * instructions, there's no reason even to suppose that a store occurs.
+ * It seems like the stop-for-GC handler must be enforcing that GC sees things
+ * stored in the correct order for out-of-order memory models */
+static void impart_mark_stickiness(lispobj word)
+{
+    // This function does not care whether 'word' points to a valid object.
+    // At worst this will spurisouly mark a card as sticky,
+    // which can happen only if it was already marked as dirty.
+    page_index_t page = find_page_index((void*)word);
+    // Unlike for gencgc, do not check page_bytes_used or generation
+    // (maybe could/should for large object pages?)
+    if (page >= 0 && page_boxed_p(page) // stores to raw bytes are uninteresting
+        && lowtag_ok_for_page_type(word, page_table[page].type)
+        && plausible_tag_p(word)) { // "plausible" is good enough
+        /* if 'word' is the correctly-tagged pointer to the base of a SIMPLE-VECTOR,
+         * then set the sticky mark on every marked card. The only other large
+         * objects are CODE (writes to which are pseudo-atomic),
+         * and BIGNUM (which aren't on boxed pages)
+         * I'm not sure if it's inadvertent that this first 'if' is taken
+         * for non-large simple-vectors. It probably can't hurt,
+         * but I think it's not necessary */
+        if (lowtag_of(word) == OTHER_POINTER_LOWTAG &&
+            widetag_of(native_pointer(word)) == SIMPLE_VECTOR_WIDETAG) {
+            generation_index_t gen = page_table[page].gen;
+            while (1) {
+                long card = page_to_card_index(page);
+                int i;
+                for(i=0; i<CARDS_PER_PAGE; ++i)
+                    if (gc_card_mark[card+i]==CARD_MARKED) gc_card_mark[card+i]=STICKY_MARK;
+                if (page_ends_contiguous_block_p(page, gen)) return;
+                ++page;
+            }
+        } else if (gc_card_mark[addr_to_card_index((void*)word)] == CARD_MARKED) {
+            gc_card_mark[addr_to_card_index((void*)word)] = STICKY_MARK;
+        }
+    }
+}
+
+#if !GENCGC_IS_PRECISE || defined LISP_FEATURE_MIPS || defined LISP_FEATURE_PPC64
 static void preserve_pointer(os_context_register_t object,
                              __attribute__((unused)) void* arg) {
     /* The mark-region GC never filters based on type tags,
@@ -467,33 +462,10 @@ static void preserve_pointer(os_context_register_t object,
 static void sticky_preserve_pointer(os_context_register_t register_word, void* arg)
 {
     uword_t word = register_word;
-    if (is_lisp_pointer(word)) {
-        page_index_t page = find_page_index((void*)word);
-        if (page >= 0 && page_boxed_p(page) // stores to raw bytes are uninteresting
-            && lowtag_ok_for_page_type(word, page_table[page].type)
-            && plausible_tag_p(word)) { // "plausible" is good enough
-            if (lowtag_of(word) == OTHER_POINTER_LOWTAG &&
-                widetag_of(native_pointer(word)) == SIMPLE_VECTOR_WIDETAG) {
-                /* if 'word' is the correctly-tagged pointer to the base of a SIMPLE-VECTOR,
-                 * then set the sticky mark on every marked page. The only other large
-                 * objects are CODE (writes to which are pseudo-atomic),
-                 * and BIGNUM (which aren't on boxed pages) */
-                generation_index_t gen = page_table[page].gen;
-                while (1) {
-                    long card = page_to_card_index(page);
-                    int i;
-                    for(i=0; i<CARDS_PER_PAGE; ++i)
-                        if (gc_card_mark[card+i]==CARD_MARKED) gc_card_mark[card+i]=STICKY_MARK;
-                    if (page_ends_contiguous_block_p(page, gen)) break;
-                    ++page;
-                }
-            } else if (gc_card_mark[addr_to_card_index((void*)word)] == CARD_MARKED) {
-                gc_card_mark[addr_to_card_index((void*)word)] = STICKY_MARK;
-            }
-        }
-    }
+    if (is_lisp_pointer(word)) impart_mark_stickiness(word);
     preserve_pointer(word, arg);
 }
+#endif
 
 #define pin_exact_root(r) mr_preserve_object(r)
 
@@ -576,6 +548,58 @@ walk_generation(uword_t (*proc)(lispobj*,lispobj*,uword_t),
     return 0;
 }
 
+static void __attribute__((unused)) maybe_pin_code(lispobj addr) {
+    page_index_t page = find_page_index((char*)addr);
+
+    if (page < 0) {
+        if (immobile_space_p(addr))
+            immobile_space_preserve_pointer((void*)addr);
+        return;
+    }
+    struct code* code = (struct code*)dynamic_space_code_from_pc((char *)addr);
+    if (code) pin_exact_root(make_lispobj(code, OTHER_POINTER_LOWTAG));
+}
+
+#if GENCGC_IS_PRECISE && !defined(reg_CODE)
+
+static int boxed_registers[] = BOXED_REGISTERS;
+
+/* Pin all (condemned) code objects pointed to by the chain of in-flight calls
+ * based on scanning from the innermost frame pointer. This relies on an exact backtrace,
+ * which some of our architectures have trouble obtaining. But it's theoretically
+ * more efficient to do it this way versus looking at all stack words to see
+ * whether each points to a code object. */
+static void pin_call_chain_and_boxed_registers(struct thread* th) {
+    lispobj *cfp = access_control_frame_pointer(th);
+
+    if (cfp) {
+      while (1) {
+        lispobj* ocfp = (lispobj *) cfp[0];
+        lispobj lr = cfp[1];
+        if (ocfp == 0)
+            break;
+        maybe_pin_code(lr);
+        cfp = ocfp;
+      }
+    }
+    int i = fixnum_value(read_TLS(FREE_INTERRUPT_CONTEXT_INDEX,th));
+    for (i = i - 1; i >= 0; --i) {
+        os_context_t* context = nth_interrupt_context(i, th);
+        maybe_pin_code((lispobj)*os_context_register_addr(context, reg_LR));
+
+        for (unsigned i = 0; i < (sizeof(boxed_registers) / sizeof(int)); i++) {
+            lispobj word = *os_context_register_addr(context, boxed_registers[i]);
+            if (is_lisp_pointer(word)) {
+                impart_mark_stickiness(word);
+                pin_exact_root(word);
+            }
+        }
+    }
+
+}
+#endif
+
+#if !GENCGC_IS_PRECISE
 extern void visit_context_registers(void (*proc)(os_context_register_t, void*),
                                     os_context_t *context, void*);
 static void NO_SANITIZE_ADDRESS NO_SANITIZE_MEMORY
@@ -673,6 +697,7 @@ conservative_stack_scan(struct thread* th,
         }
     }
 }
+#endif
 
 static void scan_explicit_pins(__attribute__((unused)) struct thread* th)
 {
@@ -713,6 +738,17 @@ __attribute__((unused)) static void check_contiguity()
       }
 }
 
+extern void verify_hash_tables(bool);
+/* This is a 3-state boolean. Set it to 1 to generally enable checking.
+ * It will get set to 2 on a particular GC cycle to actually do a one-shot check
+ * and then it resets to 1. I don't want it always on, because then we'd possibly
+ * introduce crashing other than in the test that tests for crashing.
+ * (i.e. it gets lucky enough to work most of the time because it allows
+ * silent breakage of hash-tables that never see another GETHASH operation)
+ * Also I think REMHASH can produce false positives in the checker.
+ * So don't enable unless you're looking for trouble */
+int check_hash_tables;
+
 #define PAGE_PINNED 0xFF
 
 /* Garbage collect a generation. If raise is 0 then the remains of the
@@ -770,7 +806,7 @@ garbage_collect_generation(generation_index_t generation, int raise,
 
     /* Change to a new space for allocation, resetting the alloc_start_page */
         gc_alloc_generation = new_space;
-        RESET_ALLOC_START_PAGES();
+        reset_alloc_start_pages(false);
 
         /* Don't try to allocate into pseudo-static, when we collect it */
         if (generation == PSEUDO_STATIC_GENERATION)
@@ -831,7 +867,7 @@ garbage_collect_generation(generation_index_t generation, int raise,
     // preserves the SB-THREAD:THREAD instance and its startup function,
     // neither of which will move.
 
-#if 0 // was: ifdef STARTING_THREADS
+#ifdef STARTING_THREADS
     lispobj pin_list = SYMBOL(STARTING_THREADS)->value;
     for ( ; pin_list != NIL ; pin_list = CONS(pin_list)->cdr ) {
         lispobj thing = CONS(pin_list)->car;
@@ -931,7 +967,7 @@ garbage_collect_generation(generation_index_t generation, int raise,
     }
 
     mr_collect_garbage(raise);
-    RESET_ALLOC_START_PAGES();
+    reset_alloc_start_pages(false);
     struct generation* g = &generations[generation];
     /* Set the new gc trigger for the GCed generation. */
     g->gc_trigger = g->bytes_allocated + g->bytes_consed_between_gc;
@@ -943,6 +979,11 @@ garbage_collect_generation(generation_index_t generation, int raise,
     if (generation >= verify_gens)
         hexdump_and_verify_heap(cur_thread_approx_stackptr,
                                 VERIFY_POST_GC | (generation<<1) | raise);
+    if (check_hash_tables == 2) {
+        fprintf(stderr, "Verifying hashtables\n");
+        verify_hash_tables(0);
+        check_hash_tables = 1;
+    }
 
     extern int n_unboxed_instances;
     n_unboxed_instances = 0;
@@ -962,75 +1003,6 @@ find_next_free_page(void)
     return last_page + 1;
 }
 
-/*
- * Supposing the OS can only operate on ranges of a certain granularity
- * (which we call 'gencgc_release_granularity'), then given any page rage,
- * align the lower bound up and the upper down to match the granularity.
- *
- *     |-->| OS page | OS page |<--|
- *
- * If the interior of the aligned range is nonempty,
- * perform three operations: unmap/remap, fill before, fill after.
- * Otherwise, just one operation to fill the whole range.
- */
-static void
-remap_page_range (page_index_t from, page_index_t to)
-{
-    /* There's a mysterious Solaris/x86 problem with using mmap
-     * tricks for memory zeroing. See sbcl-devel thread
-     * "Re: patch: standalone executable redux".
-     */
-    /* I have no idea what the issue with Haiku is, but using the simpler
-     * zero_pages() works where the unmap,map technique does not. Yet the
-     * trick plus a post-check that the pages were correctly zeroed finds
-     * no problem at that time. So what's failing later and why??? */
-#if defined LISP_FEATURE_SUNOS || defined LISP_FEATURE_HAIKU
-    zero_pages(from, to);
-#else
-    size_t granularity = gencgc_release_granularity;
-    // page_address "works" even if 'to' == page_table_pages-1
-    char* start = page_address(from);
-    char* end   = page_address(to+1);
-    char* aligned_start = PTR_ALIGN_UP(start, granularity);
-    char* aligned_end   = PTR_ALIGN_DOWN(end, granularity);
-
-    /* NOTE: this is largely pointless because gencgc-release-granularity
-     * is everywhere defined to be EXACTLY +backend-page-bytes+
-     * which by definition is the quantum at which we'll unmap/map.
-     * Maybe we should remove the needless complexity? */
-    if (aligned_start < aligned_end) {
-        zero_range_with_mmap(aligned_start, aligned_end-aligned_start);
-        memset(start, 0, aligned_start - start);
-        memset(aligned_end, 0, end - aligned_end);
-    } else {
-        zero_pages(from, to);
-    }
-#endif
-    page_index_t i;
-    for (i = from; i <= to; i++) set_page_need_to_zero(i, 0);
-}
-
-static void
-remap_free_pages (page_index_t from, page_index_t to)
-{
-    page_index_t first_page, last_page;
-
-    for (first_page = from; first_page <= to; first_page++) {
-        if (!page_free_p(first_page) || !page_need_to_zero(first_page))
-            continue;
-
-        last_page = first_page + 1;
-        while (page_free_p(last_page) &&
-               (last_page <= to) &&
-               (page_need_to_zero(last_page)))
-            last_page++;
-
-        remap_page_range(first_page, last_page-1);
-
-        first_page = last_page;
-    }
-}
-
 generation_index_t small_generation_limit = 1;
 
 extern int finalizer_thread_runflag;
@@ -1048,7 +1020,7 @@ void NO_SANITIZE_ADDRESS NO_SANITIZE_MEMORY
 collect_garbage(generation_index_t last_gen)
 {
     ++n_gcs;
-    THREAD_JIT(0);
+    THREAD_JIT_WP(0);
     generation_index_t gen = 0;
     bool gc_mark_only = 0;
     int raise, more = 0;
@@ -1104,7 +1076,7 @@ collect_garbage(generation_index_t last_gen)
         (lispobj*)ALIGN_DOWN((uword_t)&last_gen, N_WORD_BYTES);
     /* Verify the new objects created by Lisp code. */
     if (pre_verify_gen_0)
-        verify_heap(cur_thread_approx_stackptr, VERIFY_PRE_GC);
+        hexdump_and_verify_heap(cur_thread_approx_stackptr, VERIFY_PRE_GC);
 
     if (gencgc_verbose > 1) {
         fprintf(stderr, "Pre-GC:\n");
@@ -1269,7 +1241,7 @@ collect_garbage(generation_index_t last_gen)
         // as that causes the thread to exit.
         finalizer_thread_runflag = newval ? newval : 1;
     }
-    THREAD_JIT(1);
+    THREAD_JIT_WP(1);
     // Clear all pin bits for the next GC cycle.
     // This could be done in the background somehow maybe.
     page_index_t max_nfp = initial_nfp > next_free_page ? initial_nfp : next_free_page;
@@ -1310,6 +1282,14 @@ int small_allocation_count = 0;
 
 int gencgc_alloc_profiler;
 
+#ifdef LISP_FEATURE_DARWIN_JIT
+#define gc_memclear(pt, addr, len) \
+    do { if (gc_active_p || pt != PAGE_TYPE_CODE) memset(addr, 0, len); \
+    else { THREAD_JIT_WP(0); memset(addr, 0, len); THREAD_JIT_WP(1); } } while (0)
+#else
+#define gc_memclear(pt, addr, len) memset(addr, 0, len)
+#endif
+
 NO_SANITIZE_MEMORY lispobj*
 lisp_alloc(__attribute__((unused)) int flags,
            struct alloc_region *region, sword_t nbytes,
@@ -1343,7 +1323,7 @@ lisp_alloc(__attribute__((unused)) int flags,
     }
 
     if (try_allocate_small_after_region(nbytes, region)) {
-      memset(region->start_addr, 0, addr_diff(region->end_addr, region->start_addr));
+      gc_memclear(page_type, region->start_addr, addr_diff(region->end_addr, region->start_addr));
       return region->start_addr;
     }
 
@@ -1399,11 +1379,11 @@ lisp_alloc(__attribute__((unused)) int flags,
 #endif
 
     ensure_region_closed(region, page_type);
-    int __attribute__((unused)) ret = mutex_acquire(&free_pages_lock);
-    gc_assert(ret);
-    page_index_t alloc_start = get_alloc_start_page(page_type);
+    struct allocator_state alloc_start = get_alloc_start_page(page_type);
     bool largep = nbytes >= LARGE_OBJECT_SIZE && page_type != PAGE_TYPE_CONS;
     if (largep) {
+        int __attribute__((unused)) ret = mutex_acquire(&free_pages_lock);
+        gc_assert(ret);
         uword_t largest_hole;
         page_index_t new_page = try_allocate_large(nbytes, page_type, gc_alloc_generation,
                                                    &alloc_start, page_table_pages, &largest_hole);
@@ -1413,9 +1393,15 @@ lisp_alloc(__attribute__((unused)) int flags,
         gc_assert(ret);
         new_obj = page_address(new_page);
         set_allocation_bit_mark(new_obj);
-        memset(new_obj, 0, nbytes);
+        gc_memclear(page_type, new_obj, nbytes);
     } else {
+        /* Try to find a page before acquiring free_pages_lock. */
+        pre_search_for_small_space(nbytes, page_type, &alloc_start, page_table_pages);
+        int __attribute__((unused)) ret = mutex_acquire(&free_pages_lock);
+        gc_assert(ret);
         if (!gc_active_p) small_allocation_count++;
+        /* This search will only re-visit the page found by pre_search_for_small_space
+         * if no one else claimed the page since acquiring free_pages_lock. */
         bool success =
             try_allocate_small_from_pages(nbytes, region, page_type,
                                           gc_alloc_generation,
@@ -1425,7 +1411,7 @@ lisp_alloc(__attribute__((unused)) int flags,
         ret = mutex_release(&free_pages_lock);
         gc_assert(ret);
         new_obj = region->start_addr;
-        memset(new_obj, 0, addr_diff(region->end_addr, new_obj));
+        gc_memclear(page_type, new_obj, addr_diff(region->end_addr, new_obj));
     }
 
     return new_obj;
@@ -1558,8 +1544,8 @@ static void note_failure(lispobj thing, lispobj *where, struct verify_state *sta
     if (state->object_addr) {
         lispobj obj = compute_lispobj(state->object_addr);
         page_index_t pg = find_page_index(state->object_addr);
-        fprintf(stderr, "Ptr %p @ %"OBJ_FMTX" (lispobj %"OBJ_FMTX",pg%d) sees %s\n",
-                (void*)thing, (uword_t)where, obj, (int)pg, str);
+        fprintf(stderr, "Ptr %p @ %"OBJ_FMTX" (lispobj %"OBJ_FMTX",pg%d,h=%"OBJ_FMTX" sees %s\n",
+                (void*)thing, (uword_t)where, obj, (int)pg, *native_pointer(obj), str);
         // Record this in state->err_objs if possible
         int i;
         for(i=0; i<MAX_ERR_OBJS; ++i)
@@ -1698,11 +1684,6 @@ static int verify_headered_object(lispobj* object, sword_t nwords,
         if (layout) {
             CHECK(layout, object);
             struct bitmap bitmap = get_layout_bitmap(LAYOUT(layout));
-            if (widetag_of(object) == FUNCALLABLE_INSTANCE_WIDETAG) {
-#ifndef LISP_FEATURE_COMPACT_INSTANCE_HEADER
-                gc_assert(bitmap.bits[0] == (sword_t)-4);
-#endif
-            }
             if (lockfree_list_node_layout_p(LAYOUT(layout))) {
                 struct list_node* node = (void*)object;
                 lispobj next = node->_node_next;
@@ -1829,22 +1810,23 @@ static int verify_range(lispobj* start, lispobj* end, struct verify_state* state
             }
 #endif
             if (widetag != FILLER_WIDETAG && pg >= 0) {
+                    int pt = page_table[pg].type;
                     // Assert proper page type
                     if (state->object_header) // is not a cons
-                        gc_assert(page_table[pg].type != PAGE_TYPE_CONS);
+                        gc_assert(pt != PAGE_TYPE_CONS);
 #ifdef LISP_FEATURE_USE_CONS_REGION
-                    else if (page_table[pg].type != PAGE_TYPE_CONS) {
+                    else if (pt != PAGE_TYPE_CONS) {
                       if (is_cons_half(where[0]))
                           gc_assert(acceptable_filler_cons_p(where));
                     }
 #endif
                     if (widetag == CODE_HEADER_WIDETAG) {
-                        if (!is_code(page_table[pg].type))
+                        if (!is_code(pt))
                             lose("object @ %p is code on non-code page", where);
                     } else if (widetag == FUNCALLABLE_INSTANCE_WIDETAG) {
                         // where these reside depends on the architecture
                     } else {
-                        if (is_code(page_table[pg].type))
+                        if (is_code(pt))
                             lose("object @ %p is non-code on code page", where);
                     }
             }
@@ -2067,17 +2049,17 @@ void gc_gen_report_to_file(int filedes, FILE *file) {
   uword_t large_generation_type[PSEUDO_STATIC_GENERATION + 1][8] = { 0 };
   uword_t large_generation_total[8] = { 0 };
   for (page_index_t p = 0; p < page_table_pages; p++) {
-    unsigned char type = page_table[p].type & PAGE_TYPE_MASK;
+    unsigned char pt = page_table[p].type & PAGE_TYPE_MASK;
     if (page_free_p(p)) continue;
     else if (page_single_obj_p(p)) {
       generation_index_t gen = page_table[p].gen;
-      large_generation_type[gen][type] += GENCGC_PAGE_BYTES;
+      large_generation_type[gen][pt] += GENCGC_PAGE_BYTES;
       large_generation_total[gen] += GENCGC_PAGE_BYTES;
     } else {
       for_lines_in_page (l, p) {
         if (line_bytemap[l]) {
           generation_index_t gen = DECODE_GEN(line_bytemap[l]);
-          small_generation_type[gen][type] += LINE_SIZE;
+          small_generation_type[gen][pt] += LINE_SIZE;
           small_generation_total[gen] += LINE_SIZE;
         }
       }
@@ -2091,7 +2073,7 @@ void gc_gen_report_to_file(int filedes, FILE *file) {
   FORMAT("  %4s %6s %6s %6s %6s %6s %6s\n",
          "Gen#", "Raw", "Boxed", "Mixed", "Cons", "Code", "Total");
   OUTPUT("Small objects (in MiB):\n");
-  for (generation_index_t g = 0; g <= HIGHEST_NORMAL_GENERATION; g++)
+  for (generation_index_t g = 0; g <= PSEUDO_STATIC_GENERATION; g++)
     if (small_generation_total[g]) {
       uword_t *gen = small_generation_type[g];
       FORMAT("  %4d %6d %6d %6d %6d %6d %6d\n",
