@@ -11,16 +11,10 @@
  * files for more information.
  */
 
-/*
- * FIXME:
- *   Some of the code in here is deeply broken, depending on guessing
- *   already out-of-date values instead of getting them from sbcl.h.
- */
-
 #include <stdio.h>
 #include <string.h>
 
-#include "sbcl.h"
+#include "genesis/sbcl.h"
 #include "print.h"
 #include "runtime.h"
 #include "code.h"
@@ -359,6 +353,8 @@ static void brief_struct(lispobj obj)
 #include "genesis/defstruct-description.h"
 static bool tagged_slot_p(struct layout *layout, int slot_index)
 {
+    // Since we're doing this scan, we could return the name
+    // and exact raw type.
     if (instancep(layout->_info)) {
         struct defstruct_description* dd = (void*)(layout->_info-INSTANCE_POINTER_LOWTAG);
         lispobj slots = dd->slots;
@@ -369,7 +365,15 @@ static bool tagged_slot_p(struct layout *layout, int slot_index)
                 return (fixnum_value(dsd->bits) & DSD_RAW_TYPE_MASK) == 0;
         }
     }
-    return 0;
+    /* Revision 2b783b49 said to prefer LAYOUT-INFO vs BITMAP because the bitmap
+     * can indicate a 0 bit ("raw") for any slot that _may_ be ignored by GC, such as
+     * slots constrained to FIXNUM. Unfortunately that misses that CONDITION instances
+     * have trailing variable-length tagged data. In practice an instance may have raw
+     * words only if it has a DD, which most CONDITION subtypes do not. Therefore this
+     * could almost always return 1. But layout-of-layout is an important use of trailing
+     * raw slots. Attempting to print random words as tagged could be disastrous.
+     * Therefore, test the bitmap if the above loop failed to find slot_index. */
+    return bitmap_logbitp(slot_index, get_layout_bitmap(layout));
 }
 
 static void print_struct(lispobj obj)
@@ -487,26 +491,24 @@ static void print_slots(char **slots, int count, lispobj *ptr)
 
 lispobj symbol_function(struct symbol* symbol)
 {
-    if (symbol->fdefn) return FDEFN(symbol->fdefn)->fun;
+    if (symbol->fdefn && symbol->fdefn != NIL) return FDEFN(symbol->fdefn)->fun;
     return NIL;
 }
 
 static void print_fun_or_otherptr(lispobj obj)
 {
-    lispobj *ptr;
-    unsigned long header;
-    int count, type, index;
+    int index;
     char buffer[16];
 
-    ptr = native_pointer(obj);
+    lispobj *ptr = native_pointer(obj);
     if (ptr == NULL) {
         printf(" (NULL Pointer)");
         return;
     }
 
-    header = *ptr++;
-    count = HeaderValue(header);
-    type = header_widetag(header);
+    int count = object_size(ptr)-1;
+    uword_t header = *ptr++;
+    int type = header_widetag(header);
 
     print_obj("header: ", header);
     if (!other_immediate_lowtag_p(header)) {
@@ -535,7 +537,7 @@ static void print_fun_or_otherptr(lispobj obj)
         print_slots(ratio_slots, count, ptr);
         break;
 
-    case COMPLEX_WIDETAG:
+    case COMPLEX_RATIONAL_WIDETAG:
         print_slots(complex_slots, count, ptr);
         break;
 
@@ -556,7 +558,7 @@ static void print_fun_or_otherptr(lispobj obj)
 #ifdef LISP_FEATURE_COMPACT_SYMBOL
         // print_obj doesn't understand raw words, so make it a fixnum
         int pkgid = symbol_package_id(sym) << N_FIXNUM_TAG_BITS;
-        print_obj("package_id: ", pkgid);
+        print_obj("pkgid: ", pkgid);
 #endif
         break;
 
