@@ -8,12 +8,13 @@
  * provided with absolutely no warranty. See the COPYING and CREDITS
  * files for more information.
  */
-#include "sbcl.h"
+#include "genesis/sbcl.h"
 
 #ifdef LISP_FEATURE_SB_SAFEPOINT /* entire file */
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
+#include <stdarg.h>
 #ifndef LISP_FEATURE_WIN32
 #include <sched.h>
 #endif
@@ -35,11 +36,11 @@
 #include "genesis/cons.h"
 #include "genesis/fdefn.h"
 #include "interr.h"
-#include "alloc.h"
-#include "gc-internal.h"
-#include "getallocptr.h"
+#include "genesis/sap.h"
+#include "gc.h"
 #include "interrupt.h"
 #include "lispregs.h"
+#include "print.h"
 
 const char* gc_phase_names[GC_NPHASES] = {
     "GC_NONE",
@@ -98,7 +99,6 @@ const char* gc_phase_names[GC_NPHASES] = {
  * time will move to GC_INVOKED and then wait for the WITHOUT-GCING
  * thread to finish up, then proceed to GC_COLLECT. */
 
-#ifdef LISP_FEATURE_SB_THREAD
 #define CURRENT_THREAD_VAR(name) \
     struct thread *name = get_sb_vm_thread()
 #define THREAD_STOP_PENDING(th) \
@@ -106,14 +106,8 @@ const char* gc_phase_names[GC_NPHASES] = {
 #define SET_THREAD_STOP_PENDING(th,state) \
     write_TLS(STOP_FOR_GC_PENDING,state,th)
 #define WITH_ALL_THREADS_LOCK \
-    mutex_acquire(&all_threads_lock); \
-    RUN_BODY_ONCE(all_threads_lock, mutex_release(&all_threads_lock))
-#else
-#define CURRENT_THREAD_VAR(name)
-#define THREAD_STOP_PENDING(th) NIL
-#define SET_THREAD_STOP_PENDING(th,state)
-#define WITH_ALL_THREADS_LOCK
-#endif
+    ignore_value(mutex_acquire(&all_threads_lock)); \
+    RUN_BODY_ONCE(all_threads_lock, ignore_value(mutex_release(&all_threads_lock)))
 
 #if !defined(LISP_FEATURE_WIN32)
 /* win32-os.c covers these, but there is no unixlike-os.c, so the normal
@@ -176,8 +170,9 @@ void safepoint_init()
     for (i=GC_NONE; i<GC_NPHASES; ++i)
         InitializeConditionVariable(&gc_state.phase_cond[i]);
     InitializeCriticalSection(&gc_state.lock);
-#else
-    os_validate(NOT_MOVABLE, GC_SAFEPOINT_PAGE_ADDR, BACKEND_PAGE_BYTES, 0, 0);
+#elif !defined LISP_FEATURE_X86_64 && !defined LISP_FEATURE_RELOCATABLE_STATIC_SPACE
+    // 64-bit already allocated a trap page when the GC card mark table was made
+    os_alloc_gc_space(0, NOT_MOVABLE, GC_SAFEPOINT_PAGE_ADDR, BACKEND_PAGE_BYTES);
 #endif
     gc_state.phase = GC_NONE;
 }
@@ -186,10 +181,8 @@ void
 gc_state_lock()
 {
     odxprint(safepoints,"GC state to be locked");
-#ifdef LISP_FEATURE_SB_THREAD
-    int result = mutex_acquire(&gc_state.lock);
+    int __attribute__((unused)) result = mutex_acquire(&gc_state.lock);
     gc_assert(result);
-#endif
     if (gc_state.master) {
         fprintf(stderr,"GC state lock glitch [%p] in thread %p phase %d (%s)\n",
                 gc_state.master,get_sb_vm_thread(),gc_state.phase,
@@ -209,10 +202,8 @@ gc_state_unlock()
              gc_state.phase, gc_phase_names[gc_state.phase]);
     gc_assert(get_sb_vm_thread()==gc_state.master);
     gc_state.master = NULL;
-#ifdef LISP_FEATURE_SB_THREAD
-    int result = mutex_release(&gc_state.lock);
+    int __attribute__((unused)) result = mutex_release(&gc_state.lock);
     gc_assert(result);
-#endif
     odxprint(safepoints,"%s","GC state unlocked");
 }
 
@@ -227,13 +218,7 @@ gc_state_wait(gc_phase_t phase)
     gc_assert(gc_state.master == self);
     gc_state.master = NULL;
     while(gc_state.phase != phase && !(phase == GC_QUIET && (gc_state.phase > GC_QUIET))) {
-#ifdef LISP_FEATURE_WIN32
-        SleepConditionVariableCS(&gc_state.phase_cond[phase], &gc_state.lock, INFINITE);
-#elif defined LISP_FEATURE_SB_THREAD
-        pthread_cond_wait(&gc_state.phase_cond[phase],&gc_state.lock);
-#else
-        lose("gc_state_wait() blocks, but we're #-SB-THREAD");
-#endif
+        CONDITION_VAR_WAIT(&gc_state.phase_cond[phase], &gc_state.lock);
     }
     gc_assert(gc_state.master == NULL);
     gc_state.master = self;
@@ -282,10 +267,10 @@ static inline gc_phase_t gc_phase_next(gc_phase_t old) {
     return (old+1) % GC_NPHASES;
 }
 
-static inline boolean
+static inline bool
 thread_blocks_gc(struct thread *thread)
 {
-    return read_TLS(GC_INHIBIT,thread)==T;
+    return read_TLS(GC_INHIBIT,thread)==LISP_T;
 }
 /* set_thread_csp_access -- alter page permissions for not-in-Lisp
    flag (Lisp Stack Top) of the thread `th'. The flag may be modified
@@ -304,8 +289,8 @@ thread_blocks_gc(struct thread *thread)
    There is a full VM page reserved for this word; page permissions
    are switched to read-only for race-free examine + wait + use
    scenarios. */
-static inline boolean
-set_thread_csp_access(struct thread* th, boolean writable)
+static inline bool
+set_thread_csp_access(struct thread* th, bool writable)
 {
     os_protect((char*)th - (THREAD_HEADER_SLOTS*N_WORD_BYTES) - THREAD_CSP_PAGE_SIZE,
                THREAD_CSP_PAGE_SIZE,
@@ -335,20 +320,20 @@ static inline void gc_notify_early()
             if (p==gc_state.collector)
                 continue;
             odxprint(safepoints,"notifying thread %p csp %p",p,csp_around_foreign_call(p));
-            boolean was_in_lisp = !set_thread_csp_access(p,0);
+            bool was_in_lisp = !set_thread_csp_access(p,0);
             if (was_in_lisp) {
                 /* Threads "in-lisp" block leaving GC_MESSAGE, as we
                  * need them to hit their CSP or the GSP, and we unmap
                  * the GSP when transitioning to GC_INVOKED. */
                 gc_state.phase_wait[GC_MESSAGE]++;
-                SET_THREAD_STOP_PENDING(p,T);
+                SET_THREAD_STOP_PENDING(p, LISP_T);
             } else if (thread_blocks_gc(p)) {
                 /* Threads "in-alien" don't block leaving GC_MESSAGE,
                  * as the CSP trap is sufficient to catch them, but
                  * any thread that is WITHOUT-GCING prevents exit from
                  * GC_INVOKED. */
                 gc_state.phase_wait[GC_INVOKED]++;
-                SET_THREAD_STOP_PENDING(p,T);
+                SET_THREAD_STOP_PENDING(p, LISP_T);
             }
         }
     }
@@ -369,10 +354,10 @@ static inline void gc_notify_final()
             if (p == gc_state.collector)
                 continue;
             odxprint(safepoints,"notifying thread %p csp %p",p,csp_around_foreign_call(p));
-            boolean was_in_lisp = !set_thread_csp_access(p,0);
+            bool was_in_lisp = !set_thread_csp_access(p,0);
             if (was_in_lisp) {
                 gc_state.phase_wait[GC_SETTLED]++;
-                SET_THREAD_STOP_PENDING(p,T);
+                SET_THREAD_STOP_PENDING(p, LISP_T);
             }
         }
     }
@@ -382,12 +367,12 @@ static inline void gc_done()
 {
     CURRENT_THREAD_VAR(self);
     struct thread *p;
-    boolean inhibit = (read_TLS(GC_INHIBIT,self)==T);
+    bool inhibit = (read_TLS(GC_INHIBIT,self)==LISP_T);
 
     odxprint(safepoints,"%s","global denotification");
     WITH_ALL_THREADS_LOCK {
         for_each_thread(p) {
-            if (inhibit && (read_TLS(GC_PENDING,p)==T))
+            if (inhibit && (read_TLS(GC_PENDING,p)==LISP_T))
                 write_TLS(GC_PENDING,NIL,p);
             set_thread_csp_access(p,1);
         }
@@ -452,11 +437,7 @@ static inline void gc_advance(gc_phase_t cur, gc_phase_t old) {
         gc_state.phase = gc_phase_next(gc_state.phase);
         odxprint(safepoints,"no blockers, direct advance to %d (%s)",gc_state.phase,gc_phase_names[gc_state.phase]);
         gc_handle_phase();
-#ifdef LISP_FEATURE_WIN32
-        WakeAllConditionVariable(&gc_state.phase_cond[gc_state.phase]);
-#elif defined LISP_FEATURE_SB_THREAD
-        pthread_cond_broadcast(&gc_state.phase_cond[gc_state.phase]);
-#endif
+        CONDITION_VAR_WAKE_ALL(&gc_state.phase_cond[gc_state.phase]);
     }
     odxprint(safepoints,"going to wait for %d threads",gc_state.phase_wait[gc_state.phase]);
     gc_state_wait(cur);
@@ -469,7 +450,7 @@ thread_register_gc_trigger()
     struct thread *self = get_sb_vm_thread();
     WITH_GC_STATE_LOCK {
         if (gc_state.phase == GC_NONE &&
-            read_TLS(IN_SAFEPOINT,self)!=T &&
+            read_TLS(IN_SAFEPOINT,self)!=LISP_T &&
             !thread_blocks_gc(self)) {
             /* A thread (this thread), while doing allocation, has
              * determined that we need to run the garbage collector.
@@ -543,7 +524,7 @@ check_pending_thruptions(os_context_t *ctx)
      * "signal". */
     if (thread_extra_data(p)->pending_signal_set)
         if (__sync_fetch_and_and(&thread_extra_data(p)->pending_signal_set,0))
-            write_TLS(THRUPTION_PENDING, T, p);
+            write_TLS(THRUPTION_PENDING, LISP_T, p);
 #endif
 
     if (!thread_may_thrupt(ctx))
@@ -621,42 +602,42 @@ assert_on_stack(struct thread *th, void *esp)
 }
 
 /// Similar to the one in gc-common, but without the sigmask test.
-static boolean can_invoke_post_gc(struct thread* th)
+static bool can_invoke_post_gc(struct thread* th)
 {
     lispobj obj = th->lisp_thread;
     if (!obj) return 0;
     struct thread_instance* lispthread = (void*)(obj - INSTANCE_POINTER_LOWTAG);
-    if (!lispthread->primitive_thread) return 0;
+    if (!lispthread->uw_primitive_thread) return 0;
     return 1;
 }
 
 // returns 0 if skipped, 1 otherwise
-int check_pending_gc(__attribute((unused)) os_context_t *ctx)
+int check_pending_gc()
 {
     odxprint(misc, "check_pending_gc");
     struct thread * self = get_sb_vm_thread();
     int done = 0;
     sigset_t sigset;
 
-    if ((read_TLS(IN_SAFEPOINT,self) == T) &&
+    if ((read_TLS(IN_SAFEPOINT,self) == LISP_T) &&
         ((read_TLS(GC_INHIBIT,self) == NIL) &&
          (read_TLS(GC_PENDING,self) == NIL))) {
         write_TLS(IN_SAFEPOINT,NIL,self);
     }
     if (!thread_blocks_gc(self) && (read_TLS(IN_SAFEPOINT, self) == NIL)) {
-        if (read_TLS(GC_PENDING, self) == T) {
+        if (read_TLS(GC_PENDING, self) == LISP_T) {
             lispobj gc_happened = NIL;
 
-            bind_variable(IN_SAFEPOINT,T,self);
+            bind_variable(IN_SAFEPOINT,LISP_T,self);
             block_deferrable_signals(&sigset);
-            if(read_TLS(GC_PENDING,self)==T)
+            if(read_TLS(GC_PENDING,self)==LISP_T)
                 gc_happened = funcall1(StaticSymbolFunction(SUB_GC), 0);
             unbind(self);
             thread_sigmask(SIG_SETMASK,&sigset,NULL);
-            if (gc_happened == T) {
+            if (gc_happened == LISP_T) {
                 /* POST_GC wants to enable interrupts */
-                if ((read_TLS(INTERRUPTS_ENABLED,self) == T ||
-                     read_TLS(ALLOW_WITH_INTERRUPTS,self) == T)
+                if ((read_TLS(INTERRUPTS_ENABLED,self) == LISP_T ||
+                     read_TLS(ALLOW_WITH_INTERRUPTS,self) == LISP_T)
                     && can_invoke_post_gc(self))
                     funcall0(StaticSymbolFunction(POST_GC));
                 done = 1;
@@ -670,15 +651,15 @@ int check_pending_gc(__attribute((unused)) os_context_t *ctx)
 void thread_in_lisp_raised(os_context_t *ctxptr)
 {
     struct thread *self = get_sb_vm_thread();
-    boolean check_gc_and_thruptions = 0;
+    bool check_gc_and_thruptions = 0;
     odxprint(safepoints,"%s","thread_in_lisp_raised");
 
     /* Either we just hit the GSP trap, or we took a PIT stop and
      * there is a stop-for-GC or thruption pending. */
     WITH_GC_STATE_LOCK {
         if (gc_state.phase == GC_FLIGHT &&
-            read_TLS(GC_PENDING,self)==T &&
-            !thread_blocks_gc(self) && read_TLS(IN_SAFEPOINT,self)!=T) {
+            read_TLS(GC_PENDING,self)==LISP_T &&
+            !thread_blocks_gc(self) && read_TLS(IN_SAFEPOINT,self)!=LISP_T) {
             /* Some thread (possibly even this one) that does not have
              * GC_INHIBIT set has noticed that a GC is warranted and
              * advanced the phase to GC_FLIGHT, arming the GSP trap,
@@ -698,7 +679,7 @@ void thread_in_lisp_raised(os_context_t *ctxptr)
                 gc_advance(GC_NONE,GC_QUIET);
             } else {
                 /* ??? Isn't this already T? */
-                write_TLS(GC_PENDING,T,self);
+                write_TLS(GC_PENDING,LISP_T,self);
             }
             csp_around_foreign_call(self) = 0;
             check_gc_and_thruptions = 1;
@@ -731,7 +712,7 @@ void thread_in_lisp_raised(os_context_t *ctxptr)
                  * we'll take a PIT stop and wind up in the case
                  * above...  Or we'll call gc_stop_the_world(). */
                 gc_advance(GC_INVOKED,gc_state.phase);
-                SET_THREAD_STOP_PENDING(self,T);
+                SET_THREAD_STOP_PENDING(self,LISP_T);
                 /* Why do we not want to run thruptions here? */
             }
         }
@@ -739,7 +720,7 @@ void thread_in_lisp_raised(os_context_t *ctxptr)
     /* If we still need to GC, and it's not inhibited, call into
      * SUB-GC.  Phase is either GC_QUIET or GC_NONE. */
     if (check_gc_and_thruptions) {
-        check_pending_gc(ctxptr);
+        check_pending_gc();
 #ifdef LISP_FEATURE_SB_SAFEPOINT
         while(check_pending_thruptions(ctxptr));
 #endif
@@ -749,7 +730,7 @@ void thread_in_lisp_raised(os_context_t *ctxptr)
 void thread_in_safety_transition(os_context_t *ctxptr)
 {
     struct thread *self = get_sb_vm_thread();
-    boolean was_in_alien;
+    bool was_in_alien;
 
     odxprint(safepoints,"%s","GC safety transition");
     WITH_GC_STATE_LOCK {
@@ -794,7 +775,7 @@ void thread_in_safety_transition(os_context_t *ctxptr)
                  * logic is identical to the similar case in
                  * thread_in_lisp_raised(). */
                 gc_advance(GC_INVOKED,gc_state.phase);
-                SET_THREAD_STOP_PENDING(self,T);
+                SET_THREAD_STOP_PENDING(self,LISP_T);
             }
         }
     }
@@ -809,7 +790,7 @@ void thread_in_safety_transition(os_context_t *ctxptr)
 void thread_interrupted(os_context_t *ctxptr)
 {
     struct thread *self = get_sb_vm_thread();
-    boolean gc_active, was_in_alien;
+    bool gc_active, was_in_alien;
 
     odxprint(safepoints,"%s","pending interrupt trap");
     WITH_GC_STATE_LOCK {
@@ -825,7 +806,7 @@ void thread_interrupted(os_context_t *ctxptr)
             thread_in_lisp_raised(ctxptr);
         }
     }
-    check_pending_gc(ctxptr);
+    check_pending_gc();
 #ifdef LISP_FEATURE_SB_SAFEPOINT
     while(check_pending_thruptions(ctxptr));
 #endif
@@ -851,6 +832,7 @@ gc_stop_the_world()
         switch(gc_state.phase) {
         case GC_NONE:
             gc_advance(GC_QUIET,gc_state.phase);
+            /* FALLTHRU */
         case GC_FLIGHT:
         case GC_MESSAGE:
         case GC_INVOKED:
@@ -865,6 +847,7 @@ gc_stop_the_world()
             } else {
                 gc_state_wait(GC_QUIET);
             }
+            /* FALLTHRU */
         case GC_QUIET:
             /* Some number of threads were trying to get to GC_QUIET.
              * But this thread is sufficient to be able to leave
@@ -911,16 +894,16 @@ wake_thread_io(struct thread * thread)
 
 void wake_thread_impl(struct thread_instance *lispthread)
 {
-    struct thread* thread = (void*)lispthread->primitive_thread;
+    struct thread* thread = (void*)lispthread->uw_primitive_thread;
     wake_thread_io(thread);
 
-    if (read_TLS(THRUPTION_PENDING,thread)==T)
+    if (read_TLS(THRUPTION_PENDING,thread)==LISP_T)
         return;
 
-    write_TLS(THRUPTION_PENDING,T,thread);
+    write_TLS(THRUPTION_PENDING,LISP_T,thread);
 
-    if ((read_TLS(GC_PENDING,thread)==T)
-        ||(THREAD_STOP_PENDING(thread)==T)
+    if ((read_TLS(GC_PENDING,thread)==LISP_T)
+        ||(THREAD_STOP_PENDING(thread)==LISP_T)
         )
         return;
 
@@ -940,13 +923,13 @@ void wake_thread_impl(struct thread_instance *lispthread)
 # else
 void wake_thread_impl(struct thread_instance *lispthread)
 {
-    struct thread *thread = (void*)lispthread->primitive_thread;
+    struct thread *thread = (void*)lispthread->uw_primitive_thread;
     struct thread *self = get_sb_vm_thread();
 
     /* Must not and need not attempt to signal ourselves while we're the
      * STW initiator. */
     if (thread == self) {
-        write_TLS(THRUPTION_PENDING,T,self);
+        write_TLS(THRUPTION_PENDING,LISP_T,self);
         while (check_pending_thruptions(0 /* ignore the sigmask */))
             ;
         return;
@@ -971,16 +954,16 @@ void wake_thread_impl(struct thread_instance *lispthread)
                 WITH_ALL_THREADS_LOCK {
                     do {
                             odxprint(safepoints, "wake_thread_posix: found");
-                            write_TLS(THRUPTION_PENDING,T,thread);
-                            if (read_TLS(GC_PENDING,thread) == T
-                                || THREAD_STOP_PENDING(thread) == T)
+                            write_TLS(THRUPTION_PENDING,LISP_T,thread);
+                            if (read_TLS(GC_PENDING,thread) == LISP_T
+                                || THREAD_STOP_PENDING(thread) == LISP_T)
                                 break;
 
                             if (os_get_csp(thread)) {
                                 odxprint(safepoints, "wake_thread_posix: kill");
                                 /* ... and in foreign code.  Push it into a safety
                                  * transition. */
-                                int status = pthread_kill((pthread_t)lispthread->os_thread, SIGURG);
+                                int status = pthread_kill((pthread_t)lispthread->uw_os_thread, SIGURG);
                                 if (status)
                                     lose("wake_thread_posix: pthread_kill failed with %d",
                                          status);
@@ -991,7 +974,7 @@ void wake_thread_impl(struct thread_instance *lispthread)
             gc_advance(GC_NONE,GC_INVOKED);
         } else {
             odxprint(safepoints, "wake_thread_posix: passive");
-            write_TLS(THRUPTION_PENDING, T, thread);
+            write_TLS(THRUPTION_PENDING, LISP_T, thread);
         }
     }
     thread_sigmask(SIG_SETMASK, &oldset, 0);
@@ -1001,11 +984,6 @@ void wake_thread_impl(struct thread_instance *lispthread)
 
 void* os_get_csp(struct thread* th)
 {
-    FSHOW_SIGNAL((stderr, "Thread %p has CSP %p, stack [%p,%p]\n",
-                  th,
-                  (void*)csp_around_foreign_call(th),
-                  th->control_stack_start,
-                  th->control_stack_end));
     return (void*)csp_around_foreign_call(th);
 }
 
@@ -1053,11 +1031,6 @@ handle_safepoint_violation(os_context_t *ctx, os_vm_address_t fault_address)
 {
     struct thread *self = get_sb_vm_thread();
 
-    FSHOW_SIGNAL((stderr, "fault_address = %p, sp = %p, &csp = %p\n",
-                  fault_address,
-                  GC_SAFEPOINT_TRAP_ADDR, csp_around_foreign_call(self)));
-
-
     if (fault_address == (os_vm_address_t) GC_SAFEPOINT_TRAP_ADDR) {
 #ifdef LISP_FEATURE_C_STACK_IS_CONTROL_STACK
         /* We're on the altstack and don't want to run Lisp code. */
@@ -1086,4 +1059,52 @@ handle_safepoint_violation(os_context_t *ctx, os_vm_address_t fault_address)
 }
 #endif /* LISP_FEATURE_WIN32 */
 
+void
+vodxprint_fun(const char *fmt, va_list args)
+{
+#ifdef LISP_FEATURE_WIN32
+    DWORD lastError = GetLastError();
+#endif
+    int original_errno = errno;
+
+    char buf[1024];
+    int n = 0;
+
+    snprintf(buf, sizeof(buf), "["THREAD_ID_LABEL"] ", THREAD_ID_VALUE);
+    n = strlen(buf);
+
+    vsnprintf(buf + n, sizeof(buf) - n - 1, fmt, args);
+    /* buf is now zero-terminated (even in case of overflow).
+     * Our caller took care of the newline (if any) through `fmt'. */
+
+    /* A sufficiently POSIXy implementation of stdio will provide
+     * per-FILE locking, as defined in the spec for flockfile.  At least
+     * glibc complies with this.  Hence we do not need to perform
+     * locking ourselves here.  (Should it turn out, of course, that
+     * other libraries opt for speed rather than safety, we need to
+     * revisit this decision.) */
+    fputs(buf, stderr);
+
+#ifdef LISP_FEATURE_WIN32
+    /* stdio's stderr is line-bufferred, i.e. \n ought to flush it.
+     * Unfortunately, MinGW does not behave the way I would expect it
+     * to.  Let's be safe: */
+    fflush(stderr);
+#endif
+
+#ifdef LISP_FEATURE_WIN32
+    SetLastError(lastError);
+#endif
+    errno = original_errno;
+}
+
+void
+odxprint_fun(const char *fmt, ...)
+{
+    va_list args;
+    va_start(args, fmt);
+    vodxprint_fun(fmt, args);
+    va_end(args);
+}
+
 #endif /* LISP_FEATURE_SB_SAFEPOINT -- entire file */
