@@ -11,14 +11,17 @@
 
 #include <string.h>
 #include <ctype.h>
+#include <stdint.h>
 
-#include "sbcl.h"
-#include "runtime.h"
+#include "genesis/sbcl.h"
+#include "interr.h"
+#include "lispobj.h"
 #include "os.h"
 #include "search.h"
 #include "thread.h"
 #include "gc.h"
 #include "genesis/primitive-objects.h"
+#include "genesis/instance.h"
 #include "genesis/hash-table.h"
 #include "genesis/package.h"
 #include "genesis/split-ordered-list.h"
@@ -37,6 +40,15 @@ search_read_only_space(void *pointer)
 lispobj *
 search_static_space(void *pointer)
 {
+#ifdef LISP_FEATURE_PERMGEN // consider it part of static space
+    {
+    lispobj* found = 0;
+    if ((uword_t)pointer >= PERMGEN_SPACE_START &&
+        (lispobj*)pointer < permgen_space_free_pointer &&
+        ((found = gc_search_space((lispobj*)PERMGEN_SPACE_START, pointer)) != NULL))
+        return found;
+    }
+#endif
     lispobj *start = (lispobj *)STATIC_SPACE_OBJECTS_START;
     lispobj *end = static_space_free_pointer;
     if ((pointer < (void *)start) || (pointer >= (void *)end))
@@ -44,12 +56,31 @@ search_static_space(void *pointer)
     return gc_search_space(start, pointer);
 }
 
+#ifndef LISP_FEATURE_IMMOBILE_SPACE
+lispobj *search_codeblob_offsets(void* pointer) {
+    if ((uword_t)pointer < TEXT_SPACE_START ||
+        (uword_t)pointer >= (uword_t)text_space_highwatermark) return 0;
+    struct vector* v = (void*)TEXT_SPACE_START;
+    uint32_t* data = (void*)v->data;
+    int index = bsearch_lesseql_uint32((char*)pointer - (char*)TEXT_SPACE_START,
+                                       data, vector_len(v));
+    if (index >= 0) {
+        lispobj* base = (lispobj*)(TEXT_SPACE_START + data[index]);
+        gc_assert(widetag_of(base) == CODE_HEADER_WIDETAG);
+        return base;
+    }
+    return 0;
+}
+#endif
+
 lispobj *search_all_gc_spaces(void *pointer)
 {
     lispobj *start;
     if (((start = search_dynamic_space(pointer)) != NULL) ||
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
         ((start = search_immobile_space(pointer)) != NULL) ||
+#else
+        ((start = search_codeblob_offsets(pointer)) != NULL) ||
 #endif
         ((start = search_static_space(pointer)) != NULL) ||
         ((start = search_read_only_space(pointer)) != NULL))
@@ -156,10 +187,30 @@ struct symbol* lisp_symbol_from_tls_index(lispobj tls_index)
 }
 #endif
 
+static uword_t bruteforce_findpkg_by_id(lispobj* where, lispobj* limit, uword_t id)
+{
+    lispobj layout;
+    for ( where = next_object(where, 0, limit) ; where ;
+          where = next_object(where, object_size(where), limit) ) {
+        if (widetag_of(where) == INSTANCE_WIDETAG
+            && (layout = instance_layout(where)) != 0
+            && layout_depth2_id(LAYOUT(layout)) == PACKAGE_LAYOUT_ID
+            && ((struct package*)where)->id == id) {
+            return make_lispobj(where, INSTANCE_POINTER_LOWTAG);
+        }
+    }
+    return 0;
+}
+
 lispobj get_package_by_id(int id) {
     // lisp_package_vector is a tagged pointer to SIMPLE-VECTOR
     lispobj vector = barrier_load(&lisp_package_vector);
-    if (!vector) lose("get_package_by_id: no package vector");
+    if (!vector) {
+        // Perform a heap walk. This should never occur except in core loading/saving.
+        lispobj result =  walk_generation(bruteforce_findpkg_by_id, -1, make_fixnum(id));
+        if (is_lisp_pointer(result)) return result;
+        lose("get_package_by_id: no package vector");
+    }
     if (id >= vector_len(VECTOR(vector))) lose("can't decode package ID %d", id);
     return barrier_load(&VECTOR(vector)->data[id]);
 }
@@ -179,7 +230,7 @@ static lispobj* search_package_symbols(lispobj package, char* symbol_name)
     struct package* pkg = (void*)INSTANCE(package);
     int pass;
     for (pass = 0; pass <= 1; ++pass) {
-        struct symbol_hashset* table = (void*)
+        struct symbol_table* table = (void*)
           INSTANCE(barrier_load(pass ? &pkg->external_symbols : &pkg->internal_symbols));
         gc_assert(widetag_of(&table->header) == INSTANCE_WIDETAG);
         lispobj cells = barrier_load(&table->_cells);
@@ -324,13 +375,21 @@ int bsearch_greatereql_uint32(uint32_t item, uint32_t* array, int nelements)
     if (low < nelements) return low;
     return -1;
 }
+#else
+// these passthru stubs could probably just be linker symbol aliases
+int bsearch_lesseql_uint32(uint32_t item, uint32_t* array, int nelements) {
+    return bsearch_lesseql_uword(item, array, nelements);
+}
+int bsearch_greatereql_uint32(uint32_t item, uint32_t* array, int nelements) {
+    return bsearch_greatereql_uword(item, array, nelements);
+}
 #endif
 
 /* Find in an address-based split-ordered list
  * Unlike the lisp algorithm, this does not "assist" a pending deletion
  * by completing it with compare-and-swap - this loop simply ignores
  * any deleted nodes that haven't been snipped out yet. */
-struct split_ordered_list_node*
+struct solist_node*
 split_ordered_list_find(struct split_ordered_list* solist,
                         lispobj key)
 {
@@ -349,7 +408,7 @@ split_ordered_list_find(struct split_ordered_list* solist,
     while ((nodeptr & WIDETAG_MASK) == UNBOUND_MARKER_WIDETAG) {
         nodeptr = bins->data[--bin_index];
     }
-    struct split_ordered_list_node* node = (void*)native_pointer(nodeptr);
+    struct solist_node* node = (void*)native_pointer(nodeptr);
     lispobj hash_as_fixnum = make_fixnum(full_hash);
     while (1) {
         if (node->node_hash == hash_as_fixnum) { // possible hit
