@@ -381,3 +381,156 @@
      (loop for i below n
            sum (coerce n 'single-float)))
    (or (integer 0 0) single-float)))
+
+(with-test (:name :overflow-transform-order)
+  (checked-compile-and-assert
+      (:optimize :safe)
+      `(lambda (a m)
+         (declare (fixnum a))
+         (let ((j (* 44 a)))
+           (when m
+             (the fixnum j))))
+    ((most-positive-fixnum nil) nil)
+    ((most-positive-fixnum t) (condition 'type-error))))
+
+(with-test (:name :logtest-memref-boxed)
+  (checked-compile-and-assert
+      ()
+      `(lambda (b)
+         (declare (sb-vm:word b))
+         (when (oddp b)
+           (lambda (m)
+             (when m
+               (setf b 1))
+             b)))
+    (((expt 2 (1- sb-vm:n-word-bits))) nil)
+    (((1+ (expt 2 (1- sb-vm:n-word-bits)))) t :test (lambda (x y)
+                                                      y
+                                                      (functionp (car x))))))
+
+(with-test (:name :range-unsigned)
+  (assert-type
+   (lambda (d)
+     (declare (type (integer 1 109) d))
+     (typep (- d) '(integer -47727025476642942 -2593702250735)))
+   null))
+
+(with-test (:name :signed-byte-8-p-unsigned)
+  (checked-compile
+   `(lambda (a)
+      (declare (type (simple-array sb-vm:word (*)) a)
+               (optimize speed))
+      (the (signed-byte 8) (aref a 0)))
+   :allow-notes nil))
+
+(with-test (:name :or-chain)
+  (checked-compile-and-assert
+   ()
+   `(lambda (b)
+      (declare (fixnum b))
+      (case b ((0 -3) 1) (t 2)))
+   ((0) 1)
+   ((-3) 1)
+   ((3) 2)
+   ((1) 2)))
+
+(with-test (:name :or-chain-types)
+  (checked-compile-and-assert
+   ()
+   `(lambda (b)
+      (declare ((integer -1 1) b))
+      (case b
+        ((-1 0) 0)
+        (t 1)))
+   ((-1) 0)
+   ((0) 0)
+   ((1) 1)))
+
+(with-test (:name :or-chain-tagging)
+  (checked-compile-and-assert
+   ()
+   `(lambda (x)
+      (or (eq x -6)
+          (eq x -2)))
+   ((-6) t)
+   ((-2) t)
+   ((6) nil)
+   ((2) nil)
+   ((-12) nil)
+   ((-4) nil))
+  (checked-compile-and-assert
+   ()
+   `(lambda (x)
+      (or (eq x 0)
+          (eq x -4)))
+   ((0) t)
+   ((-4) t)
+   ((4) nil)
+   ((-8) nil))
+  (checked-compile-and-assert
+   ()
+   `(lambda (x)
+      (or (eq x 97)
+          (eq x 65)))
+   ((-4611686018427387807) nil)
+   ((97) t)
+   ((65) t))
+  (checked-compile-and-assert
+   ()
+   `(lambda (x)
+      (or (eq x -65)
+          (eq x -97)))
+   ((-97) t)
+   ((-65) t))
+  (checked-compile-and-assert
+   ()
+   `(lambda (x)
+      (case x ((-3 -2 17) t)))
+   ((4611686018427387902) nil)
+   ((-3) t)
+   ((-2) t)
+   ((17) t)))
+
+(with-test (:name :range<=-same)
+  (checked-compile-and-assert
+   ()
+   `(lambda (a c)
+      (declare (type fixnum a))
+      (let ((v7 (if c
+                    4611686018427387904
+                    -6)))
+        (if (> v7 a)
+            a
+            (if (<= a v7)
+                0
+                a))))
+    ((-7 nil) -7)
+    ((-7 t) -7)
+    ((-6 nil) 0)
+    ((-6 t) -6)
+    ((-3 nil) -3)))
+
+(with-test (:name :/-folding)
+  (checked-compile-and-assert
+      (:optimize :safe)
+      `(lambda (a)
+         (declare (bit a))
+         (/ 1 a))
+    ((1) 1)
+    ((0) (condition 'division-by-zero)))
+  (checked-compile-and-assert
+      (:optimize :safe)
+      `(lambda (a)
+         (declare (bit a))
+         (= (/ 5 a) 5))
+    ((1) t)
+    ((0) (condition 'division-by-zero))))
+
+(with-test (:name :dpb-size-overflow)
+  (checked-compile-and-assert
+   ()
+   `(lambda (a)
+      (declare ((unsigned-byte 8) a))
+      (dpb a (byte 63 8)
+           81))
+   ((90) 23121)))
