@@ -502,20 +502,6 @@
                   (eval `(let (,@(mapcar 'list vars vals))
                       ,read-form)))))))
 
-(let ((foo (cons :foo nil)))
-  (defun cas-foo (old new)
-    (cas (cdr foo) old new)))
-
-(defcas foo () cas-foo)
-
-(with-test (:name :cas-and-macroexpansion)
-  (assert (not (cas (foo) nil t)))
-  (assert (eq t (cas (foo) t nil)))
-  (symbol-macrolet ((bar (foo)))
-    (assert (not (cas bar nil :ok)))
-    (assert (eq :ok (cas bar :ok nil)))
-    (assert (not (cas bar nil t)))))
-
 (with-test (:name :atomic-push
             :skipped-on (not :sb-thread))
   (let ((store (cons nil nil))
@@ -553,17 +539,6 @@
                      (atomic-pop (symbol-value 'x))))))
       1)))
 
-#+x86-64 ; missing symbol sb-vm::signed-sap-cas-32 otherwise
-(with-test (:name :cas-sap-ref ; FIXME: remove
-            :skipped-on :interpreter)
-  (let ((v (make-array 2 :element-type '(signed-byte 32)
-                         :initial-element -1)))
-    (let ((old (sb-sys:%primitive sb-vm::signed-sap-cas-32
-                                  (sb-sys:vector-sap v)
-                                  0 -1 1)))
-      (assert (= old -1))
-      (assert (eql (aref v 0) 1)))))
-
 (in-package "SB-VM")
 
 ;;; This file defines a structure, so is an 'impure' test
@@ -605,7 +580,7 @@
     t))
 
 (test-util:with-test (:name :wide-compare-and-exchange
-                      :skipped-on (or :interpreter (not (or :x86 :x86-64))))
+                      :skipped-on (not (or :x86 :x86-64)))
   (multiple-value-bind (a b c d) (%cpu-identification 0 0)
     (declare (ignore b c d))
     ;; paranoidly check for whether we can execute function ID 1
@@ -628,12 +603,10 @@
         ;;  1. it was using a :dword move where it should have used a :qword
         ;;  2. it was emitting constants as bignums instead of inline raw constants
         (macrolet ((test (signedp nbits newval
-                          &aux (ref (symbolicate
-                                     (if signedp "SIGNED-" "")
-                                     "SAP-REF-"
-                                     (write-to-string nbits)))
-                                     (init (if signedp -1
-                                               (ldb (byte nbits 0) most-positive-word))))
+                          &aux (ref (symbolicate (if signedp "SIGNED-" "")
+                                                 "SAP-REF-" nbits))
+                               (init (if signedp -1
+                                         (ldb (byte nbits 0) most-positive-word))))
                      `(progn
                         ;; (format t "Testing ~a with initial bits ~x~%" ',ref ,init)
                         (setf (,ref sap 0) ,init)
@@ -690,14 +663,17 @@
           (assert (= (sap-ref-32 sap 0) (* n-threads n-increments))))))))
 
 (define-alien-variable "small_generation_limit" (signed 8))
-#+(or x86-64 ppc64)
+;; PPC64 shouldn't fail, but depending on the CPU revision it might not
+;; have the needed instruction, and I don't know how to test for it.
+;; And surely it doesn't really depend on endian-ness, but the machine
+;; that I'm testing on which is little-endian passes the test.
+#+(and sb-thread (or x86-64 (and ppc64 little-endian)))
+(progn
 (defun cas-an-alien-byte (x y) (cas small-generation-limit x y))
-#+(or x86-64 ppc64)
 (compile 'cas-an-alien-byte)
-(test-util:with-test (:name :cas-alien
-                            :skipped-on (not (and :sb-thread (or :ppc64 :x86-64))))
+(test-util:with-test (:name :cas-alien)
   (assert (= small-generation-limit 1))
   (assert (= (cas-an-alien-byte 0 5) 1))
   (assert (= (cas-an-alien-byte 1 6) 1))
   (assert (= small-generation-limit 6))
-  (setf small-generation-limit 1))
+  (setf small-generation-limit 1)))
