@@ -189,16 +189,15 @@
 
 (with-test (:name :type-constraint-joining-terminates)
   (checked-compile
-   sb-c::
    `(lambda (name vop)
       (block foo
-        (do* ((block (vop-block vop) (ir2-block-prev block))
-              (last vop (ir2-block-last-vop block)))
+        (do* ((block (sb-c::vop-block vop) (sb-c::ir2-block-prev block))
+              (last vop (sb-c::ir2-block-last-vop block)))
              (nil)
-          (aver (eq (ir2-block-block block) (ir2-block-block (vop-block vop))))
-          (do ((current last (vop-prev current)))
+          (sb-c::aver (eq (sb-c::ir2-block-block block) (sb-c::ir2-block-block (sb-c::vop-block vop))))
+          (do ((current last (sb-c::vop-prev current)))
               ((null current))
-            (when (eq (vop-name current) name)
+            (when (eq (sb-c::vop-name current) name)
               (return-from foo current))))))))
 
 (with-test (:name :type-constraint-joining-conflicts)
@@ -507,19 +506,7 @@
               s)))))
     '(values array &optional))))
 
-(with-test (:name :+integer-argument-constarint)
-  (assert
-   (type-specifiers-equal
-    (caddr
-     (sb-kernel:%simple-fun-type
-      (checked-compile
-       `(lambda (a x y)
-          (declare (fixnum x))
-          (aref a (+ x y))
-          y))))
-    '(values integer &optional))))
-
-(with-test (:name :+integer-argument-constarint.var)
+(with-test (:name :+integer-argument-constraint.var)
   (assert
    (type-specifiers-equal
     (caddr
@@ -530,9 +517,11 @@
           (let ((m (+ x y)))
             (aref a (+ m (aref a m)))
             y)))))
-    '(values integer &optional))))
+    `(values (integer ,(1+ most-negative-fixnum)
+                      ,(+ most-positive-fixnum array-dimension-limit))
+             &optional))))
 
-(with-test (:name :+integer-argument-constarint.typep)
+(with-test (:name :+integer-argument-constraint.typep)
   (assert
    (type-specifiers-equal
     (caddr
@@ -837,9 +826,490 @@
                     (ctu:ir1-named-calls
                      `(lambda (x v)
                         (declare (integer x)
-                                 (simple-vector v)
-                                 (optimize (debug 1)))
+                                 (simple-vector v))
+                        (loop for i below (min x (length v))
+                              sum (aref v i)))
+                     nil))
+             0))
+  (assert (= (count 'sb-kernel:%check-bound
+                    (ctu:ir1-named-calls
+                     `(lambda (x v)
+                        (declare (simple-vector v))
                         (loop for i below (min x (length v))
                               sum (aref v i)))
                      nil))
              0)))
+
+(with-test (:name :bounds-check-make-array)
+  (assert (= (count 'sb-kernel:%check-bound
+                    (ctu:ir1-named-calls
+                     `(lambda (length)
+                        (declare (integer length))
+                        (let ((array (make-array length)))
+                          (loop for i below length
+                                do (setf (aref array i) i))))
+                     nil))
+             0))
+  (assert (= (count 'sb-kernel:%check-bound
+                    (ctu:ir1-named-calls
+                     `(lambda (length)
+                        (let ((array (make-array length)))
+                          (loop for i below length
+                                do (setf (aref array i) i))))
+                     nil))
+             0))
+  (assert (= (count 'sb-kernel:%check-bound
+                    (ctu:ir1-named-calls
+                     `(lambda (type length)
+                        (let ((array (make-sequence type length)))
+                          (loop for i below length
+                                do (setf (aref array i) i))))
+                     nil))
+             0)))
+
+(with-test (:name :bounds-check-down)
+  (assert (= (count 'sb-kernel:%check-bound
+                    (ctu:ir1-named-calls
+                     `(lambda (v)
+                        (let ((end (1- (length v))))
+                          (decf end)
+                          (when (>= end 0)
+                            (svref v end))))
+                     nil))
+             0))
+  (assert (= (count 'sb-kernel:%check-bound
+                    (ctu:ir1-named-calls
+                     `(lambda (v)
+                        (declare (optimize (debug 2)))
+                        (let ((end (1- (length v))))
+                          (decf end)
+                          (when (>= end 0)
+                            (svref v end))))
+                     nil))
+             0))
+  (assert (= (count 'sb-kernel:%check-bound
+                    (ctu:ir1-named-calls
+                     `(lambda (v)
+                        (loop for i from (1- (length v)) downto 0
+                              collect (svref v i)))
+                     nil))
+             0))
+  (assert (= (count 'sb-kernel:%check-bound
+                    (ctu:ir1-named-calls
+                     `(lambda (v)
+                        (declare (optimize (debug 2)))
+                        (loop for i from (1- (length v)) downto 0
+                              collect (svref v i)))
+                     nil))
+             0)))
+
+(with-test (:name :reoptimize)
+  (assert-type
+   (lambda ()
+     (let ((m 0))
+       (incf m)
+       (incf m)
+       m))
+   (eql 2))
+  (assert-type
+   (lambda (n j)
+     (declare (integer j)
+              (optimize (debug 1)))
+     (the (integer 5) n)
+     (when (> j n)
+       j))
+   (or (integer 6) null)))
+
+(with-test (:name :+back)
+  (assert-type
+   (lambda (x y)
+     (when (> (+ x 1) y)
+       x))
+   (or real null))
+  (assert-type
+   (lambda (x y)
+     (when (= (+ x 1) y)
+       x))
+   (or number null))
+  (assert-type
+   (lambda (x y)
+     (when (integerp (+ x y))
+       x))
+   (or number null))
+  (assert-type
+   (lambda (x y)
+     (declare (real y))
+     (when (integerp (+ x y))
+       x))
+   (or rational null))
+  (assert-type
+   (lambda (x)
+     (when (integerp (+ x 1))
+       x))
+   (or integer null))
+  (assert-type
+   (lambda (x)
+     (when (typep (+ x 3) '(integer 0 20))
+       x))
+   (or (integer -3 17) null))
+  (assert-type
+   (lambda (x m)
+     (declare ((real 2 4) m)
+              (fixnum x))
+     (when (typep (+ m x) '(integer 2 4))
+       x))
+   (or (integer -2 2) null)))
+
+(with-test (:name :-back)
+  (assert-type
+   (lambda (x)
+     (when (typep (- x 4) '(integer 0 10))
+       x))
+   (or (integer 4 14) null))
+  (assert-type
+   (lambda (x)
+     (when (typep (- 3 x) '(integer 0 10))
+       x))
+   (or (integer -7 3) null))
+  (assert-type
+   (lambda (x)
+     (when (> (- 3 x) 10)
+       x))
+   (or real null)))
+
+(with-test (:name :*back)
+  (assert-type
+   (lambda (x)
+     (when (typep (* x 4) '(integer 0 10))
+       x))
+   (or (rational 0 5/2) null))
+  (assert-type
+   (lambda (x m)
+     (declare ((real 0 3) m))
+     (when (typep (* x m) '(integer 0 10))
+       x))
+   (or number null))
+  (assert-type
+   (lambda (x)
+     (when (> (* x 3) 10)
+       x))
+   (or real null))
+  (assert-type
+   (lambda (x m)
+     (declare (real x)
+              (integer m))
+     (when (typep (* x m) 'integer)
+       x))
+   (or rational null)))
+
+(with-test (:name :ignore-hairy-types)
+  (checked-compile
+   '(lambda (a)
+     (declare (optimize (debug 2)))
+     (let ((v (the (satisfies eval) a)))
+       (dotimes (i 3)
+         (print (/= a 1))
+         (the real a))
+       v)))
+  (checked-compile
+   '(lambda (d)
+     (values
+      (let ((v9 d)
+            (v1 (the (satisfies eval) (setf d 1))))
+        (if (= v1 v9) 0 1))
+      (loop sum (if (> (if (<= 220 d)
+                           (if (and (< 220 d)
+                                    (>= 1024 d))
+
+                               7)
+                           220) 0)
+                    1 2))))))
+
+(with-test (:name :set-constraint)
+  (assert-type
+   (lambda (x)
+     (when (> (decf x 10) 10)
+       x))
+   (or (real (10)) null)))
+
+(with-test (:name :set-constraint-inherit)
+  (checked-compile-and-assert
+   ()
+   `(lambda (b c)
+      (declare (type (integer 1 1625159256) b))
+      ((lambda (v2 v6)
+         (logorc1 (setf c 1448416540835)
+                  (gcd
+                   (if (>= b c)
+                       (if (< b c)
+                           0
+                           -197521449755177454)
+                       0)
+                   (decf v2 (incf v2 (incf v6 (setq v2 (setq b 1044756040))))))))
+       0 0))
+   ((572837281 0) -1448416540836)))
+
+(with-test (:name :float-eql-bits)
+  (assert-type
+   (lambda (number)
+     (declare (type double-float number))
+     (cond ((eql number 1d0) number)))
+   (or (eql 1.0d0) null)))
+
+(with-test (:name :constraint-amount)
+  (assert (= (count 'sb-kernel:%check-bound
+                    (ctu:ir1-named-calls
+                     `(lambda (v)
+                        (loop for i below (- (length v) 2) by 2
+                              sum (svref v (1+ i))))
+                     nil))
+             0))
+  (assert (= (count 'sb-kernel:%check-bound
+                    (ctu:ir1-named-calls
+                     `(lambda (v)
+                        (loop for i from (- (length v) 2) downto 0
+                              collect (svref v (1+ i))))
+                     nil))
+             0))
+  (checked-compile-and-assert
+      ()
+      `(lambda (a)
+         (declare (type fixnum a))
+         (let ((j (+ a most-positive-fixnum)))
+           (let ((bbb a))
+             (dotimes (i a)
+               (incf bbb))
+             (< bbb j))))
+    ((3) t)))
+
+(with-test (:name :constraint-multiple-eql-variables)
+  (assert-type
+   (lambda (x y)
+     (declare (integer x y)
+              (optimize (debug 2)))
+     (assert (< x y))
+     (< x y))
+   (eql t)))
+
+(with-test (:name :dead-blocks)
+  (assert-type
+   (lambda (c d)
+     (declare ((integer -31307831194 26651266057) c))
+     (let ((v5 (if d
+                   -206316434836409151
+                   17179869177)))
+       (if (/= v5 c)
+           1
+           (if (> v5 c)
+               (logior
+                c
+                (loop for lv2 below 1
+                      sum
+                      (rem lv2 128)))
+               1))))
+   (eql 1)))
+
+(with-test (:name :dead-blocks.2)
+  (checked-compile `(lambda (a b)
+                      (declare (type (integer 0 171951449) a)
+                               ((member -1965785401190741689 18014398509481983) b)
+                               (optimize (debug 2)))
+                      (let ((v3 b)
+                            (v6 116510974941639))
+                        (if (> v6 v3)
+                            (if (> v3 v6)
+                                (let ((v3 (logand a v3)))
+                                  (loop for lv1 below 3
+                                        count (let ((v8 (min lv1 v3)))
+                                                (<= v8  576460752303423490))))))))))
+
+(with-test (:name :multiple-equality-constraints)
+  (checked-compile
+   `(lambda (v)
+      (declare (simple-vector v))
+      (let* ((i (1- (length v))))
+        (declare (fixnum i))
+        (loop
+         (decf i)
+         (print (aref v i)))))))
+
+(with-test (:name :min/max-constraints-set)
+  (checked-compile-and-assert
+   ()
+   `(lambda (a b c)
+      (let ((x
+              (if (< a c)
+                  c
+                  a))
+            (y (setq a (min most-positive-fixnum b))))
+        (> y x)))
+   ((1 5 3) t)
+   ((2 4 6) nil)))
+
+(with-test (:name :mod)
+  (assert (= (count 'sb-kernel:%check-bound
+                    (ctu:ir1-named-calls
+                     `(lambda (v i)
+                        (svref v (mod i (length v))))
+                     nil))
+             0)))
+
+(with-test (:name :index-ranges)
+  (assert (= (count 'sb-kernel:%check-bound
+                    (ctu:ir1-named-calls
+                     `(lambda (a)
+                        (when (= (length a) 2)
+                          (svref a 1)))
+                     nil))
+             0))
+  (assert (= (count 'sb-kernel:%check-bound
+                    (ctu:ir1-named-calls
+                     `(lambda (a l i)
+                        (declare ((integer 2 3) l)
+                                 ((integer 0 1) i))
+                        (when (= (length a) l)
+                          (svref a i)))
+                     nil))
+             0)))
+
+(with-test (:name :constant-vector)
+  (assert (= (count 'sb-kernel:%check-bound
+                    (ctu:ir1-named-calls
+                     `(lambda (x)
+                        (declare (fixnum x))
+                        (and (<= x 2)
+                             (aref #(1 2 3) x)))
+                     nil))
+             0)))
+
+(with-test (:name :map-equality-constraints)
+  (checked-compile-and-assert
+   ()
+   `(lambda (c d)
+      (declare (notinline identity)
+               (bit d))
+      (dotimes (e 3 (identity (logior c 10 (- d -6 e))))))
+   ((2 1) 14)
+   ((9 0) 11)))
+
+(with-test (:name :local-call)
+  (assert-type
+   (lambda (a)
+     (flet ((b ()
+              a))
+       (declare (notinline b))
+       (the integer a)
+       (b)
+       (b)))
+   integer))
+
+(with-test (:name :set-vars-equal)
+  (assert-type
+   (lambda (a)
+     (incf a 10)
+     (eq a (progn (loop repeat a do (print 20)) a)))
+   (eql t))
+  (checked-compile-and-assert
+   ()
+   `(lambda (a)
+      (logxor (unwind-protect a (setq a 7)) a))
+  ((0) 7)
+  ((7) 0)))
+
+(with-test (:name :set-vars-equal-leaf-change)
+  (checked-compile-and-assert
+   ()
+   `(lambda (b)
+      (declare (integer b))
+      (let ((c 1))
+        (decf c (shiftf b b))))
+   ((3) -2)
+   ((-4) 5)))
+
+(with-test (:name :all-var-values)
+  (checked-compile
+   `(lambda (c)
+      (declare (optimize (safety 0)))
+      (loop for lv3 below 1
+            do
+            (loop for lv4 below 3
+                  sum (progv nil nil
+                        (setq c 1000)))
+            (unless (eq  lv3 -1)
+              (the integer (catch 'ct4 16385)))))))
+
+(with-test (:name :cons-cells)
+  (assert-type
+   (lambda (x)
+     (if (car x)
+         (consp x)
+         t))
+   (eql t))
+  (assert-type
+   (lambda (x)
+     (if (car x)
+         nil
+         (consp x)))
+   boolean)
+  (assert-type
+   (lambda (x)
+     (if (> (car x) 1)
+         (consp x)
+         t))
+   (eql t))
+  (assert-type
+   (lambda (x)
+     (if (> (car x) 1)
+         t
+         (consp x)))
+   (eql t)))
+
+(with-test (:name :char-code)
+  (assert-type
+   (lambda (x)
+     (let ((n (char-code x)))
+       (when (or (< 31 n 127) (= n 10))
+         x)))
+   (or null standard-char))
+  (assert-type
+   (lambda (x)
+     (declare (optimize (debug 2)))
+     (when (standard-char-p x)
+       x))
+   (or null standard-char)))
+
+(with-test (:name :vector-length-alias)
+  (checked-compile-and-assert
+      ()
+      `(lambda (x)
+         (or (typep x '(integer 3 3))
+             (typep x '(simple-array character (3)))))
+    ((3) t)
+    (("000") t)
+    ((6) nil)
+    (("") nil)))
+
+(with-test (:name :eql-var-replacement)
+  (checked-compile-and-assert
+      (:allow-notes nil)
+      `(lambda (x y)
+         (declare (float x)
+                  ((simple-array double-float (*)) y))
+         (let ((d (aref y 0)))
+           (when (and (/= d 0)
+                      (eql d x))
+             x)))
+    ((1d0 (make-array 1 :element-type 'double-float :initial-element 1d0)) 1d0 :test #'equal)
+    ((1d0 (make-array 1 :element-type 'double-float :initial-element 0d0)) nil)
+    ((0d0 (make-array 1 :element-type 'double-float :initial-element 1d0)) nil)
+    ((0d0 (make-array 1 :element-type 'double-float :initial-element 0d0)) nil)))
+
+(with-test (:name :eql-var-initial-unused)
+  (checked-compile-and-assert
+      ()
+      `(lambda (p1 p2 p3)
+         (declare (type (integer -5) p3))
+         (let ((n (the (integer 0) p3)))
+           (setq n (dpb p1 (byte p2 35) n))
+           n))
+    ((1 2 3) 34359738371)))
