@@ -45,9 +45,9 @@
                           (alien-funcall df-decode-locs (sb-kernel:get-lisp-obj-address locs)
                                          (addr offset) (addr elsewhere-pc)))))
               (assert (= res 1))
-              (multiple-value-bind (start-pc expect-elsewhere-pc form-number expect-offset)
+              (multiple-value-bind (start-pc expect-elsewhere-pc expect-offset)
                   (sb-c::cdf-decode-locs cdf)
-                (declare (ignore start-pc form-number))
+                (declare (ignore start-pc))
                 (unless (and (= offset expect-offset)
                              (= elsewhere-pc expect-elsewhere-pc))
                   (setq ok nil)
@@ -89,5 +89,6 @@
                     (assert (= valid 1))
                     (assert (= valid 0)))))))))))
 (compile 'randomly-probe-pure-boxed-objects)
-(with-test (:name :fast-valid-tagged-pointer-p)
+(with-test (:name :fast-valid-tagged-pointer-p
+                  :broken-on :mark-region-gc)
   (randomly-probe-pure-boxed-objects))
