@@ -458,3 +458,187 @@ fractional bits."
                  (when (< d 1.0d0)
                    (log d)))
                (or null (double-float * 0.0d0))))
+
+(with-test (:name (ftruncate :minus-zeros :one-arg))
+  (flet ((f (x) (declare (notinline ftruncate)) (ftruncate x)))
+    (declare (notinline f))
+    (assert (equal (multiple-value-list (f 0)) '(0.0 0)))
+    (assert (equal (multiple-value-list (f 0.0)) '(0.0 0.0)))
+    (assert (equal (multiple-value-list (f -0.0)) '(-0.0 0.0)))
+    (assert (equal (multiple-value-list (f 0.5)) '(0.0 0.5)))
+    (assert (equal (multiple-value-list (f 1/2)) '(0.0 1/2)))
+    (assert (equal (multiple-value-list (f -0.5)) '(-0.0 -0.5)))
+    (assert (equal (multiple-value-list (f -1/2)) '(-0.0 -1/2)))))
+
+(with-test (:name (ffloor :minus-zeros :one-arg))
+  (flet ((f (x) (declare (notinline ffloor)) (ffloor x)))
+    (declare (notinline f))
+    (assert (equal (multiple-value-list (f 0)) '(0.0 0)))
+    (assert (equal (multiple-value-list (f 0.0)) '(0.0 0.0)))
+    (assert (equal (multiple-value-list (f -0.0)) '(-0.0 0.0)))
+    (assert (equal (multiple-value-list (f 0.5)) '(0.0 0.5)))
+    (assert (equal (multiple-value-list (f 1/2)) '(0.0 1/2)))
+    (assert (equal (multiple-value-list (f -0.5)) '(-1.0 0.5)))
+    (assert (equal (multiple-value-list (f -1/2)) '(-1.0 1/2)))))
+
+(with-test (:name (fceiling :minus-zeros :one-arg))
+  (flet ((f (x) (declare (notinline fceiling)) (fceiling x)))
+    (declare (notinline f))
+    (assert (equal (multiple-value-list (f 0)) '(0.0 0)))
+    (assert (equal (multiple-value-list (f 0.0)) '(0.0 0.0)))
+    (assert (equal (multiple-value-list (f -0.0)) '(-0.0 0.0)))
+    (assert (equal (multiple-value-list (f 0.5)) '(1.0 -0.5)))
+    (assert (equal (multiple-value-list (f 1/2)) '(1.0 -1/2)))
+    (assert (equal (multiple-value-list (f -0.5)) '(-0.0 -0.5)))
+    (assert (equal (multiple-value-list (f -1/2)) '(-0.0 -1/2)))))
+
+(with-test (:name (fround :minus-zeros :one-arg))
+  (flet ((f (x) (declare (notinline fround)) (fround x)))
+    (declare (notinline f))
+    (assert (equal (multiple-value-list (f 0)) '(0.0 0)))
+    (assert (equal (multiple-value-list (f 0.0)) '(0.0 0.0)))
+    (assert (equal (multiple-value-list (f -0.0)) '(-0.0 0.0)))
+    (assert (equal (multiple-value-list (f 0.5)) '(0.0 0.5)))
+    (assert (equal (multiple-value-list (f 1/2)) '(0.0 1/2)))
+    (assert (equal (multiple-value-list (f -0.5)) '(-0.0 -0.5)))
+    (assert (equal (multiple-value-list (f -1/2)) '(-0.0 -1/2)))))
+
+(with-test (:name (ftruncate :minus-zeros :two-arg))
+  (flet ((f (x y) (declare (notinline ftruncate)) (ftruncate x y)))
+    (declare (notinline f))
+    (assert (equal (multiple-value-list (f 0 1)) '(0.0 0)))
+    (assert (equal (multiple-value-list (f 0 -1)) '(-0.0 0)))
+    (assert (equal (multiple-value-list (f 0.0 1)) '(0.0 0.0)))
+    (assert (equal (multiple-value-list (f 0.0 -1)) '(-0.0 0.0)))
+    (assert (equal (multiple-value-list (f -0.0 1)) '(-0.0 0.0)))
+    (assert (equal (multiple-value-list (f -0.0 -1)) '(0.0 0.0)))
+    (assert (equal (multiple-value-list (f 0.5 1)) '(0.0 0.5)))
+    (assert (equal (multiple-value-list (f 0.5 -1)) '(-0.0 0.5)))
+    (assert (equal (multiple-value-list (f 1/2 1)) '(0.0 1/2)))
+    (assert (equal (multiple-value-list (f 1/2 -1)) '(-0.0 1/2)))
+    (assert (equal (multiple-value-list (f -0.5 1)) '(-0.0 -0.5)))
+    (assert (equal (multiple-value-list (f -0.5 -1)) '(0.0 -0.5)))
+    (assert (equal (multiple-value-list (f -1/2 1)) '(-0.0 -1/2)))
+    (assert (equal (multiple-value-list (f -1/2 -1)) '(0.0 -1/2)))))
+
+(with-test (:name (ffloor :minus-zeros :two-arg))
+  (flet ((f (x y) (declare (notinline ffloor)) (ffloor x y)))
+    (declare (notinline f))
+    (assert (equal (multiple-value-list (f 0 1)) '(0.0 0)))
+    (assert (equal (multiple-value-list (f 0 -1)) '(-0.0 0)))
+    (assert (equal (multiple-value-list (f 0.0 1)) '(0.0 0.0)))
+    (assert (equal (multiple-value-list (f 0.0 -1)) '(-0.0 0.0)))
+    (assert (equal (multiple-value-list (f -0.0 1)) '(-0.0 0.0)))
+    (assert (equal (multiple-value-list (f -0.0 -1)) '(0.0 0.0)))
+    (assert (equal (multiple-value-list (f 0.5 1)) '(0.0 0.5)))
+    (assert (equal (multiple-value-list (f 0.5 -1)) '(-1.0 -0.5)))
+    (assert (equal (multiple-value-list (f 1/2 1)) '(0.0 1/2)))
+    (assert (equal (multiple-value-list (f 1/2 -1)) '(-1.0 -1/2)))
+    (assert (equal (multiple-value-list (f -0.5 1)) '(-1.0 0.5)))
+    (assert (equal (multiple-value-list (f -0.5 -1)) '(0.0 -0.5)))
+    (assert (equal (multiple-value-list (f -1/2 1)) '(-1.0 1/2)))
+    (assert (equal (multiple-value-list (f -1/2 -1)) '(0.0 -1/2)))))
+
+(with-test (:name (fceiling :minus-zeros :two-arg))
+  (flet ((f (x y) (declare (notinline fceiling)) (fceiling x y)))
+    (declare (notinline f))
+    (assert (equal (multiple-value-list (f 0 1)) '(0.0 0)))
+    (assert (equal (multiple-value-list (f 0 -1)) '(-0.0 0)))
+    (assert (equal (multiple-value-list (f 0.0 1)) '(0.0 0.0)))
+    (assert (equal (multiple-value-list (f 0.0 -1)) '(-0.0 0.0)))
+    (assert (equal (multiple-value-list (f -0.0 1)) '(-0.0 0.0)))
+    (assert (equal (multiple-value-list (f -0.0 -1)) '(0.0 0.0)))
+    (assert (equal (multiple-value-list (f 0.5 1)) '(1.0 -0.5)))
+    (assert (equal (multiple-value-list (f 0.5 -1)) '(-0.0 0.5)))
+    (assert (equal (multiple-value-list (f 1/2 1)) '(1.0 -1/2)))
+    (assert (equal (multiple-value-list (f 1/2 -1)) '(-0.0 1/2)))
+    (assert (equal (multiple-value-list (f -0.5 1)) '(-0.0 -0.5)))
+    (assert (equal (multiple-value-list (f -0.5 -1)) '(1.0 0.5)))
+    (assert (equal (multiple-value-list (f -1/2 1)) '(-0.0 -1/2)))
+    (assert (equal (multiple-value-list (f -1/2 -1)) '(1.0 1/2)))))
+
+(with-test (:name (fround :minus-zeros :two-arg))
+  (flet ((f (x y) (declare (notinline fround)) (fround x y)))
+    (declare (notinline f))
+    (assert (equal (multiple-value-list (f 0 1)) '(0.0 0)))
+    (assert (equal (multiple-value-list (f 0 -1)) '(-0.0 0)))
+    (assert (equal (multiple-value-list (f 0.0 1)) '(0.0 0.0)))
+    (assert (equal (multiple-value-list (f 0.0 -1)) '(-0.0 0.0)))
+    (assert (equal (multiple-value-list (f -0.0 1)) '(-0.0 0.0)))
+    (assert (equal (multiple-value-list (f -0.0 -1)) '(0.0 0.0)))
+    (assert (equal (multiple-value-list (f 0.5 1)) '(0.0 0.5)))
+    (assert (equal (multiple-value-list (f 0.5 -1)) '(-0.0 0.5)))
+    (assert (equal (multiple-value-list (f 1/2 1)) '(0.0 1/2)))
+    (assert (equal (multiple-value-list (f 1/2 -1)) '(-0.0 1/2)))
+    (assert (equal (multiple-value-list (f -0.5 1)) '(-0.0 -0.5)))
+    (assert (equal (multiple-value-list (f -0.5 -1)) '(0.0 -0.5)))
+    (assert (equal (multiple-value-list (f -1/2 1)) '(-0.0 -1/2)))
+    (assert (equal (multiple-value-list (f -1/2 -1)) '(0.0 -1/2)))))
+
+(with-test (:name (ftruncate double-float))
+  (flet ((f (x y)
+           (declare (type double-float x))
+           (let ((ev (eval `(ftruncate ,x)))
+                 (co (ftruncate x)))
+             (assert (eql ev y))
+             (assert (eql co y)))))
+     (f 0d0 0d0)
+     (f -0d0 -0d0)
+     (f 0.25d0 0d0)
+     (f 0.49d0 0d0)
+     (f 0.5d0 0d0)
+     (f 0.51d0 0d0)
+     (f 0.75d0 0d0)
+     (f -0.25d0 -0d0)
+     (f -0.49d0 -0d0)
+     (f -0.5d0 -0d0)
+     (f -0.51d0 -0d0)
+     (f -0.75d0 -0d0)
+     (f 1d0 1d0)
+     (f 1.5d0 1d0)
+     (f -1d0 -1d0)
+     (f -1.5d0 -1d0)
+     (f 1.0000000000000001d15 1d15)
+     (f -1.0000000000000001d15 -1d15)
+     (f 1d16 1d16)
+     (f -1d16 -1d16)))
+
+(with-test (:name (fround double-float))
+  (flet ((f (x y)
+           (declare (type double-float x))
+           (let ((ev (eval `(fround ,x)))
+                 (co (fround x)))
+             (assert (eql ev y))
+             (assert (eql co y)))))
+    (f 0d0 0d0)
+    (f -0d0 -0d0)
+    (f 0.25d0 0d0)
+    (f 0.49d0 0d0)
+    (f 0.5d0 0d0)
+    (f 0.51d0 1d0)
+    (f 0.75d0 1d0)
+    (f -0.25d0 -0d0)
+    (f -0.49d0 -0d0)
+    (f -0.5d0 -0d0)
+    (f -0.51d0 -1d0)
+    (f -0.75d0 -1d0)
+    (f 1d0 1d0)
+    (f 1.49d0 1d0)
+    (f 1.5d0 2d0)
+    (f 1.51d0 2d0)
+    (f -1d0 -1d0)
+    (f -1.49d0 -1d0)
+    (f -1.5d0 -2d0)
+    (f -1.51d0 -2d0)
+    (f 2d0 2d0)
+    (f 2.49d0 2d0)
+    (f 2.5d0 2d0)
+    (f 2.51d0 3d0)
+    (f -2d0 -2d0)
+    (f -2.49d0 -2d0)
+    (f -2.5d0 -2d0)
+    (f -2.51d0 -3d0)
+    (f 1.0000000000000001d15 1d15)
+    (f -1.0000000000000001d15 -1d15)
+    (f 1d16 1d16)
+    (f -1d16 -1d16)))
