@@ -16,8 +16,7 @@
 ;;;; It's just that sometimes we get :PINNED as a root instead of
 ;;;; the expected reference to the one and only thread.
 ;;;; And also sb-safepoint gets a crash in C.
-#-(and gencgc sb-thread (not sb-safepoint)
-       (or arm64 ppc64 x86-64))
+#-(and (or arm64 ppc64 x86-64) sb-thread (not sb-safepoint) (not gc-stress))
 (invoke-restart 'run-tests::skip-file)
 
 (setq sb-ext:*evaluator-mode* :compile)
@@ -35,17 +34,17 @@
            (ecase root
              (:tls
               (scrubstack)
-              (sb-ext:search-roots wp :criterion :oldest :gc t :print nil))
+              (sb-ext:search-roots wp :criterion :oldest :print nil))
              (:bindings ; bind *FRED* again so the old value is on the binding stack
               (let ((*fred* 1))
                 (scrubstack)
-                (sb-ext:search-roots wp :criterion :oldest :gc t :print nil)))
+                (sb-ext:search-roots wp :criterion :oldest :print nil)))
              (:stack
               ; put the OBJ back on the control stack
               ; and also ensure that *FRED* is not a root.
               (setq obj *fred* *fred* nil)
               (scrubstack)
-              (sb-ext:search-roots wp :criterion :oldest :gc t :print nil)))))
+              (sb-ext:search-roots wp :criterion :oldest :print nil)))))
       (assert paths)
       (let* ((path (cdar paths))
              (root (car path)))
@@ -59,6 +58,7 @@
            (assert (typep (cdr path) '(cons (eql *fred*) (cons nil))))))))))
 
 (with-test (:name (sb-ext:search-roots :stack-indirect)
+            :fails-on :mark-region-gc
             :fails-on :sunos)
   (let ((wp (make-weak-pointer (list 1 2 3 4))))
     (test1 wp (weak-pointer-value wp) :stack)
@@ -69,7 +69,7 @@
 (defun f0 ()
   (let* ((c (cons 1 2))
          (wp (make-weak-pointer c)))
-    (let ((paths (sb-ext:search-roots wp :criterion :static :gc t :print nil)))
+    (let ((paths (sb-ext:search-roots wp :criterion :static :print nil)))
       (assert paths)
       (let* ((path (car paths))
              (nodes (cdr path)))
@@ -81,14 +81,19 @@
   (f0))
 
 ;;; Employ circumlocution so the file loader doesn't hold on to a string "hi"
-(defvar *string-hi* (make-weak-pointer (concatenate 'string "h" "i")))
+(defvar *string-hi*)
 (defstruct s1 foo)
+;;; Ensure that *STRING-HI* does not get smashed before we ever got to both
+;;; DEFVARS (if GC executed in between the two) but that we don't hold a reference
+;;; from the stack. (which we can't really guarantee though)
 (defparameter *top*
-  `(p q r w x y ,(make-s1 :foo `#((a b c ,(weak-pointer-value *string-hi*) d))) z))
+  (let ((s (concatenate 'string "h" "i")))
+    (setq *string-hi* (make-weak-pointer s))
+    `(p q r w x y ,(make-s1 :foo `#((a b c ,s d))) z)))
 
 ;;; Sample output:
 ;;; Path to "hi":
-;;;  6       1000209AB3 [   1] a symbol-hashset
+;;;  6       1000209AB3 [   1] a symbol-table
 ;;;  1       10048F145F [  29] a (simple-vector 37)
 ;;;  1         503B403F [   2] COMMON-LISP-USER::*TOP*
 ;;;  0       1004B885B7 [   6] a cons = (P Q R ...) ; = (NTHCDR 6 object)
@@ -97,7 +102,9 @@
 ;;;  0       1004CBB93F [   2] a (simple-vector 1)
 ;;;  0       1004D28AE7 [   3] a cons = (A B C ...) ; = (NTHCDR 3 object)
 ;;;  0       1004D28B17 [   0] a cons = ("hi" D)
-(with-test (:name :traceroot-collapse-lists)
+(with-test (:name :traceroot-collapse-lists
+                  :fails-on :mark-region-gc
+                  :skipped-on :gc-stress)
   (let* ((string (with-output-to-string (*standard-output*)
                    (search-roots *string-hi* :print :verbose)))
          (lines (split-string string #\newline)))
@@ -119,15 +126,13 @@
     (something)))
 
 (defvar *foo*)
-#+gencgc
 (with-test (:name (sb-ext:search-roots :simple-fun)
-            :broken-on (and :darwin :arm64))
+                  :fails-on (and :mark-region-gc :darwin :arm64))
   ;; Tracing a path to a simple fun wasn't working at some point
   ;; because of failure to employ fun_code_header in the right place.
   (setq *foo* (compile nil '(lambda () 42)))
   (let ((wp (sb-ext:make-weak-pointer *foo*)))
     (assert (sb-ext:search-roots wp :criterion :oldest :print nil))))
 
-#+gencgc
 (with-test (:name (sb-ext:search-roots :ignore-immediate))
-  (sb-ext:search-roots (make-weak-pointer 48) :gc t :print nil))
+  (sb-ext:search-roots (make-weak-pointer 48) :print nil))
