@@ -842,7 +842,7 @@
             (sb-kernel:%simple-fun-type
              (checked-compile
               `(lambda (a) (array-rank (the (not (array t)) a))))))
-           `(values array-rank &optional))))
+           `(values (mod 129) &optional))))
 
 (with-test (:name (:rational-intersection :lp1998008))
   (flet ((bug101 ()
@@ -887,3 +887,42 @@
    (("") nil)
    ((t) t)
    ((nil) nil)))
+
+(with-test (:name :union-integer-complex)
+  (checked-compile-and-assert
+   ()
+   `(lambda (x)
+      (typep x '(or (integer 36757953510256822604)
+                 (complex fixnum))))
+   ((-1) nil)
+   ((36757953510256822603) nil)
+   ((36757953510256822604) t)
+   ((36757953510256822605) t)
+   ((#C(1d0 1d0)) nil)
+   ((#C(1 1)) t)
+   ((#C(1 #.(expt 2 300))) nil)))
+
+#+(or arm64 x86-64)
+(with-test (:name :structure-typep-fold)
+  (assert-type
+   (lambda (a b)
+     (declare (character a))
+       (sb-c::structure-typep a b))
+   null)
+  (assert-type
+   (lambda (a)
+     (declare (hash-table a))
+     (sb-c::structure-typep a #.(sb-kernel:find-layout 'condition)))
+   null)
+  (assert-type
+   (lambda (a)
+     (declare (pathname a))
+     (sb-c::structure-typep a #.(sb-kernel:find-layout 'pathname)))
+   (eql t)))
+
+(with-test (:name :typep-vector-folding)
+  (assert-type
+   (lambda (p)
+     (declare (integer p))
+     (typep p '(vector t 1)))
+   null))
